/*******************************************************************************
***
*** Filename         : script.h
*** Purpose          : Definitions for the connection scripting routines
*** Author           : Matt J. Gumbley
*** Created          : 19/01/97
*** Last updated     : 18/08/98
***
********************************************************************************
***
*** Modification Record
***
*******************************************************************************/

#ifndef SYS_SCRIPT_H
#define SYS_SCRIPT_H

/* Definition of script structures. Each interface can have a script running
   on it, and so the script interpreter context is embedded in the network
   interface.
*/
#define SI_NAMEMAXSIZE 32             /* All names can be up to 32 characters */
#define SI_STACKSIZE 32       /* The evaluation stack consists of 32 word16's */
#define SI_VARIABLELENGTH 255     /* Each variable holds up to 255 characters */
/* Note you can't go over this, since we're using a byte to index this array */
#define SI_EXPBUFMAX 255   /* Size of variable/char-constant expansion buffer */


/* Label pointers */
struct label {
  char l_name[SI_NAMEMAXSIZE];                            /* The label's name */
  char *l_ptr;     /* It points to this location in the script storage buffer */
};


/* Variable flags */
#define VAR_WRITABLE 0x1         /* Failed if we try to write to one of these */
#define VAR_ISP      0x2                         /* Stores details of the ISP */
#define VAR_IF       0x4                   /* Stores details of the interface */
#define VAR_SI       0x8          /* Stores details of the script interpreter */
#define VAR_PR       0x10            /* Stores details of the comms protocols */

#define VAR_SYSTEM   (VAR_ISP|VAR_IF|VAR_SI|VAR_PR)   /* To check system vars */


/* Variables. They're stored in a simple linked list. */
struct var {
  char v_name[SI_NAMEMAXSIZE];                         /* The variable's name */
  char v_contents[SI_VARIABLELENGTH];              /* The variable's contents */
  byte v_index;        /* The next available position in the l_contents array */
  byte v_flags;                                    /* VAR_ flags above, OR'ed */
  struct var *v_next;                         /* Pointer to the next variable */
};


/* The parser table: Command names and their interpreter functions */
struct cmd {
  char *name;
  void (*fn) (char *);
};


/* Input states */
#define SIS_DISCARD 0x1                        /* We discard all serial input */
#define SIS_BUILDUP 0x2              /* All serial data goes into $waitstring */
#define SIS_DISCARDUNTIL 0x3          /* Discard all input until si_untilchar */

/* Sleep states */
#define SSS_ACTIVE  0x1           /* The script is actively being interpreted */
#define SSS_SLEEP   0x2                  /* Waiting for some timeout to occur */
#define SSS_DOZE    0x3             /* Waiting for some serial input to occur */
#define SSS_DOZEUNTIL 0x4                        /* Dozing until si_untilchar */

/* Debug states */
#define SID_NONE    0x1
#define SID_TRACE   0x2    /* Instructions are displayed as they are executed */
#define SID_TRACEIN 0x3       /* " + $waitstring is displayed on wait timeout */

/* Script context information structure, always available, although some 
   parts of it may only be provided when the interface is up, or being 
   brought up.
*/
struct scriptcontext {
  byte si_scriptused;                     /* Whether scripting is used or not */
  char si_scriptname[P_FNAMESIZE+2];      /* Script name, or [0]='\0' if none */
  char *si_scriptbuf;            /* The whole script is stored in this buffer */
  char *si_scriptextent;           /* Pointer to the end of the script buffer */
  long si_scriptsize;                        /* Number of bytes in the script */
  char *si_IP;          /* Instruction pointer to the next script instruction */
  struct cmd *si_cmd;    /* Pointer to parser table entry for current command */
  int si_numlines;                       /* The number of lines in the script */
  char **si_lineptrs; /* Array of pointers to start of lines in script buffer */
  int si_numlabels;         /* Number of labels, found by scanning the script */
  struct label *si_labelptrs;        /* Array of labels, si_numlabels of them */
  int si_numvars;            /* Number of variables, adjusted when one occurs */
  struct var *si_varlist;     /* Linked list of variables, si_numvars of them */
  char si_expandbuf[SI_EXPBUFMAX]; /* Variable & char-constant expansion area */
  int si_expandbuflen;                       /* Length of it, after expansion */
  word16 si_stack[SI_STACKSIZE];                          /* Evaluation stack */
  byte si_SP;                             /* Index to top of evaluation stack */
  byte si_inputstate;                       /* One of the SIS_??? flags above */
  byte si_sleepstate;                       /* One of the SSS_??? flags above */
  byte si_debuglevel;                       /* One of the SID_??? flags above */
  byte si_untilchar;          /* Character to wait for if in SIS_DISCARDUNTIL */
  byte si_capturing;                               /* Whether we're capturing */
  VOID *si_captureH;                          /* File handle for capture file */
};


#endif /* SYS_SCRIPT_H */
