/*******************************************************************************
***
*** Filename         : tcp.h
*** Purpose          : Definitions for the TCP layer in PsiStack
*** Author           : Matt J. Gumbley
*** Created          : 27/01/97
*** Last updated     : 15/08/98
***
********************************************************************************
***
*** Modification Record
***
*******************************************************************************/

#ifndef SYS_TCP_H
#define SYS_TCP_H


/* TCP endpoint types */
#define TCPT_SERVER      1
#define TCPT_CONNECTION  2
#define TCPT_MASTER      3

#define TCPINCR          904
#define LOWBYTE          0377
#define TLPORT           1024 /* FROM BSD */

#define max(a,b) ((a) < (b) ? (b) : (a))
#define min(a,b) ((a) < (b) ? (a) : (b))

typedef	long int tcpseq; 


/*
 * SEQCMP - sequence space comparator
 *  This handles sequence space wrap-around. Overlow/Underflow makes
 * the result below correct ( -, 0, + ) for any a, b in the sequence
 * space. Results:  result  implies
 *        -    a < b
 *        0    a = b
 *        +    a > b
 */
#define  SEQCMP(a, b)  ((a) - (b))

/* TCP packet format */

struct tcp {
  word16 tcp_sport;                                            /* source port */
  word16 tcp_dport;                                       /* destination port */
  tcpseq tcp_seq;                                          /* sequence number */
  tcpseq tcp_ack;                             /* acknowledged sequence number */
  byte   tcp_offset;       /* top 4 bits are header length, rest are reserved */
  byte   tcp_code;                                           /* control flags */
  word16 tcp_window;                                  /* window advertisement */
  word16 tcp_cksum;                                               /* checksum */
  word16 tcp_urgptr;                        /* urgent pointer if URG flag set */
  byte   tcp_data[1];
};

/* TCP Control Bits */

#define  TCPF_URG  0x20  /* urgent pointer is valid    */
#define  TCPF_ACK  0x10  /* acknowledgement field is valid  */
#define  TCPF_PSH  0x08  /* this segment requests a push    */
#define  TCPF_RST  0x04  /* reset the connection      */
#define  TCPF_SYN  0x02  /* synchronize sequence numbers    */
#define  TCPF_FIN  0x01  /* sender has reached end of its stream  */

#define  TCPMHLEN    20  /* minimum TCP header length    */
#define  TCPHOFFSET  0x50  /* tcp_offset value for TCPMHLEN  */
#define  TCP_HLEN(ptcp)    (((ptcp)->tcp_offset & 0xf0)>>2)

/* TCP Options */

#define  TPO_EOOL  0  /* end Of Option List      */
#define  TPO_NOOP  1  /* no Operation        */
#define  TPO_MSS   2  /* maximum Segment Size      */


/* TCP input states */

#define TCPIS_FREE         0
#define TCPIS_CLOSED       1
#define TCPIS_LISTEN       2
#define TCPIS_SYNSENT      3
#define TCPIS_SYNRCVD      4
#define TCPIS_ESTABLISHED  5
#define TCPIS_FINWAIT1     6
#define TCPIS_FINWAIT2     7
#define TCPIS_CLOSEWAIT    8
#define TCPIS_LASTACK      9
#define TCPIS_CLOSING      10
#define TCPIS_TIMEWAIT     11

#define NTCPISTATES        12

/* TCP output States */

#define TCPOS_IDLE         0
#define TCPOS_PERSIST      1
#define TCPOS_XMIT         2
#define TCPOS_REXMT        3

#define NTCPOSTATES        4


/* implementation parameters */

/* FIXME: these timer durations are all expressed in 10ths of a second
          We'll have to re-do the maths involving adaptive retransmission 
          to get those timings correct WRT this clock */

#define TCP_MAXRETRIES      12        /* max retransmissions before giving up */
#define TCP_TWOMSL        1200        /* 2 minutes (2 * Max Segment Lifetime) */
#define TCP_MAXRXT         200                   /* 20 seconds max rexmt time */
#define TCP_MINRXT           5                   /* 1/2 second min rexmt time */
#define TCP_ACKDELAY         2           /* 1/5 sec ACK delay, if TCBF_DELACK */
#define TCP_MAXPRS         600                   /* 1 minute max persist time */

/* second argument to tcpsend(): */

#define TSF_NEWDATA          0                           /* send all new data */
#define TSF_REXMT            1        /* retransmit the first pending segment */

/* third argument to tcpwr(): */

#define TWF_NORMAL           0                           /* normal data write */
#define TWF_URGENT           1                           /* urgent data write */


/* The union returned by the TCP STATUS control call */
struct tcpstat {
  int  ts_type;  /* which kind of TCP status?    */
  union {
    struct {
      long  tsu_connects;  /* # connections  */
      long  tsu_aborts;  /* # aborts    */
      long  tsu_retrans;  /* # retransmissions  */
    } T_unt;
    struct {
      IPaddr  tsu_laddr;  /* local IP    */
      short  tsu_lport;  /* local TCP port  */
      IPaddr  tsu_faddr;  /* foreign IP    */
      short  tsu_fport;  /* foreign TCP port  */
      short  tsu_rwin;  /* receive window  */
      short  tsu_swin;  /* peer's window  */
      short  tsu_state;  /* TCP state    */
      long  tsu_unacked;  /* bytes unacked  */
      int  tsu_prec;  /* IP precedence  */
    } T_unc;
    struct {
      long  tsu_requests;  /* # connect requests  */
      long  tsu_qmax;  /* max queue length  */
    } T_uns;
  } T_un;
};

#define  ts_connects  T_un.T_unt.tsu_connects
#define  ts_aborts  T_un.T_unt.tsu_aborts
#define  ts_retrans  T_un.T_unt.tsu_retrans

#define  ts_laddr  T_un.T_unc.tsu_laddr
#define  ts_lport  T_un.T_unc.tsu_lport
#define  ts_faddr  T_un.T_unc.tsu_faddr
#define  ts_fport  T_un.T_unc.tsu_fport
#define  ts_rwin    T_un.T_unc.tsu_rwin
#define  ts_swin    T_un.T_unc.tsu_swin
#define  ts_state  T_un.T_unc.tsu_state
#define  ts_unacked  T_un.T_unc.tsu_unacked
#define  ts_prec    T_un.T_unc.tsu_prec

#define  ts_requests  T_un.T_uns.tsu_requests
#define  ts_qmax    T_un.T_uns.tsu_qmax



#define TCPQLEN    20       /* TCP input & output pseudo-process queue length */
#define TCPMAXURG  IP_MAXSIZE                 /* max. urgent data buffer size */
                                                   /* was MAXNETBUF (1500+20) */
#define TCPUQLEN   5                              /* TCP urgent queue lengths */
#define TCPMINSIZE (IPMHLEN + TCPMHLEN)     /* Size of the smallest TCP frame */

#define READERS             1
#define WRITERS             2

/* tcb_flags */

#define TCBF_NEEDOUT  0x01  /* we need output      */
#define TCBF_FIRSTSEND  0x02  /* no data to ACK      */
#define TCBF_GOTFIN  0x04  /* no more to receive      */
#define TCBF_RDONE  0x08  /* no more receive data to process  */
#define TCBF_SDONE  0x10  /* no more send data allowed    */
#define TCBF_DELACK  0x20  /* do delayed ACK's      */
#define TCBF_BUFFER  0x40  /* do TCP buffering (default no)  */
#define TCBF_PUSH  0x80  /* got a push; deliver what we have  */
#define TCBF_SNDFIN  0x100  /* user process has closed; send a FIN  */
#define TCBF_RUPOK  0x200  /* receive urgent pointer is valid  */
#define TCBF_SUPOK  0x400  /* send urgent pointer is valid  */

/* aliases, for user programs */

#define TCP_BUFFER  TCBF_BUFFER
#define TCP_DELACK  TCBF_DELACK

/* receive segment reassembly data */

#define NTCPFRAG  10

/* URG queue element data */

#define UQTSIZE  (tcp_config_used.tcpc_NTCBS << 1)
  /* (total) max # pending urgent segs  */

#define UQS_FREE  0
#define UQS_ALLOC  1

struct  uqe {
  byte  uq_state;  /* UQS_* above (MJG: was int) */
  tcpseq  uq_seq;    /* start sequence of this buffer  */
  int  uq_len;    /* length of this buffer    */
  char  *uq_data;  /* data (0 if on urgent hole queue)  */
};

/* compute urgent data send and receive queue keys      */
#define  SUDK(ptcb, seq)  (ptcb->tcb_sudseq - (seq))
#define  SUHK(ptcb, seq)  (ptcb->tcb_suhseq - (seq))
#define  RUDK(ptcb, seq)  (ptcb->tcb_rudseq - (seq))
#define  RUHK(ptcb, seq)  (ptcb->tcb_ruhseq - (seq))

struct tcb {
  byte    tcb_istate;  /* input state        */ /* NB: was tcb_state */
  byte    tcb_ostate;  /* output state        */
  byte    tcb_type;  /* TCP type (SERVER, CLIENT)    */
  word16  tcb_code;  /* TCP code for next packet    */
  word16  tcb_flags;  /* various TCB state flags    */
  word16  tcb_error;  /* E_NET_xxx return error for user side    */

  IPaddr  tcb_rip;  /* remote IP address      */
  word16  tcb_rport;  /* remote TCP port      */
  IPaddr  tcb_lip;  /* local IP address      */
  word16  tcb_lport;  /* local TCP port      */

  struct  netif  *tcb_pni; /* pointer to our interface    */

  tcpseq  tcb_suna;  /* send unacked        */
  tcpseq  tcb_snext;  /* send next        */
  tcpseq  tcb_slast;  /* sequence of FIN, if TCBF_SNDFIN  */
  long    tcb_swindow;  /* send window size (octets)    */
  tcpseq  tcb_lwseq;  /* sequence of last window update  */
  tcpseq  tcb_lwack;  /* ack seq of last window update  */
  word16  tcb_cwnd;  /* congestion window size (octets)  */
  word16  tcb_ssthresh;  /* slow start threshold (octets)  */
  word16  tcb_smss;  /* send max segment size (octets)  */
  tcpseq  tcb_iss;  /* initial send sequence    */

  int     tcb_srt;  /* smoothed Round Trip Time    */
  int     tcb_rtde;  /* Round Trip deviation estimator  */
  int     tcb_persist;  /* persist timeout value    */
  int     tcb_keep;  /* keepalive timeout value    */
  int     tcb_rexmt;  /* retransmit timeout value    */
  int     tcb_rexmtcount;  /* number of rexmts sent    */

  tcpseq  tcb_rnext;  /* receive next        */

  tcpseq  tcb_rudseq;  /* base sequence for rudq entries  */
  struct queue *tcb_rudq;  /* receive urgent data queue    */
  tcpseq  tcb_ruhseq;  /* base sequence for ruhq entries  */
  struct queue *tcb_ruhq;  /* receive urgent hole queue    */
  int     tcb_sudseq;  /* base sequence for sudq entries  */
  struct queue *tcb_sudq;  /* send urgent data queue    */
  int     tcb_suhseq;  /* base sequence for suhq entries  */
  struct queue *tcb_suhq;  /* send urgent hole queue    */

  int     tcb_lqsize;  /* listen queue size (SERVERs)    */
  struct queue *tcb_listenq; /* listen queue port (SERVERs)    */
  struct tcb *tcb_pptcb;  /* pointer to parent TCB (for ACCEPT)  */

  byte   *tcb_sndbuf;  /* send buffer        */
  int     tcb_sbstart;  /* start of valid data      */
  int     tcb_sbcount;  /* data character count      */
  int     tcb_sbsize;  /* send buffer size (bytes)    */

  byte   *tcb_rcvbuf;  /* receive buffer (circular)    */
  int     tcb_rbstart;  /* start of valid data      */
  int     tcb_rbcount;  /* data character count      */
  int     tcb_rbsize;  /* receive buffer size (bytes)    */

  int     tcb_rmss;  /* receive max segment size    */
  tcpseq  tcb_cwin;  /* seq of currently advertised window  */
  struct queue *tcb_rsegq;  /* segment fragment queue    */
  tcpseq  tcb_finseq;  /* FIN sequence number, or 0    */
  tcpseq  tcb_pushseq;  /* PUSH sequence number, or 0    */

  /* Process control data. Each socket can have:
     > A process waiting for an open/close, 
     > A process waiting for read data, 
     > A process waiting for write data. 
     These *could* be separate processes, which is why we store the PID for 
     each type of possible wait. I suspect we won't get more than one process 
     working on a given socket, although sockets (and their IDs) are shared
     across the whole system, so it's possible to have one process opening,
     one reading, and one writing.

     These are the equivalent of the Comer's TCB fields tcb_rsema, tcb_ssema,
     tcb_ocsem.
  */
  word16 tcb_ocpid; /* If non-NULL, refers to a process waiting on open/close */
  word16 tcb_ocstatword;    /* Address in tcb_ocpid's memory of a status word */

  word16 tcb_wrpid;    /* If non-NULL, refers to a process waiting on a write */
  word16 tcb_wrbuf;    /* The address in tcb_wrpid's memory of a write buffer */
  word16 tcb_wrlen;                       /* Length of the above write buffer */
  word16 tcb_wrstatword;    /* Address in tcb_wrpid's memory of a status word */

  word16 tcb_rdpid;     /* If non-NULL, refers to a process waiting on a read */
  word16 tcb_rdbuf;     /* The address in tcb_rdpid's memory of a read buffer */
  word16 tcb_rdlen;                        /* Length of the above read buffer */
  word16 tcb_rdstatword;    /* Address in tcb_rdpid's memory of a status word */
};

/* TCP fragment structure */

struct tcpfrag {
  tcpseq  tf_seq;
  int  tf_len;
};

/* TCP control() functions */

#define  TCPC_LISTENQ  0x01  /* set the listen queue length    */
#define  TCPC_ACCEPT  0x02  /* wait for connect after passive open  */
#define  TCPC_STATUS  0x03  /* return status info (all, for master)  */
#define  TCPC_SOPT  0x04  /* set user-selectable options    */
#define  TCPC_COPT  0x05  /* clear user-selectable options  */
#define  TCPC_SENDURG  0x06  /* write urgent data      */


#define TCPSBS  2048                    /* TCP send buffer sizes (orig. 4096) */
#define TCPRBS  2048                /* TCP receive buffer sizes (orig. 19384) */
/* FIXME: these might be a little too big? */

#define NPORTS  100 /* FIXME: need this? */

#define MINNTCBS 4            /* Minimum # transport control blocks available */
#define INITNTCBS 10          /* Initial # transport control blocks available */
#define MAXNTCBS 16           /* Maximum # transport control blocks available */

#define INIT_TCP_BSDURG FALSE
/* The above are the initial settings. You can override them by changing the
   configuration file. This structure holds the config file details:
*/
struct tcp_config {
  byte tcpc_bsdurg;               /* Whether we use BSD urgend data semantics */
  byte tcpc_NTCBS;               /* Number of transport control blocks to use */
};

#endif /* SYS_TCP_H */

