/*******************************************************************************
***
*** Filename         : udp.h
*** Purpose          : Definitions for the UDP layer in PsiStack
*** Author           : Matt J. Gumbley
*** Created          : 27/01/97
*** Last updated     : 01/03/98
***
********************************************************************************
***
*** Modification Record
***
*******************************************************************************/

#ifndef SYS_UDP_H
#define SYS_UDP_H

#define U_HLEN  8                               /* UDP header length in bytes */

/* maximum data in UDP packet   */
#define U_MAXLEN        (IP_MAXSIZE-(IP_MINHLEN<<2)-U_HLEN)

/* UDP constants */
#define ULPORT          2050               /* initial UDP local "port" number */

/* Assigned UDP port numbers */
#define UDP_ECHO         7                                     /* echo server */
#define UDP_DISCARD      9                                  /* discard packet */
#define UDP_USERS        11                                   /* users server */
#define UDP_DAYTIME      13                            /* day and time server */
#define UDP_QOTD         17                        /* quote of the day server */
#define UDP_CHARGEN      19                            /* character generator */
#define UDP_TIME         37                                    /* time server */
#define UDP_WHOIS        43               /* who is server (user information) */
#define UDP_DNAME        53                             /* domain name server */
#define UDP_TFTP         69          /* trivial file transfer protocol server */
#define UDP_RWHO         513                   /* remote who server (ruptime) */
#define UDP_RIP          520              /* route information exchange (RIP) */

/* UDP datagram format */
struct udp {
  word16 udp_srcport;                                      /* UDP source port */
  word16 udp_dstport;                                 /* UDP destination port */
  word16 udp_len;  /* Total message length (in bytes) incl. UDP header & data */
  word16 udp_cksum;         /* UDP checksum (may be zero, to reduce overhead) */
  byte   udp_data[U_MAXLEN];                           /* Data in UDP message */
};


/* Constants for port specifications on UDP open call */
#define ANYFPORT        0                      /* accept any foreign UDP port */
#define ANYLPORT        0                    /* assign a fresh local port num */

#define MINUPPS 4                                /* Minimum # ports available */
#define INITUPPS 4                               /* Initial # ports available */
#define MAXUPPS 16                               /* Maximum # ports available */

#define MINUPPLEN 20                  /* Minimum # datagrams in a demux queue */
#define INITUPPLEN 50                 /* Initial # datagrams in a demux queue */
#define MAXUPPLEN 80                  /* Minimum # datagrams in a demux queue */

/* The above are the initial settings. You can override them by changing the
   configuration file. This structure holds the config file details:
*/
struct udp_config {
  byte udpc_UPPS;
  byte udpc_UPPLEN;
};

/* Mapping of external network UDP "port" to internal Psion "port" */
struct upq {                                       /* UDP demultiplexing info */
  int    upq_inuse;                                  /* is this entry in use? */
  word16 upq_lport;                                  /* local UDP port number */
  word16 upq_fport;                                /* foreign UDP port number */
  IPaddr upq_fip;                               /* foreign machine IP address */
  struct queue *upq_q;    /* Demux queue of incoming datagrams of size UPPLEN */

  /* Process control data. Each socket can have:
     > A process waiting for an open/close, 
     > A process waiting for read data, 
     > A process waiting for write data. 
     These *could* be separate processes, which is why we store the PID for 
     each type of possible wait. I suspect we won't get more than one process 
     working on a given socket, although sockets (and their IDs) are shared
     across the whole system, so it's possible to have one process opening,
     one reading, and one writing.
  */
  word16 upq_ocpid; /* If non-NULL, refers to a process waiting on open/close */
  word16 upq_ocstatword;    /* Address in upq_ocpid's memory of a status word */

  word16 upq_wrpid;    /* If non-NULL, refers to a process waiting on a write */
  word16 upq_wrbuf;    /* The address in upq_wrpid's memory of a write buffer */
  word16 upq_wrlen;                       /* Length of the above write buffer */
  word16 upq_wrstatword;    /* Address in upq_wrpid's memory of a status word */

  word16 upq_rdpid;     /* If non-NULL, refers to a process waiting on a read */
  word16 upq_rdbuf;     /* The address in upq_rdpid's memory of a read buffer */
  word16 upq_rdlen;                        /* Length of the above read buffer */
  word16 upq_rdstatword;    /* Address in upq_rdpid's memory of a status word */
};


#endif /* SYS_UDP_H */

