/*******************************************************************************
***
*** Filename         : asy.c
*** Purpose          : PsiStack Asynchronous Serial routines
*** Author           : Matt J. Gumbley
*** Created          : 09/04/97
*** Last updated     : 19/01/98
***
********************************************************************************
***
*** Modification Record
***
*******************************************************************************/

#include "stack.h"

GLDEF_D struct asy asy;

LOCAL_C INT asy_calcparams(P_SRCHAR *);
LOCAL_C INT asy_setparams(P_SRCHAR *);

/*******************************************************************************
***
*** Function         : asy_init
*** Preconditions    : None
*** Postconditions   : The asynchronous layer is initialised
***
*******************************************************************************/

VOID asy_init(VOID)
{
  p_scpy(asy.asy_serDeviceName,"TTY:A");
  asy.asy_speed=P_BAUD_19200;
  asy.asy_parity=NOPAR;
  asy.asy_databits=P_DATA_8;
  asy.asy_handshake=RTSCTS;
}


/*******************************************************************************
***
*** Function         : asy_calcparams(p)
*** Preconditions    : the various serial settings have been stored in asy_*,
***                    p is a pointer to a P_SRCHAR struct.
*** Postconditions   : This sets the P_SRCHAR type containing the serial
***                    settings, in a Psion-readable form :-)
***                    It returns TRUE on success, and FALSE on failure.
***
*******************************************************************************/

LOCAL_C INT asy_calcparams(P_SRCHAR *p)
{
LOCAL_D TEXT fname[]="asy_calcparams: %s\r\n";

  /* Convert baud rate into the correct P_SRCHAR flag */
  p->rbaud = p->tbaud = asy.asy_speed;
  p->frame = 0;
  p->flags = 0;
  p->xon = 0x11;
  p->xoff = 0x13;
  p->tmask = 0L;

  /* Convert the parity setting */
  switch (asy.asy_parity) {
    case IGNORE:
      p->parity = 0; 
      p->flags |= P_IGNORE_PARITY;
      break;
    case NOPAR:
      p->parity = 0; 
      p->flags &= ~P_IGNORE_PARITY;
      break;
    case EVENPAR:
      p->parity = P_PAR_EVEN; 
      p->frame |= P_PARITY;
      p->flags &= ~P_IGNORE_PARITY;
      break;
    case ODDPAR:
      p->parity = P_PAR_ODD; 
      p->frame |= P_PARITY;
      p->flags &= ~P_IGNORE_PARITY;
      break;
    default:
      console_printf(fname, (UINT) "Bad parity setting");
      return FALSE;
  }

  /* Convert the data and stop bits setting */
  p->frame |= (asy.asy_databits | asy.asy_stopbits);

  /* Convert the handshake setting */
  switch (asy.asy_handshake) {
    case IGNORE:
      p->hand = P_IGN_CTS;
      break;
    case XONXOFF: 
      p->hand = P_OBEY_XOFF | P_SEND_XOFF | P_IGN_CTS;
      break;
    case RTSCTS: 
      p->hand = 0;
      break;
    default:
      p->hand = 0;
      console_printf(fname, (UINT) "Bad handshake setting");
      return FALSE;
  }
  p->hand |= (asy.asy_devicedetect==1 ? (P_FAIL_DSR | P_OBEY_DSR) : 0) |
             (asy.asy_carrierdetect==1 ? (P_FAIL_DCD | P_OBEY_DCD) : 0);
  return TRUE;
}


/*******************************************************************************
***
*** Function         : asy_setparams(p)
*** Preconditions    : The serial port is open, p is a pointer to a P_SRCHAR 
***                    structure, as created by a call to asy_calcparams
*** Postconditions   : asy_setparams returns TRUE on success, and the port
***                    is set as required, otherwise, FALSE is returned.
***
*******************************************************************************/

LOCAL_C INT asy_setparams(P_SRCHAR *p)
{
INT retval;
register TEXT *err;
  if ((retval=p_iow(asy.asy_serH, P_FSET, p)) != 0) {
    switch (retval) {
      case E_GEN_ARG   : err= "bad settings"; break;
      case E_GEN_NSUP  : err= "not supported"; break;
      case E_FILE_LINE : err= "no DSR/DTR or DCD"; break;
      default          : err= "huh?"; break;
    }
    console_printf("asy_setparams: %s (%d)\r\n", (UINT) err, retval);
    return FALSE;
  }
  return TRUE;
}


/*******************************************************************************
***
*** Function         : asy_open opens the specified port with the specified 
***                    characteristics, or, if the port is already open, 
***                    changes the port characteristics.
*** Preconditions    : asy_serDeviceName is a string like "TTY:A"
***                    asy_speed is P_BAUD_50 to P_BAUD_115000
***                    asy_parity is IGNORE, NOPAR, EVENPAR or ODDPAR
***                    asy_databits is P_DATA_7 or P_DATA_8
***                    asy_stopbits is P_TWOSTOP or 0 (=> One stop bit)
***                    asy_handshake is IGNORE, XONXOFF, RTSCTS
*** Postconditions   : asy_open is TRUE, and the open worked.
***                    asy_open is negative, and the open failed.
***
*******************************************************************************/

INT asy_open(VOID)
{
P_SRCHAR params;
INT retval;
LOCAL_D TEXT fname[]="asy_open: %s %s\r\n";
register TEXT *err;

  if (asy.asy_serDeviceName[0]=='\0') {
    console_printf(fname,(UINT)"","No device name set");
    return E_FILE_DEVICE;
  }

  /* Need to convert the ASY interface's definitions of the port 
     characteristics into things EPOC can work with: P_SRCHAR flags, etc. */
  if (!asy_calcparams(&params))
    return E_FILE_INVALID;

  /* Open the port, if it isn't already open */
  if (asy.asy_serH == NULL) {
    if ((retval=p_open(&asy.asy_serH, asy.asy_serDeviceName, -1)) != 0) {
      switch (retval) {
        case E_GEN_NOMEMORY: err= "no memory"; break;
        case E_GEN_INUSE   : err= "in use"; break;
        case E_FILE_DEVICE : err= "illegal device"; break;
        case E_FILE_LOCKED : err= "is PsiWin link active?"; break;
        default            : err= "bug"; break;
      }
      console_printf(fname,(UINT)"open",err);
      asy.asy_serH = NULL;
      return retval;
    }
  }

  /* Now change the parameters */
  if (!asy_setparams(&params)) {
    asy_close();
    return E_FILE_INVALID;
  }

  /* Queue a read, abandoning any previously queued read */
  p_iow2(asy.asy_serH, P_FCANCEL);
  p_waitstat(&asy.asy_serStatus);
  if ((retval=asy_queue()) != 0) {
    switch (retval) {
      case E_FILE_PARITY   : err= "parity err"; break;
      case E_FILE_FRAME  : err= "framing err"; break;
      case E_FILE_OVERRUN: err= "overrun err"; break;
      case E_FILE_LINE : err= "line fail"; break;
      case E_GEN_OVER : err= "input discard"; break;
      default          : err= "bug"; break;
    }
    console_printf(fname,(UINT)"read",err);
    asy_close();
    return retval;
  }

  if (netif->ni_dlp == NIDLP_PPP && !sc.si_scriptused)
    ppp_getevent(EVENT_UP); 
  /* If we're using PPP protocol and not using a script, then 
     opening the serial link trigger an UP event to signal PPP that the 
     hardware link layer is ready to process PPP packets. This is probably 
     not enough, we should be sure that the physical link is established. 
     Any idea about this, Matt?
     I need such a signal for the PPP FSM. [Patrick] 

     The physical layer is up at this point - the p_open worked, asy_setparams
     worked, and a read has been queued. This should be sufficient. Or do you 
     mean that a modem connection has been made? You'd be using a script if 
     this is the case. "No script" means "direct connection", and if the 
     above worked, then the RS232 connection is OK.  [Matt] 
   */

  return TRUE;
}



/*******************************************************************************
***
*** Function         : asy_close()
*** Precondition     : The serial device is open.
*** Postcondition    : The device is closed.
***
*******************************************************************************/

VOID asy_close(VOID)
{
  if (asy.asy_serH == NULL)
    return;
  p_iow2(asy.asy_serH, P_FCANCEL);
  p_waitstat(&asy.asy_serStatus);
  p_close(asy.asy_serH);
  asy.asy_serH = NULL;
  msg_event(EVT_LINEDROPPED, NULL, 0);
}


/*******************************************************************************
***
*** Function         : asy_write
*** Precondition     : fd is open.
*** Postcondition    : asy_write is positive and the data has been sent. The
***                    return value indicates how many bytes have been sent.
***                    asy_write is FALSE and the data has not been sent.
***
*******************************************************************************/

INT asy_write(BYTE *Data, INT len)
{
register int retval;
UBYTE ctrl[3];
LOCAL_D TEXT fname[]="asy_write: %s\r\n";

  if (asy.asy_serH == NULL) {
    console_printf(fname, (UINT)"Port not open");
    return FALSE;
  }

#ifdef SLIPDUMP
  printf( "asy_write: dump:\n");
  hexdump(Data, len);
#endif

  /* If we're using hardware handshaking, is CTS active? Checking here will
     help prevent the line blocking... Just pretend we've sent the data,
     but warn the user. */
  if (asy.asy_handshake == RTSCTS) {
    ctrl[1]=0x00;
    if (p_iow(asy.asy_serH, P_FCTRL, ctrl)==0 && (!ctrl[0] & P_SRCTRL_CTS)) {
      console_printf(fname,(UINT)"CTS down");
      return len;
    }
  }

  /* OK, this should be done asynchronously in a "quality system". */
  retval=p_write(asy.asy_serH, Data, len);
  switch (retval) {
    case E_FILE_PARITY:
    case E_FILE_FRAME:
    case E_FILE_OVERRUN:
      console_printf(fname,(UINT)"frame error - check baud rate etc.");
      break;
    case E_FILE_LINE:
      console_printf(fname,(UINT)"line was dropped");
      /*msg_event(EVT_LINEDROPPED, NULL, 0);*/
      break;
    case 0: /* All OK */
      break;
    default:
      console_printf("asy_write: write failed err %d\r\n", retval);
  }
  return (retval==0); /* Don't think this indicates how much was sent! */
}


/*******************************************************************************
***
*** Function         : asy_writechar
*** Precondition     : fd is open, Data is a character
*** Postcondition    : asy_writechar is positive and the data has been sent. 
***                    asy_writechar is FALSE and the data has not been sent.
***
*******************************************************************************/

INT asy_writechar(BYTE Data)
{
BYTE Buffer = Data;
  return asy_write(&Buffer, 1);
}


/*******************************************************************************
***
*** Function         : asy_queue
*** Preconditions    : 
*** Postconditions   : Returns 0 on OK, and a read is queued, or negative on 
***                    error.
***
*******************************************************************************/

INT asy_queue(VOID)
{
INT retval;
LOCAL_D TEXT fname[]="asy_queue: %s\r\n";
register TEXT *err;
  if (asy.asy_serH == NULL)
    return E_FILE_LINE;
  asy.asy_serLength = 1;
  retval= p_ioa5(asy.asy_serH, P_FREAD, &asy.asy_serStatus,
                 &asy.asy_serBuffer, &asy.asy_serLength);
  if (retval != 0) {
    switch (retval) {
      case E_FILE_PARITY   : err= "parity err"; break;
      case E_FILE_FRAME  : err= "framing err"; break;
      case E_FILE_OVERRUN: err= "overrun err"; break;
      case E_FILE_LINE : err= "line fail"; break;
      case E_GEN_OVER : err= "input discard"; break;
      default          : err= "huh?"; break;
    }
    console_printf(fname, (UINT)err);
    asy_close();
  }
  return retval;
}


/*******************************************************************************
***
*** Function         : asy_process()
*** Preconditions    : at least one octet of serial data is available.
*** Postconditions   : Any further outstanding serial data is read into the
***                    serial input buffer.
***
*******************************************************************************/

VOID asy_process(VOID)
{
  if (asy.asy_serLength) { /* Append further data to buffer */
    p_iow(asy.asy_serH, P_FTEST, (UWORD *)&asy.asy_serLength);

    /* Limit the amount that can be processed due to small buffer */
    if (asy.asy_serLength > (ASY_SERBUFSIZE-1)) {
      /* If you get one of these in your log, please mail me.. */
      console_printf("Mail Matt: Set ASY_SERBUFSIZE>%d\r\n", asy.asy_serLength);
      asy.asy_serLength = (ASY_SERBUFSIZE-1);
    }

    if (asy.asy_serLength > 0) /* If there was some more data.. */
      /* Read as much as we can into the serial buffer */
      p_read(asy.asy_serH, &asy.asy_serBuffer[1], asy.asy_serLength);

    /* Don't forget that first byte */
    asy.asy_serLength++;
  }
}


/*******************************************************************************
***
*** Function         : discard_handler()
*** Preconditions    : netif has serial data queued
*** Postconditions   : The data is discarded (nothing is done)
***
*******************************************************************************/

void discard_handler(void)
{
  /* Contemplate navel */
}


/*******************************************************************************
***
*** Function         : asy_sethandler()
*** Preconditions    : The network interface is in a particular state
***                    The ni_state, ni_dlp, and term ctrlchars flags
***                    indicate which handler this routine will change 
***                    the handler pointer to.
*** Postconditions   : The function pointer asy_serHandler is set to point
***                    to one of the designated serial handler routines, 
***                    depending on the state of the interface, and, if up,
***                    the data link protocol in use. Returns TRUE or FALSE
***                    depending on OK/fail.
***
*******************************************************************************/

GLDEF_C INT asy_sethandler(void)
{
LOCAL_D TEXT fname[]="asy_sethandler: %s\r\n";

  asy.asy_serHandler = discard_handler;
  switch (netif->ni_state) {
    case NIS_DOWN:
      break;
    case NIS_UP:
      switch (netif->ni_dlp) {
        case NIDLP_LOCAL:
          break;
        case NIDLP_SLIP:
          asy.asy_serHandler = slip_inprocess;
          break;
        case NIDLP_PPP:
          asy.asy_serHandler = ppp_inprocess;
          break;
        case NIDLP_CSLIP:
          asy.asy_serHandler = cslip_inprocess;
          break;
        default:
          console_printf(fname, (UINT)"unknown link protocol");
          return FALSE;
      }
      break;
    case NIS_TERM:
      if (flags.misc_flag & MISC_FLAG_TERMCTRLCHARS)
        asy.asy_serHandler = term_expanded_handler;
      else
        asy.asy_serHandler = term_raw_handler;
      break;
    case NIS_SCRIPT:
      asy.asy_serHandler = script_handler;
      break;
    case NIS_PPP_UP:
      asy.asy_serHandler = ppp_inprocess;
      break;
    case NIS_PPP_DOWN:
      asy.asy_serHandler = ppp_inprocess;
      break;
    default:
      console_printf(fname, (UINT)"unknown interface state");
      return FALSE;
  }
  return TRUE;
}


