/*******************************************************************************
***
*** Filename         : buffer.c
*** Purpose          : Linked memory buffer ADT
*** Author           : Matt J. Gumbley
*** Created          : 02/02/97
*** Last updated     : 18/08/98
***
********************************************************************************
***
*** Modification Record
***
*******************************************************************************/

#include "stack.h"

/* The internal buffer pool structure, allocated by buf_init() */
static struct buf *buf_pool;

/* How many entries it has */
static int buf_pool_entries;
static int buf_pool_size;

/* Statistics */
static UINT buf_stat_allocs; /* # allocations satisfied */
static UINT buf_stat_fails; /* # allocation failures */
static UINT buf_stat_usage; /* % usage during session */
static ULONG buf_stat_bytes_allocated; /* How many bytes allocated in total */
static UINT buf_stat_largest_bytes_allocated; /* Largest buffer */


/* Here's the hard-coded distribution of buffer sizes, and quantities.
   This could be read from a file, if the need arises. For the moment, this is
   hard-coded, and will be adjusted to optimise memory usage when the stack is
   fully working, and we can analyse its behaviour
   NB: buffers must be at least as large as the largest possible IP datagram 
       header for the ip_process routine to do fast processing of the IP 
       header
*/

struct bufdist bufdist_table[] = {
  { SLIP_MTU, 10 },
  { PPP_MTU, 10 },
  { 8192, 2 },
  /*{ 16384, 2 },*/
  { 0, 0 } /* Sentinel to end the table */
};

/*******************************************************************************
***
*** Function         : buf_init()
*** Preconditions    : bufdist_table is a buffer size distribution structure.
*** Postconditions   : buf_init is TRUE, the buffer pool is initialised, 
***                    memory allocated and statistics about its size recorded.
***                    buf_init is FALSE, something failed.
***
*******************************************************************************/

int buf_init()
{
struct bufdist *bd;
struct buf *b;
int i;
int j;

  /* Scan through the bufdist_table, calculating how many entries to allocate */
  buf_pool_entries = buf_pool_size = 0;
  buf_stat_bytes_allocated = 0L;
  buf_stat_largest_bytes_allocated = 0;


  for (bd = bufdist_table; bd->size != 0; bd++) {
    buf_pool_entries += bd->amount;
    buf_pool_size += bd->amount * bd->size;
  }

  /* And allocate the table */
  buf_pool = (struct buf *)p_alloc( buf_pool_entries * sizeof(struct buf) );
#ifdef ALLCHK
  p_allchk(1);
#endif
  if (buf_pool != NULL)
    p_bfil(buf_pool, buf_pool_entries * sizeof(struct buf), 0);
  else
    return FALSE;

  /* Allocate buffers of each size */
  for (j = 0, b = buf_pool, bd = bufdist_table; bd->size != 0; bd++) {
    /* Allocate buffers for this size */
    for (i = 0; i < bd->amount; i++,b++,j++) {
      b->data = (byte *)p_alloc(bd->size);
#ifdef ALLCHK
      p_allchk(2);
#endif
      if (b->data == NULL)
        goto buf_init_failed;
      b->allocsize = bd->size;
    }
  }

  /* All allocated! O happy day! */
  return TRUE;

  /* An allocation failed.
     OK, if you want to replace this goto/label, be my guest. 
     This works.. 
  */

buf_init_failed:
  /* Free all allocated buffers */
  for (i = 0, b = buf_pool; i < buf_pool_entries; i++, b++)
    if (b->data) {
      p_free(b->data);
#ifdef ALLCHK
      p_allchk(3);
#endif
    }
  p_free(buf_pool);
#ifdef ALLCHK
  p_allchk(4);
#endif
  return FALSE;
}


/*******************************************************************************
***
*** Function         : buf_alloc(size)
*** Preconditions    : The buffer pool is initialised
*** Postconditions   : buf_alloc is non-NULL, and a buffer has been allocated
***                    of the required size. Either:
***                    1) A free buffer was found directly
***                    2) No free buffer was found directly, and a buffer had 
***                       to be created by chaining one or more smaller ones
***                       together
***                    The buffer header is initialised.
***                    buf_alloc is NULL, and no buffer space is available.
***
*******************************************************************************/


struct buf *buf_alloc(int size)
{
struct buf *b;
int i;
  buf_stat_bytes_allocated += (long) size;
  if (size > buf_stat_largest_bytes_allocated)
    buf_stat_largest_bytes_allocated = size;

  /* Scan through the buffer pool, looking for an available buffer of a usable
     size. */
  for (i = 0, b = buf_pool; i < buf_pool_entries; i++, b++) 
    if (b->allocsize >= size && !(b->flags & BUF_INUSE) ) {
      buf_stat_allocs++;
      /* Clear down this buffer for storage */
      b->flags |= BUF_INUSE;
      b->usedsize = 0; /********** YOU NEED TO FILL THIS IN ***********/
      b->curr = b;
      return b;
    }

  /* There just isn't enough memory. If this is a big problem, we should alter
     the number of buffers in the distribution table. */
  buf_stat_fails++;
  return NULL;
}


/*******************************************************************************
***
*** Function         : buf_free(buf)
*** Preconditions    : The buffer has been created by buf_alloc, and is the 
***                    first buffer in a chain, or a single buffer.
*** Postconditions   : It, along with any others in its chain is returned to the
***                    buffer pool. 
*** Note             : No checking is done to ensure that the buffer is the
***                    first one in the chain. If you want to release buffers
***                    half-way through a chain, use buf_shrink, which 
***                    terminates the end of the chain properly.
***
*******************************************************************************/

void buf_free(struct buf *b)
{
  while (b != NULL) {
    b->usedsize = 0;
    b->curr = b;
    b->flags &= ~BUF_INUSE;
    b = b->next;
  }
}


/*******************************************************************************
***
*** Function         : buf_chain(b1, b2)
*** Preconditions    : The two buffers have been created by buf_alloc
*** Postconditions   : They are chained together. If b1 is already chained to
***                    a buffer, b2 is inserted between b1 and that buffer.
***
*******************************************************************************/

void buf_chain(struct buf *b1, struct buf *b2)
{
  b2->next = b1->next; /* Could be NULL or otherwise */
  b1->next = b2;
}


/*******************************************************************************
***
*** Function         : buf_copy(b1, offset, len, b2)
*** Preconditions    : b1 and b2 have been created by buf_alloc, and the
***                    region b1:(offset, offset+len) lies wholly within b1
*** Postconditions   : len bytes from position offset in buffer chain b1 is
***                    copied to the start of buffer b2

           I may not have to do this routine, if my idea for fast
           IP datagram fragmentation works... See Comer v2, ch7...
           do I absolutely have to copy each fragment to a new buffer,
           or can I simply have a header buffer, which I modify for each
           fragment, sending this, then the fragment data? I suppose
           having a copy, with fragments being queued for transmission
           *does* decouple the IP and data link layer, but it also makes
           for more memory moving, etc. 

           ... probably have to implement it after all, but hold ^^ that
               thought ...

***
*******************************************************************************/

void buf_copy(struct buf *b1, int offset, int len, struct buf *b2)
{
int b1offset=0;
int b1start;
struct buf *origb2;

  /* Find the buffer in the chain b1 which holds position 'offset' */
  while (offset >= (b1offset+b1->usedsize)) {
    /* b1offset is the offset in the chain of the first byte of the buffer b1 */
    b1offset += b1->usedsize;
    b1 = b1->next;
    if (!b1)
      fatal("buf_copy: region outside chain (initial)");
  }

  /* b1 points to the buffer holding position 'offset' and b1offset is the 
     offset in the chain of the first byte of the buffer b1. (Thinking of
     the chain of buffer data as an array for a moment).

     Assert: offset < b1offset+b1->usedsize by above while condition

     The region to copy starts inside buffer b1 at pos offset-b1offset.
  */
  b1start = offset-b1offset; /* Position of 'offset' in b1 */

  /* Initialise b2, so the add byte algorithm below can do its magic */
  b2->usedsize = 0;
  origb2 = b2;

  /* Copy from b1start to the end of this buffer, adding the bytes to b2.
     After the first iteration of this loop, this generalises from the
     start of the buffer to its end, or the end of the data to be copied.
  */


  while (len > 0) {
    /* FIXME: this for loop copy could be optimised, using memcpy()...
              this would have to detect end of source and destination buffer,
              and the call to memcpy() would probably be more expensive than
              it's worth, but hey, I could be proved wrong :)
              It's still faster than calling buf_addbyte...
    */
    for (; (len > 0) && (b1start < b1->usedsize); len--, b1start++) {
      if (b2->usedsize == b2->allocsize) {
        b2 = b2->next;
        if (!b2)
          fatal("buf_copy: unexpected end of destn chain");
        b2->usedsize = 0;
        origb2->curr = b2;
      }
      b2->data[b2->usedsize++] = b1->data[b1start];
    }

    /* Now add the next buffer's worth of data in the chain b1 to chain b2 */
    if (len > 0) {
      b1 = b1->next;
      if (!b1)
        fatal("buf_copy: region outside chain (in copy)");
      b1start = 0;
    }
  }
}


/*******************************************************************************
***
*** Function         : buf_insert(b1, b2)
*** Preconditions    : b1 and b2 have both been created by buf_alloc, b1 may be
***                    a single buffer or a chain, b2 is a single buffer
*** Postconditions   : b2 is inserted in the correct position in the chain b1.
***                    Ordering is done by the ID number of the buffer. This is
***                    used for IP datagram fragment reassembly, where the ID
***                    number is the fragment offset in the datagram.
***
*******************************************************************************/

void buf_insert(struct buf *b1, struct buf *b2)
{
}


/*******************************************************************************
***
*** Function         : buf_destroy()
*** Preconditions    : The buffer pool has been initialised, and some buffers
***                    may be allocated.
*** Postconditions   : The buffer pool, and any allocated buffers, are returned
***                    to the OS.
***
*******************************************************************************/

void buf_destroy(void)
{
struct buf *b;
int i;

  if (buf_pool) {
    /* Free all allocated buffers */
    for (i = 0, b = buf_pool; i < buf_pool_entries; i++, b++)
      if (b->data) {
        p_free(b->data);
#ifdef ALLCHK
        p_allchk(6);
#endif
      }
    p_free(buf_pool);
#ifdef ALLCHK
    p_allchk(7);
#endif
  }
}


/*******************************************************************************
***
*** Function         : buf_size()
*** Preconditions    : The buffer system has been initialsied.
*** Postconditions   : The return is the size of the buffer table and buffers 
***                    allocated in bytes.
***
*******************************************************************************/

int buf_size(void)
{
  return buf_pool_size + 
         (buf_pool_entries * sizeof (struct buf));
}


/*******************************************************************************
***
*** Function         : buf_addbyte(buf, b)
*** Preconditions    : buf is a buffer chain. b is a byte.
*** Postconditions   : buf_addbyte is TRUE and b has been appended to the end 
***                    of the buffer chain. buf_addbyte is FALSE, and the buffer
***                    chain is full.
***
*******************************************************************************/

int buf_addbyte(struct buf *buf, byte b)
{
struct buf *origbuf = buf;

  if (buf->usedsize == buf->allocsize) {
    buf = buf->next;
    if (!buf)
      return FALSE;
    buf->usedsize = 0;
    origbuf->curr = buf;
  }

  buf->data[buf->usedsize++] = b;
  return TRUE;
}


/*******************************************************************************
***
*** Function         : buf_shrink(buf, size)
*** Preconditions    : buf is a buffer chain, size is positive.
*** Postconditions   : Any buffers chained in after size bytes have been stored
***                    in the chain are released to the pool. The chain is
***                    terminated properly.
***
*******************************************************************************/

void buf_shrink(struct buf *buf, int size)
{
int bufoffset=0;
struct buf *rest;

  /* Find the buffer in the chain buf which holds position 'size' */
  while (size >= (bufoffset+buf->usedsize)) {
    /* bufoffset is the offset in the chain of the first byte of buffer buf */
    bufoffset += buf->usedsize;
    buf = buf->next;
    if (!buf)
      fatal("buf_shrink: size outside chain");
  }

  /* buf points to the buffer holding position 'size' and bufoffset is the 
     offset in the chain of the first byte of the buffer b1. (Thinking of
     the chain of buffer data as an array for a moment).

     Terminate the chain here, and free the rest of the chain
  */
  rest = buf->next;
  buf->next = NULL;
  if (rest != NULL)
    buf_free(rest);
}


/*******************************************************************************
***
*** Function         : buf_autoshrink(buf)
*** Preconditions    : buf is a buffer chain of some size
*** Postconditions   : Any buffers chained in after size bytes have been stored
***                    in the chain are released to the pool. The chain is
***                    terminated properly. (Length of the resulting buffer is
***                    determined.)
***
*******************************************************************************/

void buf_autoshrink(struct buf *buf)
{
int len;
  len = buf_length(buf);
  console_printf("buf_autoshrink: shrinking %d byte buffer to %d bytes.\r\n",
              buf->allocsize, len);
  buf_shrink(buf, len);
}


/*******************************************************************************
***
*** Function         : buf_length(buf)
*** Preconditions    : buf is a buffer chain.
*** Postconditions   : buf_length is the used length of the whole chain.
***
*******************************************************************************/

int buf_length(struct buf *buf)
{
int buflength = 0;

  while (buf != NULL) {
    if (buf->usedsize == 0)
      break;
    buflength += buf->usedsize;
    buf = buf->next;
  }
  return buflength;
}


/*******************************************************************************
***
*** Function         : buf_reset(buf)
*** Preconditions    : buf is a buffer chain.
*** Postconditions   : All flags and data structure innards are reset so the 
***                    buffer chain can be used for chained storage
***
*******************************************************************************/

void buf_reset(struct buf *buf)
{
  while (buf != NULL) {
    buf->usedsize = 0;
    buf->curr = buf;
    buf = buf->next;
  }
}


/*******************************************************************************
***
*** Function         : buf_hexdump(buf)
*** Preconditions    : buf is a buffer chain 
*** Postconditions   : An offset | hexadecimal | printable ASCII dump of the
***                    designated buffer chain is output to the log device
***
*******************************************************************************/

void buf_hexdump(struct buf *buf)
{ 
char LINE[80];
long offset=0L;
long left;
int i,upto16,x,bufindex;
unsigned char b;
  bufindex = 0;
  left= buf_length(buf);
  while (left>0) {
    for (i=0; i<78; i++)
      LINE[i]=' ';
    LINE[9]=LINE[59]='|';
    LINE[77]='\r';
    LINE[78]='\n';
    LINE[79]='\0';
    p_atos(LINE,"%08lx",offset);
    LINE[8]=' ';
    upto16 = (left>16) ? 16 : (int)left;
    for (x=0; x<upto16; x++) {
      if (bufindex == buf->usedsize) {
        bufindex = 0;
        buf = buf->next;
      }
      /*else
        bufindex++;*/
      b = buf->data[bufindex++];
      LINE[11+(3*x)]=hexdig((b&0xf0)>>4);
      LINE[12+(3*x)]=hexdig(b&0x0f);
      LINE[61+x]=p_isprint((char)b) ? ((char)b) : '.';
    }
    console_print(LINE);
    offset+=16L;
    left-=16L;
  }
}


/*******************************************************************************
***
*** Function         : buf_dup(data, len)
*** Preconditions    : data is some data, len bytes long
*** Postconditions   : buf_dup is NULL, and there isn't a buffer available to
***                    hold a copy of the data.
***                    buf_dup is not NULL and the data is copied into a new
***                    buffer.
***
*******************************************************************************/

struct buf *buf_dup(byte *data, int len)
{
struct buf *b;
  b = buf_alloc(len);
  if (b == NULL)
    return NULL;
#ifdef DEBUG
  console_puts("buf_dup: duplicating...");
#endif
  p_bcpy(b->data, data, len);
  b->usedsize = len;
#ifdef DEBUG
  console_puts("done...");
#endif
  return b;
}


/*******************************************************************************
***
*** Function         : buf_stats()
*** Preconditions    : none
*** Postconditions   : The buffer allocator statistics are output
***
*******************************************************************************/

void buf_stats(void)
{
int i;
struct buf *b;

  console_puts("buf_stats:");

  console_printf("average buffer request size: %ld\r\n", 
                  (UINT)(buf_stat_allocs == 0 ? 0 :  /* BUG? */
                  buf_stat_bytes_allocated / (long) buf_stat_allocs));

  console_printf("largest buffer request size: %ld\r\n",
                  buf_stat_largest_bytes_allocated);

  console_printf("allocation successes: %d\r\n", buf_stat_allocs);

  console_printf("allocation failures:  %d\r\n", buf_stat_fails);

  console_printf("memory usage:         %u%%\r\nBUF  SIZE  USED INUSE\r\n",
                  buf_stat_usage); 
  for (i = 0, b = buf_pool; i < buf_pool_entries; i++, b++) 
    console_printf(" %2d %5d %5d %c\r\n", i+1, b->allocsize, b->usedsize, 
               b->flags & BUF_INUSE ? 'Y' : 'N');
}


int  buf_order(struct buf *b1, struct buf *b2, int x)
{
  console_puts("buf_order: *** THIS ROUTINE NOT WRITTEN YET");
  return NULL;
}

