/*******************************************************************************
***
*** Filename         : buffer.h
*** Purpose          : Linked memory buffer ADT definitions and prototypes
*** Author           : Matt J. Gumbley
*** Created          : 02/02/97
*** Last updated     : 18/08/98
***
********************************************************************************
***
*** Modification Record
***
*******************************************************************************/

#ifndef BUFFER_H
#define BUFFER_H

/* The buffer system in the PSION TCP/IP stack is quite simple. It pre-allocates
   a section of the memory, and organises it into a pool of available chunks.
   The size and number of these chunks is given by buffer distribution info.
   These chunks are usually of a small size, typically the size of protocol
   headers, MTU's, and small multiples of a kilobyte. Buffers may be chained
   together in a singly-linked list to create larger buffers or chains for the
   purposes of data-link layer frame buildup and IP datagram reassembly. Also,
   this mechanism reduces the amount of data copying when progressing a datagram
   through the stack: lower-level protocols can prefix higher-level datagrams
   with their protocol headers by chaining another small buffer to the start of
   an existing chain. 
   For more details, see Comer vol II, sec 3.5, "Buffer Management"
*/


/* Structure describing the size, quantity of each section of the buffer 
   pool's buffers:
*/

struct bufdist {
  int size;                          /* Size of this type of buffer, in bytes */
  int amount;           /* Number of buffers that are pre-allocated for these */
};



/* The individual buffer descriptors, and entries in the buffer pool
   are taken from \sibosdk\include\psistack\sys\buffer.h
*/

/* A note about buffer chaining:
   From the buffer pool, a number of these buffers are chained together using
   their next fields. The usedsize fields of all buffers in such a chain is
   set to zero, and the curr field of each buffer in the chain points to
   itself. The curr field only makes sense in the first buffer in a chain.
   It is used when writing data sequentially into a buffer chain. It points to
   the current buffer, and the usedsize field of this current buffer points to
   the next available space in that buffer. buf_addbyte takes care of all the
   updating of these fields for you. The total amount of data stored in a chain
   is found by the buf_length function. In the case of data-link layer protocol
   buildup buffers, it is customary to allocate a chain of at least the MTU.
   In the case of SLIP, the worst case frame size is twice the MTU. Not all of
   this chain would usually be used, so when the frame has been fully built up,
   superflous buffers can be released into the pool with buf_shrink. There may
   be a need for buf_expand, but I can't see one yet.
*/

/* Function prototypes ********************************************************/

extern int  buf_init(void);
extern struct buf *buf_alloc(int);
extern void buf_free(struct buf *);
extern void buf_chain(struct buf *, struct buf *);
extern void buf_copy(struct buf *, int, int, struct buf *);
extern void buf_insert(struct buf *, struct buf *);
extern void buf_destroy(void);
extern int  buf_size(void);
extern int  buf_addbyte(struct buf *, byte);
extern void buf_shrink(struct buf *, int);
extern void buf_autoshrink(struct buf *);
extern int  buf_length(struct buf *);
extern void buf_reset(struct buf *);
extern void buf_hexdump(struct buf *);
extern struct buf *buf_dup(byte *, int);
extern void buf_stats(void);
extern int  buf_order(struct buf *, struct buf *, int);

#endif /* BUFFER_H */
