/*******************************************************************************
***
*** Filename         : cslip.c
*** Purpose          : Definitions for the CSLIP data-link layer in PsiStack
*** Author           : Matt J. Gumbley
*** Created          : 20/01/97
*** Last updated     : 01/03/98
*** Note             : This is a direct copy of the SLIP code, with slip 
***                    renamed to cslip.
***
********************************************************************************
***
*** Modification Record
***
*******************************************************************************/

#include "stack.h"

/* Some variables to help me tune the CSLIP layer */
struct logiostat cslip_iostat[] = {
  "bytes", 0, 0,
  "frames", 0, 0,
  "largestframe", 0, 0,
  NULL, 0, 0
};
#define cslipstat_bytes cslip_iostat[0]
#define cslipstat_frames cslip_iostat[1]
#define cslipstat_largestframe cslip_iostat[2]


/*******************************************************************************
***
*** Function         : cslip_outprocess takes a frame from netif->ni_devoutq 
***                    and sends it.
*** Preconditions    : netif possibly has frames queued for output.
*** Postconditions   : cslip_outprocess is positive and the first frame in the 
***                    output queue has been sent. The return value is the 
***                    size of the CSLIP-encoded frame sent.
***                    cslip_outprocess is FALSE and either the output queue is 
***                    empty, or an error occurred.
***
*******************************************************************************/

int cslip_outprocess(void)
{
register int i;
register byte *d, s;
register struct buf *rawbuf, *origrawbuf;
INT errcode;
UBYTE ctrl[3];

  /* Encode any queued frame. */
  origrawbuf = rawbuf = queue_dequeue(netif->ni_devoutq);
  if (rawbuf == NULL) 
    return FALSE;

  if (flags.log_flag & LOG_FLAG_FLOWTRACE)
    console_puts("cslip_outprocess: encapsulating datagram");

  /* Now encode */
  d = dls.cs.cs_txbuildup;
  *d++ = SLIP_END;

  /* Encode every buffer in the chain */
  while (rawbuf != NULL) {
    for (i = 0; i < rawbuf->usedsize; i++) {
      switch (s = rawbuf->data[i]) {
        case SLIP_ESC:
          *d++ = SLIP_ESC;
          *d++ = SLIP_ESCESC;
          break;
        case SLIP_END:
          *d++ = SLIP_ESC;
          *d++ = SLIP_ESCEND;
          break;
        default:
          *d++ = s;
          break;
      }
    }
    rawbuf = rawbuf->next;
  }

  *d++ = SLIP_END;
  i = d - dls.cs.cs_txbuildup;
  if (flags.log_flag & LOG_FLAG_LINKDUMP) {
    console_printf("cslip_outprocess: CSLIP-encoded datagram: 0x%04x bytes:\r\n",i);
    hexdump(dls.cs.cs_txbuildup, i);
  }

  /* If we're using hardware handshaking, is CTS active? */
  /* FIXME: This checking of CTS is copied from the other data link code -
            it should be made common later... MJG */
  if (asy.asy_handshake == RTSCTS) {
    ctrl[1]=0x00;
    if (p_iow(asy.asy_serH, P_FCTRL, ctrl)==0 && (!ctrl[0] & P_SRCTRL_CTS)) {
      console_puts("cslip_outprocess: CTS inactive");
      buf_free(origrawbuf);
      nif_close(netif);
      return FALSE;
    }
  }

  errcode=p_write(asy.asy_serH, dls.cs.cs_txbuildup, i);
  if (errcode != 0)
    goto cslipsenderr;

  buf_free(origrawbuf);

  netif->ni_ooctets += i;
  cslipstat_frames.ls_output++;
  cslipstat_bytes.ls_output += i;
  if (i > cslipstat_largestframe.ls_output)
    cslipstat_largestframe.ls_output = i;

  if (flags.log_flag & LOG_FLAG_FLOWTRACE)
    console_puts("cslip_outprocess: done encapsulating datagram");

  return i;

cslipsenderr:
  switch(errcode) {
    case E_FILE_PARITY:
    case E_FILE_FRAME:
    case E_FILE_OVERRUN:
      console_puts("cslip_outprocess: frame error - check baud rate etc.");
      break;
    case E_FILE_LINE:
      console_puts("cslip_outprocess: line was dropped");
      buf_free(origrawbuf);
      nif_close(netif);
      return FALSE;
    default:
      console_printf("cslip_outprocess: write failed err %d\r\n", errcode);
  }
  buf_free(origrawbuf);
  return d - dls.cs.cs_txbuildup;
}


/*******************************************************************************
***
*** Function         : cslip_inprocess decodes a CSLIP frame 
*** Preconditions    : netif is an open network interface, with
***                    asy_serLength bytes of received serial data stored in
***                    asy_serBuffer
*** Postconditions   : Either a frame has been read, and placed on 
***                    netif->ni_ipinq for retrieval and processing by the 
***                    IP layer, or either an empty frame has been built, 
***                    and so can be ignored, or the frame has not been 
***                    fully built.
***                    If we're out of memory to allocate a fresh buffer, 
***                    bring the interface down.
***
*******************************************************************************/

void cslip_inprocess(void)
{
register UWORD len = asy.asy_serLength;
register UBYTE *serbuf = &asy.asy_serBuffer[0];
register UBYTE ch;


  netif->ni_ioctets += len;

  /* Process each byte in the input stream */
  for (; len>0 ; len--,serbuf++) {
    ch = *serbuf;
    switch (ch) {
      case SLIP_END: /* Start of new frame, or end of one */
        if (dls.cs.cs_rcnt == 0) 
          continue; /* Empty frame */

        /* Enqueue it for IP but not if IP's input queue is full */
        if (queue_full(netif->ni_ipinq)) {
          console_puts("cslip_inprocess: IP inq full; discarding");
          netif->ni_idiscard++;
          /* Queue full, so re-use buildup buffer */
          dls.cs.cs_rcnt = 0;
          continue;
        }
        else {
          /* Queue the frame for IP. */
          cslipstat_bytes.ls_input += dls.cs.cs_rcnt;
          cslipstat_frames.ls_input++;
          if (dls.cs.cs_rcnt > cslipstat_largestframe.ls_input)
            cslipstat_largestframe.ls_input = dls.cs.cs_rcnt;

          dls.cs.cs_rxbuildup->usedsize = dls.cs.cs_rcnt;
          queue_enqueue(netif->ni_ipinq, dls.cs.cs_rxbuildup);
          dls.cs.cs_rxbuildup = buf_alloc( SLIP_MTU );
          if (dls.cs.cs_rxbuildup == NULL) {
            console_puts("cslip_inprocess: couldn't alloc new rxbuf");
            nif_close(netif);
            return;
          }
          dls.cs.cs_rcnt = 0;
          dls.cs.cs_escaped = FALSE;
          attn_signal(ATTN_IP);
          if (flags.log_flag & LOG_FLAG_FLOWTRACE)
            console_puts("cslip_inprocess: datagram enqueued");
          continue;
        }

      case SLIP_ESC:
        dls.cs.cs_escaped = TRUE;
        continue;

      default:
        break;
    }

    if (dls.cs.cs_escaped) {
      /* Translate 2-char escape sequence into original char */
      dls.cs.cs_escaped = FALSE;
      switch (ch) {
        case SLIP_ESCESC:
          ch = SLIP_ESC;
          break;
        case SLIP_ESCEND:
          ch = SLIP_END;
          break;
        default:
          netif->ni_ierrors++;
          break;
      }
    }

    /* We reach here with a character for the buildup buffer. Make
       sure there's space for it.
    */
    if (dls.cs.cs_rcnt == SLIP_MTU) {
      console_puts("cslip_inprocess: buildup buffer full; discarding");
      netif->ni_idiscard++;
      dls.cs.cs_rcnt = 0;
    }
    else 
      dls.cs.cs_rxbuildup->data[dls.cs.cs_rcnt++] = ch;
  }
}


/*******************************************************************************
***
*** Function         : cslip_stats
*** Preconditions    : None
*** Postconditions   : The average frame size for send and receive is output.
***
*******************************************************************************/

void cslip_stats(void)
{
  console_puts("cslip_stats:");
  console_printf("avg frame size in octets: rcv: %d snd: %d\r\n",
    cslipstat_frames.ls_input==0 ? 0 : 
                              cslipstat_bytes.ls_input/cslipstat_frames.ls_input,
    cslipstat_frames.ls_output==0 ? 0 : 
                            cslipstat_bytes.ls_output/cslipstat_frames.ls_output);
  console_writeiostats(cslip_iostat);
}
