/*******************************************************************************
***
*** Filename         : icmp.c
*** Purpose          : Internet Control Message Protocol for PsiStack
*** Author           : Matt J. Gumbley
*** Created          : 21/01/97
*** Last updated     : 27/02/98
***
********************************************************************************
***
*** Modification Record
***
*******************************************************************************/

#include "stack.h"

struct ping_info ping_info = { {127,0,0,1}, 64, 1 };
word16 icmp_ident = 0;
word16 icmp_sequence = 0;

/* Some stats about the ICMP layer */
#define	ICMP_TYPES        19
struct logstat icmp_stat[] = {
  "checksum", 0,                                      /* ICMP Checksum errors */
  "noloop", 0,                              /* No ICMP in response to an ICMP */
  "bdcsts", 0,                                      /* Ignore broadcast ICMPs */
  "unsupp", 0,                  /* # ICMP frames received we don't handle yet */
  "weirdoes", 0,                             /* # frames with odd frame types */
  "outerrors", 0,                                    /* # errors on ICMP send */
  NULL, 0
};
#define icmpstat_checksum icmp_stat[0].ls_value
#define icmpstat_noloop icmp_stat[1].ls_value
#define icmpstat_bdcsts icmp_stat[2].ls_value
#define icmpstat_unsupp icmp_stat[3].ls_value
#define icmpstat_weirdoes icmp_stat[4].ls_value
#define icmpstat_outerrors icmp_stat[5].ls_value

struct logiostat icmp_iostat[] = {
  "echo reply", 0, 0,
  "type 1", 0, 0,
  "type 2", 0, 0,
  "dest unreachable", 0, 0,
  "source quench", 0, 0,
  "redirect", 0, 0,
  "type 6", 0, 0,
  "type 7", 0, 0,
  "echo request", 0, 0,
  "type 9", 0, 0,
  "type 10", 0, 0,
  "ttl exceeded", 0, 0,
  "parameter problem", 0, 0,
  "timestamp request", 0, 0,
  "timestamp reply", 0, 0,
  "information request", 0, 0,
  "information reply", 0, 0,
  "mask request", 0, 0,
  "mask reply", 0, 0,
  NULL, 0, 0
};
#define icmpstat_input(x) icmp_iostat[x].ls_input
#define icmpstat_output(x) icmp_iostat[x].ls_output
#define protonames(x) icmp_iostat[x].ls_name


static char *desturs[]={
  "Network unreachable",
  "Host unreachable",
  "Protocol unreachable",
  "Port unreachable",
  "Fragmentation needed and DF set",
  "Source route failed",
  "Destination network unknown",
  "Destination host unknown",
  "Source host isolated",
  "Dest. network comms admin. prohibited",
  "Dest. host comms admin. prohibited",
  "Network unreachable for type of service",
  "Host unreachable for type of service"
};


/*******************************************************************************
***
*** Function         : icmp_setsrc
*** Preconditions    : 
*** Postconditions   : The source address of the ICMP packet is set.
***
*******************************************************************************/

void icmp_setsrc(struct ip *pip)
{
int i;

  for (i=0; i<NI_MAXINTERFACES; ++i) {
    if (i == NI_LOCAL)
      continue;
    if (net_match(pip->ip_dst,nif[i].ni_ip,nif[i].ni_mask,FALSE))
      break;
  }
  if (i == NI_MAXINTERFACES)
    p_bcpy(pip->ip_src, ip_anyaddr, IP_ALEN);
  else
    p_bcpy(pip->ip_src, nif[i].ni_ip, IP_ALEN);
}


/*******************************************************************************
***
*** Function         : icmp_setdata
*** Preconditions    : 
*** Postconditions   : Sets the data section of the ICMP packet.
***                    Return value is the data length.
***
*******************************************************************************/

/* ECHOMAX must be an even number */
#define ECHOMAX(pip) (IP_MAXSIZE-ICMP_HLEN-IP_HLEN(pip))
#define isodd(x) (x & 0x01)

int icmp_setdata(int type, struct ip *pip, char *pa2)
{
struct icmp *pic = (struct icmp *)pip->ip_data;
int    i, len;
byte   *pdata;
ULONG  *ptime;

  switch (type) {
    case ICT_ECHORP:
      len = pip->ip_len - IP_HLEN(pip) - ICMP_HLEN;
      if (isodd(len))
        pic->icmp_data[len] = 0; /* so cksum works */
      return len;
    case ICT_DESTUR:
    case ICT_SRCQ:
    case ICT_TIMEX:
      pic->icmp_mbz = 0;    /* must be 0 */
      break;
    case ICT_REDIRECT:
      p_bcpy(pic->icmp_gw, pa2, IP_ALEN);
      break;
    case ICT_PARAMP:
      pic->icmp_ptr = (char) pa2;
      for (i=0; i<ICMP_PADLEN; ++i)
        pic->icmp_pad[i] = 0;
      break;
    case ICT_MASKRP:
      p_bcpy(pic->icmp_data, pa2, IP_ALEN);
      break;
    case ICT_ECHORQ:
      if ((int)pa2 > ECHOMAX(pip))
        pa2 = (char *)ECHOMAX(pip);
      /* Initialise ICMP payload */
      for (i=0,pdata=pic->icmp_data; i<(int)pa2; i++,pdata++)
        *pdata = (byte)((i % 0x100) & 0x00ff);
      ptime=(ULONG *)pic->icmp_data;
      if (machine_canreturntickcount) 
        *ptime = p_returntickcount();
      else
        *ptime = timer_ticks;
      if (isodd((int)pa2))
        pic->icmp_data[(int)pa2] = 0;
      return (int)pa2;
    case ICT_MASKRQ:
      p_bcpy(pic->icmp_data, ip_anyaddr, IP_ALEN);
      return IP_ALEN;
  }
  return 0;
}


/*******************************************************************************
***
*** Function         : icmp_process
*** Preconditions    : netif is the open network interface, b is a buffer chain 
***                    holding an an ICMP packet
*** Postconditions   : TRUE and it's been processed. FALSE and there was a
***                    checksum error.
***
*******************************************************************************/

void icmp_process(struct buf *b)
{
struct ip  *pip;
struct icmp *pic;
int    len;
char   timebuf[40];
char   textbuf[200];
ULONG diff, *ptime;

  pip = (struct ip *)b->data;
  pic = (struct icmp *) pip->ip_data;

  len = pip->ip_len - IP_HLEN(pip);
  if (cksum((word16 *)pic, len>>1)) {
    icmpstat_checksum++;
    buf_free(b);
    return;
  }

  if (pic->icmp_type >= 0 && pic->icmp_type <= ICMP_TYPES)
    icmpstat_input(pic->icmp_type)++;

  if (flags.log_flag & LOG_FLAG_FLOWTRACE) {
    console_puts("icmp_process: datagram received");
    if (flags.log_flag & LOG_FLAG_ICMPDUMP) {
      console_puts("icmp_process: dump");
      icmp_dump(b);
    }
  }

  switch(pic->icmp_type) {

    case ICT_ECHORQ:
      if (flags.log_flag & LOG_FLAG_FLOWTRACE)
        console_puts("icmp_process: ping.. pong");
      icmp_send(ICT_ECHORP, 0, pip->ip_src, b, (char *)0);
      return;

    case ICT_MASKRQ:
      if (flags.log_flag & LOG_FLAG_FLOWTRACE)
        console_puts("icmp_process: mask request");
      buf_free(b); /* We're not a gateway */
      icmpstat_unsupp++;
      return;

    case ICT_MASKRP:
      if (flags.log_flag & LOG_FLAG_FLOWTRACE)
        console_puts("icmp_process: mask reply");
      buf_free(b); /* We use 'static configuration information' for our
                      address masks. RFC 1122, Sec. 3.2.2.9 */
      icmpstat_unsupp++;
      return;

    case ICT_ECHORP:
      /* Ping on the Psion would have generated an ICT_ECHORQ. This is the 
         reply, so send it on to the Psion "Ping application".... */
      if (flags.log_flag & LOG_FLAG_FLOWTRACE)
        console_puts("icmp_process: echo reply");

      /* Calculate time difference */
      ptime=(ULONG *)pic->icmp_data;
      if (machine_canreturntickcount) {
        diff = p_returntickcount() - (*ptime); /* In 1/32ths of a second */
        if (diff>32)
          p_atos(timebuf, "%ds, %dms (%d 32ths/sec)", 
                 (int)(diff/32L), (int)((diff%32L)*32), (int)diff);
        else
          p_atos(timebuf, "%dms (%d 32ths/sec)", (int)(diff*32), (int)diff);
      }
      else {
        /* BUG: timer_ticks is updated every half-second */
        diff = timer_ticks - (*ptime); /* In 1/10ths of a second */
        if (diff>10)
          p_atos(timebuf, "%ds, %dms (%d 10ths/sec)", 
                 (int)(diff/10L), (int)((diff%10L)*10), (int)diff);
        else
          p_atos(timebuf, "%dms (%d 10ths/sec)", (int)(diff*10), (int)diff);
      }
      /* A bit of a kludge. We should probably have a PING event... */
      p_atos(textbuf, "ping: %d bytes %s seq=%d id=%d ttl=%d time=%s\r\n", 
               pip->ip_len, szConvertIPaddr(pip->ip_src), 
               net2hs(pic->icmp_seq), net2hs(pic->icmp_id), 
               (int)pip->ip_ttl, timebuf);
      msg_event(EVT_UI_TERMDATA, textbuf, p_slen(textbuf));

      /* Send next ping. The attn_process() routine will pick up on this, and
         send it, unless ping_info->p_numpings >= 0 */
      if (ping_info.p_numpings>0)
        attn_signal(ATTN_PING);
      buf_free(b);
      return;

    case ICT_REDIRECT:
      console_printf("icmp_process: redirect %s",
                 (UINT) szConvertIPaddr(pic->icmp_gw));
      buf_free(b);
      icmpstat_unsupp++;
      return;

    case ICT_DESTUR: 
      if (flags.log_flag & LOG_FLAG_FLOWTRACE)
        console_puts("icmp_process: destination unreachable");
      buf_free(b);
      icmpstat_unsupp++;
      return;

    case ICT_SRCQ:
      if (flags.log_flag & LOG_FLAG_FLOWTRACE)
        console_puts("icmp_process: source quench");
      buf_free(b);
      icmpstat_unsupp++;
      return;

    case ICT_TIMEX:
      if (flags.log_flag & LOG_FLAG_FLOWTRACE)
        console_puts("icmp_process: time exceeded");
      buf_free(b);
      icmpstat_unsupp++;
      return;

    case ICT_PARAMP:
      if (flags.log_flag & LOG_FLAG_FLOWTRACE)
        console_puts("icmp_process: parameter problem");
      buf_free(b);
      icmpstat_unsupp++;
      return;

    case ICT_TIMERQ:
      if (flags.log_flag & LOG_FLAG_FLOWTRACE)
        console_puts("icmp_process: time request");
      buf_free(b);
      icmpstat_unsupp++;
      return;

    case ICT_TIMERP:
      if (flags.log_flag & LOG_FLAG_FLOWTRACE)
        console_puts("icmp_process: time reply (eh?)");
      buf_free(b);
      icmpstat_unsupp++;
      return;

    default:
      if (flags.log_flag & LOG_FLAG_FLOWTRACE)
        console_printf("icmp_process: weird frame type 0x%02x\r\n", pic->icmp_type);
      icmpstat_weirdoes++;
      buf_free(b);
      return;
  }
  /* Should really never get here */
  console_puts("icmp_process: assertion failed - mail Matt");
  buf_free(b);
  return;
}


/*******************************************************************************
***
*** Function         : icmp_setbuf
*** Preconditions    : type is an ICMP packet type. pa1 is the old packet.
*** Postconditions   : The ICMP response is returned, or NULL if OOM.
***                    pisresp is true if the packet is a response.
***                    piserr is true if the packet is an error
***
*******************************************************************************/

struct buf *icmp_setbuf(int type, struct buf *pa1, int *pisresp, int *piserr)
{
struct buf *b;
int bufsize;

  *pisresp = *piserr = FALSE;

  switch (type) {
    case ICT_DESTUR:
    case ICT_SRCQ:
    case ICT_TIMEX:
    case ICT_PARAMP:
      b = pa1;
      *piserr = TRUE;
      break;
    case ICT_ECHORP:
    case ICT_INFORP:
    case ICT_MASKRP:
      b = pa1;
      *pisresp = TRUE;
      break;
    case ICT_ECHORQ:
      bufsize = IPMHLEN + ICMP_HLEN + (UINT)(ping_info.p_size & 0xffff);
      b = buf_alloc(bufsize);
      if (b)
        b->usedsize = bufsize;
      else
        console_printf("icmp_setbuf: echo request: out of memory (%d bytes)\r\n", 
                   bufsize);
      break;
    case ICT_REDIRECT:
    case ICT_TIMERQ:
    case ICT_INFORQ:
    case ICT_MASKRQ:
      return NULL; /* We don't do anything with these */
      break;
    case ICT_TIMERP:    /* Not Implemented */
      /* IcmpOutTimestampsReps++; */
      /* icmpstat_outerrors--;   Kludge: we increment above */
      buf_free(pa1);
      return NULL;
  }
  if (*piserr)
    *pisresp = TRUE;
  return b;
}


/*******************************************************************************
***
*** Function         : icmp_errok
*** Preconditions    : 
*** Postconditions   : Returns boolean, indicating whether it is OK to send an
***                    error response.
***
*******************************************************************************/

int icmp_errok(struct buf *buf)
{
struct ip  *pip = (struct ip *)buf->data;
struct icmp *pic = (struct icmp *)pip->ip_data;

  /* don't send errors about error packets... */

  if (pip->ip_proto == IPT_ICMP)
    switch(pic->icmp_type) {
      case ICT_DESTUR:
      case ICT_REDIRECT:
      case ICT_SRCQ:
      case ICT_TIMEX:
      case ICT_PARAMP:
        return FALSE;
      default:
        break;
    }
  /* ...or other than the first of a fragment */

  if (pip->ip_fragoff & IP_FRAGOFF)
    return FALSE;
  /* ...or broadcast packets */

  if (isbrc(pip->ip_dst))
    return FALSE;
  return TRUE;
}


/*******************************************************************************
***
*** Function         : icmp_dump(b)
*** Preconditions    : b is a buffer holding an ICMP frame in NETWORK byte order
***                    The IP datagram header is in HOST byte order.
*** Postconditions   : A dump and analysis of the frame is sent to the log 
***                    device.
***
*******************************************************************************/

void icmp_dump(struct buf *b)
{
struct ip *i;
struct icmp *ic;
int len;
word16 c;
char szbuf[256];
char *cksumok;
byte *seq;


  i = (struct ip *)b->data;
  ic = (struct icmp *)i->ip_data;

  /* Is it duff? */
  len = i->ip_len - IP_HLEN(i);
  if ((c=cksum((word16 *)ic, len >> 1)) != 0) 
    cksumok="BAD";
  else
    cksumok="OK";

  console_printf("icmp_dump: ICMP frame len 0x%04x (offset 0x%04x in dump)\r\n", 
             i->ip_len-IP_HLEN(i), IP_HLEN(i));
  console_printf("ICMP Type: 0x%02x (%s) Code: 0x%02x Checksum: 0x%04x (%s)\r\n",
             ic->icmp_type, protonames(ic->icmp_type), ic->icmp_code, 
             ic->icmp_cksum, cksumok);

  /* Type-specific stuff - remember the ICMP payload is in NETWORK b.o. */

  switch (ic->icmp_type) {

    case ICT_ECHORP:
    case ICT_ECHORQ:
      console_printf("ID: 0x%04x SEQ: 0x%04x +data\r\n", 
                 net2hs(ic->icmp_id), net2hs(ic->icmp_seq));
      break;

    case ICT_DESTUR:
      console_puts(desturs[ic->icmp_code]);
      /* FIXME: might need ip_dump(ic->icmp_data), but ip_dump takes a buf */
      break;

    case ICT_SRCQ:
    case ICT_TIMEX:
      /* FIXME: might need ip_dump(ic->icmp_data), but ip_dump takes a buf */
      break;

    case ICT_REDIRECT:
      p_atos(szbuf,"Gateway: %s\n",szConvertIPaddr(ic->icmp_gw));
      console_print(szbuf);
      /* FIXME: might need ip_dump(ic->icmp_data), but ip_dump takes a buf */
      break;

    case ICT_PARAMP:
      /* FIXME: might need ip_dump(ic->icmp_data), but ip_dump takes a buf */
      break;

    case ICT_TIMERQ:
    case ICT_TIMERP:
      console_printf("ID: 0x%04x SEQ: 0x%04x +data\r\n", 
                 net2hs(ic->icmp_id), net2hs(ic->icmp_seq));
      /* FIXME: need to output timestamps. Adding them to the icmp union
         will mess up other stuff probably. Need to redo the union I guess. */
      break;


    case ICT_INFORP:
    case ICT_INFORQ:
      /* Obsolete */
      break;

    case ICT_MASKRQ:
    case ICT_MASKRP:
      console_printf("ID: 0x%04x SEQ: 0x%04x +data\r\n", 
                 net2hs(ic->icmp_id), net2hs(ic->icmp_seq));
      seq=(((byte*)ic->icmp_seq)+2);
      console_printf("Mask: %s\r\n",
                 (UINT)szConvertIPaddr(seq));
      break;

    default:
      break;
  }
}



/*******************************************************************************
***
*** Function         : icmp_stats()
*** Preconditions    : none
*** Postconditions   : The ICMP stats are displayed
***
*******************************************************************************/

void icmp_stats(void)
{
  console_puts("icmp_stats:");
  console_writestats(icmp_stat);
  console_writeiostats(icmp_iostat);
}


/*******************************************************************************
***
*** Function         : icmp_send(ICTtype, ICCtype, IPaddr dst, struct buf *)
*** Preconditions    : The datagram gets reused, but is in NETWORK byte order.
***                    ICType implies:
***                    ICT_REDIRECT - pa2 == gateway address
***                    ICT_PARAMP - pa2 == (packet) pointer to parameter error
***                    ICT_MASKRP - pa2 == mask address
***                    ICT_ECHORQ - pa1 == unused, pa2 == data size
***                    ICT_ECHORP - pa1 == buf to reuse, pa2 == 0
***                    ICT_DESTUR - pa1 == src addr, pa2 == buf to reuse
*** Postconditions   : An ICMP datagram of the type specified is sent, reusing
***                    the buffer. See IP.C for some usage of this.
***                    Returns TRUE/FALSE depending on whether the datagram is 
***                    send.
*** Humour           : You are lost in a maze of dark, twisty passages...
***
*******************************************************************************/

int icmp_send(byte type, byte code, IPaddr dst, struct buf *pa1, char *pa2)
{
struct buf *buf;
struct ip  *pip;
struct icmp *pic;
int    isresp, iserr;
IPaddr tdst;
int    datalen;

  icmpstat_output(type)++;
  p_bcpy(tdst, dst, IP_ALEN); /* worry free pass by value  */
  if (flags.log_flag & LOG_FLAG_FLOWTRACE) {
    console_printf("icmp_send: sending %s code %d\r\n", 
               (UINT)protonames(type), code);
  }

  buf = icmp_setbuf(type, pa1, &isresp, &iserr);
  if (buf == NULL) {
    console_puts("icmp_send: icmp_setbuf was NULL");
    icmpstat_outerrors++;
    return FALSE;
  }
  pip = (struct ip *)buf->data;
  pic = (struct icmp *) pip->ip_data;

  datalen = ICMP_HLEN;
  /* we fill in the source here, so routing won't break it */

  if (isresp) {
    if (iserr) {
      if (!icmp_errok(buf)) {
        if (flags.log_flag & LOG_FLAG_FLOWTRACE)
          console_puts("icmp_send: can't send error about error/response");
        buf_free(buf);
        return TRUE;
      }
      /* Fill in the IP header and copy of 8 bytes for the response */
      p_bcpy(pic->icmp_data, pip, IP_HLEN(pip)+8);
      datalen += IP_HLEN(pip)+8;
    }
    icmp_setsrc(pip);
  }
  else
    p_bcpy(pip->ip_src, ip_anyaddr, IP_ALEN);

  p_bcpy(pip->ip_dst, tdst, IP_ALEN);

  pic->icmp_type = type;
  pic->icmp_code = code;
  if (!isresp) {
    icmp_ident++;     /* Need to increment these here */
    icmp_sequence++;  /* since hs2net is a macro :(   */
    pic->icmp_id = hs2net(icmp_ident);
    pic->icmp_seq = hs2net(icmp_sequence);
  }
  datalen += icmp_setdata(type, pip, pa2);
  pic->icmp_cksum = 0;
  pic->icmp_cksum = cksum((word16 *)pic, (datalen+1)>>1);

  pip->ip_proto = IPT_ICMP;  /* for generated packets */
  ip_send(tdst, buf, datalen); 

  return TRUE;
}


/*******************************************************************************
***
*** Function         : ping_process()
*** Preconditions    : A ping attention event was signalled by the main loop,
***                    and so we have to handle it.
*** Postconditions   : A ping may be generated.
***
*******************************************************************************/

void ping_process()
{
  if (flags.log_flag & LOG_FLAG_FLOWTRACE)
    console_printf("ping_process: numpings is %d\r\n", (UINT) ping_info.p_numpings);
  if (ping_info.p_numpings>0) {
    ping_info.p_numpings--;
    icmp_send(ICT_ECHORQ, 0, ping_info.p_ip, 
              (struct buf *)0, (char *)ping_info.p_size);
    /* The next ping will be sent when the reply is received */
  }
}


/*******************************************************************************
***
*** Function         : trace_process()
*** Preconditions    : A trace attention event was signalled by the main loop,
***                    and so we have to handle it.
*** Postconditions   : Traceroute data may be send.
***
*******************************************************************************/

void trace_process()
{
  /* Nothing yet */
}
