/*******************************************************************************
***
*** Filename         : ip.c
*** Purpose          : The IP layer in PsiStack
*** Author           : Matt J. Gumbley
*** Created          : 19/01/97
*** Last updated     : 27/02/98
***
********************************************************************************
***
*** Modification Record
***
*******************************************************************************/

#include "stack.h"

IPaddr ip_maskall   = {255,255,255,255};
IPaddr ip_anyaddr   = {0,0,0,0};
IPaddr ip_loopback  = {127,0,0,1};
IPaddr ip_loopnet   = {127,0,0,0};
IPaddr ip_loopmask  = {255,0,0,0};
IPaddr ip_loopbrc   = {127,255,255,255};

struct ip_fq ip_fq[IPFQ_SIZE];
word16 ip_ident = 0;

/* Some stats about the IP layer */
struct logstat ip_stat[] = {
  "checksum", 0,                                      /* # IP Checksum errors */
  "version", 0,             /* # datagrams received with version other than 4 */
  "badclass", 0,                         /* # class D or E datagrams received */
  "unknetwork", 0,             /* # datagrams not destined for our interfaces */
  "ttlexpiry", 0,           /* # datagrams that didn't get sent 'cos TTL == 0 */
  "fnadf", 0,           /* # datagrams that needed fragmenting but had DF set */
  "options", 0,                       /* # datagrams received with IP options */
  "unkproto", 0,    /* # datagrams rxed by local interface for unk. protocols */
  "fragcreates", 0,                                    /* # fragments created */
  "fragOKs", 0,                                           /* # fragmentations */
  "reassemOKs", 0,                                          /* # reassemblies */
  "reassemfails", 0,                                 /* # reassembly failures */
  "hugereassemfails", 0,                       /* # reassemblies > IP_MAXSIZE */
  "outdiscards", 0,                        /* # datagrams discarded on output */
  "outrequests", 0,                       /* # datagrams requested for output */
  NULL, 0
};
#define ipstat_checksum ip_stat[0].ls_value
#define ipstat_version ip_stat[1].ls_value
#define ipstat_badclass ip_stat[2].ls_value
#define ipstat_unknetwork ip_stat[3].ls_value
#define ipstat_ttlexpiry ip_stat[4].ls_value
#define ipstat_fnadf ip_stat[5].ls_value
#define ipstat_options ip_stat[6].ls_value
#define ipstat_unkproto ip_stat[7].ls_value
#define ipstat_fragcreates ip_stat[8].ls_value
#define ipstat_fragOKs ip_stat[9].ls_value
#define ipstat_reassemOKs ip_stat[10].ls_value
#define ipstat_reassemfails ip_stat[11].ls_value
#define ipstat_hugereassemfails ip_stat[12].ls_value
#define ipstat_outdiscards ip_stat[13].ls_value
#define ipstat_outrequests ip_stat[14].ls_value

/*******************************************************************************
***
*** Function         : ip_options(buf)
*** Preconditions    : Buf holds an IP datagram which may hold IP options.
*** Postconditions   : The options are deleted; we don't process them.
***
*******************************************************************************/

void ip_options(struct buf *b)
{
register struct ip *i = (struct ip *)b->data;
/* char *popt, *popend; Might need these if option processing added */
int len;

  if (IP_HLEN(i) == IPMHLEN)
    return;
  /* popt = i->ip_data;
   * popend = &b->data[IP_HLEN(i)];
   */

  /* NOTE: options not implemented yet */

  /* delete the options */
  len = i->ip_len-IP_HLEN(i); /* data length  */
  if (len)
    p_bcpy(i->ip_data, &b->data[IP_HLEN(i)], len);
  i->ip_len = IPMHLEN + len;
  i->ip_verlen = (i->ip_verlen & 0xf0) | IP_MINHLEN;
}


/*******************************************************************************
***
*** Function         : ip_frag_join(fq)
*** Preconditions    : fq is a fragment queue
*** Postconditions   : Non-NULL is returned, and the fragments in the queue 
***                    are concatenated, if they have all been collected.
***                    NULL is returned if the fragments cannot be joined.
***
*******************************************************************************/

struct buf *ip_frag_join(struct ip_fq *fq)
{
register struct buf *b;
register struct ip *i;
int off, packoff, seq;
struct buf *fullbuf;
struct ip *fullip;

  if (fq->ipfq_state == IPFQS_BOGUS)
    return NULL;

  /* See if we have the whole datagram */

  off = 0;
  b = fq->ipfq_q;
  while (b != NULL) {
    i = (struct ip *)b->data;
    packoff =  (i->ip_fragoff & IP_FRAGOFF)<<3;
    if (off < packoff) /* Is there a missing fragment? */
      return NULL;
    off = packoff + i->ip_len - IP_HLEN(i);
    b = b->next;
  }

  if (off > IP_MAXSIZE) { /* Is this too big for us to handle? */
    buf_free(fq->ipfq_q);
    fq->ipfq_state = IPFQS_FREE;
    ipstat_hugereassemfails++;
    return NULL;
  }

  if ((i->ip_fragoff & IP_MF) == 0) {
    /* Now reconstruct the fragment. See ipfcons, Comer II, p121 */
    fullbuf = buf_alloc(off);
    if (fullbuf == NULL) {
      buf_free(fq->ipfq_q);
      fq->ipfq_state = IPFQS_FREE;
      ipstat_reassemfails++;
      return NULL;
    }

    /* Copy the IP header */
    b = fq->ipfq_q;
    i = (struct ip *)b->data;
    fullip = (struct ip *)fullbuf->data;
    off = IP_HLEN(i);
    seq = 0;
    p_bcpy(fullip, i, off);

    /* Copy the data */
    while (b != NULL) {
      int dlen, doff;
      i = (struct ip *)b->data;
      doff = IP_HLEN(i) + seq - ((i->ip_fragoff & IP_FRAGOFF)<<3);
      dlen = i->ip_len - doff;
      p_bcpy(fullip + off, i + doff, dlen);
      off += dlen;
      seq += dlen;
      b = b->next;
    }

    /* Fix the large packet header */
    fullip = (struct ip *)fullbuf->data;
    fullip->ip_len = off;
    fullip->ip_fragoff = 0;

    /* Release resources */
    buf_free(fq->ipfq_q);
    fq->ipfq_state = IPFQS_FREE;
    ipstat_reassemOKs++;
    return fullbuf;
      
  }

  return NULL;
}


/*******************************************************************************
***
*** Function         : ip_frag_add(fq, b)
*** Preconditions    : fq is a fragment queue table entry, b holds a fragment
*** Postconditions   : Returns TRUE, and the fragment has been added to the
***                    reassembly queue.
***                    Returns FALSE, and the fragment has not been added.
***
*******************************************************************************/

int ip_frag_add(struct ip_fq *fq, struct buf *b)
{
struct ip *i;
int fragoff;

  if (fq->ipfq_state != IPFQS_VALID) {
    buf_free(b);
    return FALSE;
  }
  i = (struct ip *)b->data;
  fragoff = i->ip_fragoff & IP_FRAGOFF;

  if (buf_order(fq->ipfq_q, b, -fragoff) < 0) { /* FIXME: buf_order not written */

/* NB: buf_order will probably use the order field of the buf structure. This is also used to store TCP sequence numbers, and they are long. So, you'll have to extend the third parameter of buf_order to long. */

    /* overflow-- free all frags and drop */
    buf_free(b);
    ipstat_reassemfails++;
    buf_free(fq->ipfq_q); /* probably - will this release an entire ordered chain? */
    fq->ipfq_state = IPFQS_BOGUS;
    return FALSE;
  }
  fq->ipfq_ttl = IPFQ_TTL;    /* restart timer */
  return TRUE;
}


/*******************************************************************************
***
*** Function         : ip_frag_reass(b)
*** Preconditions    : b contains an IP datagram
*** Postconditions   : Returns non-null, and points to a complete datagram
***                    iff the fragment has been completely reassembled.
***                    Returns null if the fragment has been added to the
***                    reassembly queue, or if an error occurs. 
***
*******************************************************************************/

struct buf *ip_frag_reass(struct buf *b)
{
int firstfree,n;
struct ip *i;
struct ip_fq *piq;
  i = (struct ip *)b->data;
  if ((i->ip_fragoff & (IP_FRAGOFF | IP_MF)) == 0)
    return b; /* No reassembly required */


  /* FIXME: need buf_order.... */
  console_puts("ip_frag_reass: reassembly isn't quite finished!");
  return NULL;

  ipstat_reassemOKs++;
  firstfree = -1;
  for (n=0; n<IPFQ_SIZE; ++n) {
    piq = &ip_fq[n];

    if (piq->ipfq_state == IPFQS_FREE) {
      if (firstfree == -1)
        firstfree = n;
      continue;
    }
    if (piq->ipfq_id != i->ip_id)
      continue;
    if (p_bcmp(piq->ipfq_src, IP_ALEN, i->ip_src, IP_ALEN) != 0)
      continue;
    /* found a match */
    if (ip_frag_add(piq, b) == 0) 
      return NULL;
    return ip_frag_join(piq);
  }
  /* no match */

  if (firstfree < 0) {
    /* no room-- drop */
    buf_free(b);
    return NULL;
  }
  ip_fq[firstfree].ipfq_q = b;
  if (ip_fq[firstfree].ipfq_q == NULL) {
    buf_free(b);
    return NULL;
  }
  p_bcpy(ip_fq[firstfree].ipfq_src, i->ip_src, IP_ALEN);
  ip_fq[firstfree].ipfq_id = i->ip_id;
  ip_fq[firstfree].ipfq_ttl = IPFQ_TTL;
  ip_fq[firstfree].ipfq_state = IPFQS_VALID;
  ip_frag_add(&ip_fq[firstfree], b);
  return NULL;
}


/*******************************************************************************
***
*** Function         : ip_h2net(i)
*** Preconditions    : i is an IP datagram with header in host byte order
*** Postconditions   : It is in network byte order
***
*******************************************************************************/

void ip_h2net(struct ip *i)
{
  /* Note: does not convert IP options */
  i->ip_len = hs2net(i->ip_len);
  i->ip_id = hs2net(i->ip_id);
  i->ip_fragoff = hs2net(i->ip_fragoff);
}


/*******************************************************************************
***
*** Function         : ip_net2h(i)
*** Preconditions    : i is an IP datagram with header in network byte order
*** Postconditions   : It is in host byte order
***
*******************************************************************************/

void ip_net2h(struct ip *i)
{
  /* Note: does not convert IP options */
  i->ip_len = net2hs(i->ip_len);
  i->ip_id = net2hs(i->ip_id);
  i->ip_fragoff = net2hs(i->ip_fragoff);
}


/*******************************************************************************
***
*** Function         : ip_send
*** Preconditions    : 
*** Postconditions   : 
***
*******************************************************************************/

void ip_send(IPaddr faddr, struct buf *b, int datalen)
{
struct ip *pip = (struct ip *)b->data;

  pip->ip_verlen = (IP_VERSION<<4) | IP_MINHLEN;
  pip->ip_tos = 0;
  pip->ip_len = datalen+IP_HLEN(pip);
  pip->ip_id = ip_ident++;
  pip->ip_fragoff = 0;
  pip->ip_ttl = IP_INIT_TTL;
  p_bcpy(pip->ip_dst, faddr, IP_ALEN);

  /*
   * Special case for ICMP, so source matches destination
   * with multiple interfaces.
   */
  if (pip->ip_proto != IPT_ICMP)
    p_bcpy(pip->ip_src, ip_anyaddr, IP_ALEN);
  if (!queue_enqueue(nif[NI_LOCAL].ni_ipinq, b)) {
    console_puts("ip_send: local IP in q full");
    buf_free(b);
    ipstat_outdiscards++;
  }
  ipstat_outrequests++;
  attn_signal(ATTN_IP);
}


/*******************************************************************************
***
*** Function         : net_num computes the network portion of a given IP
***                    address.
*** Preconditions    : ipa is an IP address.
*** Postconditions   : net is its network address
***
*******************************************************************************/

void net_num(IPaddr net, IPaddr ipa)
{
int bc = IP_ALEN;

  p_bcpy(net, ipa, IP_ALEN);
  if (IP_CLASSA(net)) bc = 1;
  if (IP_CLASSB(net)) bc = 2;
  if (IP_CLASSC(net)) bc = 3;
  for (; bc < IP_ALEN; bc++)
    net[bc] = 0;
}


/*******************************************************************************
***
*** Function         : brc_num computes the broadcast address of a given IP
***                    address.
*** Preconditions    : ipa is an IP address.
*** Postconditions   : brc is its network address
***
*******************************************************************************/

void brc_num(IPaddr brc, IPaddr ipa)
{
int bc = IP_ALEN;

  net_num(brc, ipa);
  if (IP_CLASSA(brc)) bc = 1;
  if (IP_CLASSB(brc)) bc = 2;
  if (IP_CLASSC(brc)) bc = 3;
  for (; bc < IP_ALEN; bc++)
    brc[bc] = ~0;
}


/*******************************************************************************
***
*** Function         : net_mask computes the network mask of a given net
***                    address.
*** Preconditions    : net is a net address.
*** Postconditions   : mask is its mask
***
*******************************************************************************/

void net_mask(IPaddr mask, IPaddr net)
{
IPaddr netpart;
int isdefault = TRUE;
int i;
int bc = IP_ALEN;
  for (i=0; i<IP_ALEN; i++) {
    mask[i]= ~0;
    isdefault &= net[i] == 0;
  }
  if (isdefault) {
    p_bcpy(mask, net, IP_ALEN);
    return;
  }
  /* check for net match (for subnets) */
  net_num(netpart, net);
  for (i=0; i<NI_MAXINTERFACES; i++) {
    if (nif[i].ni_state != NIS_DOWN)
      if (p_bcmp(nif[i].ni_net, IP_ALEN, netpart, IP_ALEN)==0) {
        p_bcpy(mask, nif[i].ni_mask, IP_ALEN);
        return;
      }
  }
  if (IP_CLASSA(net)) bc = 1;
  if (IP_CLASSB(net)) bc = 2;
  if (IP_CLASSC(net)) bc = 3;
  for (; bc<IP_ALEN; bc++)
    mask[bc] = 0;
}


/*******************************************************************************
***
*** Function         : isbrc determines whether dest is a broadcast address
*** Preconditions    : dest is an IP address.
*** Postconditions   : isbrc is boolean.
***
*******************************************************************************/

int isbrc(IPaddr dest)
{
int inum;
  /* All 0's and 1's are broadcast */
  if (p_bcmp(dest, IP_ALEN, ip_anyaddr, IP_ALEN) ||
      p_bcmp(dest, IP_ALEN, ip_maskall, IP_ALEN))
    return TRUE;
  /* Check real broadcast address and BSD_style for net and subnet */
  for (inum=0; inum < NI_MAXINTERFACES; inum++) 
    if (nif[inum].ni_state == NIS_UP)
      if (p_bcmp(dest, IP_ALEN, nif[inum].ni_brc, IP_ALEN) ||
          p_bcmp(dest, IP_ALEN, nif[inum].ni_nbrc, IP_ALEN) ||
          p_bcmp(dest, IP_ALEN, nif[inum].ni_subnet, IP_ALEN) ||
          p_bcmp(dest, IP_ALEN, nif[inum].ni_net, IP_ALEN))
        return TRUE;
  return FALSE;
}


/*******************************************************************************
***
*** Function         : net_metch(dst, net, mask, islocal)
*** Preconditions    : dst is an address, net is a net address, mask is a net 
***                    mask, islocal is boolean indicating a locally generated 
***                    frame
*** Postconditions   : net_match is boolean: is 'dst' on 'net'?
***
*******************************************************************************/

int net_match(IPaddr dst, IPaddr net, IPaddr mask, int islocal)
{
int i;
  for (i=0; i<IP_ALEN; i++)
    if ((dst[i] & mask[i]) != (net[i] & mask[i]))
      return FALSE;

  /* Local sources should only match unicast addresses (host routes).
     We don't do IP multicast, so ignore class D */
  if (islocal && isbrc(dst))
    return (p_bcmp(mask, IP_ALEN, ip_maskall, IP_ALEN) != 0); 
  /* != 255.255.255.255 */

  return TRUE;
}


/*******************************************************************************
***
*** Function         : which_nif(addr, local)
*** Preconditions    : addr is a destination IP address, islocal is boolean,
***                    indicating whether the datagram is locally generated
***                    (whether or not it came from the higher level protocols
***                    on this machine).
*** Postconditions   : which_nif is -1, and there is no route for this
***                    datagram to take.
***                    which_nif is positive, and gives the interface over 
***                    which to send this datagram.
***
*******************************************************************************/

int which_nif(IPaddr addr, int islocal)
{
int ifnum,gwnum,retifnum;

  /* FIXME: this can very likely be simplified, and is currently broken. */

  if (flags.log_flag & LOG_FLAG_FLOWTRACE) {
    console_printf("which_nif: IP %s %s\r\n", (UINT)szConvertIPaddr(addr),
                   (UINT)(islocal ? "(locally generated) " : ""));
  }

  for (ifnum=0; ifnum < NI_MAXINTERFACES; ifnum++) {
    if (nif[ifnum].ni_state == NIS_UP) {
      if (net_match(addr, nif[ifnum].ni_net, nif[ifnum].ni_mask, islocal)) {
        /* If the datagram has come from upper layers (i.e. the local 
           interface), route it out to a real interface, otherwise, send 
           upwards to the local interface */
        retifnum = islocal ? ifnum : NI_LOCAL;
        if (flags.log_flag & LOG_FLAG_FLOWTRACE) 
          console_printf("net match NI_IFACE%d\r\n", retifnum);
        return retifnum;
      }

      /* forall gateways on this interface,
           if gateway.network and interface.network match
             route over this interface
             return
       */
      for (gwnum = 0; gwnum < NI_GATEWAYS; gwnum++)
        if (net_match(addr, nif[ifnum].ni_gateways[gwnum], 
                      nif[ifnum].ni_mask, islocal)) {
          retifnum = islocal ? ifnum : NI_LOCAL;
          if (flags.log_flag & LOG_FLAG_FLOWTRACE) 
            console_printf("gw match NI_IFACE%d\r\n", retifnum);
          return retifnum;
        }
    }
  }
  /* Addr does not match with any of the interface's network IDs, nor does it
     match with any of the network ID's reachable through any interfaces (the
     so-called gateways). So if the default interface is up, route it 
     thataway */

  if (netif->ni_state == NIS_UP) {
    if (flags.log_flag & LOG_FLAG_FLOWTRACE) 
      console_puts("def match NI_NETWORK");
    return NI_NETWORK;
  }

  if (flags.log_flag & LOG_FLAG_FLOWTRACE) 
    console_puts("no match");
  return -1;
}


/*******************************************************************************
***
*** Function         : ip_dump outputs an interpreted datagram trace to the 
***                    log device
*** Preconditions    : b is a buffer holding an IP datagram 
*** Postconditions   : ip_verlen indicates that this is an IPv4 frame and
***                    a trace is output. Otherwise, nothing.
***
*******************************************************************************/

void ip_dump(struct buf *b)
{
int prevtime;
char sc,dc;
struct ip *i;
word16 fragoff;
TEXT buf[100];
  i = (struct ip *)b->data;
  prevtime = console_timestamp(FALSE);
  console_setprefix(NULL);

  p_atos(&buf[0], "ip_dump: raw frame len buf_length()=0x%04x", buf_length(b));
  console_puts(buf);
  buf_hexdump(b);

  if ((i->ip_verlen & 0xf0) != 0x40) {
    console_puts("ip_dump: not an IPv4 frame!");
    /* Normally n->ni_ierrors++; */
  }
  else {
    p_atos(buf,"IHL: 0x%02x longs (0x%04x bytes) TOS: 0x%02x ID: 0x%04x",
      i->ip_verlen & 0x0f, IP_HLEN(i), i->ip_tos, i->ip_id);
    console_puts(buf);

    p_atos(buf,"Total length: 0x%04x Payload: 0x%04x bytes",
      i->ip_len, i->ip_len - IP_HLEN(i));
    console_puts(buf);

    fragoff = i->ip_fragoff;
    p_atos(buf,"Frag offset: 0x%04x [%c] [%c] TTL: 0x%02x",
      fragoff & IP_FRAGOFF, 
      fragoff & IP_DF ? 'D' : ' ',
      fragoff & IP_MF ? 'M' : ' ', i->ip_ttl);
    console_puts(buf);

    p_atos(buf,"Checksum: 0x%04x Protocol: ",i->ip_cksum);
    console_print(buf);

    switch (i->ip_proto) {
      case IPT_ICMP: 
        console_puts("ICMP"); 
        break;
      case IPT_IGMP:
        console_puts("IGMP");
        break;
      case IPT_GGP:
        console_puts("GGP");
        break;
      case IPT_TCP:  
        console_puts("TCP"); 
        break;
      case IPT_EGP:
        console_puts("EGP");
        break;
      case IPT_IGP:
        console_puts("IGP");
        break;
      case IPT_UDP:  
        console_puts("UDP"); 
        break;
      case IPT_ISOTP4:
        console_puts("ISO TP4");
        break;
      case IPT_OSPFIGP:
        console_puts("OSPF");
        break;
      default:
        p_atos(buf, "Unknown 0x%02x", i->ip_proto); 
        console_puts(buf);
        /* Normally n->ni_iunkproto++; */
        break;
    }
    if (IP_CLASSA(i->ip_src))
      sc='A';
    if (IP_CLASSB(i->ip_src))
      sc='B';
    if (IP_CLASSC(i->ip_src))
      sc='C';
    if (IP_CLASSD(i->ip_src))
      sc='D';
    if (IP_CLASSE(i->ip_src))
      sc='E';
    p_atos(buf,"SRC [%c]: %s ",
      sc,szConvertIPaddr(i->ip_src));
    console_print(buf);
    if (IP_CLASSA(i->ip_dst))
      dc='A';
    if (IP_CLASSB(i->ip_dst))
      dc='B';
    if (IP_CLASSC(i->ip_dst))
      dc='C';
    if (IP_CLASSD(i->ip_dst))
      dc='D';
    if (IP_CLASSE(i->ip_dst))
      dc='E';
    p_atos(buf,"DST [%c]: %s\r\n***",
      dc,szConvertIPaddr(i->ip_dst));
    console_puts(buf);

    /* Detailed protocol dump must be done when the datagram reaches higher 
       layers, and is converted to host byte order by those layers.*/
  }
  console_timestamp(prevtime);
  console_setprefix(szSTACK);
}



/*******************************************************************************
***
*** Function         : ip_fragheader_copy(buffrom, bufto, offindg)
*** Preconditions    : buffrom and bufto are buffers, buffrom contains a 
***                    datagram. offindg is an offset into the datagram.
*** Postconditions   : The IP header and options are copied. The length of the
***                    header are returned.
***
*******************************************************************************/

int ip_fragheader_copy(struct buf *buffrom, struct buf *bufto, int offindg)
{
register struct ip *ipfrom;
word16 i, maxhlen;
byte olen, otype;
int hlen = IPMHLEN;


  ipfrom = (struct ip *)buffrom->data;
  if (offindg == 0) {
    p_bcpy(bufto->data, ipfrom, IP_HLEN(ipfrom));
    return IP_HLEN(ipfrom);
  }

  /* Copy basic header */
  p_bcpy(bufto->data, ipfrom, hlen);

  /* Copy options */
  maxhlen = IP_HLEN(ipfrom);
  i = hlen;
  while (i < maxhlen) {
    otype = buffrom->data[i];
    olen = buffrom->data[++i];
    if (otype & IPO_COPY) {
      p_bcpy(&bufto->data[hlen], &buffrom->data[i-1], olen);
      hlen += olen;
    }
    else if (otype == IPO_NOP || otype == IPO_EOOP) {
      bufto->data[hlen++] = otype;
      olen = 1;
    }
    i += olen-1;

    if (otype == IPO_EOOP)
      break;
  }
  /* Pad to a multiple of 4 octets */
  while (hlen % 4)
    bufto->data[hlen++] = IPO_NOP;
  return hlen;
}


/*******************************************************************************
***
*** Function         : ip_frag_send(n, dst, buf, offset, maxdlen, offindg)
*** Preconditions    : n is a network interface to send the fragment over.
***                    dst is the destination IP address; buf is the whole
***                    (unfragmented) buffer; maxdlen is the maximum size of
***                    a datagram on this network, and offindg is the offset
***                    in the datagram of the start of this fragment.
*** Postconditions   : The return value is TRUE or FALSE depending on whether 
***                    the fragment is sent or not.
***
*******************************************************************************/

int ip_frag_send(struct netif *n, IPaddr dst, struct buf *buf, int offset,
                 int maxdlen, int offindg)
{
struct buf *bufnew;
struct ip *ip, *ipnew;
int hlen, len, newlen;


  ip = (struct ip *)buf->data;

  newlen = maxdlen + IP_HLEN(ip);

  bufnew = buf_alloc(newlen);
  /* bufnew->usedsize is 0 */
  if (bufnew == NULL) {
    console_printf("ip_frag_send: frag. alloc. %d bytes failed\r\n", newlen);
    return FALSE;
  }

  /* Copy the IP header and options */
  hlen = ip_fragheader_copy(buf, bufnew, offindg);
  bufnew->usedsize = hlen;

  ipnew = (struct ip *)bufnew->data;

  ipnew->ip_fragoff = IP_MF | (offindg>>3);
  ipnew->ip_len = len = maxdlen + hlen;
  ipnew->ip_cksum = 0;
  ip_h2net(ipnew);
  ipnew->ip_cksum = cksum((word16 *)ipnew, hlen >> 1);
  p_bcpy(bufnew->data + hlen, buf->data + IP_HLEN(ip) + offset, maxdlen);
  bufnew->usedsize = hlen + maxdlen;
  /* Do something with dst? */
  return nif_send(n, bufnew);
}


/*******************************************************************************
***
*** Function         : ip_put(ifnum, dstaddr, buf)
*** Preconditions    : ifnum is an interface number, dstaddr is the destination
***                    that the datagram should be sent to, and buf is the 
***                    datagram in HOST byte order
*** Postconditions   : The datagram is fragmented if necessary, into fragments
***                    which safely fit under the network MTU, and placed onto
***                    the device output queue, where the data link layer will
***                    pick them up for encoding. ip_put returns TRUE or FALSE
***                    depending on whether the datagram is sent OK.
***
*******************************************************************************/

int ip_put(int ifnum, IPaddr dst, struct buf* buf)
{
register struct netif *n;
register struct ip *i;
int hlen, maxdlen, tosend, offset, offindg;

  n = &nif[ifnum];
  i = (struct ip *)buf->data;
  if (n->ni_state != NIS_UP) {
    console_puts("ip_put: interface is not up");
    buf_free(buf);
    return FALSE;
  }

  /* Send it, or fragment it */
  if (i->ip_len <= n->ni_mtu) {
    i->ip_cksum = 0;
    ip_h2net(i);
    i->ip_cksum = cksum((word16 *)i, IP_HLEN(i) >> 1);
    if (flags.log_flag & LOG_FLAG_FLOWTRACE) {
      console_printf("ip_put: direct output to net interface %d\r\n", ifnum);
    }
    return nif_send(n, buf);
  }

  if (flags.log_flag & LOG_FLAG_FLOWTRACE) {
    console_printf("ip_put: fragmenting: ip_len is %d, MTU is %d\r\n", 
           i->ip_len, n->ni_mtu);
  }

  /* Fragmentation required */
  if (i->ip_fragoff & IP_DF) {
    /* Don't Fragment set. Set MTU in ICMP FNADF args, send ICMP */
    ipstat_fnadf++;
    /* icmp_args = n->ni_mtu; */
    /* FIXME: do the storing of the ICMP args */
    if (flags.log_flag & LOG_FLAG_FLOWTRACE)
      console_puts("ip_put: fnadf");
    /* FIXME: I think there should be an ip_h2net(i) here, but Comer
       disagrees */
    icmp_send(ICT_DESTUR, ICC_FNADF, i->ip_src, buf, (char *)0); /* Reuse buf */
    return FALSE;
  }

  /* Create fragments and send */
  maxdlen = (n->ni_mtu - IP_HLEN(i)) & ~7;
  offset = 0;
  offindg = (i->ip_fragoff & IP_FRAGOFF) << 3;
  tosend = i->ip_len - IP_HLEN(i);

  if (flags.log_flag & LOG_FLAG_IPDUMP) {
    console_puts("ip_put: original datagram buf (before fragmentation):");
    ip_dump(buf);
  }

  /* Divide the datagram into a sequence of maximum-sized fragments plus
     a final fragment of whatever remains. */
  while (tosend > maxdlen) {
    if (!ip_frag_send(n, dst, buf, offset, maxdlen, offindg)) {
      ipstat_outdiscards++;
      buf_free(buf);
      console_puts("ip_put: a fragment send failed");
      return FALSE;
    }
    ipstat_fragcreates++;
    tosend -= maxdlen;
    offset += maxdlen;
    offindg += maxdlen;
  }
  ipstat_fragOKs++;

  /* Send the final fragment - modify the existing buffer */
  ipstat_fragcreates++;
  hlen = ip_fragheader_copy(buf, buf, offindg);
  /*i = buf->data; * why? */
  /* Slide the residual fragment to the start of the buffer (just after the
     header). */
  p_bcpy(&buf->data[hlen], &buf->data[IP_HLEN(i)+offset], tosend);
  buf->usedsize = hlen + tosend;
  /* Keep More Fragments, if this was a fragment to start with */
  i->ip_fragoff = (i->ip_fragoff & IP_MF) | (offindg >> 3);
  i->ip_len = tosend + hlen;
  i->ip_cksum = 0;
  ip_h2net(i);
  i->ip_cksum = cksum((word16 *)i, hlen >> 1);
  /* Supposed to do something with dst here? */
  return nif_send(n, buf);
}



/*******************************************************************************
***
*** Function         : ip_process is the main IP pseudo-process
*** Preconditions    : ATTN_IP is set, and there may be an open network 
***                    interface.
*** Postconditions   : All open network interfaces are processed. If there is 
***                    an incoming datagram, it is converted from network byte
***                    order to host byte order, and dealt with. 
***                    Datagrams may be placed on the device output queue 
***                    for the data link layer to process as it wants to. They
***                    are converted to network byte order before doing so.
***
*******************************************************************************/

void ip_process(void)
{
struct netif *n;
struct buf *inbuf; /*, *dbcbuf;*/
struct ip *i;
int rtin; /*,rtin2;*/
int in;
word16 c;

  /* Deal with every open network interface */
  for (in=0,n=nif; in<NI_MAXINTERFACES; in++,n++)
    if (n->ni_state == NIS_UP) {

      /* Process datagrams coming from the interface. */
      if ((inbuf=queue_dequeue(n->ni_ipinq)) == NULL)
        continue;

      if (flags.log_flag & LOG_FLAG_FLOWTRACE) {
        console_printf("ip_process: datagram dequeued from interface %d\r\n", in);
      }

      /* The following cast makes an assumption: a buf will never be 
         smaller than the maximum-sized IP header */
      i = (struct ip *)inbuf->data;

      /* TODO: Is this datagram shorter than a legal datagram? */
      /* TODO: Is this header too short? */
      
      /* Only bother processing IPv4 datagrams */
      if ((i->ip_verlen & 0xf0) != 0x40) {
        if (flags.log_flag & LOG_FLAG_FLOWTRACE) {
          console_printf("ip_process: not IPv4 datagram (0x%2x)\r\n",
                 i->ip_verlen & 0xf0);
        }
        n->ni_iunkproto++;
        ipstat_version++;
        buf_free(inbuf);
        continue;
      }

      /* Ditch class D (multicast) and E (reserved) datagrams for now... */
      if (IP_CLASSD(i->ip_dst) || IP_CLASSE(i->ip_dst)) {
        if (flags.log_flag & LOG_FLAG_FLOWTRACE) 
          console_puts("ip_process: got class D/E datagram");
        if (flags.log_flag & LOG_FLAG_IPDUMP) {
          ip_net2h(i); /* So the dump is sensible */
          ip_dump(inbuf);
        }
        n->ni_iaddress++;
        ipstat_badclass++;
        buf_free(inbuf);
        continue;
      }

      /* TODO: If we're running low on memory, issue an ICMP source quench */
      /* If this datagram has come from anywhere other than the local
         interface, it needs its checksum examining, and converting to host
         byte order before processing any further. */
      if (in != NI_LOCAL) {
        /* Is the header checksum correct? */
        if ((c=cksum((word16 *)i, IP_HLEN(i) >> 1)) != 0) {
          if (flags.log_flag & LOG_FLAG_FLOWTRACE) {
            console_printf("ip_process: bad cksum 0x%04x should be 0x%04x\r\n",
                   c, i->ip_cksum);
            if (flags.log_flag & LOG_FLAG_IPDUMP) {
              ip_net2h(i); /* So the dump is sensible */
              ip_dump(inbuf);
            }
          }
          n->ni_ierrors++;
          ipstat_checksum++;
          buf_free(inbuf);
          continue;
        }
        ip_net2h(i);
      }

      /* Datagrams are in host byte order after this point */

      if (flags.log_flag & LOG_FLAG_IPDUMP) {
        console_puts("ip_process: dump");
        ip_dump(inbuf);
      }

      /* Is this datagram destined for one of our network interfaces, or
         is it a broadcast datagram? 
         [which should be treated as if it had been addressed to one of our
         interface's IP addresses - RFC 1122, p31]
      */
      rtin = which_nif(i->ip_dst, (in == NI_LOCAL));
      if (flags.log_flag & LOG_FLAG_FLOWTRACE) {
        console_printf("ip_process: which_nif returned %d\r\n", rtin);
      }
      if (rtin == -1) {
        /* This datagram has come in to one of our interfaces, but isn't
           meant for any of them, and it's not a broadcast either. 
           Sorry, can't get there from here... */
        if (flags.log_flag & LOG_FLAG_FLOWTRACE)
          console_puts("ip_process: datagram not for us, and not broadcast");
        n->ni_ierrors++;
        ipstat_unknetwork++;
        ip_h2net(i); 
        icmp_send(ICT_DESTUR, ICC_NETUR, i->ip_src, inbuf, (char *)0);
        continue;
      }

      /* The datagram is destined for us */

      /* TODO: check IP options here. Are we doing anything with these? */
      if ((i->ip_verlen & 0x0f) > IP_MINHLEN) {
        ipstat_options++;
        if (flags.log_flag & LOG_FLAG_FLOWTRACE) {
          console_printf("ip_process: datagram has IP options (len 0x%x)\r\n",
                         i->ip_verlen & 0x0f);
        }
        /* Option processing is unimplemented yet, although RFC 1122 says
           I MUST support at least source routing. This is not normally
           used, and most firewalls would ignore source routed frames for
           security reasons, so I'm not spending time doing this.. 
        */
        buf_free(inbuf);
        continue; /* Assume silent discard */
      }
      

      /* Route the datagram. Fill in Source IP address, if we're the sender */
      if (in == NI_LOCAL) {
        /* This datagram has come from the local machine - has the higher 
           level protocol specified any fixed source address?
        */
        if (p_bcmp(i->ip_src,IP_ALEN,ip_anyaddr,IP_ALEN) == 0) { /* 0.0.0.0 */
          /* No, it hasn't. If this datagram needs to go back to the local 
             machine, set the source to be the same as the destination, 
             otherwise, set the source to be that of the routed interface.
          */
          if (rtin == NI_LOCAL)
            p_bcpy(i->ip_src, i->ip_dst, IP_ALEN);
          else
            p_bcpy(i->ip_src, nif[rtin].ni_ip, IP_ALEN);
        }
      }
      else {
        /* This datagram came from another interface */
        if (flags.log_flag & LOG_FLAG_FLOWTRACE)
          console_puts("ip_process: adjusting checksum");
        /* Adjust header checksum for modified TTL */
        i->ip_cksum += 0x100;
        if ((i->ip_cksum & 0xff00) == 0)
          i->ip_cksum++; 

        /* Decrement TTL and discard if zero, sending an ICMP "Time Exceeded"
           message
        */
        if (--(i->ip_ttl) == 0 && rtin != NI_LOCAL ) {
          n->ni_ierrors++;
          ipstat_ttlexpiry++;
          if (flags.log_flag & LOG_FLAG_FLOWTRACE)
            console_puts("ip_process: time exceeded");
          ip_h2net(i);
          icmp_send(ICT_TIMEX, ICC_TIMEX, i->ip_src, (struct buf *)inbuf, 
                    (char *)0);
          continue;
        }
      }


      /* Process directed broadcasts (i.e. a valid netid, with hostid of 
         all 1's [Comer v1, p62]) : create copy of directed broadcast
         datagrams destined for the local machine, and send a copy to the
         local interface.
      */
      /*************************************
      if (rtin == NI_LOCAL && isbrc(i->ip_dst)) {
      if (flags.log_flag & LOG_FLAG_FLOWTRACE)
        console_puts("ip_process: directed broadcast");
        * Don't need copies of locally generated datagrams or nonbroadcasts *
        rtin2 = which_nif(i->ip_dst, TRUE); * TRUE => Locally generated *
        if (rtin2 != -1 && rtin2 != rtin) {
        if (flags.log_flag & LOG_FLAG_FLOWTRACE)
          console_puts("ip_process: (duplicating)");
          * It's a directed broadcast - make a copy *
          dbcbuf = buf_dup((byte *)i, i->ip_len);
          if (dbcbuf != NULL) {
            if (!ip_put(rtin2, i->ip_dst, dbcbuf)) {
              console_puts("ip_process: directed broadcast copy send failed");
            }
          }
        }
      }*************************************/

      /* Now put this datagram on one of the network interface output
         queues. (It's still in host byte order)
      */
     
      if (flags.log_flag & LOG_FLAG_FLOWTRACE) {
        console_printf("ip_process: datagram for %s -> netif %d output q\r\n",
               (UINT)szConvertIPaddr(i->ip_dst), rtin);
      }
      if (!ip_put(rtin, i->ip_dst, inbuf)) {
        console_puts("ip_process: datagram output failed");
      }

    }
}


/*******************************************************************************
***
*** Function         : ip_stats()
*** Preconditions    : none
*** Postconditions   : The IP stats are output.
***
*******************************************************************************/

void ip_stats(void) 
{
  console_puts("ip_stats:");
  console_writestats(ip_stat);
}

