/*******************************************************************************
***
*** Filename         : local.c
*** Purpose          : Local 'loopback' pseudo-interface. Interface between 
***                    IP and upper layers of PsiStack.
*** Author           : Matt J. Gumbley
*** Created          : 01/03/97
*** Last updated     : 27/02/98
***
********************************************************************************
***
*** Modification Record
***
*******************************************************************************/

#include "stack.h"

/*******************************************************************************
***
*** Function         : local_outprocess is the main local input handler. Output 
***                    is handled by ip_process(). 
*** Preconditions    : locif is the open local interface, possibly with IP 
***                    datagrams on its output queue (which is actually the
***                    input queue for the higher level protocols). 
***                    This routine is called by nif_process in response to
***                    an ATTN_NIF.
*** Postconditions   : Datagrams are sent to the proper upper-level protocols.
***
*******************************************************************************/

VOID local_outprocess(VOID)
{
register struct buf *rawbuf;
register struct ip *i;

  /* Process any pending output */
  rawbuf = queue_dequeue(locif->ni_devoutq);
  if (rawbuf == NULL)
    return;

  i = (struct ip *) rawbuf->data;
  ip_net2h(i);
  locif->ni_ooctets += i->ip_len;

  /* Reassemble this fragment into a datagram, if it is a fragment */
  if (ip_frag_reass(rawbuf) == NULL)
    return;

  /* Handle IP options processing - delete any options present */
  ip_options(rawbuf);

  if (flags.log_flag & LOG_FLAG_FLOWTRACE) {
    console_puts("local_outprocess: sending upwards:");
    if (flags.log_flag & LOG_FLAG_IPDUMP)
      ip_dump(rawbuf);
  }

  /* This datagram either came from higher layers (got looped back by 
     ip_process), or from another interface. It came from one device input 
     queue and got deposited on this one. Anyway, if it was internally 
     generated, we *hope* it's valid. If it came from another interface, 
     it was vetted before being enqueued for processing here.
  */

  /* OK, Who is this IPv4 datagram for? */

  /* Don't forget i and rawbuf are local to this function. It is the 
     responsibility of the higher protocol layers to free this, as I do in the 
     default arm in this switch.
  */
  switch (i->ip_proto) {
    case IPT_ICMP:
      icmp_process(rawbuf);
      break;
    case IPT_TCP:
      tcpi_process(rawbuf);
      break;
    case IPT_UDP:
      udp_process(rawbuf);
      break;
    default:
      locif->ni_iunkproto++; 
      if (flags.log_flag & LOG_FLAG_FLOWTRACE) {
        console_printf("local_outprocess: unknown protocol 0x%2x\r\n",
                       i->ip_proto);
      }
      /* Reuse the buffer */
      icmp_send(ICT_DESTUR, ICC_PROTUR, i->ip_src, 
                (struct buf *)rawbuf, (char *)0);
      break;
  }
}



/*******************************************************************************
***
*** Function         : local_send - should be used instead of nif_send
*** Preconditions    : i is an IP datagram. The IP datagram header is in HOST 
***                    byte order. It is the responsibility of whatever higher 
***                    layer protocol is sending this datagram to convert its
***                    particular payload (TCP segment, UDP message) to NETWORK
***                    byte order, as this layer has no knowledge of upper
***                    layers.
*** Postconditions   : nif[NI_LOCAL]'s ni_ipinq is full, and i is discarded, or
***                    nif[NI_LOCAL]'s ni_ipinq is not full, and i is enqueued 
***
*******************************************************************************/

void local_send(struct buf *i)
{
struct ip *p;

  /* Length of i should be net2hs(i->ip_len) */
  if (!queue_enqueue(locif->ni_ipinq, i)) {
    console_puts("local_send: output q full\n");
    locif->ni_odiscard++;
    return;
  }

  if (flags.log_flag & LOG_FLAG_FLOWTRACE) {
    console_puts("local_send: sending downwards:");
    if (flags.log_flag & LOG_FLAG_IPDUMP)
      ip_dump(i);
  }

  p = (struct ip *)i->data;
  locif->ni_ioctets += p->ip_len;
  attn_signal(ATTN_IP);
}


