/*******************************************************************************
***
*** Filename         : msg.c
*** Purpose          : IPC Messaging for PsiStack
*** Author           : Matt J. Gumbley, Patrick Robbe
*** Created          : 28/07/97
*** Last updated     : 24/08/98
***
********************************************************************************
***
*** Modification Record
***
*******************************************************************************/

#include "stack.h"

#include "..\include\sys\socket.h"
#include "..\include\netdb.h"

/* The format of messages received by the stack */
typedef struct stkrxmsg {
  E_MESSAGE sm_e_message;
  STKMSG    sm_msg;
} STKRXMSG;

#define sm_type   sm_e_message.type                /* Of type SMT_??? below */
#define sm_pid    sm_e_message.pid
#define sm_uwords sm_msg.sm_uwords

GLDEF_D WORD    msg_status; /* When 0 msg_this points to a message to process */
LOCAL_D STKRXMSG *msg_this;

#define MAX_CLIENTS (E_MAX_PROCESSES -4)

/* Clients requesting notification of stack events get their buffer and status 
   word info stored here */
typedef struct stkeventclient {
  UWORD sec_pid;
  UWORD sec_eventmask;
  VOID *sec_pstatusword;
  VOID *sec_pbuffer;
  UWORD sec_length;
  VOID *sec_plenbuffer;
} STKEVENTCLIENT;

LOCAL_D STKEVENTCLIENT secs[MAX_CLIENTS];

UWORD   msg_interestmasks[MAX_CLIENTS];
UWORD   msg_interestpids[MAX_CLIENTS];

struct msgname {
  int params;
  char *name;
};

struct msgname msgnames0_30[] = {
  /*   0 */  0,  "free0",
  /*   1 */  4,  "QUEUE_EVENT_REQUEST",
  /*   2 */  0,  "CANCEL_EVENT_REQUEST",
  /*   3 */  0,  "UNLOAD",
  /*   4 */  0,  "GET_NETIFSTATE",
  /*   5 */  2,  "OPENCLOSE_NETIF",
  /*   6 */  1,  "SET_MISC",
  /*   7 */  0,  "GET_MISC",
  /*   8 */  1,  "SET_NETIF",
  /*   9 */  1,  "GET_NETIF",
  /*  10 */  1,  "SET_ASY",
  /*  11 */  1,  "GET_ASY",
  /*  12 */  1,  "OPENCLOSE_ASY",
  /*  13 */  1,  "SET_SC",
  /*  14 */  1,  "GET_SC",
  /*  15 */  1,  "SET_ISPC",
  /*  16 */  1,  "GET_ISPC",
  /*  17 */  1,  "SET_LOG",
  /*  18 */  0,  "GET_LOG",
  /*  19 */  1,  "GET_PING",
  /*  20 */  1,  "SET_PING",
  /*  21 */  0,  "SEND_PING",
  /*  22 */  1,  "SEND_KEY",
  /*  23 */  0,  "STATS",
  /*  24 */  2,  "SEND_RAWDATA",
  /*  25 */  1,  "GET_UDP_CONFIG",
  /*  26 */  1,  "SET_UDP_CONFIG",
  /*  27 */  1,  "GET_TCP_CONFIG",
  /*  28 */  1,  "SET_TCP_CONFIG",
  /*  29 */  1,  "GET_PPP_CONFIG",
  /*  30 */  1,  "SET_PPP_CONFIG",
};

struct msgname msgnames128_143[] = {
  /* 128 */  4,  "SL_ACCEPT",
  /* 129 */  4,  "SL_BIND",
  /* 130 */  4,  "SL_CONNECT",
  /* 131 */  3,  "SL_LISTEN",
  /* 132 */  7,  "SL_RECVFROM",
  /* 133 */  0,  "SL_RECVMSG",
  /* 134 */  5,  "SL_SEND",
  /* 135 */  0,  "SL_SENDMSG",
  /* 136 */  7,  "SL_SENDTO",
  /* 137 */  3,  "SL_SHUTDOWN",
  /* 138 */  4,  "SL_SOCKET",
  /* 139 */  2,  "SL_CLOSE",
  /* 140 */  6,  "SL_GETSOCKOPT",
  /* 141 */  6,  "SL_SETSOCKOPT",
  /* 142 */  4,  "SL_GETSOCKNAME",
  /* 143 */  4,  "SL_GETPEERNAME",
};

/*******************************************************************************
***
*** Function         : msg_init()
*** Preconditions    : 
*** Postconditions   : The message queue is initialised, and msg_init() is TRUE.
***                    Or, some error occurred, and it's FALSE.
***
*******************************************************************************/

GLDEF_C UINT msg_init()
{
register int i;

  for (i=0; i<MAX_CLIENTS; i++)
    secs[i].sec_eventmask=secs[i].sec_pid=0;

  /* Initialise the message queue */
  switch (p_minit(MAX_CLIENTS, sizeof(STKMSG) )) {
    case E_GEN_NOMEMORY: 
      console_puts("No memory for messages"); 
      return FALSE;
    case E_GEN_NOSEM:
      console_puts("No semaphores for messages");
      return FALSE;
  }

  /* Queue our first asynchronous read */
  msg_queue();
  return TRUE;
}


/*******************************************************************************
***
*** Function         : msg_queue
*** Preconditions    : An asynchronous message receive request is queued
*** Postconditions   : 
***
*******************************************************************************/

GLDEF_C VOID msg_queue()
{
  p_mreceive(&msg_status, &msg_this);
}




/*******************************************************************************
***
*** Function         : msg_process()
*** Preconditions    : A message has been received, and msg_this is pointing to
***                    it.
*** Postconditions   : It is dealt with.
***
*******************************************************************************/

GLDEF_C VOID msg_process()
{
register int i;
register int needsave = FALSE;
register STKEVENTCLIENT *sec;
struct msgname *msgs;
int msgno;
int retval; 

  /* Since another process generated this message, it will be blocked, waiting
   * for us to reply to it. So we can't send any events to that process, or
   * everything will deadlock. So, if any of these routines being called in
   * this switch can generate events, they must queue them up, and then send
   * them after this function has replied to the message with p_mfree.
   */

  /* Assume that it'll complete successfully */
  retval = 0;

  if (flags.log_flag & LOG_FLAG_MSG) {
    if (msg_this->sm_type >= 128) {
      msgs = msgnames128_143;
      msgno = msg_this->sm_type - 128;
    }
    else {
      msgs = msgnames0_30;
      msgno = msg_this->sm_type;
    }
    console_printf("msg_process: SMT_%s( ", (UINT)msgs[msgno].name);
    for (i=0; i<msgs[msgno].params; i++)
      console_printf("0x%04x ", msg_this->sm_uwords[i]);
    console_puts(")");
  }

  switch (msg_this->sm_type) {

    case SMT_QUEUE_EVENT_REQUEST : 
      /* console_printf("registering pid 0x%04x, mask 0x%04x\r\n", msg_this->sm_pid, msg_this->sm_uwords[0]); */
      /* Store this process ID, and its interests */
      for (i=0, sec=secs; i<MAX_CLIENTS; sec++,i++)
        if (sec->sec_pid == 0)
          break;
      if (i == MAX_CLIENTS) /* No more space in the interest table 
                               - shouldn't happen */
        retval = FALSE;
      else {
        sec->sec_pid = msg_this->sm_pid;
        sec->sec_eventmask = msg_this->sm_uwords[0];
        sec->sec_pstatusword = (VOID *)msg_this->sm_uwords[1];
        sec->sec_pbuffer = (VOID *)msg_this->sm_uwords[2];
        sec->sec_length = msg_this->sm_uwords[3];
        sec->sec_plenbuffer = (VOID *)msg_this->sm_uwords[4];
        retval = TRUE;
      }
      break;
      
    case SMT_CANCEL_EVENT_REQUEST :
      for (i=0, sec=secs; i<MAX_CLIENTS; sec++,i++)
        if (sec->sec_pid == msg_this->sm_pid)
          break;
      if (i == MAX_CLIENTS) /* This client was not registered */
        retval = FALSE;
      else {
        sec->sec_pid = sec->sec_eventmask = 0;
        retval = TRUE;
      }
      break;

    case SMT_UNLOAD :
      /* The setup program has closed, and wants us to close as well.
         This will cause the stack to close at the end of the event loop. */
      stack_active = FALSE;
      retval = TRUE;
      break;

    case SMT_GET_NETIFSTATE :
      /* Return the current interface state */
      retval = (UWORD) netif->ni_state;
      break;
      
    case SMT_SET_NETIF :
      /* Set the internal settings of the network interface */
      if (netif->ni_state != NIS_DOWN)
        retval = FALSE;
      else {
        p_pcpyfr(msg_this->sm_pid, (VOID *)msg_this->sm_uwords[0],  netif, 
                 sizeof(struct netif));
        needsave = TRUE;
        retval = TRUE;
      }
      break;

    case SMT_GET_NETIF :
      /* Get the internal settings of the network interface */
      p_pcpyto(msg_this->sm_pid, (VOID *)msg_this->sm_uwords[0], netif, 
               sizeof(struct netif));
      retval = TRUE;
      break;

    case SMT_OPENCLOSE_NETIF :
      /* Open or close the network layer */
      if (msg_this->sm_uwords[0] == TRUE)
        retval = nif_open(msg_this->sm_uwords[1]);
      else {
        retval = TRUE;
        nif_close(netif);
      }
      break;

    case SMT_SET_ASY :
      /* Set the internal settings of the serial interface */
      if (netif->ni_state != NIS_DOWN)
        retval = FALSE;
      else {
        p_pcpyfr(msg_this->sm_pid, (VOID *)msg_this->sm_uwords[0], &asy, 
                 sizeof(struct asy));
        needsave = TRUE;
        retval = TRUE;
      }
      break;

    case SMT_GET_ASY :
      /* Get the internal settings of the serial interface */
      p_pcpyto(msg_this->sm_pid, (VOID *)msg_this->sm_uwords[0], &asy, 
               sizeof(struct asy));
      retval = TRUE;
      break;

    case SMT_OPENCLOSE_ASY :
      /* Open or close the asy layer, or reset serial handler */
      switch (msg_this->sm_uwords[0]) {
        case TRUE:
          retval = asy_open();
          break;
        case FALSE:
          asy_close();
          retval = TRUE;
          break;
        case 0xBEEF: 
          retval = asy_sethandler();
      }
      break;

    case SMT_SET_SC :
      /* Set the internal settings of the script interpreter */
      if (netif->ni_state != NIS_DOWN)
        retval = FALSE;
      else {
        p_pcpyfr(msg_this->sm_pid, (VOID *)msg_this->sm_uwords[0], &sc, 
                 sizeof(struct scriptcontext));
        needsave = TRUE;
        retval = TRUE;
      }
      break;

    case SMT_GET_SC :
      /* Get the internal settings of the script interpreter */
      p_pcpyto(msg_this->sm_pid, (VOID *)msg_this->sm_uwords[0], &sc, 
               sizeof(struct scriptcontext));
      retval = TRUE;
      break;

    case SMT_SET_ISPC :
      /* Set the internal settings of the ISP */
      if (netif->ni_state != NIS_DOWN)
        retval = FALSE;
      else {
        p_pcpyfr(msg_this->sm_pid, (VOID *)msg_this->sm_uwords[0], &ispc, 
                 sizeof(struct ISPcontext));
        needsave = TRUE;
        retval = TRUE;
      }
      break;

    case SMT_GET_ISPC :
      /* Get the internal settings of the ISP */
      p_pcpyto(msg_this->sm_pid, (VOID *)msg_this->sm_uwords[0], &ispc, 
               sizeof(struct ISPcontext));
      retval = TRUE;
      break;

    case SMT_SET_LOG :
      /* Set new logging flags */
      flags.log_flag = msg_this->sm_uwords[0]; /* log flags */
      /* Adjust timestamping */
      console_timestamp(flags.log_flag & LOG_FLAG_TIMESTAMP);
      needsave = TRUE;
      retval = TRUE;
      break;

    case SMT_GET_LOG :
      /* Get the current logging level */
      retval = flags.log_flag;
      break;

    case SMT_SET_MISC :
      /* Set new misc flags */
      flags.misc_flag = msg_this->sm_uwords[0]; /* misc flags */
      needsave = TRUE;
      retval = TRUE;
      break;

    case SMT_GET_MISC :
      /* Get the current misc flags */
      retval = flags.misc_flag;
      break;

    case SMT_SET_PING :
      /* Set the internal ping settings */
      p_pcpyfr(msg_this->sm_pid, (VOID *)msg_this->sm_uwords[0], &ping_info, 
               sizeof(struct ping_info));
      needsave = TRUE;
      retval = TRUE;
      break;

    case SMT_GET_PING :
      /* Get the internal ping settings */
      p_pcpyto(msg_this->sm_pid, (VOID *)msg_this->sm_uwords[0], &ping_info, 
               sizeof(struct ping_info));
      retval = TRUE;
      break;

    case SMT_SEND_PING :
      attn_signal(ATTN_PING);
      retval = TRUE;
      break;

    case SMT_SEND_KEY :
      retval = term_sendkey((byte) (msg_this->sm_uwords[0] & 0x00ff));
      break;

    case SMT_SEND_RAWDATA :
      retval = FALSE;
      break;

    case SMT_STATS :
      stack_stats();
      retval = TRUE;
      break;

    case SMT_SET_UDP_CONFIG :
      /* Set the internal UDP settings */
      if (netif->ni_state != NIS_DOWN)
        retval = FALSE;
      else {
        p_pcpyfr(msg_this->sm_pid, (VOID *)msg_this->sm_uwords[0], &udp_config, 
                 sizeof(struct udp_config));
        needsave = TRUE;
        retval = TRUE;
      }
      break;

    case SMT_GET_UDP_CONFIG :
      /* Get the internal UDP settings */
      p_pcpyto(msg_this->sm_pid, (VOID *)msg_this->sm_uwords[0], &udp_config, 
               sizeof(struct udp_config));
      retval = TRUE;
      break;

    case SMT_SET_TCP_CONFIG :
      /* Set the internal TCP settings */
      if (netif->ni_state != NIS_DOWN)
        retval = FALSE;
      else {
        p_pcpyfr(msg_this->sm_pid, (VOID *)msg_this->sm_uwords[0], &tcp_config, 
                 sizeof(struct tcp_config));
        needsave = TRUE;
        retval = TRUE;
      }
      break;

    case SMT_GET_TCP_CONFIG :
      /* Get the internal TCP settings */
      p_pcpyto(msg_this->sm_pid, (VOID *)msg_this->sm_uwords[0], &tcp_config, 
               sizeof(struct tcp_config));
      retval = TRUE;
      break;

    case SMT_SET_PPP_CONFIG :
      /* Set the internal settings of PPP */
      if (netif->ni_state != NIS_DOWN)
        retval = FALSE;
      else {
        p_pcpyfr(msg_this->sm_pid, (VOID *)msg_this->sm_uwords[0], &ppp_config, 
                 sizeof(struct ppp_config));
        needsave = TRUE;
        retval = TRUE;
      }
      break;

    case SMT_GET_PPP_CONFIG :
      /* Get the internal settings of PPP */
      p_pcpyto(msg_this->sm_pid, (VOID *)msg_this->sm_uwords[0], &ppp_config, 
               sizeof(struct ppp_config));
      retval = TRUE;
      break;

    case SMT_SL_ACCEPT :
      retval = socket_accept(msg_this->sm_pid,           /* client process ID */
                 msg_this->sm_uwords[0],                            /* sockfd */
                 msg_this->sm_uwords[1],               /* client DS *peeraddr */
                 msg_this->sm_uwords[2],               /* client DS *paddrlen */
                 msg_this->sm_uwords[3]);              /* client DS *statword */
      break;

    case SMT_SL_BIND :
      retval = socket_bind(msg_this->sm_pid,             /* client process ID */
                 msg_this->sm_uwords[0],                            /* sockfd */
                 msg_this->sm_uwords[1],                  /* client DS *saddr */
                 msg_this->sm_uwords[2],                /* client DS *addrlen */
                 msg_this->sm_uwords[3]);              /* client DS *statword */
      break;

    case SMT_SL_CLOSE :
      retval = socket_close(msg_this->sm_pid,            /* client process ID */
                 msg_this->sm_uwords[0],                            /* sockfd */
                 msg_this->sm_uwords[1]);              /* client DS *statword */
      break;

    case SMT_SL_CONNECT :
      retval = socket_connect(msg_this->sm_pid,          /* client process ID */
                 msg_this->sm_uwords[0],                            /* sockfd */
                 msg_this->sm_uwords[1],              /* client DS *serv_addr */
                 msg_this->sm_uwords[2],                /* client DS *addrlen */
                 msg_this->sm_uwords[3]);              /* client DS *statword */
      break;

    case SMT_SL_GETPEERNAME :
      retval = socket_getpeername(msg_this->sm_pid,      /* client process ID */
                 msg_this->sm_uwords[0],                            /* sockfd */
                 msg_this->sm_uwords[1],                   /* client DS *addr */
                 msg_this->sm_uwords[2],               /* client DS *paddrlen */
                 msg_this->sm_uwords[3]);              /* client DS *statword */
      break;

    case SMT_SL_GETSOCKOPT :
      retval = socket_getsockopt(msg_this->sm_pid,       /* client process ID */
                 msg_this->sm_uwords[0],                            /* sockfd */
                 msg_this->sm_uwords[1],                             /* level */
                 msg_this->sm_uwords[2],                           /* optname */
                 msg_this->sm_uwords[3],                 /* client DS *optval */
                 msg_this->sm_uwords[4],                 /* client DS *optlen */
                 msg_this->sm_uwords[5]);              /* client DS *statword */
      break;

    case SMT_SL_GETSOCKNAME :
      retval = socket_getsockname(msg_this->sm_pid,      /* client process ID */
                 msg_this->sm_uwords[0],                            /* sockfd */
                 msg_this->sm_uwords[1],                   /* client DS *addr */
                 msg_this->sm_uwords[2],               /* client DS *paddrlen */
                 msg_this->sm_uwords[3]);              /* client DS *statword */
      break;

    case SMT_SL_LISTEN :
      retval = socket_listen(msg_this->sm_pid,           /* client process ID */
                 msg_this->sm_uwords[0],                            /* sockfd */
                 msg_this->sm_uwords[1],                                 /* n */
                 msg_this->sm_uwords[2]);              /* client DS *statword */
      break;

    case SMT_SL_RECVFROM :
      retval = socket_recvfrom(msg_this->sm_pid,         /* client process ID */
                 msg_this->sm_uwords[0],                            /* sockfd */
                 msg_this->sm_uwords[1],                   /* client DS *buff */
                 msg_this->sm_uwords[2],                               /* len */
                 msg_this->sm_uwords[3],                             /* flags */
                 msg_this->sm_uwords[4],                   /* client DS *from */
                 msg_this->sm_uwords[5],                /* client DS *fromlen */
                 msg_this->sm_uwords[6]);              /* client DS *statword */
      break;

    case SMT_SL_RECVMSG :
      break;

    case SMT_SL_SEND :
      retval = socket_send(msg_this->sm_pid,             /* client process ID */
                 msg_this->sm_uwords[0],                            /* sockfd */
                 msg_this->sm_uwords[1],                   /* client DS *buff */
                 msg_this->sm_uwords[2],                               /* len */
                 msg_this->sm_uwords[3],                             /* flags */
                 msg_this->sm_uwords[4]);              /* client DS *statword */
      break;

    case SMT_SL_SENDMSG :
      break;

    case SMT_SL_SENDTO :
      retval = socket_sendto(msg_this->sm_pid,           /* client process ID */
                 msg_this->sm_uwords[0],                            /* sockfd */
                 msg_this->sm_uwords[1],                   /* client DS *buff */
                 msg_this->sm_uwords[2],                               /* len */
                 msg_this->sm_uwords[3],                             /* flags */
                 msg_this->sm_uwords[4],                     /* client DS *to */
                 msg_this->sm_uwords[5],                  /* client DS *tolen */
                 msg_this->sm_uwords[6]);              /* client DS *statword */
      break;

    case SMT_SL_SETSOCKOPT :
      retval = socket_setsockopt(msg_this->sm_pid,       /* client process ID */
                 msg_this->sm_uwords[0],                            /* sockfd */
                 msg_this->sm_uwords[1],                             /* level */
                 msg_this->sm_uwords[2],                           /* optname */
                 msg_this->sm_uwords[3],                 /* client DS *optval */
                 msg_this->sm_uwords[4],                            /* optlen */
                 msg_this->sm_uwords[5]);              /* client DS *statword */
      break;

    case SMT_SL_SHUTDOWN :
      retval = socket_shutdown(msg_this->sm_pid,         /* client process ID */
                 msg_this->sm_uwords[0],                            /* sockfd */
                 msg_this->sm_uwords[1],                               /* how */
                 msg_this->sm_uwords[2]);              /* client DS *statword */
      break;

    case SMT_SL_SOCKET :
      retval = socket_socket(msg_this->sm_pid,           /* client process ID */
                 msg_this->sm_uwords[0],                            /* domain */
                 msg_this->sm_uwords[1],                              /* type */
                 msg_this->sm_uwords[2],                          /* protocol */
                 msg_this->sm_uwords[3]);              /* client DS *statword */
      break;

    default:
      /* The stack kernel can't handle database message (for instance) */
      console_printf("msg_process: bad msg type %d\r\n", msg_this->sm_type);
      break;
  }
  /* Finished with this message */
  p_mfree(msg_this, retval);

  /* Save the config ? */
  if (needsave) {
    config_save();
  }
}


/*******************************************************************************
***
*** Function         : msg_cancel()
*** Preconditions    : An asynchronous read request may be outstanding.
*** Postconditions   : It is cancelled.
***
*******************************************************************************/

GLDEF_C VOID msg_cancel()
{
  p_mcancel();
}



/*******************************************************************************
***
*** Function         : msg_event(event, buffer, length)
*** Preconditions    : A major event has occurred in the stack which must be 
***                    broadcast to all interested parties.. event is an 
***                    EVT_ code, buffer,length is some data. 
*** Postconditions   : msg_event is TRUE, and the event is given to those 
***                    clients that are interested in it. (They'll have to 
***                    queue another request to get more.)
***                    msg_event is FALSE, and something spooky is goin' down.
***
*******************************************************************************/

GLDEF_C UINT msg_event(UWORD event, VOID *buffer, UWORD length)
{
LOCAL_D TEXT fname[]="msg_event: %s\r\n";
register STKEVENTCLIENT *sec;
register int i;
register UWORD pid;
UWORD retlen;

  if (flags.log_flag & LOG_FLAG_FLOWTRACE) 
    switch (event) {
      case EVT_LINEDROPPED:
        console_printf(fname, (UINT)"line dropped");
        break;
      case EVT_SCRIPTING:
        console_printf(fname, (UINT)"starting a script");
        break;
      case EVT_LINK_UP:
        console_printf(fname, (UINT)"starting link-up negotiation");
        break;
      case EVT_NET_UP:
        console_printf(fname, (UINT)"connected to the network");
        break;
      case EVT_LINK_DOWN:
        console_printf(fname, (UINT)"starting link-down negotiation"); 
        break;
      case EVT_NET_DOWN:
        console_printf(fname, (UINT)"disconnected from the network");
        break;
      case EVT_NEW_METHOD:
        console_printf(fname, (UINT)"selected a method for connection");
        break;
      case EVT_UI_MESSAGE:
        console_printf(fname, (UINT)"display this:");
        console_puts(buffer);
        break;
      case EVT_UI_TERMDATA:
        console_printf(fname, (UINT)"term data");
        break;
      case EVT_TERM:
        console_printf(fname, (UINT)"entered terminal mode");
        break;
      default:
        console_printf("msg_event: unknown event 0x%04x\r\n", event);
        return FALSE;
    }

  /* Complete the requests of any clients that are interested in this event */
  for (i=0, sec=secs; i<MAX_CLIENTS; sec++,i++)
    if ((pid=sec->sec_pid) && (sec->sec_eventmask & event)) {
      /* console_printf("msg_event: informing pid 0x%04x event 0x%04x\r\n", pid, event); */
      /* FIXME? Should this bit be done with interrupts disabled to preserve
                atomicity? */

      /* Copy any data to the client buffer (if they specified a buffer) */
      if (buffer && sec->sec_pbuffer) {
        retlen = min(length, sec->sec_length);
        p_pcpyto(pid, sec->sec_pbuffer, buffer, retlen);
      }
      else
        retlen = 0;

      /* Copy the length to the client buffer (if they specified a buffer) */
      if (sec->sec_plenbuffer) 
        p_pcpyto(pid, sec->sec_plenbuffer, &retlen, 2);

      /* Store the event type in the client's status word */
      p_pcpyto(pid, sec->sec_pstatusword, &event, 2);

      /* FIXME: could context switch here. Any other signal for the client 
                will cause it to read its status word, and handle the request. 
                When we get the CPU back, we'll signal the client, and cause 
                a stray signal.
                However, the stack runs at a higher priority than the other 
                clients (hopefully), so the above should never happen. 
                Sure would like to disable/enable interrupts round this bit 
                though, just for my own piece of mind :-) */

      /* Remove it from the secs table */
      sec->sec_pid = sec->sec_eventmask = 0;

      /* Signal the process */
      p_iosignalbypid(pid);

      /* Applications that listen for terminal mode data need to get it on
       * screen FAST. So, if you're a terminal watcher, you'll get switched to.
       * Aren't I kind? */
      if (event & EVT_UI_TERMDATA) /* May be more later */
        p_presume(pid);
    }
  return TRUE;
}


