/*******************************************************************************
***
*** Filename         : nif.c
*** Purpose          : Network Interface Layer
*** Author           : Matt J. Gumbley, Patrick H. Robbe
*** Created          : 19/01/97
*** Last updated     : 06/03/98
***
********************************************************************************
***
*** Modification Record: 
***
*******************************************************************************/

#include "stack.h"

struct netif nif[NI_MAXINTERFACES];
struct netif *locif = &nif[NI_LOCAL];
struct netif *netif = &nif[NI_NETWORK];

/* Save memory - why not! */
static char szFail[] = "Couldn't %s %s %s %s\r\n";
static char szAllocate[] = "allocate";
static char szCreate[] = "create";
static char szScriptInterpreter[] = "script interpreter";
static char szPPP[] = "PPP";
static char szSLIP[] = "SLIP";
static char szCSLIP[] = "CSLIP";
static char szRx[] = "Rx";
static char szTx[] = "Tx";
static char szQueue[] = "queue";
static char szBuffer[] = "buffer";
static char szContext[] = "context";
static char szNull[] = "";


/*******************************************************************************
***
*** Function         : nif_errmsg(verb, word1, word2, word3)
*** Preconditions    : 
*** Postconditions   : 
***
*******************************************************************************/

void nif_errmsg(TEXT *verb, TEXT *word1, TEXT *word2, TEXT *word3)
{
  console_printf(szFail, (UINT)verb, word1, word2, word3);
}


/*******************************************************************************
***
*** Function         : nif_init initialises the network interface structures
*** Preconditions    : None.
*** Postconditions   : nif_init returns TRUE, and the nif[] array is 
***                    initialised with only the local loopback interface 
***                    configured.
***                    nif_init returns FALSE and the queues for the local
***                    loopback interface could not be created, or the ISP 
***                    context could not be created.
*** Note             : Since logging has not been initialised, we must log
***                    straight to the console. No log_* functions.
***
*******************************************************************************/

int nif_init(void)
{
int i;
struct netif *n;

  /* Clear out both interface structures */
  for (i=0, n=nif; i<NI_MAXINTERFACES; i++, n++) {
    p_bfil(n, sizeof(struct netif), 0);
    n->ni_state=NIS_DOWN;
    n->ni_dlp=NIDLP_SLIP; /* SLIP is the default so far */

    /* Allocate two queues */
    if ((n->ni_ipinq=queue_create(IP_INQSZ)) == NULL) {
      nif_errmsg(szCreate, "local IP input", szQueue, szNull);
      goto initfail;
    }

    if ((n->ni_devoutq=queue_create(NI_INQSZ)) == NULL) {
      nif_errmsg(szCreate, "local device output", szQueue, szNull);
      goto initfail;
    }
  } 

  /* Configure the local loopback interface */
  n=&nif[NI_LOCAL];
  n->ni_state=NIS_UP;
  n->ni_dlp=NIDLP_LOCAL;
  p_bcpy(n->ni_ip,ip_loopback,IP_ALEN);
  p_bcpy(n->ni_net,ip_loopnet,IP_ALEN);
  p_bcpy(n->ni_subnet,ip_loopnet,IP_ALEN);
  p_bcpy(n->ni_mask,ip_loopmask,IP_ALEN);
  p_bcpy(n->ni_brc,ip_loopbrc,IP_ALEN);
  p_bcpy(n->ni_nbrc,ip_loopbrc,IP_ALEN);
  n->ni_mtu=3584;            /* This is what Linux uses */
  n->ni_oqlen=NI_INQSZ;
  n->ni_maxreasm=0L; /* FIXME if I ever use this */

  /* Configure the main network interface */
  n=&nif[NI_NETWORK];
  asy_sethandler();

  return TRUE;

initfail:
  for (i=0, n=nif; i<NI_MAXINTERFACES; i++, n++) {
    if (n->ni_ipinq != NULL)
      queue_delete(n->ni_ipinq);
    if (n->ni_devoutq != NULL)
      queue_delete(n->ni_devoutq);
  }
  return FALSE;
}


/*******************************************************************************
***
*** Function         : nif_buildupfree()
*** Preconditions    : The network interface may have data-link layer
***                    buildup buffers allocated.
*** Postconditions   : Those buffers are freed.
***
*******************************************************************************/

void nif_buildupfree(void)
{
  switch (netif->ni_dlp) {
    case NIDLP_SLIP:
      if (dls.ss.ss_txbuildup)
        p_free(dls.ss.ss_txbuildup);
      if (dls.ss.ss_rxbuildup)
        buf_free(dls.ss.ss_rxbuildup);
      dls.ss.ss_rxbuildup = NULL;
      dls.ss.ss_txbuildup = NULL;
      break;
    case NIDLP_PPP:
      if (dls.ps.ps_txbuildup)
        p_free(dls.ps.ps_txbuildup);
      if (dls.ps.ps_rxbuildup)
        buf_free(dls.ps.ps_rxbuildup);
      dls.ps.ps_rxbuildup = NULL;
      dls.ps.ps_txbuildup = NULL;
      if (lcs.lcp_txbuildup)
        p_free(lcs.lcp_txbuildup);
      if (lcs.lcp_rxbuildup)
        buf_free(lcs.lcp_rxbuildup);
      lcs.lcp_rxbuildup = NULL;
      lcs.lcp_txbuildup = NULL;
      if (dls.ps.ps_inframes) 
        queue_delete(dls.ps.ps_inframes);
      dls.ps.ps_inframes = NULL;
      if (lcs.lcp_inframes) 
        queue_delete(lcs.lcp_inframes);
      lcs.lcp_inframes = NULL;

      break;
    case NIDLP_CSLIP:
      if (dls.cs.cs_txbuildup)
        p_free(dls.cs.cs_txbuildup);
      if (dls.cs.cs_rxbuildup)
        buf_free(dls.cs.cs_rxbuildup);
      dls.cs.cs_rxbuildup = NULL;
      dls.cs.cs_txbuildup = NULL;
      break;
  }
}
 

/*******************************************************************************
***
*** Function         : nif_open opens the network interface with the specified
***                    characteristics, to go either NIS_UP or NIS_TERM.
*** Preconditions    : sc->si_scriptname is either '\0', meaning no script 
***                    will be used to negotiate the connection, or will contain
***                    a filename containing a script to use.
***                    ni_ip is the IP address of this device. ATM, the
***                    net mask is automatically determined. Subnetting isn't
***                    supported.
*** Postconditions   : nif_open is TRUE, and the open worked.
***                    nif_open is FALSE, and the open failed.
***
*******************************************************************************/

int nif_open(int newstate)
{
LOCAL_D TEXT szScriptFailed[]="Script failed";

  /* FIXME ? If the new state is the same as the current state, return 
             unchanged? */

  /* FIXME ? If the current state is TERM, and the change is to go UP, 
             don't drop the line first. */

  /* Open the port */
  asy_close();
  if (asy_open() < 0) {
    /* asy_open will report to the console about serial port open failures */
    goto openfail; 
  }

  /* Fill in the netif structure */
  net_num(netif->ni_net,netif->ni_ip);
  net_num(netif->ni_subnet,netif->ni_ip);
  net_mask(netif->ni_mask,netif->ni_net);
  brc_num(netif->ni_brc,netif->ni_ip);
  brc_num(netif->ni_nbrc,netif->ni_ip);

  switch (netif->ni_dlp) {
    case NIDLP_SLIP:
      netif->ni_mtu = SLIP_MTU; /* Or whatever RFC 1055 says */

      dls.ss.ss_txbuildup = NULL;
      dls.ss.ss_rxbuildup = NULL;

      dls.ss.ss_txbuildup = p_alloc( (SLIP_MTU * 2)+2 );
      if (dls.ss.ss_txbuildup == NULL) {
        nif_errmsg(szAllocate, szSLIP, szTx, szBuffer);
        goto openfail;
      }
      /* Now we allocate a buffer rather than a memory cell, and it only has
         to be SLIP_MTU bytes long, not SLIP_MTU * 2, since we perform 
         character unstuffing on this, not stuffing. */
      dls.ss.ss_rxbuildup = buf_alloc( SLIP_MTU );
      if (dls.ss.ss_rxbuildup == NULL) {
        nif_errmsg(szAllocate, szSLIP, szRx, szBuffer);
        goto openfail;
      }
      dls.ss.ss_rcnt = 0;
      dls.ss.ss_escaped = FALSE;
      break;

    case NIDLP_PPP:
      /* Allocate PPP incoming queues */
      if ((dls.ps.ps_inframes=queue_create(PPP_INQSZ)) == NULL) {
        nif_errmsg(szCreate, "local PPP input", szQueue, szNull);
        goto openfail;
      }
      netif->ni_mtu = PPP_MTU;
      /* Or whatever RFCs 1161, 1548, 1570, 1547, 1332, 1172 say */

      if ((lcs.lcp_inframes=queue_create(PPP_INQSZ)) == NULL) {
        nif_errmsg(szCreate, "local LCP input", szQueue, szNull);
        goto openfail;
      }

      dls.ps.ps_txbuildup = NULL;
      dls.ps.ps_rxbuildup = NULL;

      dls.ps.ps_txbuildup = p_alloc( PPP_MTU );
      if (dls.ps.ps_txbuildup == NULL) {
        nif_errmsg(szAllocate, szPPP, szTx, szBuffer);
        goto openfail;
      }

      dls.ps.ps_rxbuildup = p_alloc( PPP_MTU );
      if (dls.ps.ps_rxbuildup == NULL) {
        nif_errmsg(szAllocate, szPPP, szRx, szBuffer);
        goto openfail;
      }

      lcs.lcp_txbuildup = NULL;
      lcs.lcp_rxbuildup = NULL;

      lcs.lcp_txbuildup = p_alloc( PPP_MTU );
      if (lcs.lcp_txbuildup == NULL) {
        nif_errmsg(szAllocate, szPPP, szTx, szBuffer);
        goto openfail;
      }

      lcs.lcp_rxbuildup = p_alloc( PPP_MTU );
      if (lcs.lcp_rxbuildup == NULL) {
        nif_errmsg(szAllocate, szPPP, szRx, szBuffer);
        goto openfail;
      }
      break;

    case NIDLP_CSLIP:
      netif->ni_mtu = CSLIP_MTU;
      /* Or whatever RFCs 1055 and 1144 say */

      dls.cs.cs_txbuildup = NULL;
      dls.cs.cs_rxbuildup = NULL;

      dls.cs.cs_txbuildup = p_alloc( CSLIP_MTU );
      if (dls.cs.cs_txbuildup == NULL) {
        nif_errmsg(szAllocate, szCSLIP, szTx, szBuffer);
        goto openfail;
      }

      dls.cs.cs_rxbuildup = p_alloc( CSLIP_MTU );
      if (dls.cs.cs_rxbuildup == NULL) {
        nif_errmsg(szAllocate, szCSLIP, szRx, szBuffer);
        goto openfail;
      }
      break;
  }

  netif->ni_ioctets=
  netif->ni_iucast=
  netif->ni_inucast=
  netif->ni_idiscard=
  netif->ni_ierrors=
  netif->ni_iunkproto=
  netif->ni_ooctets=
  netif->ni_oucast=
  netif->ni_onucast=
  netif->ni_odiscard=
  netif->ni_oerrors=
  netif->ni_oqlen=0;
  netif->ni_maxreasm=576; /* FIXME: see Comer v2, p96 */


  /* Which state do we want to go to? UP or TERM? */
  switch (newstate) {
    case NIS_UP:
      /* Negotiate with the remote data link protocol before connecting,
	 if needed. This may alter whatever network interface parameters are
         negotiated with the remote end of the connection. */
      if (sc.si_scriptused) {
        /* script_start will initialise the ni_sc structure
           if the script can be started. It returns immediately.
           The script is processed asynchronously. */
        if (script_start() == FALSE) {
          console_puts(szScriptFailed);
          msg_event(EVT_UI_MESSAGE, szScriptFailed, p_slen(szScriptFailed));
          goto openfail;
        }
        else {
          netif->ni_state=NIS_SCRIPT;
          msg_event(EVT_SCRIPTING, NULL, 0);
        }
      }
      else {
        /* No script is involved. If the data link layer is PPP, we need to
           send the ACTION_OPEN, and work through the PPP finite
           state machine, to bring the PPP link into 'STATE_OPENED' state,
           whereupon, it'll bring the interface UP. If the data link layer
           is not PPP, then bring the interface up directly */
        if (netif->ni_dlp == NIDLP_PPP) {
          msg_event(EVT_LINK_UP, NULL, 0); /* Should be after ppp_getevent? */
          ppp_getevent(EVENT_OPEN);
          netif->ni_state = NIS_PPP_UP;
        }
        else {
          netif->ni_state = NIS_UP;
          msg_event(EVT_NET_UP, NULL, 0);
        }
      }
      break;
    case NIS_TERM:
      netif->ni_state=NIS_TERM;
      msg_event(EVT_TERM, NULL, 0);
      break;
    default:
      console_puts("nif_open: not asked for UP or TERM!");
      goto openfail;
  }
  asy_sethandler();
  return TRUE;


  /* We have failed. Release all resources */
openfail:
  nif_buildupfree();
  asy_close();
  asy_sethandler();
  return FALSE;
}


/*******************************************************************************
***
*** Function         : nif_close closes a network interface and frees resources.
*** Preconditions    : n is an open network interface.
*** Postconditions   : The data link is not PPP, and the network interface
***                    is closed, and reset to 'useable'.
***                    If the data link is PPP, the PPP finite state machine is
***                    used to gracefully bring the link down, whereupon the
***                    interface is brought down and set to 'useable'.
***
*******************************************************************************/

void nif_close(struct netif *n)
{
  /* Close down whatever we were doing */
  switch (n->ni_state) {
    case NIS_UP:
      if (n->ni_dlp == NIDLP_PPP) { /* Must be NI_NETWORK */
        ppp_close_gracefully();
        /* Don't do any more processing here, as we will call nif_close again 
           from ppp_close_gracefully, but this time the state will be NIS_DOWN 
        */
        return; 
      }
      break;
    case NIS_PPP_UP:
      if (n->ni_dlp == NIDLP_PPP) { /* Must be NI_NETWORK */
        ppp_close_gracefully();
        /* Don't do any more processing here, as we will call nif_close again 
           from ppp_close_gracefully, but this time the state will be NIS_DOWN 
        */
        n->ni_state = NIS_PPP_DOWN;
        msg_event(EVT_LINK_DOWN, NULL, 0);
        return; 
      }
      break;
  }

  if (n->ni_ipinq)
    queue_discard(n->ni_ipinq);
  if (n->ni_devoutq)
    queue_discard(n->ni_devoutq);
  n->ni_state = NIS_DOWN;
  msg_event(EVT_NET_DOWN, NULL, 0);
  if (n == netif) {
    asy_close();
    nif_buildupfree();
    asy_sethandler();
  }
}


/*******************************************************************************
***
*** Function         : nif_destroy()
*** Preconditions    : Frees absolutely everything connected with all interfaces
*** Postconditions   : The interfaces are naked.
***
*******************************************************************************/

void nif_destroy(void)
{
int i;
struct netif *n;
  asy_sethandler();
  for (i=0,n=nif; i<NI_MAXINTERFACES; i++,n++) {
    nif_close(n);
  }
}


/*******************************************************************************
***
*** Function         : nif_process handles incoming and outgoing frames on all
***                    open network interfaces. It isn't called inprocess or 
***                    outprocess, since although it's really an output 
***                    process, the local interface's output is the higher 
***                    layer's input. It gets confusing...
*** Preconditions    : There may be a network interface open.
*** Postconditions   : Frames get processed.
***
*******************************************************************************/

void nif_process(void)
{
  local_outprocess();
  if (netif->ni_state == NIS_UP) {
    /* Send outgoing frames. Which data-link layer? */
    switch (netif->ni_dlp) {
      case NIDLP_SLIP:  slip_outprocess(); break;
      case NIDLP_CSLIP: cslip_outprocess(); break;
      case NIDLP_PPP:   ppp_outprocess(); break;
      default: break;
    }
  }
}


/*******************************************************************************
***
*** Function         : nif_send
*** Preconditions    : n is an open network interface, i is an IP datagram 
***                    that has been converted into NETWORK byte order.
*** Postconditions   : nif_send is FALSE, n's ni_devoutq is full, and i is 
***                    discarded, or
***                    nif_send is TRUE, n's ni_devoutq is not full, and i is 
***                    enqueued 
***
*******************************************************************************/

int nif_send(struct netif *n, struct buf *i)
{

  /* Length of i should be i->ip_len */
  if (!queue_enqueue(n->ni_devoutq, i)) {
    console_puts("nif_send: output q full");
    n->ni_odiscard++;
    return FALSE;
  }
  /* Wake up! Time to process! */
  attn_signal(ATTN_NIF);
  return TRUE;
}


/*******************************************************************************
***
*** Function         : nif_stats()
*** Preconditions    : 
*** Postconditions   : All network interface stats are displayed
***
*******************************************************************************/

void nif_stats()
{
int i;
  for (i=0; i<NI_MAXINTERFACES; i++)
    nif_stat(&nif[i]);
}


/*******************************************************************************
***
*** Function         : nif_stat(n)
*** Preconditions    : n is a network interface
*** Postconditions   : n's statistics are displayed
***
*******************************************************************************/

void nif_stat(struct netif *n)
{
  console_printf("nif_stat: IP address %s\r\ninput output description\r\n", 
                  (UINT)szConvertIPaddr(n->ni_ip));
  console_printf("%5d %5d  octets\r\n", n->ni_ioctets, n->ni_ooctets);
  console_printf("%5d %5d  unicasts\r\n", n->ni_iucast, n->ni_oucast);
  console_printf("%5d %5d  non-unicasts\r\n", n->ni_inucast, n->ni_onucast);
  console_printf("%5d %5d  discards\r\n", n->ni_idiscard, n->ni_odiscard);
  console_printf("%5d %5d  errors\r\n", n->ni_ierrors, n->ni_oerrors);
  console_printf("%5d        unk. protos\r\n", n->ni_iunkproto);
  console_printf("%5d        D/E addrs\r\n", n->ni_iaddress);
}


