/*******************************************************************************
***
*** Filename         : ppp.c
*** Purpose          : Point-to-Point Protocol support for PsiStack
*** Author           : Patrick H. Robbe, Matt J. Gumbley
*** Created          : 11/07/97
*** Last updated     : 24/07/98
***
********************************************************************************
***
*** Modification Record: ppp_inprocess changed to fix the escaped character bug.
***                      Added the different out-queues to handle LCP/IP/...
***                      packets.
***
*******************************************************************************/

#include "stack.h"

/* Here's the initial PPP configuration 
 *
 * May be completely wrong ;-) -- MJG
 */
struct ppp_config ppp_config = { 30, PPP_IPHDR_UNCOMPRESSED, PPP_AUTH_NONE };

/*
 * FCS lookup table as calculated by the table generator.
 * See RFC 1662 page 19-20 for 16 bits FCS computation method. The 
 * generator program is given on pages 18-19.
 */
static unsigned int fcstab[256] = {
    0x0000, 0x1189, 0x2312, 0x329b, 0x4624, 0x57ad, 0x6536, 0x74bf,
    0x8c48, 0x9dc1, 0xaf5a, 0xbed3, 0xca6c, 0xdbe5, 0xe97e, 0xf8f7,
    0x1081, 0x0108, 0x3393, 0x221a, 0x56a5, 0x472c, 0x75b7, 0x643e,
    0x9cc9, 0x8d40, 0xbfdb, 0xae52, 0xdaed, 0xcb64, 0xf9ff, 0xe876,
    0x2102, 0x308b, 0x0210, 0x1399, 0x6726, 0x76af, 0x4434, 0x55bd,
    0xad4a, 0xbcc3, 0x8e58, 0x9fd1, 0xeb6e, 0xfae7, 0xc87c, 0xd9f5,
    0x3183, 0x200a, 0x1291, 0x0318, 0x77a7, 0x662e, 0x54b5, 0x453c,
    0xbdcb, 0xac42, 0x9ed9, 0x8f50, 0xfbef, 0xea66, 0xd8fd, 0xc974,
    0x4204, 0x538d, 0x6116, 0x709f, 0x0420, 0x15a9, 0x2732, 0x36bb,
    0xce4c, 0xdfc5, 0xed5e, 0xfcd7, 0x8868, 0x99e1, 0xab7a, 0xbaf3,
    0x5285, 0x430c, 0x7197, 0x601e, 0x14a1, 0x0528, 0x37b3, 0x263a,
    0xdecd, 0xcf44, 0xfddf, 0xec56, 0x98e9, 0x8960, 0xbbfb, 0xaa72,
    0x6306, 0x728f, 0x4014, 0x519d, 0x2522, 0x34ab, 0x0630, 0x17b9,
    0xef4e, 0xfec7, 0xcc5c, 0xddd5, 0xa96a, 0xb8e3, 0x8a78, 0x9bf1,
    0x7387, 0x620e, 0x5095, 0x411c, 0x35a3, 0x242a, 0x16b1, 0x0738,
    0xffcf, 0xee46, 0xdcdd, 0xcd54, 0xb9eb, 0xa862, 0x9af9, 0x8b70,
    0x8408, 0x9581, 0xa71a, 0xb693, 0xc22c, 0xd3a5, 0xe13e, 0xf0b7,
    0x0840, 0x19c9, 0x2b52, 0x3adb, 0x4e64, 0x5fed, 0x6d76, 0x7cff,
    0x9489, 0x8500, 0xb79b, 0xa612, 0xd2ad, 0xc324, 0xf1bf, 0xe036,
    0x18c1, 0x0948, 0x3bd3, 0x2a5a, 0x5ee5, 0x4f6c, 0x7df7, 0x6c7e,
    0xa50a, 0xb483, 0x8618, 0x9791, 0xe32e, 0xf2a7, 0xc03c, 0xd1b5,
    0x2942, 0x38cb, 0x0a50, 0x1bd9, 0x6f66, 0x7eef, 0x4c74, 0x5dfd,
    0xb58b, 0xa402, 0x9699, 0x8710, 0xf3af, 0xe226, 0xd0bd, 0xc134,
    0x39c3, 0x284a, 0x1ad1, 0x0b58, 0x7fe7, 0x6e6e, 0x5cf5, 0x4d7c,
    0xc60c, 0xd785, 0xe51e, 0xf497, 0x8028, 0x91a1, 0xa33a, 0xb2b3,
    0x4a44, 0x5bcd, 0x6956, 0x78df, 0x0c60, 0x1de9, 0x2f72, 0x3efb,
    0xd68d, 0xc704, 0xf59f, 0xe416, 0x90a9, 0x8120, 0xb3bb, 0xa232,
    0x5ac5, 0x4b4c, 0x79d7, 0x685e, 0x1ce1, 0x0d68, 0x3ff3, 0x2e7a,
    0xe70e, 0xf687, 0xc41c, 0xd595, 0xa12a, 0xb0a3, 0x8238, 0x93b1,
    0x6b46, 0x7acf, 0x4854, 0x59dd, 0x2d62, 0x3ceb, 0x0e70, 0x1ff9,
    0xf78f, 0xe606, 0xd49d, 0xc514, 0xb1ab, 0xa022, 0x92b9, 0x8330,
    0x7bc7, 0x6a4e, 0x58d5, 0x495c, 0x3de3, 0x2c6a, 0x1ef1, 0x0f78
};

/* Some variables to help me tune the PPP layer */
struct logiostat ppp_iostat[] = {
  "bytes", 0, 0,
  "frames", 0, 0,
  "largestframe", 0, 0,
  NULL, 0, 0
};
#define pppstat_bytes ppp_iostat[0]
#define pppstat_frames ppp_iostat[1]
#define pppstat_largestframe ppp_iostat[2]

/* Function prototypes */
void ppp_ip_enqueue(void);
void ppp_lcp_enqueue(void);

/*******************************************************************************
***
*** Function         : ppp_close_gracefully()
*** Preconditions    : netif is a PPP network interface which is currently up.
***                    nif_close() was called, and so it is this function's
***                    job to gracefully close the link, and then call
***                    nif_close 'properly'.
*** Postconditions   : The link is brought down, and the interface closed.
***
*******************************************************************************/

VOID ppp_close_gracefully(void)
{
  /* Bring the interface down */

  /* Set the state to down. nif_close will do this as part of it's close, but
     if it is NIS_UP when you call nif_close, and netif->ni_dlp is NIDLP_PPP,
     this routine gets called to close gracefully again, resulting in an
     infinite loop... */
  netif->ni_state=NIS_DOWN;
  /* We're only 'half-down' at this point, AFAICR, so broadcast that the link 
     layer is going down. nif_close will set it to NIS_DOWN anyway */
  msg_event(EVT_LINK_DOWN, NULL, 0);
  ppp_getevent(EVENT_CLOSE);
  nif_close(netif);
}


/*******************************************************************************
***
*** Function         : ppp_outprocess() takes a frame from netif->devoutq 
***                    and sends it.
*** Preconditions    : netif is a network interface, possibly with frames queued
***                    for output.
*** Postconditions   : ppp_outprocess is positive and the first frame in the 
***                    output queue has been sent. The return value is the 
***                    size of the PPP-encoded frame sent.
***                    ppp_outprocess is FALSE and either the output queue is 
***                    empty, or an error occurred.
***
*******************************************************************************/

int ppp_outprocess(void)
{
register int i;
register byte *d, s;
register struct buf *rawbuf, *origrawbuf;
INT errcode;
UBYTE ctrl[3];
unsigned int fcs;


  if (flags.log_flag & LOG_FLAG_FLOWTRACE)
    console_puts("ppp_outprocess: encapsulating datagram");

  /* Encode the frame. */
  origrawbuf = rawbuf = queue_dequeue(netif->ni_devoutq);
  if (rawbuf == NULL)
    return FALSE;

  /* Now encode */
  d = dls.ps.ps_txbuildup;
  *d++ = PPP_FLAG;
  fcs=PPP_INITFCS16;
  /* Start adding into the FCS from here... */
  if (lcs.lcp_acfc==0) { /* Address and Control Fields not compressed */
    *d++ = PPP_ADDRESS;
    fcs = ppp_fcs16(fcs, *d);
    *d++ = PPP_CONTROL;
    fcs = ppp_fcs16(fcs, *d);
    }

  /* Encode every buffer in the chain */
  while (rawbuf != NULL) {
    for (i = 0; i < rawbuf->usedsize; i++) {
      switch (s = rawbuf->data[i]) {
        case PPP_FLAG:
          fcs = ppp_fcs16(fcs, s);
          *d++ = PPP_ESC;
          *d++ = PPP_XOR ^ s;
          break;
        case PPP_ESC:
          fcs = ppp_fcs16(fcs, s);
          *d++ = PPP_ESC;
          *d++ = PPP_XOR ^ PPP_ESC;
          break;
        default:
          *d++ = s;
          fcs = ppp_fcs16(fcs, s);
          break;
      }
    }
    rawbuf = rawbuf->next;
  }

  /* End the FCS calculation */
  *d++ = fcs & 0xff;           /* FCS byte 1 */;
  *d++ = ((fcs >> 8) & 0xff);  /* FCS byte 2 */;
  *d++ = PPP_FLAG;
  i = d - dls.ps.ps_txbuildup;
  if (flags.log_flag & LOG_FLAG_IPDUMP) {
    /* Patrick: you may want to change this to LOG_FLAG_LINKDUMP? --Matt */
    console_printf("ppp_outprocess: ppp-encoded datagram: 0x%04x bytes:\r\n", i);
    hexdump(dls.ps.ps_txbuildup, i);
  }

  /* If we're using hardware handshaking, is CTS active? */
  /* FIXME: This checking of CTS is copied from the other data link code -
            it should be made common later... MJG */
  if (asy.asy_handshake == RTSCTS) {
    ctrl[1]=0x00;
    if (p_iow(asy.asy_serH, P_FCTRL, ctrl)==0 && (!ctrl[0] & P_SRCTRL_CTS)) {
      console_puts("ppp_outprocess: CTS inactive");
      ppp_getevent(EVENT_DOWN);
      buf_free(origrawbuf);
      nif_close(netif);
      return FALSE;
    }
  }

  errcode=p_write(asy.asy_serH, dls.ps.ps_txbuildup, i);
  if (errcode != 0)
    goto pppsenderr;

  buf_free(origrawbuf);

  netif->ni_ooctets += i;
  pppstat_frames.ls_output++;
  pppstat_bytes.ls_output += i;
  if (i > pppstat_largestframe.ls_output)
    pppstat_largestframe.ls_output = i;

  if (flags.log_flag & LOG_FLAG_FLOWTRACE)
    console_puts("ppp_outprocess: done encapsulating datagram");

  return i;

pppsenderr:
  //  ppp_getevent(EVENT_DOWN);  ???
  switch(errcode) {
    case E_FILE_PARITY:
    case E_FILE_FRAME:
    case E_FILE_OVERRUN:
      console_puts("ppp_outprocess: frame error - check baud rate etc.");
      break;
    case E_FILE_LINE:
      console_puts("ppp_outprocess: line was dropped");
      buf_free(origrawbuf);
      nif_close(netif);
      return FALSE;
    default:
      console_printf("ppp_outprocess: write failed err %d\r\n", errcode);
  }
  buf_free(origrawbuf);
  return d - dls.ps.ps_txbuildup;
}


/*******************************************************************************
***
*** Function         : ppp_inprocess decodes an encapsulated PPP frame
*** Preconditions    : netif is an open network interface, with
***                    asy_serLength bytes of received serial data stored in
***                    asy_serBuffer
*** Postconditions   : Either a frame has been read, and placed on 
***                    netif->ni_ipinq for retrieval and processing by the IP 
***                    layer, or on lcs.lcp_inframes for retrieval and 
***                    processing by the LCP layer, or an empty frame has 
***                    been built, and so can be ignored, or the frame has 
***                    not been fully built.
***
*******************************************************************************/

void ppp_inprocess(void)
{
register UWORD len = asy.asy_serLength;
register UBYTE *serbuf = &asy.asy_serBuffer[0];
register UBYTE ch,ch1,ch2;

  netif->ni_ioctets += len;

  /* Process each byte in the input stream */
  for (; len>0 ; len--,serbuf++) {
    ch = *serbuf;

    if (ch==PPP_ESC) { /* The character is escaped, so we read the next */
      ch=*serbuf++;    /* one, and un-escape it. */
      ch = ch ^ PPP_XOR;
    }

    /* Now that we are sure to have only un-secaped characters, we proceed. */
    if (ch==PPP_FLAG) { /* Start of new frame, or end of one */

      /* Start of a new frame */
      if (dls.ps.ps_rcnt == 0) {
        /* After the start flag, we have the adress and the control bytes - */
        /*   if they are not compressed. */
        ch1 = *serbuf++; /* it should be the Address byte (0xFF) */
        ch2 = *serbuf++; /* and this one the Control byte (0x03) */
        if (ch1!=0xFF || ch2!=0x03) {
          if (lcs.lcp_acfc==0) { /* ACFC not negociated */
            console_puts("ppp_inprocess: Bad Address or Control Field; discarding");
            netif->ni_idiscard++;
            dls.ps.ps_rcnt = 0;
          }
          else {
            *serbuf--;  /* we have to read again the first two bytes, */
            *serbuf--;  /* as they were not the Address and Control Field */
          }
        }

        /* Then we have 1 or 2 byte(s) protocol field */
        dls.ps.ps_protocol = TRUE;
      }
      else {
        /* This is the end of the frame, so we transmit it to the appropriate */
        /* layer */
        switch (dls.ps.ps_prot) {
          case 0x0021: /* This is an Internet Protocol datagram, enqueue it */
            ppp_ip_enqueue();
            break;
          case 0x8021: /* This is an Internet Protocol Control Protocol datagram */
            break;
          case 0xc021: /* This is a Link Control Protocol datagram */
          case 0xc023: /* This is a Password Authentication Protocol datagram */
          case 0xc025: /* This is a Link Quality Protocol datagram */
          case 0xc223: /* This is a Challenge Handshake Authentication Prot. datagram */
            ppp_lcp_enqueue();
            /* We send anything LCP-related to the lcp handler which will */
            /* dispatch the datagram to the appropriate sub-handler */
            break;
          default: /* This is an unknown (or more probably unsuported) protocol */
            ppp_action(ACTION_SPJ); /* We'll be supporting CHAP, PAP, LQP! [Patrick] */
            break;
        }
      }
    }

    /* If we are still in the protocol field, get the protocol value */
    if (dls.ps.ps_protocol) {
      /* We are in the protocol field */
      if (ch && 0x01) {
        /* this is the last protocol field byte */
        dls.ps.ps_prot = dls.ps.ps_prot + ch;
        dls.ps.ps_protocol = FALSE; // No more protocol byte
      }
      else
        /* this is the first byte, and we have a second one coming next */
        dls.ps.ps_prot = dls.ps.ps_prot + (ch * 256);
    }
    else {
      /* We reach here with a character for the buildup buffer. Make
         sure there's space for it.
      */
      if (dls.ps.ps_rcnt == PPP_MTU) {
        console_puts("ppp_inprocess: buildup buffer full; discarding");
        netif->ni_idiscard++;
        dls.ps.ps_rcnt = 0;
      }
      else
        dls.ps.ps_rxbuildup->data[dls.ps.ps_rcnt++] = ch;
    }
  }
}


/*******************************************************************************
***
*** Function         : ppp_ip_enqueue
*** Preconditions    : the ppp handler has received a complete frame
*** Postconditions   : the frame is enqueued for IP
***
*******************************************************************************/

void ppp_ip_enqueue(void)
{
  /* Enqueue it for IP but not if IP's input queue is full */
  if (queue_full(netif->ni_ipinq)) {
    console_puts("ppp_ip_enqueue: IP inq full; discarding");
    netif->ni_idiscard++;
    /* Queue full, so re-use buildup buffer */
    dls.ps.ps_rcnt = 0;
  }
  else {
    /* Queue the frame for IP. */
    pppstat_bytes.ls_input += dls.ps.ps_rcnt;
    pppstat_frames.ls_input++;
    if (dls.ps.ps_rcnt > pppstat_largestframe.ls_input)
      pppstat_largestframe.ls_input = dls.ps.ps_rcnt;
    dls.ps.ps_rxbuildup->usedsize = dls.ps.ps_rcnt;
    queue_enqueue(netif->ni_ipinq, dls.ps.ps_rxbuildup);
    dls.ps.ps_rxbuildup = buf_alloc( PPP_MTU );
    if (dls.ps.ps_rxbuildup == NULL) {
      console_puts("ppp_ip_enqueue: couldn't alloc new rxbuf");
      nif_close(netif);
      return;
    }
    dls.ps.ps_rcnt = 0;
    dls.ps.ps_escaped = FALSE;
    attn_signal(ATTN_IP);
    if (flags.log_flag & LOG_FLAG_FLOWTRACE)
      console_puts("ppp_ip_enqueue: datagram enqueued");
  }
}


/*******************************************************************************
***
*** Function         : ppp_lcp_enqueue
*** Preconditions    : the ppp handler has received a LCP frame
*** Postconditions   : the frame is enqueued for LCP
***
*******************************************************************************/

void ppp_lcp_enqueue(void)
{
  /* Enqueue it for LCP but not if LCP's input queue is full */
  if (queue_full(lcs.lcp_inframes)) {
    console_puts("ppp_lcp_enqueue: PPP inframes full; discarding");
    /* Patrick: do we need to maintain that count of discarded 
                LCP frames? -- Matt */
    /* Queue full, so re-use buildup buffer */
    lcs.lcp_rcnt = 0;
  }
  else {
    /* Queue the frame for LCP. */
    pppstat_bytes.ls_input += lcs.lcp_rcnt;
    pppstat_frames.ls_input++;
    if (lcs.lcp_rcnt > pppstat_largestframe.ls_input)
      pppstat_largestframe.ls_input = lcs.lcp_rcnt;
    lcs.lcp_rxbuildup->usedsize = lcs.lcp_rcnt;
    queue_enqueue(lcs.lcp_inframes, lcs.lcp_rxbuildup);
    lcs.lcp_rxbuildup = buf_alloc( PPP_MTU );
    if (lcs.lcp_rxbuildup == NULL) {
      console_puts("ppp_lcp_enqueue: couldn't alloc new rxbuf");
      nif_close(netif);
      return;
    }
    lcs.lcp_rcnt = 0;
    attn_signal(ATTN_LCP);
    if (flags.log_flag & LOG_FLAG_FLOWTRACE)
      console_puts("ppp_lcp_enqueue: datagram enqueued");
  }
}


/*******************************************************************************
***
*** Function         : ppp_stats
*** Preconditions    : None
*** Postconditions   : The average frame size for send and receive is output.
***
*******************************************************************************/

void ppp_stats(void)
{
  console_puts("ppp_stats:");
  console_printf("avg frame size in octets: rcv: %d snd: %d\r\n",
                 pppstat_frames.ls_input==0 ? 0 : 
                 pppstat_bytes.ls_input/pppstat_frames.ls_input,
                 pppstat_frames.ls_output==0 ? 0 : 
                 pppstat_bytes.ls_output/pppstat_frames.ls_output);
  console_writeiostats(ppp_iostat);
}


/*******************************************************************************
***
*** Function         : ppp_init
*** Preconditions    : None
*** Postconditions   : Leaves the Finite State Machine in the "initial" state
***
*******************************************************************************/

void ppp_init(void)
{
  console_puts("ppp_init");
  dls.ps.ps_state = STATE_INITIAL;
  dls.ps.ps_rtsts = Timer_stopped; 
  dls.ps.ps_rtval = Restart_Timer_def;
  dls.ps.ps_rtmax = Restart_Timer_def;

  /* Patrick: shouldn't ^^^ this be ppp_config.restart_timer? -- MJG*/

  dls.ps.ps_mtval = Max_Terminate_def;
  dls.ps.ps_mcval = Max_Configure_def;
  dls.ps.ps_mfval = Max_Failure_def;
  dls.ps.ps_mru = PPP_MRU;
  dls.ps.ps_rcval = Max_Configure_def;

  lcs.lcp_mru=1500;                         /* Maximum-Receive-Unit (default 1500) */
  lcs.lcp_auth_prot=0x0000;                 /* Authentication Protocol (PAP or CHAP) */
  lcs.lcp_qual_prot=0x0000;                 /* Quality Monitoring Protocol (LQR) */
  lcs.lcp_magic_number=lcp_new_magic_nb();  /* Magic Number */
  lcs.lcp_pfc=0x00;                         /* Protocol Field Compression */
  lcs.lcp_acfc=0x00;                        /* Address and Control Field Compression */
  lcs.lcp_cur_tx_id=0x00;                   /* Identifier for last emitted LCP packets */
  lcs.lcp_cur_rx_id=0x00;                   /* Identifier for last received LCP packet */

}


/*******************************************************************************
***
*** Function         : ppp_getevent
*** Preconditions    : 
*** Postconditions   :
***
*******************************************************************************/

void ppp_getevent(byte event)
{
TEXT *str;
static TEXT fname[]="ppp_event: ";

  /* Do diagnostic messages */
  if (flags.log_flag & LOG_FLAG_FLOWTRACE) {
    switch (event) {
      case EVENT_UP:
        str = "Up";
        break;
      case EVENT_DOWN:
        str = "Down";
        break;
      case EVENT_OPEN:
        str = "Open";
        break;
      case EVENT_CLOSE:
        str = "Close";
        break;
      case EVENT_TOP:
        str = "Time Out +";
        break;
      case EVENT_TOM:
        str = "Time Out -";
        break;
      case EVENT_RCRP:
        str = "Receive Configure Request +";
        break;
      case EVENT_RCRM:
        str = "Receive Configure Request -";
        break;
      case EVENT_RCA:
        str = "Receive Configure Ack";
        break;
      case EVENT_RCN:
        str = "Receive Configure Nak/Rej";
        break;
      case EVENT_RTR:
        str = "Receive Terminate Request";
        break;
      case EVENT_RTA:
        str = "Receive Terminate Ack";
        break;
      case EVENT_RUC:
        str = "Receive Unknown Code";
        break;
      case EVENT_RXJP:
        str = "Receive Code/Protocol Reject +";
        break;
      case EVENT_RXJM:
        str = "Receive Code/Protocol Reject -";
        break;
      case EVENT_RXR:
        str = "Receive Echo Request/Reply/Discard Request";
        break;
      default:
        str = "Unknown event"; /* Added by MJG */
        break;
    }
    console_printf("%s%s\r\n", (UINT)fname, str);
  }

  /* Now process the event */
  switch (dls.ps.ps_state) {
    /* If the Finite State Machine is in the INITIAL state */
    case STATE_INITIAL:
      switch (event) {
        case EVENT_UP:
          dls.ps.ps_state = STATE_CLOSED;
          break;
        case EVENT_OPEN:
          ppp_action(ACTION_TLS); // This Layer Started
          ppp_change_state(STATE_STARTING);
          break;
        case EVENT_CLOSE:
          dls.ps.ps_state = STATE_INITIAL;
          break;
        default:
          break;
      }
      break;
    /* If the Finite State Machine is in the STARTING state */
    case STATE_STARTING:
      switch (event) {
        case EVENT_UP:
          ppp_action(ACTION_IRC);      // Initialize Restart Count
          ppp_action(ACTION_SCR);      // Send Configure Request
          ppp_change_state(STATE_REQ_SENT);
          break;
        case EVENT_OPEN:
          ppp_change_state(STATE_STARTING);
          break;
        case EVENT_CLOSE:
          ppp_action(ACTION_TLF);      // This Layer Finished
          ppp_change_state(STATE_INITIAL);
          break;
        default:
          break;
      }
      break;
    /* If the Finite State Machine is in the CLOSED state */
    case STATE_CLOSED:
      switch (event) {
        case EVENT_DOWN:
          ppp_change_state(STATE_INITIAL);
          break;
        case EVENT_OPEN:
          ppp_action(ACTION_IRC);      // Initialize Restart Count
          ppp_action(ACTION_SCR);      // Send Configure Request
          ppp_change_state(STATE_REQ_SENT);
          break;
        case EVENT_CLOSE:
          ppp_change_state(STATE_CLOSED);
          break;
        case EVENT_RCRP:
        case EVENT_RCRM:
        case EVENT_RCA:
        case EVENT_RCN:
        case EVENT_RTR:
          ppp_action(ACTION_STA);      // Send Terminate Ack
          ppp_change_state(STATE_CLOSED);
          break;
        case EVENT_RTA:
        case EVENT_RXJP:
        case EVENT_RXR:
          ppp_change_state(STATE_CLOSED);
          break;
        case EVENT_RUC:
          ppp_action(ACTION_SCJ);      // Send Code Reject
          ppp_change_state(STATE_CLOSED);
          break;
        case EVENT_RXJM:
          ppp_action(ACTION_TLF);      // This Layer Finished
          ppp_change_state(STATE_CLOSED);
          break;
        default:
          break;
      }
      break;
    /* If the Finite State Machine is in the STOPPED state */
    case STATE_STOPPED:
      switch (event) {
        case EVENT_DOWN:
          ppp_action(ACTION_TLS);      // This Layer Started
          ppp_change_state(STATE_STARTING);
          break;
        case EVENT_OPEN:
          ppp_change_state(STATE_STOPPED);
          break;
        case EVENT_CLOSE:
          ppp_change_state(STATE_CLOSED);
          break;
        case EVENT_RCRP:
          ppp_action(ACTION_IRC);      // Initialize Restart Count
          ppp_action(ACTION_SCR);      // Send Configure Request
          ppp_action(ACTION_SCA);      // Send Configure Ack
          ppp_change_state(STATE_ACK_SENT);
          break;
        case EVENT_RCRM:
          ppp_action(ACTION_IRC);      // Initialize Restart Count
          ppp_action(ACTION_SCR);      // Send Configure Request
          ppp_action(ACTION_SCN);      // Send Configure Nack
          ppp_change_state(STATE_REQ_SENT);
          break;
        case EVENT_RCA:
        case EVENT_RCN:
        case EVENT_RTR:
          ppp_action(ACTION_STA);      // Send Terminate Ack
          ppp_change_state(STATE_STOPPED);
          break;
        case EVENT_RTA:
        case EVENT_RXJP:
        case EVENT_RXR:
          ppp_change_state(STATE_STOPPED);
          break;
        case EVENT_RUC:
          ppp_action(ACTION_SCJ);      // Send Code Reject
          ppp_change_state(STATE_STOPPED);
          break;
        case EVENT_RXJM:
          ppp_action(ACTION_TLF);      // This Layer Finished
          ppp_change_state(STATE_STOPPED);
          break;
        default:
          break;
      }
      break;
    /* If the Finite State Machine is in the CLOSING state */
    case STATE_CLOSING:
      switch (event) {
        case EVENT_DOWN:
          ppp_change_state(STATE_INITIAL);
          break;
        case EVENT_OPEN:
          ppp_change_state(STATE_STOPPING);
          break;
        case EVENT_CLOSE:
          ppp_change_state(STATE_CLOSING);
          break;
        case EVENT_TOP:
          ppp_action(ACTION_STR);      // Send Terminate Request
          ppp_change_state(STATE_CLOSING);
          break;
        case EVENT_TOM:
          ppp_action(ACTION_TLF);      // This Layer Finished
          ppp_change_state(STATE_CLOSED);
          break;
        case EVENT_RCRP:
        case EVENT_RCRM:
        case EVENT_RCA:
        case EVENT_RCN:
        case EVENT_RXJP:
        case EVENT_RXR:
          ppp_change_state(STATE_CLOSING);
          break;
        case EVENT_RTR:
          ppp_action(ACTION_STA);      // Send Terminate Ack
          ppp_change_state(STATE_CLOSING);
          break;
        case EVENT_RTA:
          ppp_action(ACTION_TLF);      // This Layer Finished
          ppp_change_state(STATE_CLOSED);
          break;
        case EVENT_RUC:
          ppp_action(ACTION_SCJ);      // Send Code Reject
          ppp_change_state(STATE_CLOSING);
          break;
        case EVENT_RXJM:
          ppp_action(ACTION_TLF);      // This Layer Finished
          ppp_change_state(STATE_CLOSED);
          break;
        default:
          break;
      }
      break;
    /* If the Finite State Machine is in the STOPPING state */
    case STATE_STOPPING:
      switch (event) {
        case EVENT_DOWN:
          ppp_change_state(STATE_STARTING);
          break;
        case EVENT_OPEN:
          ppp_change_state(STATE_STOPPING);
          break;
        case EVENT_CLOSE:
          ppp_change_state(STATE_CLOSING);
          break;
        case EVENT_TOP:
          ppp_action(ACTION_STR);      // Send Terminate Request
          ppp_change_state(STATE_STOPPING);
          break;
        case EVENT_TOM:
          ppp_action(ACTION_TLF);      // This Layer Finished
          ppp_change_state(STATE_STOPPED);
          break;
        case EVENT_RCRP:
        case EVENT_RCRM:
        case EVENT_RCA:
        case EVENT_RCN:
        case EVENT_RXJP:
        case EVENT_RXR:
          ppp_change_state(STATE_STOPPING);
          break;
        case EVENT_RTR:
          ppp_action(ACTION_STA);      // Send Terminate Ack
          ppp_change_state(STATE_STOPPING);
          break;
        case EVENT_RTA:
          ppp_action(ACTION_TLF);      // This Layer Finished
          ppp_change_state(STATE_STOPPED);
          break;
        case EVENT_RUC:
          ppp_action(ACTION_SCJ);      // Send Code Reject
          ppp_change_state(STATE_STOPPING);
          break;
        case EVENT_RXJM:
          ppp_action(ACTION_TLF);      // This Layer Finished
          ppp_change_state(STATE_STOPPED);
          break;
        default:
          break;
      }
      break;
    /* If the Finite State Machine is in the REQ-SENT state */
    case STATE_REQ_SENT:
      switch (event) {
        case EVENT_DOWN:
          ppp_change_state(STATE_STARTING);
          break;
        case EVENT_OPEN:
          ppp_change_state(STATE_REQ_SENT);
          break;
        case EVENT_CLOSE:
          ppp_action(ACTION_IRC);      // Initialize Restart Count
          ppp_action(ACTION_STR);      // Send Terminate Request
          ppp_change_state(STATE_CLOSING);
          break;
        case EVENT_TOP:
          ppp_action(ACTION_SCR);      // Send Configure Request
          ppp_change_state(STATE_REQ_SENT);
          break;
        case EVENT_TOM:
          ppp_action(ACTION_TLF);      // This Layer Finished
          ppp_change_state(STATE_STOPPED);
          break;
        case EVENT_RCRP:
          ppp_action(ACTION_SCA);      // Send Configure Ack
          ppp_change_state(STATE_ACK_SENT);
          break;
        case EVENT_RCRM:
          ppp_action(ACTION_SCN);      // Send Configure Nack
          ppp_change_state(STATE_REQ_SENT);
          break;
        case EVENT_RCA:
          ppp_action(ACTION_IRC);      // Initialize Restart Count
          ppp_change_state(STATE_ACK_RCVD);
          break;
        case EVENT_RCN:
          ppp_action(ACTION_IRC);      // Initialize Restart Count
          ppp_action(ACTION_SCR);      // Send Configure Request
          ppp_change_state(STATE_REQ_SENT);
          break;
        case EVENT_RTR:
          ppp_action(ACTION_STA);      // Send Terminate Ack
          ppp_change_state(STATE_REQ_SENT);
          break;
        case EVENT_RTA:
          ppp_change_state(STATE_REQ_SENT);
          break;
        case EVENT_RUC:
          ppp_action(ACTION_SCJ);      // Send Code Reject
          ppp_change_state(STATE_REQ_SENT);
          break;
        case EVENT_RXJP:
          ppp_change_state(STATE_REQ_SENT);
          break;
        case EVENT_RXJM:
          ppp_action(ACTION_TLF);      // This Layer Finished
          ppp_change_state(STATE_STOPPED);
          break;
        case EVENT_RXR:
          ppp_change_state(STATE_REQ_SENT);
          break;
        default:
          break;
      }
      break;
    /* If the Finite State Machine is in the ACK-RCVD state */
    case STATE_ACK_RCVD:
      switch (event) {
        case EVENT_DOWN:
          ppp_change_state(STATE_STARTING);
          break;
        case EVENT_OPEN:
          ppp_change_state(STATE_ACK_RCVD);
          break;
        case EVENT_CLOSE:
          ppp_action(ACTION_IRC);      // Initialize Restart Count
          ppp_action(ACTION_STR);      // Send Terminate Request
          ppp_change_state(STATE_CLOSING);
          break;
        case EVENT_TOP:
          ppp_action(ACTION_SCR);      // Send Configure Request
          ppp_change_state(STATE_REQ_SENT);
          break;
        case EVENT_TOM:
          ppp_action(ACTION_TLF);      // This Layer Finished
          ppp_change_state(STATE_STOPPED);
          break;
        case EVENT_RCRP:
          ppp_action(ACTION_SCA);      // Send Configure Ack
          ppp_action(ACTION_TLU);      // This Layer Up
          ppp_change_state(STATE_OPENED);
          break;
        case EVENT_RCRM:
          ppp_action(ACTION_SCN);      // Send Configure Nack
          ppp_change_state(STATE_ACK_RCVD);
          break;
        case EVENT_RCA:
          ppp_action(ACTION_SCR);      // Send Configure Request
          ppp_change_state(STATE_REQ_SENT);
          break;
        case EVENT_RCN:
          ppp_action(ACTION_SCR);      // Send Configure Request
          ppp_change_state(STATE_REQ_SENT);
          break;
        case EVENT_RTR:
          ppp_action(ACTION_STA);      // Send Terminate Ack
          ppp_change_state(STATE_REQ_SENT);
          break;
        case EVENT_RTA:
          ppp_change_state(STATE_REQ_SENT);
          break;
        case EVENT_RUC:
          ppp_action(ACTION_SCJ);      // Send Code Reject
          ppp_change_state(STATE_ACK_RCVD);
          break;
        case EVENT_RXJP:
          ppp_change_state(STATE_REQ_SENT);
          break;
        case EVENT_RXJM:
          ppp_action(ACTION_TLF);      // This Layer Finished
          ppp_change_state(STATE_STOPPED);
          break;
        case EVENT_RXR:
          ppp_change_state(STATE_ACK_RCVD);
          break;
        default:
          break;
      }
      break;
    /* If the Finite State Machine is in the ACK-SENT state */
    case STATE_ACK_SENT:
      switch (event) {
        case EVENT_DOWN:
          ppp_change_state(STATE_STARTING);
          break;
        case EVENT_OPEN:
          ppp_change_state(STATE_ACK_SENT);
          break;
        case EVENT_CLOSE:
          ppp_action(ACTION_IRC);      // Initialize Restart Count
          ppp_action(ACTION_STR);      // Send Terminate Request
          ppp_change_state(STATE_CLOSING);
          break;
        case EVENT_TOP:
          ppp_action(ACTION_SCR);      // Send Configure Request
          ppp_change_state(STATE_ACK_SENT);
          break;
        case EVENT_TOM:
          ppp_action(ACTION_TLF);      // This Layer Finished
          ppp_change_state(STATE_STOPPED);
          break;
        case EVENT_RCRP:
          ppp_action(ACTION_SCA);      // Send Configure Ack
          ppp_change_state(STATE_ACK_SENT);
          break;
        case EVENT_RCRM:
          ppp_action(ACTION_SCN);      // Send Configure Nack
          ppp_change_state(STATE_REQ_SENT);
          break;
        case EVENT_RCA:
          ppp_action(ACTION_IRC);      // Initialize Restart Count
          ppp_action(ACTION_TLU);      // This Layer Up
          ppp_change_state(STATE_OPENED);
          break;
        case EVENT_RCN:
          ppp_action(ACTION_IRC);      // Initialize Restart Count
          ppp_action(ACTION_SCR);      // Send Configure Request
          ppp_change_state(STATE_ACK_SENT);
          break;
        case EVENT_RTR:
          ppp_action(ACTION_STA);      // Send Terminate Ack
          ppp_change_state(STATE_REQ_SENT);
          break;
        case EVENT_RTA:
          ppp_change_state(STATE_ACK_SENT);
          break;
        case EVENT_RUC:
          ppp_action(ACTION_SCJ);      // Send Code Reject
          ppp_change_state(STATE_ACK_SENT);
          break;
        case EVENT_RXJP:
          ppp_change_state(STATE_ACK_SENT);
          break;
        case EVENT_RXJM:
          ppp_action(ACTION_TLF);      // This Layer Finished
          ppp_change_state(STATE_STOPPED);
          break;
        case EVENT_RXR:
          ppp_change_state(STATE_ACK_SENT);
          break;
        default:
          break;
      }
      break;
    /* If the Finite State Machine is in the OPENED state */
    case STATE_OPENED:
      switch (event) {
        case EVENT_UP:
          break;
        case EVENT_DOWN:
          ppp_action(ACTION_TLD);      // This Layer Down
          ppp_change_state(STATE_STARTING);
          break;
        case EVENT_OPEN:
          ppp_change_state(STATE_OPENED);
          break;
        case EVENT_CLOSE:
          ppp_action(ACTION_TLD);      // This Layer Down
          ppp_action(ACTION_IRC);      // Initialize Restart Count
          ppp_action(ACTION_STR);      // Send Terminate Request
          ppp_change_state(STATE_CLOSING);
          break;
        case EVENT_TOP:
          break;
        case EVENT_TOM:
          break;
        case EVENT_RCRP:
          ppp_action(ACTION_TLD);      // This Layer Down
          ppp_action(ACTION_SCR);      // Send Configure Request
          ppp_action(ACTION_SCA);      // Send Configure Ack
          ppp_change_state(STATE_ACK_SENT);
          break;
        case EVENT_RCRM:
          ppp_action(ACTION_TLD);      // This Layer Down
          ppp_action(ACTION_SCR);      // Send Configure Request
          ppp_action(ACTION_SCN);      // Send Configure Nack
          ppp_change_state(STATE_REQ_SENT);
          break;
        case EVENT_RCA:
          ppp_action(ACTION_TLD);      // This Layer Down
          ppp_action(ACTION_SCR);      // Send Configure Request
          ppp_change_state(STATE_REQ_SENT);
          break;
        case EVENT_RCN:
          ppp_action(ACTION_TLD);      // This Layer Down
          ppp_action(ACTION_SCR);      // Send Configure Request
          ppp_change_state(STATE_REQ_SENT);
          break;
        case EVENT_RTR:
          ppp_action(ACTION_TLD);      // This Layer Down
          ppp_action(ACTION_ZRC);      // Zero Restart Count
          ppp_action(ACTION_STA);      // Send Terminate Ack
          ppp_change_state(STATE_STOPPING);
          break;
        case EVENT_RTA:
          ppp_action(ACTION_TLD);      // This Layer Down
          ppp_action(ACTION_SCR);      // Send Configure Request
          ppp_change_state(STATE_REQ_SENT);
          break;
        case EVENT_RUC:
          ppp_action(ACTION_SCJ);      // Send Code Reject
          ppp_change_state(STATE_OPENED);
          break;
        case EVENT_RXJP:
          ppp_change_state(STATE_OPENED);
          break;
        case EVENT_RXJM:
          ppp_action(ACTION_TLD);      // This Layer Down
          ppp_action(ACTION_IRC);      // Initialize Restart Count
          ppp_action(ACTION_STR);      // Send Terminate Request
          ppp_change_state(STATE_STOPPING);
          break;
        case EVENT_RXR:
          ppp_action(ACTION_SER);      // Send Echo Reply
          ppp_change_state(STATE_OPENED);
          break;
        default:
          break;
      }
      break;
    default:
      break;
  }
}


/*******************************************************************************
***
*** Function         : ppp_change_state
*** Preconditions    : newstate is a defined PPP automaton state
*** Postconditions   : if new state is different from the old state, the
***                    new state is achieved, stopping or starting the
***                    restart timer if needed
***
*******************************************************************************/

void ppp_change_state(byte newstate)
{
TEXT *str;
static TEXT fname[]="ppp_state: ";

  if (newstate != dls.ps.ps_state) {
    dls.ps.ps_state = newstate;
    switch (dls.ps.ps_state) {
      case STATE_INITIAL:
      case STATE_STARTING:
      case STATE_CLOSED:
      case STATE_STOPPED:
      case STATE_OPENED:
        dls.ps.ps_rtsts = Timer_stopped;
        if (flags.log_flag & LOG_FLAG_FLOWTRACE)
          console_printf("%s Timer stopped\r\n", (UINT)fname);
        break;
      default:
        dls.ps.ps_rtsts = Timer_running;
        if (flags.log_flag & LOG_FLAG_FLOWTRACE)
          console_printf("%s Timer running\r\n", (UINT)fname);
    }
    if (dls.ps.ps_state == STATE_OPENED) {
      console_puts("PPP: link is up");
      msg_event(EVT_LINK_UP, NULL, 0);
    }
  }

  /* Output a diagnostic message showing new state */
  if (flags.log_flag & LOG_FLAG_FLOWTRACE) {
    switch (dls.ps.ps_state) {
      case STATE_INITIAL:
        str = "Initial (0)";
        break;
      case STATE_STARTING:
        str = "Starting (1)";
        break;
      case STATE_CLOSED:
        str = "Closed (2)";
        break;
      case STATE_STOPPED:
        str = "Stopped (3)";
        break;
      case STATE_CLOSING:
        str = "Closing (4)";
        break;
      case STATE_STOPPING:
        str = "Stopping (5)";
        break;
      case STATE_REQ_SENT:
        str = "Req-Sent (6)";
        break;
      case STATE_ACK_RCVD:
        str = "Ack-Rcvd (7)";
        break;
      case STATE_ACK_SENT:
        str = "Ack-Sent (8)";
        break;
      case STATE_OPENED:
        str = "Opened (9)";
        break;
      default:
        str = "Unknown state";
        break;
    }
    console_printf("%s%s\r\n", (UINT)fname, str);
  }
}


/*******************************************************************************
***
*** Function         : ppp_timer
*** Preconditions    : nil
*** Postconditions   : the restart timer is decreased
***
*******************************************************************************/

void ppp_timer(void)
{
  if (netif->ni_dlp == NIDLP_PPP && dls.ps.ps_rtsts == Timer_running) {
    dls.ps.ps_rtval = dls.ps.ps_rtval - 5;
    /* This is a 'quick and dirty' way to set up the timer. I know this,
       but it's fine for me because a) it works even on a 3a, and b) because
       we do not really need to have a high precision for these timers
    */

    /* FIXME: integrate this with the TCP timer? What about overflows? */

    if (dls.ps.ps_rtval <= 0) {
      attn_signal(ATTN_FSM);
      console_puts("ppp_timer: Restart Timer reached zero");
    }
  }
}


/*******************************************************************************
***
*** Function         : fsm_process
*** Preconditions    : 
*** Postconditions   : 
***
*******************************************************************************/

void fsm_process(void)
{
  if ((netif->ni_dlp == NIDLP_PPP) && (dls.ps.ps_rtval <= 0)) {
    /* This PPP interface has reached a time-out condition. We must find out
       if it's a TO+ or TO- condition, to send the appropriate event to the 
       FSM.
    */
    dls.ps.ps_rtval=dls.ps.ps_rtmax;
    if (dls.ps.ps_rcval > 0) {
      /* The restart counter is higher than zero, so we try again */
      dls.ps.ps_rcval--;
      console_printf("ppp_timer: Restart Counter = %d\r\n", dls.ps.ps_rcval);
      ppp_getevent(EVENT_TOP);
    }
    else
      /* The restart counter is zero, so we stop trying */
      ppp_getevent(EVENT_TOM);
  }
}


/*******************************************************************************
***
*** Function         : ppp_action
*** Preconditions    : action is a defined PPP automaton action
*** Postconditions   : the action is processed (hopefully!)
***
*******************************************************************************/

void ppp_action(byte action)
{
TEXT *str;
static TEXT fname[]="ppp_action: %s\r\n";

  /* Output a diagnostic message */
  if (flags.log_flag & LOG_FLAG_FLOWTRACE) {
    switch (action) {
      case ACTION_IE:
        str = "Illegal Event";
        break;
      case ACTION_TLU:
        str = "This Layer Up";
        break;
      case ACTION_TLD:
        str = "This Layer Down";
        break;
      case ACTION_TLS:
        str = "This Layer Started";
        break;
      case ACTION_TLF:
        str = "This Layer Finished";
        break;
      case ACTION_IRC:
        str = "Initialize Restart Count";
        break;
      case ACTION_ZRC:
        str = "Zero Restart Count";
        break;
      case ACTION_SCR:
        str = "Send Configure Request";
        break;
      case ACTION_SCA:
        str = "Send Configure Ack";
        break;
      case ACTION_SCN:
        str = "Send Configure Nack";
        break;
      case ACTION_STR:
        str = "Send Terminate Request";
        break;
      case ACTION_STA:
        str = "Send Terminate Ack";
        break;
      case ACTION_SCJ:
        str = "Send Code Reject";
        break;
      case ACTION_SER:
        str = "Send Echo Reply";
        break;
      case ACTION_SPJ:
        str = "Send Protocol Reject";
        break;
      default:
        str = "Unknown action";
        break;
    }
    console_printf(fname, (UINT)str);
  }

  /* Now process the action */
  switch (action) {
    case ACTION_IE:
      break;

    case ACTION_TLU:
      netif->ni_state = NIS_UP;
      msg_event(EVT_NET_UP, NULL, 0);
      break;

    case ACTION_TLD:
      netif->ni_state = NIS_DOWN;
      msg_event(EVT_NET_DOWN, NULL, 0);
      break;

    case ACTION_TLS:
      break;

    case ACTION_TLF:
      break;

    case ACTION_IRC:
      dls.ps.ps_rcval = Max_Configure_def;
      /* FIXME: need to fix this as it must be either
         max-configure, max-terminate or max failure, depending on the 
         situation.  max configure is the highest, so (hopefully) it 
         shouldn't hurt. */
      break;

    case ACTION_ZRC:
      dls.ps.ps_rtval = dls.ps.ps_rtmax;
      dls.ps.ps_rcval = 0;
      break;

    case ACTION_SCR:
      dls.ps.ps_rtval = dls.ps.ps_rtmax;
      ppp_send_configure_request();
      break;

    case ACTION_SCA:
      ppp_send_configure_ack();
      break;

    case ACTION_SCN:
      ppp_send_configure_nack();
      break;

    case ACTION_STR:
      dls.ps.ps_rtval = dls.ps.ps_rtmax;
      ppp_send_terminate_request();
      break;

    case ACTION_STA:
      ppp_send_terminate_ack();
      break;

    case ACTION_SCJ:
      ppp_send_code_reject();
      break;

    case ACTION_SER:
      ppp_send_echo_reply();
      break;

    case ACTION_SPJ:
      ppp_send_protocol_reject();
      break;

    default:
      break;
  }
}
  

/*******************************************************************************
***
*** Function         : ppp_fcs16
*** Preconditions    : fcs is the initial fcs value, c is the next char included
***                    in the checksum.
*** Postconditions   : the new fcs is returned
***
*******************************************************************************/

unsigned int ppp_fcs16(unsigned int fcs, unsigned char c) {
  fcs = (fcs >> 8) ^ fcstab[(fcs ^ c) & 0xff];
  return fcs;
}

