/*******************************************************************************
***
*** Filename         : ppp.h
*** Purpose          : Point-to-Point Protocol definitions and prototypes.
*** Author           : Patrick H. Robbe, Matt J. Gumbley
*** Created          : 06/02/97
*** Last updated     : 24/08/98
***
********************************************************************************
***
*** Modification Record: moved LCP-relative defines to LCP.H [Patrick]
***
*******************************************************************************/

#ifndef PPP_H
#define PPP_H

/* Note: we may have to have a user-interface structure, i.e. one that the user
 * edits, that gets loaded into the "working" structure at load-time. We do
 * this with TCP and UDP config structures. */

GLREF_D struct ppp_config ppp_config;                    /* Declared in ppp.c */

/* PPP framing constants */
#define PPP_FLAG         0x7E
#define PPP_ADDRESS      0xFF
#define PPP_CONTROL      0x03
#define PPP_ESC          0x7D
#define PPP_XOR          0x20
#define PPP_INITFCS16    0xffff  /* Initial FCS value */
#define PPP_GOODFCS16    0xf0b8  /* Good final FCS value */

/* The Protocols field values */
#define IP   0x0021  // Internet Protocol
#define IPCP 0x8021  // Internet Protocol Control Protocol
#define LCP  0xc021  // Link Control Protocol
#define PAP  0xc023  // Password Authentication Protocol
#define LQR  0xc025  // Link Quality Protocol
#define CHAP 0xc223  // Challenge Handshake Authentication Protocol

/* The different states of the automaton */
#define STATE_INITIAL  0
#define STATE_STARTING 1
#define STATE_CLOSED   2
#define STATE_STOPPED  3
#define STATE_CLOSING  4
#define STATE_STOPPING 5
#define STATE_REQ_SENT 6
#define STATE_ACK_RCVD 7
#define STATE_ACK_SENT 8
#define STATE_OPENED   9

/* The different events causing transitions of the automaton */
#define EVENT_UP             100
#define EVENT_DOWN           101
#define EVENT_OPEN           102
#define EVENT_CLOSE          103
#define EVENT_TOP            104  // Time Out +
#define EVENT_TOM            105  // Time Out -
#define EVENT_RCRP           106  // Receive Configure Request +
#define EVENT_RCRM           107  // Receive Configure Request -
#define EVENT_RCA            108  // Receive Configure Ack
#define EVENT_RCN            109  // Receive Configure Nak/Rej
#define EVENT_RTR            110  // Receive Terminate Request
#define EVENT_RTA            111  // Receive Terminate Ack
#define EVENT_RUC            112  // Receive Unknown Code
#define EVENT_RXJP           113  // Receive Code (or Protocol) Reject +
#define EVENT_RXJM           114  // Receive Code (or Protocol) Reject -
#define EVENT_RXR            115  // Receive Echo Request, Receive-Echo-Reply
                                  // or Receive-Discard-Request

/* The different actions of the automaton */
#define ACTION_IE  200  // Illegal Event (-)
#define ACTION_TLU 201  // This Layer Up
#define ACTION_TLD 202  // This Layer Down
#define ACTION_TLS 203  // This Layer Started
#define ACTION_TLF 204  // This Layer Finished
#define ACTION_IRC 205  // Initialize Restart Count
#define ACTION_ZRC 206  // Zero Restart Count
#define ACTION_SCR 207  // Send Configure Request
#define ACTION_SCA 208  // Send Configure Ack
#define ACTION_SCN 209  // Send Configure Nack
#define ACTION_STR 210  // Send Terminate Request
#define ACTION_STA 211  // Send Terminate Ack
#define ACTION_SCJ 212  // Send Code Reject
#define ACTION_SER 213  // Send Echo Reply

#define ACTION_SPJ 214  // Send Protocol Reject (not a standard FSM action)

/* The different LCP Code Field values */
#define Configure_Request  1
#define Configure_Ack      2
#define Configure_Nak      3
#define Configure_Reject   4
#define Terminnate_Request 5
#define Terminate_Ack      6
#define Code_Reject        7
#define Protocol_Reject    8
#define Echo_Request       9
#define Echo_Reply         10
#define Discard_Request    11

/* The Timer status */
#define Timer_stopped 0
#define Timer_running 1

/* Timer & Counter related default parameters */
#define Restart_Timer_def 30  // Wait 3 seconds before sending TO event if we
                              // don't receive an Ack or Nack after sending a
                              // Request packet (Configure or Terminate)
                              // note: 3 seconds = 30 ticks (I must check that)
#define Max_Terminate_def 2   // Transmit 2 Terminate-Request without receiving
                              // a Terminate Ack packet, before assuming that
                              // the peer is unable to respond
#define Max_Configure_def 10  // Transmit 10 Configure-Request without receiving
                              // a Configure Ack, Nack or Reject packet, before
                              // assuming that the peer is unable to respond
#define Max_Failure_def   5   // Transmit 5 Configure-Nack without sending
                              // a Configure-Ack packet, before assuming that
                              // the configuration is not converging

/* The Maximum Transfer Unit for PPP networks */
#define PPP_MTU 500 /* In an ideal world, this would be 1500 */

/* The Maximum Receive Unit for PPP networks */
#define PPP_MRU 500 /* This too... */

#define PPP_INQSZ 10                                  /* IP Input queue size */


/* Storage requirements for PPP */
struct ppp_storage {
  byte *ps_txbuildup;                      /* Tx frame buildup buffer for PPP */
  struct buf *ps_rxbuildup;                /* Rx frame buildup buffer for PPP */
  int ps_rcnt;                       /* Next posision in frame being built up */
  int ps_escaped;                                     /* Are we in an escape? */
  int ps_protocol;                           /* Are we in the protocol field? */
  int ps_prot;                                              /* Protocol field */

  /* The PPP parameters used by each PPP interface */
  ULONG ps_rt;                  /* Restart Timer (value in 1/10th of seconds) */
  byte  ps_state;                        /* State of the Finite State Machine */
  byte  ps_rtsts;                              /* Status of the Restart Timer */
  byte  ps_rtval;                              /* Restart Timer current value */
  byte  ps_rtmax;                              /* Restart Timer maximum value */
  byte  ps_rcval;                                    /* Restart Counter value */
  byte  ps_mtval;                              /* MAX-TERMINATE current value */
  byte  ps_mcval;                              /* MAX-CONFIGURE current value */
  byte  ps_mfval;                                /* MAX-FAILURE current value */
  int   ps_mru;                                                        /* MRU */

  struct queue *ps_inframes;        /* Incoming PPP frames from ppp_inhandler */
};

extern int  ppp_outprocess(void);
extern void ppp_inprocess(void);
extern void ppp_init(void);
extern void ppp_getevent(byte event);
extern void ppp_change_state(byte newstate);
extern void ppp_timer(void);                                   //used in timer.c
extern void fsm_process(void);                                 //used in attn.c
extern void ppp_action(byte action);
extern void ppp_close_gracefully(void);
extern void ppp_stats(void);
extern unsigned int ppp_fcs16(unsigned int fcs, unsigned char c);

#endif /* PPP_H */
