/*******************************************************************************
***
*** Filename         : queue.c
*** Purpose          : FIFO queue routines based on a circular array, storing 
***                    buffer chain structures, from buffer.[ch]
***
***                    Can also be used to store integers, but you must use the
***                    iqueue* routines for that, NOT the queue* routines,
***                    since these might interpret the queue node contents.
***
*** Author           : Matt J. Gumbley
*** Created          : 19/01/97
*** Last updated     : 11/09/97
***
********************************************************************************
***
*** Modification Record
***
*******************************************************************************/

#include "stack.h"

/*******************************************************************************
***
*** Function         : queue_full(q), iqueue_full(q)
*** Preconditions    : q is a queue created by queue_create
*** Postconditions   : queue_full is TRUE and the queue is full.
***                    queue_full is FALSE and the queue is not full
***
*******************************************************************************/

int queue_full(struct queue *q)
{
  return q->size == q->maxsize;
}


/*******************************************************************************
***
*** Function         : queue_empty(q), iqueue_empty(q)
*** Preconditions    : q is a queue created by queue_create
*** Postconditions   : queue_empty is TRUE and the queue is empty.
***                    queue_full is FALSE and the queue is not empty.
***
*******************************************************************************/

int queue_empty(struct queue *q)
{
  return q->size == 0;
}


/*******************************************************************************
***
*** Function         : queue_create(size)
*** Preconditions    : size is a positive integer
*** Postconditions   : queue_create is NULL, and allocation failed.
***                    queue_create is non-NULL, and a queue has been created
***                    which will hold 'size' number of entries. The return
***                    is a pointer to the queue, which has been initialised.
***
*******************************************************************************/

struct queue *queue_create(int size)
{
struct queue *q;
  q = p_alloc(sizeof(struct queue) + ((size-1)*sizeof(struct buf *)));
  if (q == NULL)
    return NULL;
  q->head = 0;
  q->tail = size - 1;
  q->size = 0;
  q->maxsize = size;
  return q;
}


/*******************************************************************************
***
*** Function         : queue_delete(q)
*** Preconditions    : q is a queue created by queue_create.
*** Postconditions   : All buffers on the queue are released, and the queue
***                    itself is freed.
***
*******************************************************************************/

void queue_delete(struct queue *q)
{
  queue_discard(q);
  p_free(q);
}


/*******************************************************************************
***
*** Function         : queue_discard(q)
*** Preconditions    : q is a queue created by queue_create.
*** Postconditions   : All buffers on the queue are released.
***
*******************************************************************************/

void queue_discard(struct queue *q)
{
struct buf *b;
  /* Free all buffers on the queue */
  while (q->size > 0) {
    b=queue_dequeue(q);
    buf_free(b);
  }
}


/*******************************************************************************
***
*** Function         : queue_enqueue(q, b)
*** Preconditions    : q is a queue created by queue_create, b is a buffer (or
***                    chain of buffers) created by buf_alloc.
*** Postconditions   : queue_enqueue is FALSE, and b could not be enqueued,
***                    queue_enqueue is TRUE, and b is added to the end of the
***                    queue.
***
*******************************************************************************/

int queue_enqueue(struct queue *q, struct buf *b)
{
  if (queue_full(q)) 
    return FALSE;
  q->tail = (q->tail+1) % q->maxsize;
  q->size++;
  q->q_nodes[q->tail] = b;
  return TRUE;
}


/*******************************************************************************
***
*** Function         : queue_next(q,n)
*** Preconditions    : q is a queue, n ins an index into the queue
*** Postconditions   : the next position in the queue is returned, taking 
***                    into account the possibility of 'wrapping' past the 
***                    end of the queue.
***
*******************************************************************************/

static int queue_next(struct queue *q, int num)
{
  return (num+1) % q->maxsize;
}


/*******************************************************************************
***
*** Function         : queue_prev(q,n)
*** Preconditions    : q is a queue, n ins an index into the queue
*** Postconditions   : the previous position in the queue is returned, taking 
***                    into account the possibility of 'wrapping' past the 
***                    beginning of the queue.
***
*******************************************************************************/

static int queue_prev(struct queue *q, int num)
{
  if (num == 0)
    return q->maxsize;
  else
    return (num-1);
}

/*******************************************************************************
***
*** Function         : queue_enqueue_ordered(q, b, i)
*** Preconditions    : q is a queue created by queue_create, b is a buffer (or
***                    chain of buffers) created by buf_alloc, i is an integer
***                    giving the ordering of buffers on the queue
*** Postconditions   : queue_enqueue_ordered is FALSE, and b could not be 
***                    enqueued, since the queue is full.
***                    queue_enqueue_ordered is TRUE, and b is inserted at the
***                    relevant place in the queue, as given by i. The queue
***                    is ordered by ascending values of the integer i. Other
***                    entries in the queue may have been shuffled up (i lies 
***                    between other values in the queue), or b is inserted 
***                    at the end (i is greater than all other queue entries).
*** Note             : Queues must either use this ordered enqueue, OR the
***                    unordered enqueue, but these must not be mixed, since
***                    the unordered one doesn't do anything with the order
***                    field of the buffer - values are determined by the 
***                    caller, and this could confuse the ordering routine.
***
*******************************************************************************/


int queue_enqueue_ordered(struct queue *q, struct buf *b, int i)
{
int ipos,shufidx,size;
  if (queue_full(q)) 
    return FALSE;

  /* Need to find the right place, and shuffle some elements up */
  for (ipos=q->head, size=q->size; size != 0; ipos=queue_next(q,ipos), size--)
    if (q->q_nodes[ipos]->order > i)
      break; /* ipos is the position to insert at, and size is the number of 
                elements to shuffle up */

  if (size != 0) {
    /* The insert position is in the middle. Size is set to indicate how
       many elements to shuffle up. */
    for (shufidx=q->tail; size != 0; shufidx=queue_prev(q,shufidx), size--)
      q->q_nodes[queue_next(q,shufidx)]=q->q_nodes[shufidx];
  }
  /* else  nothing needs doing */
  /* Add an element to the queue */
  q->tail = queue_next(q, q->tail);
  /* Store the ordering information in the buffer */
  b->order = i;
  q->q_nodes[ipos]=b;
  return TRUE;
}


/*******************************************************************************
***
*** Function         : queue_dequeue(q)
*** Preconditions    : q is a queue created by queue_create
*** Postconditions   : queue_dequeue is NULL and q is empty.
***                    queue_dequeue is non-NULL, and a buffer has been dequeued
***                    the return is a pointer to this buffer.
***
*******************************************************************************/

struct buf *queue_dequeue(struct queue *q)
{
struct buf *b;
  if (queue_empty(q))
    return NULL;
  b = q->q_nodes[q->head];
  q->head = (q->head+1) % q->maxsize;
  q->size--;
  return b;
}


/*******************************************************************************
***
*** Function         : iqueue_create(size)
*** Preconditions    : size is a positive integer
*** Postconditions   : queue_create is NULL, and allocation failed.
***                    queue_create is non-NULL, and a queue has been created
***                    which will hold 'size' number of entries. The return
***                    is a pointer to the queue, which has been initialised.
***
*******************************************************************************/

struct queue *iqueue_create(int size)
{
struct queue *q;
  q = p_alloc(sizeof(struct queue) + ((size-1)*sizeof(int)));
  if (q == NULL)
    return NULL;
  q->head = 0;
  q->tail = size - 1;
  q->size = 0;
  q->maxsize = size;
  return q;
}

/*******************************************************************************
***
*** Function         : iqueue_delete(q)
*** Preconditions    : q is a queue created by iqueue_create.
*** Postconditions   : The queue is freed.
***
*******************************************************************************/

void iqueue_delete(struct queue *q)
{
  p_free(q);
}


/*******************************************************************************
***
*** Function         : iqueue_enqueue(q, i)
*** Preconditions    : q is a queue created by iqueue_create, i is an int
*** Postconditions   : iqueue_enqueue is FALSE, and i could not be enqueued,
***                    iqueue_enqueue is TRUE, and i is added to the end of the
***                    queue.
***
*******************************************************************************/

int iqueue_enqueue(struct queue *q, int i)
{
  if (iqueue_full(q))  
    return FALSE;
  q->tail = (q->tail+1) % q->maxsize;
  q->size++;
  q->q_values[q->tail] = i;
  return TRUE;
}


/*******************************************************************************
***
*** Function         : iqueue_dequeue(q)
*** Preconditions    : q is a non-empty queue created by iqueue_create
***                    You **MUST** check that the queue is not empty before
***                    calling this.
*** Postconditions   : An int has been dequeued and is the return value.
***
*******************************************************************************/

int iqueue_dequeue(struct queue *q)
{
int i;
  i = q->q_values[q->head];
  q->head = (q->head+1) % q->maxsize;
  q->size--;
  return i;
}

