/*******************************************************************************
***
*** Filename         : script.c
*** Purpose          : PsiStack dial-up scripting language interpreter.
*** Author           : Matt J. Gumbley
*** Created          : 31/03/97
*** Last updated     : 18/03/98
***
********************************************************************************
***
*** Modification Record 
***
*******************************************************************************/

#include "stack.h"


/* Function prototypes for interpreter routines */
static void cmd_baudrate(char *);
static void cmd_beep(char *);
static void cmd_capture(char *);
static void cmd_clear(char *);
static void cmd_compare(char *);
static void cmd_databits(char *);
static void cmd_detectcarrier(char *);
static void cmd_detectdevice(char *);
static void cmd_discard(char *);
static void cmd_discarduntil(char *);
static void cmd_down(char *);
static void cmd_doze(char *);
static void cmd_dozeuntil(char *);
static void cmd_drop(char *);
static void cmd_dump(char *);
static void cmd_echo(char *);
static void cmd_eval(char *);
static void cmd_falsegoto(char *);
static void cmd_flush(char *);
static void cmd_goto(char *);
static void cmd_handshaking(char *);
static void cmd_isempty(char *);
static void cmd_infomsg(char *);
static void cmd_nextphone(char *);
static void cmd_parity(char *);
static void cmd_send(char *);
static void cmd_set(char *);
static void cmd_setstring(char *);
static void cmd_sleep(char *);
static void cmd_splitip(char *);
static void cmd_stopbits(char *);
static void cmd_term(char *);
static void cmd_truegoto(char *);
static void cmd_up(char *);
static void cmd_wait(char *);
static void cmd_wait_compare();

/* Command interpreter name and function pointer table */
/* Since this is searched with a linear scan, I've sorted the table, with
   the probably-most-frequently-used commands first */
static struct cmd cmdtable[] = {
  { "wait",          cmd_wait },
  { "send",          cmd_send },
  { "echo",          cmd_echo },
  { "infomsg",       cmd_infomsg },
  { "sleep",         cmd_sleep },
  { "down",          cmd_down },
  { "up",            cmd_up },
  { "flush",         cmd_flush },
  { "discard",       cmd_discard },
  { "discarduntil",  cmd_discarduntil },
  { "doze",          cmd_doze },
  { "dozeuntil",     cmd_dozeuntil },
  { "truegoto",      cmd_truegoto },
  { "falsegoto",     cmd_falsegoto },
  { "goto",          cmd_goto },
  { "eval",          cmd_eval },
  { "term",          cmd_term },
  { "nextphone",     cmd_nextphone },
  { "isempty",       cmd_isempty },
  { "splitip",       cmd_splitip },

  { "baudrate",      cmd_baudrate },
  { "databits",      cmd_databits },
  { "detectcarrier", cmd_detectcarrier },
  { "detectdevice",  cmd_detectdevice },
  { "handshaking",   cmd_handshaking },
  { "parity",        cmd_parity },
  { "stopbits",      cmd_stopbits },

  { "compare",       cmd_compare },
  { "beep",          cmd_beep },
  { "capture",       cmd_capture },
  { "clear",         cmd_clear },
  { "drop",          cmd_drop },
  { "dump",          cmd_dump },
  { "set",           cmd_set },
  { "setstring",     cmd_setstring },
  { NULL,            NULL }
};


/* Some commonly used text */
GLDEF_D TEXT szNumeric[]="numeric";
GLDEF_D TEXT szVariable[]="variable";
GLDEF_D TEXT szName[]="name";
GLDEF_D TEXT szContents[]="contents";
GLDEF_D TEXT szTooBig[]="too big";
GLDEF_D TEXT szNotDelimited[]="name not delimited correctly";
GLDEF_D TEXT szExpected[]="expected";
GLDEF_D TEXT sz3PercentS[]="%s %s %s\r\n";
#define sz2PercentS (sz3PercentS+3)
#define szCRLF (sz3PercentS+8)
GLDEF_D TEXT szOOM[]="out of memory";
GLDEF_D TEXT szNotFound[]="not found";
GLDEF_D TEXT szNotNumeric[]="not numeric";
GLDEF_D TEXT szEquals[]=" = ";
GLDEF_D TEXT szStackOverflow[]="stack overflow";
GLDEF_D TEXT szStackUnderflow[]="stack underflow";
GLDEF_D TEXT szRPN[]="RPN";
GLDEF_D TEXT szAsyChangeFailure[]="couldn't change port settings";
GLDEF_D TEXT szIgnoreParam[]="ignore";


GLDEF_D struct scriptcontext sc;

static char hexbuf[]="<xx>";

/*******************************************************************************
***
*** Function         : script_init
*** Preconditions    : None
*** Postconditions   : The script context is initialised
***
*******************************************************************************/

VOID script_init(VOID)
{

  /* Initialise script interpreter context */
  p_bfil(&sc, sizeof(struct scriptcontext), 0);

  /* Fill in the default script names */
  p_scpy(sc.si_scriptname, "LOC::M:\\PSISTACK\\*.SCP");
  /* sc.si_scriptused = FALSE by default */
}


/*******************************************************************************
***
*** Function         : namecmp(source, destination)
*** Preconditions    : source and destination are strings. source contains a 
***                    single word, and is null-terminated immediately after
***                    this word. (An example being the command table above)
*** Postconditions   : Taking len as the length of source, namecmp returns
***                    TRUE if the first len characters of destination match
***                    source, or FALSE otherwise. (It actually returns a
***                    pointer to the character after the word in destination)
***
*******************************************************************************/

static char *namecmp(char *s, char *d)
{
  for (; (*s!='\0') && *s==*d; s++, d++);
  if (*s=='\0' && (*d=='\0' || p_isspace(*d))) {
    return d; /* TRUE */
  }
  else {
    return NULL; /* FALSE */
  }
}


/*******************************************************************************
***
*** Function         : ntizeVarname(name, buf)
*** Purpose          : Detects a variable name from a string containing it.
*** Preconditions    : name is a string containing one or whitespace-delimited
***                    words, the first of which is supposed to be a variable 
***                    name. 
***                    buf is a buffer, at least SI_NAMEMAXSIZE characters, or 
***                    may be NULL.
*** Postconditions   : 1) The variable name is longer than SI_NAMEMAXSIZE chars,
***                       and this returns NULL.
***                    2) The variable name is OK, buf is NULL. The variable 
***                       name is copied to a static buffer, null terminated 
***                       at the first whitespace. The address of the buffer 
***                       is returned.
***                    3) As above, buf is non-null, so is used to store the
***                       null-terminated variable.
***
*******************************************************************************/

static char *ntizeVarname(char *name, char *buf)
{
static char ntVarnamebuf[SI_NAMEMAXSIZE];
register char *d,*s;
register int i;
  if (buf==NULL)
    buf=ntVarnamebuf;
  for (i=0,d=buf,s=name; i<SI_NAMEMAXSIZE; i++,s++,d++)
    if (p_isspace(*s)) {
      *d='\0';
      return buf;
    }
    else 
      *d = *s;
  return NULL;
}


/*******************************************************************************
***
*** Function         : find_variable(name)
*** Purpose          : Searches through the variable list for this variable.
*** Preconditions    : scriptcontext exists, name isn't null, and may or not be
***                    null-terminated
*** Postconditions   : If the variable exists, a pointer to it is returned,
***                    else NULL
***
*******************************************************************************/

static struct var *find_variable(char *name)
{
register struct var *v;
  if (sc.si_varlist == NULL) {
    return NULL;
  }

  for (v=sc.si_varlist; v; v=v->v_next) 
    if (namecmp(v->v_name, name)!=0) {
      return v;
    }
  return NULL;
}


/*******************************************************************************
***
*** Function         : create_variable(name, contents, flags)
*** Preconditions    : scriptcontext exists, name isn't null, contents can
***                    be anything you like, flags are VAR_ flags.
*** Postconditions   : The variable gets created and a pointer to it is
***                    returned.
***                    If memory runs out, or the name or contents exceed 
***                    their respective lengths, failed gets performed, and 
***                    NULL is returned. Note: this will find variables if
***                    they already exist
***
*******************************************************************************/

static struct var *create_variable(char *name, char *contents, int flags)
{
register struct var *v;
struct var *newvar;
  if (p_slen(name) > SI_NAMEMAXSIZE)
    console_printf(sz3PercentS,(UINT)szVariable,szName,szTooBig);
  else
    if (p_slen(contents) > SI_VARIABLELENGTH) 
      console_printf(sz3PercentS,(UINT)szVariable,szContents,szTooBig);
    else {
      /* If it already exists, use it, otherwise, allocate */
      if ((newvar=find_variable(name)) == NULL) {
        if ((newvar=p_alloc(sizeof(struct var))) != NULL) {
          p_scpy(newvar->v_name, name);
          newvar->v_next = NULL;
          /* Attach the variable to the end of the variable list */
          if (sc.si_varlist == NULL) {
            sc.si_varlist = newvar;
          }
          else {
            for (v=sc.si_varlist; v; v=v->v_next) {
              if (v->v_next == NULL) {
                v->v_next = newvar;
                break;
              }
            }
          }
        }
        else {
          console_printf(sz2PercentS,(UINT)szVariable,szOOM);
        }
      }
      else {
      }
      if (newvar != NULL) {
        p_scpy(newvar->v_contents, contents);
        newvar->v_index = p_slen(contents);
        newvar->v_flags = flags;
        return newvar;
      }
    }
  /* Something went wrong */
  cmd_down(NULL);
  return NULL;
}


/*******************************************************************************
***
*** Function         : numeric_variable(name)
*** Preconditions    : name is the name of an existing variable, which may 
***                    be null-terminated
*** Postconditions   : Returns the value of the variable, or runs cmd_down 
***                    if it doesn't exist, or isn't numeric.
*** Kludge warning   : This code is cursed. This return value being NULL.. Tch.
***
*******************************************************************************/

static word16 numeric_variable(char *name)
{
register struct var *v;
register int index;
register word16 value;
  if ((v=find_variable(name)) == NULL) {
    console_printf(sz3PercentS,(UINT)szVariable,ntizeVarname(name,NULL),szNotFound);
    cmd_down(NULL);
    return NULL; /* Don't check for this, since the variable could be zero */
  }
  for (index=0,value=0; index<v->v_index; index++)
    if (!p_isdigit(v->v_contents[index])) {
      console_printf(sz3PercentS,(UINT)szVariable,name,szNotNumeric);
      cmd_down(NULL);
      return NULL; /* Don't check for this, since the variable could be zero */
    }
    else {
      value*=10;
      value+=(v->v_contents[index]-'0');
    }
  return value;
}


/*******************************************************************************
***
*** Function         : next_line()
*** Preconditions    : sc.si_IP points into the script.
*** Postconditions   : It points to the start of the next line, or the end 
***                    of script.
***
*******************************************************************************/

static void next_line(void)
{
  while (sc.si_IP != NULL && *sc.si_IP && 
         sc.si_IP < (sc.si_scriptbuf + sc.si_scriptsize - 1)) 
    sc.si_IP++;
  sc.si_IP++;
}


/*******************************************************************************
***
*** Function         : numeric_parameter(p, store)
*** Preconditions    : p points to a string containing a number, terminated 
***                    by something non-numeric.
*** Postconditions   : The integer value of the string is stored at store, and
***                    the address of the non-numeric which terminated the 
***                    search is returned.
***
*******************************************************************************/

char *numeric_parameter(char *p, int *store)
{
char ch;
  *store = 0;
  while (p_isdigit(ch = *p)) {
    *store *= 10;
    *store += (ch - '0');
    p++;
  }
  return p;
}

/*******************************************************************************
***
*** Function         : trace_hexdump(title, buf, len)
*** Preconditions    : title is a descriptinve string, buffer/len is some data
*** Postconditions   : A hexdump-style string is sent as terminal mode data
***                    to all who are interested in such things.
***
*******************************************************************************/

void trace_hexdump(char *title, char *buf, int len)
{
register int i, j;
register char *s, *d;
char expandbuf[(SI_VARIABLELENGTH * 4) + 40]; /* Assume expand all + title */
  /* Copy title */
  for (i=2,s=title,d=expandbuf;*s;i++)
    *d++=*s++;
  *d++='[';
  for (j=0,s=buf;j<len;i++,j++,s++) {
    if (!p_isprint(*s)) {
      *d++='<';
      *d++=hexdigs[(*s&0xf0) >> 4];
      *d++=hexdigs[*s&0x0f];
      *d++='>';
      i+=3;
    }
    else
      *d++=*s;
  }
  *d++=']';
  msg_event(EVT_UI_TERMDATA, expandbuf, i);
}

/*******************************************************************************
***
*** Function         : expandCC(str, &buf)
*** Preconditions    : str is some text, which may contain a character constant.
***                    buf is a pointer to a buffer.
*** Postconditions   : The first character of str is \ and may be followed by
***                    a character constant, or the first character can be 
***                    anything else. The ASCII value of the character is 
***                    stored at buf, and a pointer to the character after 
***                    the (normal character | expanded constant) in str is
***                    returned. 
***
*******************************************************************************/

static char *expandCC(char *str, byte *buf)
{
  if (*str == '\\') {
    str++;
    switch (*str) {
      case 'a': *buf= '\a'; break;
      case 'b': *buf= '\b'; break;
      case 'f': *buf= '\f'; break;
      case 'n': *buf= '\n'; break;
      case 'r': *buf= '\r'; break;
      case 't': *buf= '\t'; break;
      case 'v': *buf= '\v'; break;
      case '0': *buf= '\0'; break;
      default: *buf = '\\'; break;
      case 'x':
        *buf = (hexdig2dec(*(str+1)) << 4) + hexdig2dec(*(str+2));
        str+=2;
        break;
    }
  }
  else
    *buf = *str;
  return ++str;
}


/*******************************************************************************
***
*** Function         : expandTEXT(p)
*** Preconditions    : scriptcontext exists, p is some TEXT, which may 
***                    contain variables and character constants.
*** Postconditions   : si_expandbuf is filled with the expanded TEXT,
***                    and si_expandbuflen is set accordingly. Do not rely on 
***                    p_slen to give you the length of the expand buffer, 
***                    as there have might have been some \0 character 
***                    constants expanded...
***                    If any non-existant variables are referenced, or the
***                    expand buffer is exhausted, a message is displayed, 
***                    and failed is performed.
***                    TRUE or FALSE are returned accordingly.
***
*******************************************************************************/

static int expandTEXT(char *p)
{
char *e;
char *v, *varname;
int i;
struct var *currvar;

  for (e=sc.si_expandbuf; (*p!='\0') && e<(sc.si_expandbuf+SI_EXPBUFMAX);) {
    if (*p == '$') {
      /* Expand variables */
      p++; /* Stop over $ */
      if ((currvar = find_variable(p)) == NULL) {
        varname=ntizeVarname(p,NULL);
        if (varname == NULL)
          console_printf(sz2PercentS,(UINT)szVariable,szNotDelimited);
        else
          console_printf(sz3PercentS,(UINT)szVariable,varname,szNotFound);
        cmd_down(NULL);
        return FALSE;
      }
      for (i=currvar->v_index,v=currvar->v_contents; 
           i>0 && *v && e<(sc.si_expandbuf+SI_EXPBUFMAX);i--)
        /* Expand character constants in the variable */
        v=expandCC(v,(byte *)e++);
      p+=p_slen(currvar->v_name)+1; /* +1 to take care of delimiting space */
    }
    else
      /* Expand character constants */
      p=expandCC(p,(byte *)e++);
  }
  if (e>=(sc.si_expandbuf+SI_EXPBUFMAX)) {
    console_puts("expand buffer exhausted");
    cmd_down(NULL);
    return FALSE;
  }
  *e='\0';
  sc.si_expandbuflen = e - sc.si_expandbuf;
  return TRUE;
}


/*******************************************************************************
***
*** Function         : push(num)
*** Preconditions    : scriptcontext exists, num is a number
*** Postconditions   : The stack is full, and push returns FALSE, else the
***                    number is pushed onto the stack, and push returns TRUE.
***
*******************************************************************************/

static int push(word16 num)
{
  if (sc.si_SP == SI_STACKSIZE) 
    return FALSE;
/*p_printf("pushed %d at pos %d", num, sc.si_SP);*/
  sc.si_stack[sc.si_SP++]=num;
  return TRUE;
}


/*******************************************************************************
***
*** Function         : pop(numbuf)
*** Preconditions    : scriptcontext exists, numbuf is a buffer for the popped
***                    value.
*** Postconditions   : The stack is empty, and pop returns FALSE, else, the 
***                    top value is popped off the top of the stack, and 
***                    stored in numbuf, and TRUE is returned.
***
*******************************************************************************/

static int pop(word16 *numbuf)
{
  if (sc.si_SP == 0) 
    return FALSE;
  *numbuf = sc.si_stack[--sc.si_SP];
/*p_printf("popped %d from pos %d", *numbuf, sc.si_SP);*/
  return TRUE;
}


/*******************************************************************************
***
*** Function         : evaluateRPN(rpn)
*** Preconditions    : scriptcontext exists, rpn is some null-terminated RPN,
***                    which may contain variables, numbers and operators.
*** Postconditions   : si_stack is used to evaluate the RPN.
***                    si_stack[si_SP] contains the result.
***                    If the stack overflows or underflows, or some other
***                    error occurs, a message is displayed, failed is 
***                    performed, and FALSE is returned. Else TRUE is returned.
***
*******************************************************************************/

static int evaluateRPN(char *rpn)
{
char *r,*varname;
word16 num1, num2;
struct var *currvar;

  for (r=rpn; (*r) ;) {

    /* Handle whitespace by skipping over it */
    if (p_isspace(*r)) {
      r++;
      continue;
    }

    /* Handle numbers */
    if (p_isdigit(*r)) {
      for (num1=0; (*r) && p_isdigit(*r); r++) {
        num1*=10;
        num1+=(*r - '0');
      }
      /* Push onto stack */
      if (!push(num1)) {
        console_printf(sz3PercentS,(UINT)szRPN,szStackOverflow,"(number)");
        goto rpnevalfail;
      }
      continue;
    }

    /* Handle variables */
    if (*r == '$') {
      r++;
      if ((currvar = find_variable(r)) == NULL) {
        varname=ntizeVarname(r,NULL);
        if (varname == NULL)
          console_printf(sz2PercentS,(UINT)szVariable,szNotDelimited);
        else
          console_printf(sz3PercentS,(UINT)szVariable,varname,szNotFound);
        goto rpnevalfail;
      }
      num1 = numeric_variable(r);
      /* Num1 can be zero, so to check whether numeric_variable
         failed, check the interface state */
      if (netif->ni_state == NIS_DOWN)
        goto rpnevalfail;
      /* Push onto stack */
      if (!push(num1)) {
        console_printf(sz3PercentS,(UINT)szRPN,szStackOverflow,"(variable)");
        goto rpnevalfail;
      }
      /* Skip over the variable name in the RPN */
      while (*r && p_isalnum(*r)) r++;
      continue;
    }

    /* Handle operators */
    /* Get binary operands or unary operator operand */
    switch (*r) {
      case '!':
        if (!pop(&num1))
          goto rpnunderflow;
        break;
      case '+':
      case '-':
      case '*':
      case '/':
      case '%':
      case '=':
      case '<':
      case '>':
        if (!pop(&num1) || !pop(&num2))
          goto rpnunderflow;
        break;
      default:
        /* eh? wossat? terminate */
        console_printf("RPN syntax error '%c'\r\n",(*r));
        goto rpnevalfail;
    }

    /* Now perform operation */
    switch (*r) {
      case '+':
        push(num2+num1);
        break;
      case '-':
        push(num2-num1);
        break;
      case '*':
        push(num2*num1);
        break;
      case '/':
        if (num1 == 0) 
          goto rpndivzero;
        push(num2/num1);
        break;
      case '%':
        if (num1 == 0) 
          goto rpndivzero;
        push(num2%num1);
        break;
      case '=':
        push(num2 == num1);
        break;
      case '!':
        push(! num1);
        break;
      case '<':
        switch(*(r+1)) {
          case '=':
            r++; /* Skip an additional character of operator */
            push(num2 <= num1);
            break;
          default:
            /* < */
            push(num2 < num1);
            break;
        }
        break;
      case '>':
        switch(*(r+1)) {
          case '=':
            r++; /* Skip an additional character of operator */
            push(num2 >= num1);
            break;
          default:
            /* > */
            push(num2 > num1);
            break;
        }
        break;
    }

    /* Skip over the operator in the RPN */
    r++;
  }
  return TRUE;

rpndivzero:
  console_puts("RPN divide by zero");
  goto rpnevalfail;

rpnunderflow:
  console_printf(sz2PercentS,(UINT)szRPN,szStackUnderflow);

rpnevalfail:
  cmd_down(NULL);
  return FALSE;
}


/*******************************************************************************
***
*** Function         : cmd_whatever(p) (the command interpreters)
*** Preconditions    : p is a pointer to the parameter string for the command
*** Postconditions   : The command is processed, and IP is the position in 
***                    the script where interpretation commences, or NULL to 
***                    terminate interpretation.
***
*******************************************************************************/


static void cmd_baudrate(char *p)
{
  if (namecmp(szBaudRates[BR_300],p))
    asy.asy_speed = P_BAUD_300;
  else if (namecmp(szBaudRates[BR_600],p))
    asy.asy_speed = P_BAUD_600;
  else if (namecmp(szBaudRates[BR_1200],p))
    asy.asy_speed = P_BAUD_1200;
  else if (namecmp(szBaudRates[BR_2400],p))
    asy.asy_speed = P_BAUD_2400;
  else if (namecmp(szBaudRates[BR_4800],p))
    asy.asy_speed = P_BAUD_4800;
  else if (namecmp(szBaudRates[BR_9600],p))
    asy.asy_speed = P_BAUD_9600;
  else if (namecmp(szBaudRates[BR_19200],p))
    asy.asy_speed = P_BAUD_19200;
  else if (namecmp(szBaudRates[BR_38400],p))
    asy.asy_speed = P_BAUD_38400;
  else if (namecmp(szBaudRates[BR_57600],p))
    asy.asy_speed = P_BAUD_56000;
  else if (namecmp(szBaudRates[BR_115200],p))
    asy.asy_speed = P_BAUD_115000;
  else {
    console_printf("baud rate \"%s\" not supported\r\n",(UINT)p);
    cmd_down(NULL);
    return;
  }
  if (asy_open() < 0) {
    console_puts(szAsyChangeFailure);
    cmd_down(NULL);
  }
  else
    next_line();
}


static void cmd_beep(char *p)
{
  beep();
  next_line();
}


static void cmd_capture(char *p)
{
  /* If we're already capturing, close the old file */
  if (sc.si_capturing) {
    p_close(sc.si_captureH);
    sc.si_captureH = NULL;
    sc.si_capturing = FALSE;
  }
  /* Append to the new capture file, if one is specified */
  if (p_slen(p)) {
    if (p_open(&sc.si_captureH, p, P_FAPPEND) != 0) {
      console_printf("could not capture to %s\r\n", (UINT)p);
      cmd_down(NULL);
      return;
    }
    sc.si_capturing = TRUE;
  }
  next_line();
}


static void cmd_clear(char *p)
{
struct var *currvar, *parentvar;
char *varname;

  if (*p!='$') {
    console_printf(sz3PercentS,(UINT)"clear:",(UINT)szVariable,szExpected);
    cmd_down(NULL);
    return;
  }

  p++; /* Get past the $ */

  if ((currvar = find_variable(p)) == NULL) {
    varname=ntizeVarname(p,NULL);
    if (varname == NULL)
      console_printf(sz2PercentS,(UINT)szVariable,szNotDelimited);
    else
      console_printf(sz3PercentS,(UINT)szVariable,varname,szNotFound);
    cmd_down(NULL);
  }
  else {
    /* We can't clear system variables */
    if (currvar->v_flags & VAR_SYSTEM) {
      varname=ntizeVarname(p,NULL);
      if (varname == NULL)
        console_printf(sz2PercentS,(UINT)szVariable,szNotDelimited);
      else
        console_printf(sz3PercentS,(UINT)szVariable,varname,"cannot be cleared");
      cmd_down(NULL);
    }
    else {
      /* Is it the head, or in the tail? */
      if (sc.si_varlist == currvar)
        sc.si_varlist = currvar->v_next;
      else {
        for (parentvar=sc.si_varlist; parentvar->v_next != currvar; 
             parentvar=parentvar->v_next);
        parentvar->v_next = currvar->v_next;
      }
      p_free(currvar);
    }
  }
  next_line();
}


static void cmd_compare(char *p)
{
struct var *currvar;
char *varname, *varname2;
  p++; /* Get past the $ */
  varname = p;
  /* Advance to the character after the whitespace after the end of the 
     variable name */
  while (*p && !p_isspace(*p))
    p++;
  if (*p) 
    p++;

  currvar = find_variable(varname);
  if (currvar == NULL) {
    varname2=ntizeVarname(varname,NULL);
    if (varname2 == NULL)
      console_printf(sz2PercentS,(UINT)szVariable,szNotDelimited);
    else
      console_printf(sz3PercentS,(UINT)szVariable,varname2,szNotFound);
    cmd_down(NULL);
  }
  else {
    if (!push(p_scmp(currvar->v_contents, p)==0)) {
      console_printf(sz2PercentS,(UINT)szStackOverflow,"on compare");
      cmd_down(NULL);
    }
    else
      next_line();
  }
}


static void cmd_databits(char *p)
{
  if (*p == '7')
    asy.asy_databits = P_DATA_7;
  else if (*p == '8')
    asy.asy_databits = P_DATA_8;
  else {
    console_printf("data bits \"%s\" not supported\r\n",(UINT)p);
    cmd_down(NULL);
    return;
  }
  if (asy_open() < 0) {
    console_puts(szAsyChangeFailure);
    cmd_down(NULL);
  }
  else
    next_line();
}


static void cmd_detectcarrier(char *p)
{
  switch (p_toupper(*p)) {
    case 'Y': asy.asy_carrierdetect = 1; break;
    case 'N': asy.asy_carrierdetect = 2; break;
    default:
      console_puts("carrierdetect must be Y or N");
      cmd_down(NULL);
      return;
  }
  if (asy_open() < 0) {
    console_puts(szAsyChangeFailure);
    cmd_down(NULL);
  }
  else
    next_line();
}


static void cmd_detectdevice(char *p)
{
  switch (p_toupper(*p)) {
    case 'Y': asy.asy_devicedetect = 1; break;
    case 'N': asy.asy_devicedetect = 2; break;
    default:
      console_puts("devicedetect must be Y or N");
      cmd_down(NULL);
      return;
  }
  if (asy_open() < 0) {
    console_puts(szAsyChangeFailure);
    cmd_down(NULL);
  }
  else
    next_line();
}


static void cmd_discard(char *p)
{
  sc.si_inputstate = SIS_DISCARD;
  next_line();
}


static void cmd_discarduntil(char *p)
{
  sc.si_inputstate = SIS_DISCARDUNTIL;
  expandCC(p, &sc.si_untilchar);
  next_line();
}


static void cmd_dump(char *p)
{
register struct var *v;
  for (v=sc.si_varlist; v; v=v->v_next) {
    console_printf(sz3PercentS, (UINT)v->v_name, szEquals, v->v_contents);
    /* Don't forget capturing */
    if (sc.si_capturing) {
      p_write(sc.si_captureH, v->v_name, p_slen(v->v_name));
      p_write(sc.si_captureH, szEquals, p_slen(szEquals));
      p_write(sc.si_captureH, v->v_contents, p_slen(v->v_contents));
      p_write(sc.si_captureH, szCRLF, p_slen(szCRLF));
    }
  }
  next_line();
}


static void cmd_drop(char *p)
{
word16 tos;
  pop(&tos);
  next_line();
}


static void cmd_echo(char *p)
{
  if (!expandTEXT(p)) 
    return;
  msg_event(EVT_UI_TERMDATA, sc.si_expandbuf, sc.si_expandbuflen);
  /* Of course, this means that your application will have to support handling
   * of TERMDATA, but then, you're following the example code, aren't you? */
  if (sc.si_capturing) 
    p_write(sc.si_captureH, sc.si_expandbuf, sc.si_expandbuflen);
  next_line();
}


static void cmd_eval(char *p)
{
  if (!evaluateRPN(p))
    return;
  next_line();
}


static void cmd_down(char *p)
{
int tidy = netif->ni_state != NIS_DOWN;
  /* Kludge warning: this code is cursed */
  /* Can use the interface state to detect whether cmd_down has been used.
     Some routines, like numeric_variable() can return 0 for a variable
     holding zero (or a failure and a return of NULL), so you have to check 
     this... */
  if (netif->ni_dlp == NIDLP_PPP) 
    ppp_getevent(EVENT_CLOSE);
  else
    netif->ni_state = NIS_DOWN;
  asy_sethandler();

  /* Tidy up the script interpreter context 
     cmd_down could be called more than once on a failure, for example, when
     a variable is used in an RPN expression where it is not numeric. This
     routine will be called once by numeric_variable and then by evaluateRPN
     so we don't want to tidy up twice. So, check the interace state.
     See. Told you it was cursed. */
  if (tidy) {
    if (sc.si_IP < (sc.si_scriptbuf+sc.si_scriptsize)) 
      console_printf("cmd_down: Failed on command: %s\r\n", (UINT)sc.si_IP);
    else
      console_puts("cmd_down: End of script");
    script_finish();
    nif_close(netif);
    /* nif_close will do a msg_event(EVT_NET_DOWN, NULL, 0); */
  }
}


static void cmd_doze(char *p)
{
  sc.si_inputstate = SIS_BUILDUP;
  sc.si_sleepstate = SSS_DOZE;
  next_line();
}


static void cmd_dozeuntil(char *p)
{
  sc.si_inputstate = SIS_BUILDUP;
  sc.si_sleepstate = SSS_DOZEUNTIL;
  expandCC(p, &sc.si_untilchar);
  next_line();
}


static void cmd_falsegoto(char *p)
{
word16 tos;
  if (!pop(&tos)) {
    console_printf(sz3PercentS, (UINT)"falsegoto", p, szStackUnderflow);
    cmd_down(NULL);
  }
  if (!tos)
    cmd_goto(p);
  else
    next_line();
}


static void cmd_flush(char *p)
{
  sc.si_varlist->v_index = 0;
  next_line();
}


static void cmd_goto(char *p)
{
int i;
struct label *label;
  for (label=sc.si_labelptrs, i=0; i<sc.si_numlabels; i++,label++) {
    if (namecmp(label->l_name,p)) {
      sc.si_IP = label->l_ptr;
      return;
    }
  }
  console_printf("goto label \"%s\" %s", (UINT)p, szNotFound);
  cmd_down(NULL);
}


static void cmd_handshaking(char *p)
{
  if (namecmp("rtscts",p))
    asy.asy_handshake = RTSCTS;
  else if (namecmp("xonxoff",p))
    asy.asy_speed = XONXOFF;
  else if (namecmp("ignore",p))
    asy.asy_speed = XONXOFF;
  else {
    console_printf("handshaking \"%s\" not supported\r\n",(UINT)p);
    cmd_down(NULL);
    return;
  }
  if (asy_open() < 0) {
    console_puts(szAsyChangeFailure);
    cmd_down(NULL);
  }
  else
    next_line();
}

static void cmd_infomsg(char *p)
{
  if (!expandTEXT(p)) 
    return;
  sc.si_expandbuf[sc.si_expandbuflen]='\0';
  msg_event(EVT_UI_MESSAGE, sc.si_expandbuf, sc.si_expandbuflen+1);
  /* Of course, this means that your application will have to support handling
   * of MESSAGES, but then, you're following the example code, aren't you? */
  if (sc.si_capturing) 
    p_write(sc.si_captureH, sc.si_expandbuf, sc.si_expandbuflen);
  next_line();
}

static void cmd_isempty(char *p)
{
struct var *currvar;
char *varname;

  if (*p!='$') {
    console_printf(sz3PercentS,(UINT)"isempty:",szVariable,szExpected);
    cmd_down(NULL);
    return;
  }

  p++; /* Get past the $ */

  if ((currvar = find_variable(p)) == NULL) {
    varname=ntizeVarname(p,NULL);
    if (varname == NULL)
      console_printf(sz2PercentS,(UINT)szVariable,szNotDelimited);
    else
      console_printf(sz3PercentS,(UINT)szVariable,varname,szNotFound);
    cmd_down(NULL);
  }
  push(currvar->v_index == 0);
  next_line();
}


static void cmd_nextphone(char *p)
{
register struct var *vpn, *vp, *vpi;

  vpn=find_variable("phoneNumber");
  vp=find_variable("phone");
  if (++vpn->v_contents[0] == '4')
    vpn->v_contents[0]='1';
  switch (vpn->v_contents[0]) {
    case '1': vpi=find_variable("phone1"); break;
    case '2': vpi=find_variable("phone2"); break;
    case '3': vpi=find_variable("phone3"); break;
  }
  p_scpy(vp->v_contents, vpi->v_contents);
  vp->v_index = vpi->v_index;
  next_line();
}


static void cmd_parity(char *p)
{
  if (namecmp(szIgnoreParam,p))
    asy.asy_speed = IGNORE;
  else if (namecmp("even",p))
    asy.asy_speed = EVENPAR;
  else if (namecmp("odd",p))
    asy.asy_speed = ODDPAR;
  else if (namecmp("none",p))
    asy.asy_speed = NOPAR;
  else {
    console_printf("parity \"%s\" not supported\r\n",(UINT)p);
    cmd_down(NULL);
    return;
  }
  if (asy_open() < 0) {
    console_puts(szAsyChangeFailure);
    cmd_down(NULL);
  }
  else
    next_line();
}


static void cmd_send(char *p)
{
  if (!expandTEXT(p)) 
    cmd_down(NULL);
  if (asy_write((BYTE *)sc.si_expandbuf, sc.si_expandbuflen) >= 0) 
    next_line();
  else {
    console_puts("line failure on send");
    cmd_down(p);
  }
}


static void cmd_set(char *p)
{
struct var *currvar;
char varname[SI_NAMEMAXSIZE+1];
char szbuf[20];
word16 tos;

  if (*p!='$') {
    console_printf(sz3PercentS,(UINT)"set:",szVariable,szExpected);
    cmd_down(NULL);
    return;
  }

  p++; /* Get past the $ */
  if (ntizeVarname(p, varname) == NULL) {
    console_printf(sz2PercentS,(UINT)szVariable, szNotDelimited);
    cmd_down(NULL);
    return;
  }

  /* Advance to the character after the whitespace after the end of the 
     variable name */
  while (*p && !p_isspace(*p))
    p++;
  if (*p) 
    p++;

  currvar = find_variable(varname);
  if (currvar == NULL) {
    if ((currvar=create_variable(varname, "", VAR_WRITABLE)) == NULL) {
      console_printf(sz3PercentS,(UINT)szVariable,varname,szOOM);
      /* cmd_down(NULL); already called by create_variable */
      return;
    }
  }
  if (!evaluateRPN(p))
    return;
  if (!pop(&tos)) {
    console_printf(sz3PercentS,(UINT)"set", p, szStackUnderflow);
    cmd_down(NULL);
    return;
  }
  p_atos(szbuf,"%d",tos);
  p_scpy(currvar->v_contents, szbuf);
  currvar->v_index = p_slen(szbuf);
  next_line();
}


static void cmd_setstring(char *p)
{
struct var *currvar;
char varname[SI_NAMEMAXSIZE+1];

  if (*p!='$') {
    console_printf(sz3PercentS,(UINT)"setstring:",szVariable,szExpected);
    cmd_down(NULL);
    return;
  }

  p++; /* Get past the $ */
  if (ntizeVarname(p, varname) == NULL) {
    console_printf(sz2PercentS,(UINT)szVariable, szNotDelimited);
    cmd_down(NULL);
    return;
  }

  /* Advance to the character after the whitespace after the end of the 
     variable name */
  while (*p && !p_isspace(*p))
    p++;
  if (*p) 
    p++;

  currvar = find_variable(varname);
  if (currvar == NULL) {
    if ((currvar = create_variable(varname, p, VAR_WRITABLE)) == NULL) {
      console_printf(sz3PercentS,(UINT)szVariable,varname,szOOM);
      /* cmd_down(NULL); already called by create_variable */
      return;
    }
  }
  else {
    p_scpy(currvar->v_contents, p);
    currvar->v_index = p_slen(p);
  }
  next_line();
}


static void cmd_sleep(char *p)
{
int timeout;

  if (!p_isdigit(*p)) 
    goto cmd_sleep_fail;

  p = numeric_parameter(p, &timeout);

  /* Go to sleep for 'timeout' seconds */
  sc.si_inputstate = SIS_DISCARD;
  sc.si_sleepstate = SSS_SLEEP;
  timer_set(TQ_SCRIPT_WAKEUP, timeout*(10/(int)(timer_interval&0x0000ffff)));
  next_line();
  return;

cmd_sleep_fail:
  console_puts("sleep expects a timeout");
  cmd_down(NULL);
}


static void cmd_splitip(char *p)
{
register struct var *waitstring = sc.si_varlist;
int fieldno,found=FALSE;
struct var *currvar;
char varname[SI_NAMEMAXSIZE+1];
char *iptext, *iptextend, *scan;
int dotcount;

  if (!p_isdigit(*p)) 
    goto cmd_splitip_fail;

  p = numeric_parameter(p, &fieldno);
  if (fieldno<1) {
    console_printf(sz3PercentS,(UINT)"splitip:",szNumeric,szExpected);
    cmd_down(NULL);
    return;
  }

  p++; /* p must have pointeed to a space. 
          The variable name starts immediately afterwards */

  if (*p!='$') {
    console_printf(sz3PercentS,(UINT)"splitip:",szVariable,szExpected);
    cmd_down(NULL);
    return;
  }

  p++; /* Get past the $ */
  if (ntizeVarname(p, varname) == NULL) {
    console_printf(sz2PercentS, (UINT)szVariable, szNotDelimited);
    cmd_down(NULL);
    return;
  }

  currvar = find_variable(varname);
  if (currvar == NULL) {
    if ((currvar = create_variable(varname, "", VAR_WRITABLE)) == NULL) {
      console_printf(sz3PercentS,(UINT)szVariable,varname,szOOM);
      /* cmd_down(NULL); already called by create_variable */
      return;
    }
  }

  /* Find the fieldno'th IP address in $waitstring */
  iptext = scan = waitstring->v_contents;
  iptextend = iptext + waitstring->v_index;


  while (scan < iptextend && fieldno > 0) {

    /* If we're not on a digit, advance */
    if (!p_isdigit(*scan)) {
      scan++;
      continue;
    }

    /* Now on a digit */
    if (!isdottedquad(scan, iptextend-scan)) {
      scan++;
      continue;
    }

    /* Now on a dotted quad */
    if (fieldno > 1) {
      /* Not this DQ. Advance to character after dotted quad */
      while (scan < iptextend && isdqchar(*scan))
        scan++;
      fieldno--;
      continue;
    }
    else {
      /* This is the one to copy */
      found = TRUE;
      currvar->v_index = 0;
      dotcount = 0;
      for (; scan < iptextend && isdqchar(*scan); 
             scan++, currvar->v_index++) {
        if (*scan == '.')
          dotcount++;
        if (dotcount == 4)
          break;
        currvar->v_contents[currvar->v_index] = *scan;
      }
      break;
    }
  }
  if (!found) {
    console_puts("splitip: couldn't find IP address");
    cmd_down(NULL);
    return;
  }
  next_line();
  return;

cmd_splitip_fail:
  console_puts("splitip expects a field number and a variable name");
  cmd_down(NULL);
}


static void cmd_stopbits(char *p)
{
  if (*p == '1')
    asy.asy_stopbits = 0;
  else if (*p == '2')
    asy.asy_stopbits = P_TWOSTOP;
  else {
    console_printf("stop bits \"%s\" not supported\r\n",(UINT)p);
    cmd_down(NULL);
    return;
  }
  if (asy_open() < 0) {
    console_puts(szAsyChangeFailure);
    cmd_down(NULL);
  }
  else
    next_line(); 
}


static void cmd_up(char *p)
{
  if (netif->ni_state == NIS_SCRIPT) {

    if (netif->ni_dlp == NIDLP_PPP) {
      if (flags.log_flag & LOG_FLAG_FLOWTRACE)
        console_puts("cmd_up: Starting PPP negotiation");
      ppp_getevent(EVENT_OPEN);
    }
    else {
      netif->ni_state = NIS_UP;
      msg_event(EVT_NET_UP, NULL, 0);
    }
    asy_sethandler();

    /* Store variable contents back into the interface etc. */
    /* FIXME */

    /* Tidy up the script interpreter context */
    script_finish();
  }
}


static void cmd_term(char *p)
{
  if (netif->ni_state == NIS_SCRIPT) {
    netif->ni_state = NIS_TERM;
    msg_event(EVT_TERM, NULL, 0);
    asy_sethandler();
    /* Store variable contents back into the interface etc. */
    /* FIXME */

    /* Tidy up the script interpreter context */
    script_finish();
  }
}


static void cmd_truegoto(char *p)
{
word16 tos;
  if (!pop(&tos)) {
    console_printf(sz3PercentS,(UINT)"truegoto", p, szStackUnderflow);
    cmd_down(NULL);
  }
  if (tos)
    cmd_goto(p);
  else
    next_line();
}


static void cmd_wait_compare(void)
{
register struct var *waitstring;
register int found;

  /* Compare waitstring against the expand buffer and push 
     a boolean accordingly */
  waitstring = sc.si_varlist;

  if (sc.si_debuglevel >= SID_TRACEIN) 
    trace_hexdump("trace: $waitstring=", waitstring->v_contents,
                                         waitstring->v_index);

  found = p_bsub(waitstring->v_contents, waitstring->v_index,
                 sc.si_expandbuf, sc.si_expandbuflen) >= 0;
  push (found);
}



static void cmd_wait(char *p)
{
int timeout;

  if (!p_isdigit(*p)) 
    goto cmd_wait_fail;

  p = numeric_parameter(p, &timeout);
  p++; /* p must have pointeed to a space. 
          The string starts immediately afterwards */
  if (!expandTEXT(p)) 
    goto cmd_wait_fail;

  /* Only perform the wait if timeout > 0 */
  if (timeout == 0) 
    cmd_wait_compare();
  else {
    /* Go to sleep for 'timeout' seconds */
    sc.si_inputstate = SIS_BUILDUP;
    sc.si_sleepstate = SSS_SLEEP;
    timer_set(TQ_SCRIPT_WAKEUP, timeout*(10/(int)(timer_interval&0x0000ffff)));
    /* cmd_wait_compare will get called later in script_sleep */
  }
  next_line();
  return;

cmd_wait_fail:
  console_puts("wait expects a timeout and a string");
  cmd_down(NULL);
}




/*******************************************************************************
***
*** Function         : script_start()
*** Preconditions    : The network interface is currently down.
***                    sc.si_scriptname is the name of a file 
***                    containing a script.
*** Postconditions   : script_start is TRUE and the interface's script
***                    context has been successfully initialised. It is the
***                    responsibility of the caller to change the state
***                    variable to NIS_SCRIPT.
***                    script_start is FALSE, and something went wrong.
***
*******************************************************************************/

int script_start(void)
{
TEXT fullname[P_FNAMESIZE];
VOID *fd = NULL;
char *scan,*labelstart;
int i,j;
P_INFO info;

  console_printf("Starting \"%s\"\r\n", (UINT)sc.si_scriptname);

  sc.si_scriptbuf = NULL;
  sc.si_lineptrs = NULL; 
  sc.si_labelptrs = NULL;
  sc.si_varlist = NULL;

  /* Open the script file */
  if (p_fparse(sc.si_scriptname,"LOC::M:\\NET\\",fullname, NULL) != 0) {
    console_printf("Couldn't parse script filename \"%s\"\r\n", (UINT)sc.si_scriptname);
    goto scriptfail;
  }

  if (p_open(&fd,fullname,P_FSTREAM|P_FSHARE) != 0) {
    console_printf("Couldn't open script file \"%s\"\r\n", (UINT)fullname);
    goto scriptfail;
  }

  if (p_finfo(fullname, &info) != 0) { 
    console_printf("Couldn't get script file \"%s\" size\r\n", (UINT)fullname);
    goto scriptfail;
  }

  sc.si_scriptsize = info.size;
  sc.si_scriptbuf = p_alloc((UINT)sc.si_scriptsize);
  if (sc.si_scriptbuf == NULL) {
    console_printf("Couldn't allocate memory (%d bytes) to load script file\r\n",
               (UINT)sc.si_scriptsize);
    goto scriptfail;
  }

  if (sc.si_scriptsize != p_read(fd, (VOID *)sc.si_scriptbuf, (UINT)sc.si_scriptsize)) {
    console_printf("Couldn't read %d bytes of script file\r\n",
               (UINT)sc.si_scriptsize);
    goto scriptfail;
  }

  sc.si_scriptextent = (sc.si_scriptbuf + sc.si_scriptsize);

  p_close(fd);
  fd = NULL; /* So subsequent errors don't re-close it */

  /* How many lines in the script? */
  for (sc.si_numlines=0, scan=sc.si_scriptbuf; scan < sc.si_scriptextent; scan++)
    if (*scan == '\n')
      sc.si_numlines++;

  /* Allocate the line pointer table */
  sc.si_lineptrs = p_alloc (sizeof(char *) * sc.si_numlines);

  if (sc.si_lineptrs == NULL) {
    console_printf("Couldn't allocate %d bytes for line table\r\n",
               (UINT)sizeof(char *) * sc.si_numlines);
    goto scriptfail;
  }

  /* Fill in line ptr table, and convert \n's and \r's to nulls */
  i=0;
  scan=sc.si_scriptbuf;
  do {
    sc.si_lineptrs[i] = scan;
    /* Advance to start of next line */
    while ((scan < sc.si_scriptextent) && (*scan != '\r'))
      scan++;
    /* If we're now on a \r, change it, and the following \n to \0 */
    if (*scan == '\r') {
      *scan++='\0';
      *scan++='\0';
      /* console_printf("line %d [%s]\r\n",i, (UINT)sc.si_lineptrs[i]); */
    }
    i++;
  }
  while (scan < sc.si_scriptextent);

  /* How many labels in the script? */
  for (sc.si_numlabels=0, scan=sc.si_scriptbuf; scan < sc.si_scriptextent; scan++) {
    /* Skip white space */
    if (p_isspace(*scan))
      continue;
    /* Skip comments */
    if (*scan == '#') {
      while (scan < sc.si_scriptextent && *scan)
        scan++;
      scan++; /* Skip over CRLF */
      continue;
    }
    if (*scan == ':' && p_isalnum(*(scan+1))) {
      sc.si_numlabels++;
/*console_printf("found label %d [%s]\r\n", (UINT)sc.si_numlabels, scan);*/
      while (scan < sc.si_scriptextent && *scan)
        scan++;
      scan++; /* Skip over CRLF */
    }
    else {
      while (scan < sc.si_scriptextent && *scan != '\0')
        scan++;
      if (scan < sc.si_scriptextent)
        scan++;
    }
  }
/*console_printf("there are %d labels\r\n",(UINT)sc.si_numlabels);*/
  /* Allocate the label pointer table */
  sc.si_labelptrs = p_alloc (sizeof(struct label) * sc.si_numlabels);

  if (sc.si_labelptrs == NULL) {
    console_printf("Couldn't allocate %d bytes for label table\r\n", 
               (UINT)sizeof(struct label) * sc.si_numlabels);
    goto scriptfail;
  }

  /* Fill in label ptr table */
  for (i=0, scan=sc.si_scriptbuf; scan < sc.si_scriptextent; scan++) {
    /* Skip white space */
    if (p_isspace(*scan))
      continue;
    /* Skip comments */
    if (*scan == '#') {
      while (scan < sc.si_scriptextent && *scan)
        scan++;
      scan++; /* Skip over CRLF */
      continue;
    }
    if (*scan == ':' && p_isalnum(*(scan+1))) {
      labelstart = scan;
      scan++; /* Point to first character */
      for (j=0; j<SI_NAMEMAXSIZE && (p_isalnum(*scan)); j++)
        sc.si_labelptrs[i].l_name[j]=*scan++;
      if (j == SI_NAMEMAXSIZE) {
        console_printf("Label exceeds %d alphanumerics\r\n", (UINT)SI_NAMEMAXSIZE);
        goto scriptfail;
      }
      sc.si_labelptrs[i].l_name[j]='\0';
      sc.si_labelptrs[i].l_ptr = labelstart;
/*console_printf("label %d [%s] [%s]\r\n",(UINT)i,sc.si_labelptrs[i].l_name, labelstart);*/
      while (scan < sc.si_scriptextent && *scan != '\0')
        scan++;
      if (scan < sc.si_scriptextent)
        scan++; /* Added these two */
      i++;
    }
  }


  /* Initialise the variables */
  /* Get the variables from the interface, ISP, and protocol info. */
  /* 'Cos we use it all the time, $waitstring is the first variable */
  if (!create_variable("waitstring", "", FALSE) ||
      !create_variable("IPDNS1", szConvertIPaddr(netif->ni_dnss[0]), TRUE) ||
      !create_variable("IPDNS2", szConvertIPaddr(netif->ni_dnss[1]), TRUE) ||
      !create_variable("IPgateway1", szConvertIPaddr(netif->ni_gateways[0]), TRUE) ||
      !create_variable("IPgateway2", szConvertIPaddr(netif->ni_gateways[1]), TRUE) ||
      !create_variable("IPgateway3", szConvertIPaddr(netif->ni_gateways[2]), TRUE) ||
      !create_variable("IPgateway4", szConvertIPaddr(netif->ni_gateways[3]), TRUE) ||
      !create_variable("IPaddress", szConvertIPaddr(netif->ni_ip), TRUE) ||
      !create_variable("subnetMask", szConvertIPaddr(netif->ni_mask), TRUE) ||
      !create_variable("username", ispc.isp_username, FALSE) ||
      !create_variable("password", ispc.isp_password, FALSE) ||
      !create_variable("phone", "", FALSE) || /* Set by cmd_nextphone */
      !create_variable("phone1", ispc.isp_phone1, FALSE) ||
      !create_variable("phone2", ispc.isp_phone2, FALSE) ||
      !create_variable("phone3", ispc.isp_phone3, FALSE) ||
      !create_variable("phoneNumber", "0", FALSE))
    goto scriptfail;

  cmd_nextphone(NULL);
  
  /* Initialise the instruction and stack pointer */
  sc.si_IP = sc.si_scriptbuf;
  sc.si_SP = 0;

  /* We're not discarding, or capturing, and we're active */
  sc.si_inputstate = SIS_BUILDUP;
  sc.si_sleepstate = SSS_ACTIVE;
  sc.si_capturing = FALSE;

  /* And...... GO! */
  attn_signal(ATTN_SCRIPT);
  return TRUE;


  /* Something has gone wrong */

  scriptfail:

  if (fd != NULL)
    p_close(fd);

  script_finish();

  return FALSE;
}


/*******************************************************************************
***
*** Function         : script_step()
*** Preconditions    : The network interface is in NIS_SCRIPT mode.
*** Postconditions   : The next instruction in the script is stepped through,
***                    and the state of the script interpreter, and interface
***                    is changed as a result.
***                    This routine may change the interface state to either
***                    NIS_UP or NIS_DOWN.
***
*******************************************************************************/

void script_step(void)
{
register struct cmd *cmd;
char *d;
char cmdbuf[255];
char *cs, *cd;
int cl;

  /* Has execution been terminated? */
  if (sc.si_IP == NULL) {
    cmd_down(NULL);
    return; 
  }

  /* Advance IP to the first non-whitespace character or end-of-script */
  while (sc.si_IP < (sc.si_scriptbuf+sc.si_scriptsize) &&
         (*sc.si_IP == '\0' || p_isspace(*sc.si_IP)))
    sc.si_IP++;

  /* Have we dropped off the end of the script? */
  if (sc.si_IP >= (sc.si_scriptbuf+sc.si_scriptsize)) {
    cmd_down(NULL);
    return; 
  }

  /* Trace this command, if necessary */
  if (sc.si_debuglevel >= SID_TRACE) { /* KLUDGE vv */
    for (cl=2,cs=sc.si_IP, cd=cmdbuf; *cs && cl<250; cl++,cs++,cd++)
      *cd=*cs;
    *cd++='\r';
    *cd++='\n';
    msg_event(EVT_UI_TERMDATA, cmdbuf, cl);
  }

  /* Is this a comment or label? Advance to the next line and ignore 
     it if so. */
  if (*sc.si_IP == '#' || *sc.si_IP == ':') 
    next_line();
  else {
    /* Commands are of the form:
       command [ws parameters-to-eol]
       So match the command against the command handling table */
    for (cmd=cmdtable; (cmd->name != NULL); cmd++) {
      if ((d=namecmp(cmd->name, sc.si_IP)) != NULL) {
        /* Match!! d now points to the character after the command.
           Advance d to point to the character after the whitespace
           which follows the command, or EOL. */
        while (*d && p_isspace(*d))
          d++;
        /* Store a pointer to the command definition in the script
           interpreter context, for use in error reporting */
        sc.si_cmd = cmd;
        /* Call the appropriate function. Each command interpreter function 
           will do its own parsing, and bring the interface down, if there's 
           anything wrong. IP will be set to the next instruction. */
        cmd->fn(d);
        break;
      }
    }
    if (cmd->name == NULL) {
      console_print("Unknown command \"");
      console_print(sc.si_IP);
      console_puts("\"");
      cmd_down(NULL);
      return;
    }
  }
}


/*******************************************************************************
***
*** Function         : script_finish()
*** Preconditions    : The network interface has been executing a script.
*** Postconditions   : It isn't any more. The scriptcontext details are freed.
***
*******************************************************************************/

void script_finish(void)
{
register struct var *v,*vnext;

  /* Close capturing */
  if (sc.si_capturing) {
    p_close(sc.si_captureH);
    sc.si_captureH = NULL;
  }

  if (sc.si_scriptbuf != NULL) {
    p_free(sc.si_scriptbuf);
    sc.si_scriptbuf = NULL;
  }

  if (sc.si_lineptrs != NULL) {
    p_free(sc.si_lineptrs);
    sc.si_lineptrs = NULL;
  }

  if (sc.si_labelptrs != NULL) {
    p_free(sc.si_labelptrs);
    sc.si_labelptrs = NULL;
  }

  for (v=sc.si_varlist; v != NULL;) {
    vnext=v->v_next;
    p_free(v);
    v=vnext;
  }
  sc.si_varlist = NULL;

  /* There is no more interpreting to do */
  attn_remove(ATTN_SCRIPT);

}


/*******************************************************************************
***
*** Function         : script_handler()
*** Preconditions    : The network interface is executing a script,
***                    asy_serLength bytes of received serial data are stored
***                    in asy_serBuffer.
*** Postconditions   : If the script interpreter is discarding, the chars are
***                    discarded, otherwise they are added to $waitstring. If 
***                    this fills up, terminate.
***
*******************************************************************************/

void script_handler(void)
{
register struct var *waitstring;
register UWORD len = asy.asy_serLength;
register UBYTE *serbuf = asy.asy_serBuffer;
register int i;

  /* Dozing? */
  if (sc.si_sleepstate == SSS_DOZE) {
    sc.si_sleepstate = SSS_ACTIVE;
    attn_signal(ATTN_SCRIPT);
  }

  /* Doze until? */
  if (sc.si_sleepstate == SSS_DOZEUNTIL) {
    /* Scan the input data waiting for si_untilchar */
    for (i=0; i<len; i++)
      if (serbuf[i] == sc.si_untilchar) {
        sc.si_sleepstate = SSS_ACTIVE;
        attn_signal(ATTN_SCRIPT);
        break;
      }
  }

  /* Discard Until? */
  if (sc.si_inputstate == SIS_DISCARDUNTIL) {
    /* Scan the input data waiting for si_untilchar */
    for (i=0; i<len; i++)
      if (serbuf[i] == sc.si_untilchar) {
        sc.si_inputstate = SIS_BUILDUP;
        /* And adjust the serbuf and len to the character after untilchar */
        len -= i+1;
        serbuf += i+1;
        break;
      }
    if (i == len) /* Not found */
      return;
  }

  /* Discarding? */
  if (sc.si_inputstate == SIS_DISCARD)
    return;

  /* Capturing? */
  if (sc.si_capturing) 
    p_write(sc.si_captureH, serbuf, len);

  /* Trace + Input? */
  if (sc.si_debuglevel >= SID_TRACEIN) 
    trace_hexdump("trace: received ", (char *)serbuf, len);

  /* Overflow receive buffer? Remember, waitstring is the first variable in
     the list... */
  waitstring = sc.si_varlist;
  if ((waitstring->v_index + len) < SI_VARIABLELENGTH) {
    p_bcpy(waitstring->v_contents + waitstring->v_index, serbuf, len);
    waitstring->v_index += len;
  }
  else {
    console_puts("$waitstring overflow");
    cmd_down(NULL);
  }
}




/*******************************************************************************
***
*** Function         : script_process()
*** Preconditions    : A script event has been received by the main loop, and
***                    this is indicated by this call. 
*** Postconditions   : The script currently being interpreted is stepped.
***
*******************************************************************************/

void script_process()
{
  if (netif->ni_state == NIS_SCRIPT && sc.si_sleepstate == SSS_ACTIVE) {
    script_step();
    attn_signal(ATTN_SCRIPT);
  }
}



/*******************************************************************************
***
*** Function         : script_wakeup()
*** Preconditions    : The script currently sleeping now needs to wake up.
*** Postconditions   : The script currently sleeping is woken.
***                    Sleep commands are sleep and wait: they are processed 
***                    here.
*** Notes            : This is called from script_sleep() below, when it is
***                    the right time to wake up. The approach to timers will
***                    change so that all stack timer events are queued, so the
***                    "poll every half second" approach will be removed. Then,
***                    this routine will be called when a TQ_SCRIPT_WAKE timer
***                    queue event comes to the head of the timer queue.
***
*******************************************************************************/

void script_wakeup()
{
  /* Sleepytime no more - Wakey Wakey!! */
  sc.si_sleepstate = SSS_ACTIVE;
  attn_signal(ATTN_SCRIPT);

  /* wait and sleep are different. sleep discards, wait builds up. 

     Now we've woken up, if we were in a wait, the input state will be 
     SIS_BUILDUP, and we need to compare the contents of waitstring 
     against the expand buffer, and push TRUE or FALSE onto the stack 
     depending on whether the expand buffer is a subset of waitstring.

     If we were in a sleep, we've been discarding, and so must now go
     back to building up.
  */

  switch (sc.si_inputstate) {
      
    case SIS_BUILDUP:  /* Was waiting */
      cmd_wait_compare();
      break;

    case SIS_DISCARD: /* Was sleeping */
      /* Just go back to collecting serial data into $waitstring */
      sc.si_inputstate = SIS_BUILDUP;
      break;
  }
}
