/*******************************************************************************
***
*** Filename         : socket.c
*** Purpose          : Handlers for Sockets IPC Messages
*** Author           : Matt J. Gumbley
*** Created          : 09/01/98
*** Last updated     : 19/04/98
***
********************************************************************************
***
*** Modification Record
***
*******************************************************************************/

#include "stack.h"
#include "..\include\sys\socket.h"


/* There is an array of sockets, pointing to UDP or TCP socket data */
struct socket {
  BYTE s_inuse;    /* NULL, or SOCK_DGRAM, SOCK_STREAM, etc */
  VOID *s_control; /* Points to UPQ, or TCB */
  /* Other socket options can go here */
};

LOCAL_D struct socket *sockets;
LOCAL_D int sockets_max;

GLDEF_C INT socket_init(void)
{

  sockets_max = udp_config_used.udpc_UPPS + tcp_config_used.tcpc_NTCBS;
  sockets = NULL;
  sockets = p_alloc(sizeof(struct socket) * sockets_max);
  if (sockets == NULL) {
    console_puts("Couldn't allocate sockets");
    goto initfail;
  }

  p_bfil(sockets, sockets_max * sizeof(struct socket), 0x00);
  return TRUE;

initfail:
  if (sockets != NULL) 
    p_free(sockets);
  return FALSE;
}


/*******************************************************************************
***
*** Function         : socket_unblockpassiveopen(ptcb)
*** Preconditions    : This tcb has a process waiting on an accept. Get a 
***                    connection from the listen queue (there is guaranteed to
***                    be one stored there), create a socket for it, and 
***                    return the descriptor in the client process' status word,
***                    then clear down the PID/statword stores.
*** Postconditions   : 
*** Notes            : This code is only called by socket_accept() and 
***                    tcpi_synrcvd().
***
*******************************************************************************/

GLDEF_C INT socket_unblockpassiveopen(struct tcb *ptcb)
{
int sockno;
word16 pid;
struct socket *sock;
  /* Find a free socket */
  for (sockno=0; sockno<sockets_max; sockno++)
    if (sockets[sockno].s_inuse == 0)
      break;
  if (sockno==sockets_max) {
    return E_NET_MFILE;
  }
  /* Initialise this new socket, pointing to the child TCB */
  sock = &sockets[sockno];
  sock->s_inuse = SOCK_STREAM;
  sock->s_control = queue_dequeue(ptcb->tcb_listenq);
  /* Return the new socket descriptor to the client */
  p_pcpyto((pid=ptcb->tcb_ocpid), (VOID *)ptcb->tcb_ocstatword, 
           (VOID *)&sockno, sizeof(WORD));
  /* Clear these so the next accept that occurs before a connection 
     arrives will be "queued" until one does. */
  ptcb->tcb_ocpid = 0;
  ptcb->tcb_ocstatword = 0;
  p_iosignalbypid(pid);
  return 0; /* Successful completion */
}

/* Notes on these kernel-internal sockets calls.
   1) Parameters starting with cd_ are offsets into the Client Data segment,
      and so the pid parameter should be used with p_pcpyfr to access the 
      client's data segment.
   2) Routines return either 0 to indicate that their actions have been 
      queued, and that the calling program should wait for an asynchronous 
      completion, OR they should return an E_NET error code.
   3) Any call that returns E_NET_UNIMPLEMENTED is not yet finished
*/


GLDEF_C INT socket_accept(UWORD pid, UWORD sockfd, UWORD cd_peeraddr, 
                          UWORD cd_paddrlen, UWORD cd_statword)
{
struct tcb *ptcb;

  switch (sockets[sockfd].s_inuse) {
    case SOCK_DGRAM:
      return E_NET_OPNOTSUPP;
    case SOCK_STREAM:
      ptcb = sockets[sockfd].s_control;
      if (queue_empty(ptcb->tcb_listenq)) {
        /* The client called accept() before any connections had been received.
           So, there are no TCB pointers in the listenq. Block the client and 
           record its PID and status word address. 
        */
        ptcb->tcb_ocpid = pid;
        ptcb->tcb_ocstatword = cd_statword;
        /* Return immediately, and tcpi_synrcvd will write to the client's
           status word when it gets a SYN.
        */
        return 0; /* Successful completion */
      }
      else {
        /* tcpi_synrcvd in tcpin.c has received an incoming SYN, and created a 
           TCB for the connection (if there are any spaces left in the 
           listenq). The address of the new TCB has been enqueued on the 
           listenq, so we just get it, and create a new socket for it, then 
           pass it back to the client, in the status word.
        */
        return socket_unblockpassiveopen(ptcb);
      }
  }
  return E_NET_BADF;
}


GLDEF_C INT socket_bind(UWORD pid, UWORD sockfd, UWORD cd_saddr, 
                        UWORD addrlen, UWORD cd_statword)
{
struct sockaddr_in sin;
struct upq *pupq;
struct tcb *ptcb;
int retval;

TEXT buf[80];

  p_pcpyfr(pid, (VOID *)cd_saddr, &sin, addrlen);
  switch (sockets[sockfd].s_inuse) {
    case SOCK_DGRAM:
      pupq = sockets[sockfd].s_control;
      pupq->upq_lport = sin.sin_port;
      /* See Comer III page 103, "Binding to a well-known address using
         INADDR_ANY". We only store the local port number in the UPQ, since
         the IP layer takes care of whether incoming datagrams are addressed
         to us or not. */
      break;
    case SOCK_STREAM:
      ptcb = sockets[sockfd].s_control;
      ptcb->tcb_lport = sin.sin_port;
      /* So why can we have a local TCP-based IP address? ;-~ */
      p_bcpy(ptcb->tcb_lip, (VOID *)sin.sin_addr, IP_ALEN);

  p_atos(buf,"socket_bind: binding to %s:%d", szConvertIPaddr(ptcb->tcb_lip), ptcb->tcb_lport);
  console_puts(buf);
      break;
    default:
      return E_NET_BADF;
  }
  retval = 0;
  p_pcpyto(pid, (VOID *)cd_statword, &retval, sizeof(WORD));
  p_iosignalbypid(pid);
  return 0; /* Successful completion */
}


GLDEF_C INT socket_close(UWORD pid, UWORD sockfd, UWORD cd_statword)
{
  switch (sockets[sockfd].s_inuse) {
    case SOCK_DGRAM:
      break;
    case SOCK_STREAM:
      break;
    default:
      return E_NET_BADF;
  }
  return E_NET_UNIMPLEMENTED;
}


GLDEF_C INT socket_connect(UWORD pid, UWORD sockfd, UWORD cd_serv_addr, 
                           UWORD addrlen, UWORD cd_statword)
{
struct sockaddr_in sin;
struct upq *pupq;
struct tcb *ptcb;
int retval, ifnum, mss;

  switch (sockets[sockfd].s_inuse) {
    case SOCK_DGRAM:
      pupq = sockets[sockfd].s_control;
      if (pupq->upq_fport == -1)
        return E_NET_ISCONN;
      /* cd_serv_addr specifies the remote endpoint but does not transfer any
         datagrams to it. */
      p_pcpyfr(pid, (VOID *)cd_serv_addr, &sin, addrlen);
      if (sin.sin_family != AF_INET)
        return E_NET_AFNOSUPPORT;
      pupq->upq_lport = udp_nextport();
      pupq->upq_fport = sin.sin_port;
      p_bcpy(pupq->upq_fip, (VOID *)sin.sin_addr, IP_ALEN);
      /* Where do we initialise upq_lip? */
      retval = 0;
      p_pcpyto(pid, (VOID *)cd_statword, &retval, sizeof(WORD));
      p_iosignalbypid(pid);
      return 0; /* Successful completion */
    case SOCK_STREAM:
      ptcb = sockets[sockfd].s_control;
      p_pcpyfr(pid, (VOID *)cd_serv_addr, &sin, addrlen);

      ptcb->tcb_rport = sin.sin_port;
      p_bcpy(ptcb->tcb_rip, (VOID *)&sin.sin_addr, IP_ALEN);

      ptcb->tcb_lport = tcp_nextport();
      ifnum = which_nif((VOID*) &sin.sin_addr, TRUE);
      if (ifnum == NI_LOCAL)
        p_bcpy(ptcb->tcb_lip, ptcb->tcb_rip, IP_ALEN);
      else
        p_bcpy(ptcb->tcb_lip, (VOID *)netif->ni_ip, IP_ALEN);
      ptcb->tcb_pni = &nif[ifnum];

      if (!tcp_sync(ptcb)) {
        return E_NET_NOMEM;
      }

      /* From tcpcon.c */
      if (ifnum == NI_LOCAL)
        mss = ptcb->tcb_pni->ni_mtu-IPMHLEN-TCPMHLEN;
      else
        mss = 536;                            /* RFC 1122 default */
      ptcb->tcb_smss = mss;
      ptcb->tcb_rmss = ptcb->tcb_smss;
      ptcb->tcb_swindow = ptcb->tcb_smss;     /* conservative */
      ptcb->tcb_cwnd = ptcb->tcb_smss;        /* 1 segment */
      ptcb->tcb_ssthresh = 0xFFFF;             /* IP Max window size */
      ptcb->tcb_rnext = 0;
      ptcb->tcb_finseq = 0;
      ptcb->tcb_pushseq = 0;
      ptcb->tcb_flags = TCBF_NEEDOUT|TCBF_FIRSTSEND;
      /*ptcb->tcb_type = TCPT_CONNECTION; * Could be right?!?! */
      ptcb->tcb_ostate = TCPOS_IDLE;
      ptcb->tcb_istate = TCPIS_SYNSENT;
      ptcb->tcb_listenq = NULL;
      tcpo_stat_activeopens++;
      /* Store the process data so we can wake it up when the connection
         either times out, or is made. */
      ptcb->tcb_ocpid = pid;
      ptcb->tcb_ocstatword = cd_statword;
      tcp_kick(ptcb);
console_puts("*** socket_connect returning to application");
      return 0; /* Successful completion */
  }
  return E_NET_BADF;
}


GLDEF_C INT socket_getpeername(UWORD pid, UWORD sockfd, UWORD cd_addr, 
                               UWORD cd_paddrlen, UWORD cd_statword)
{
  if (sockets[sockfd].s_inuse == 0)
    return E_NET_BADF;
  return E_NET_UNIMPLEMENTED;
}


GLDEF_C INT socket_getsockopt(UWORD pid, UWORD sockfd, UWORD level, 
                              UWORD optname, UWORD cd_optval, UWORD cd_optlen, 
                              UWORD cd_statword)
{
  if (sockets[sockfd].s_inuse == 0)
    return E_NET_BADF;
  return E_NET_UNIMPLEMENTED;
}


GLDEF_C INT socket_getsockname(UWORD pid, UWORD sockfd, UWORD cd_addr, 
                               UWORD cd_paddrlen, UWORD cd_statword)
{
  if (sockets[sockfd].s_inuse == 0)
    return E_NET_BADF;
  return E_NET_UNIMPLEMENTED;
}


GLDEF_C INT socket_listen(UWORD pid, UWORD sockfd, UWORD n, UWORD cd_statword)
{
struct tcb *ptcb;
int retval;
  switch (sockets[sockfd].s_inuse) {
    case SOCK_DGRAM:
      return E_NET_OPNOTSUPP;
    case SOCK_STREAM:
      ptcb = sockets[sockfd].s_control;
      if ((ptcb->tcb_listenq = queue_create(n)) != NULL) {
        ptcb->tcb_lqsize = n;
        retval = 0;
        ptcb->tcb_type = TCPT_SERVER;
        ptcb->tcb_istate = TCPIS_LISTEN;
        ptcb->tcb_smss = 0; /* Comer does this in his tcpserver.c code */
      }
      else
        return E_NET_NOMEM;
      p_pcpyto(pid, (VOID *)cd_statword, (VOID *)&retval, sizeof(WORD));
      p_iosignalbypid(pid);
      return 0; /* Successful completion */
  }
  return E_NET_BADF;
}


GLDEF_C INT socket_recvfrom(UWORD pid, UWORD sockfd, UWORD cd_buff, UWORD len,
                            UWORD flags, UWORD cd_from, UWORD cd_fromlen, 
                            UWORD cd_statword)
{
struct tcb *ptcb;
  if (sockets[sockfd].s_inuse == 0)
    return E_NET_BADF;
  switch (sockets[sockfd].s_inuse) {
    case SOCK_DGRAM:
      return E_NET_UNIMPLEMENTED;
    case SOCK_STREAM:
      ptcb = sockets[sockfd].s_control;
      if (ptcb->tcb_istate != TCPIS_ESTABLISHED && 
          ptcb->tcb_istate != TCPIS_CLOSEWAIT)
        return E_NET_NOTCONN;
      break;
  }
  return E_NET_BADF;
}


GLDEF_C INT socket_send(UWORD pid, UWORD sockfd, UWORD cd_buff, UWORD len,
                        UWORD flags, UWORD cd_statword)
{
struct upq *pupq;
/*struct tcb *ptcb;*/
struct buf *pbuf;
struct ip  *pip;
struct udp *pudp;
byte       *ppayload;
int retval;
  switch (sockets[sockfd].s_inuse) {
    case SOCK_DGRAM:
      pupq = sockets[sockfd].s_control;
      if (pupq->upq_fport == -1)
        return E_NET_ISCONN;
      /* Allocate a buf to copy the data to */
      pbuf = buf_alloc(len + IPMHLEN + U_HLEN);
      if (pbuf == NULL)
        return E_NET_NOBUFS;
      pip = (struct ip *)pbuf->data;
      pudp = (struct udp *)pip->ip_data;
      ppayload = (byte *)pudp->udp_data;
      p_pcpyfr(pid, (VOID *)ppayload, (VOID *)cd_buff, len);
      if (!udp_send(pupq->upq_fip, pupq->upq_fport, pupq->upq_lport, 
                    pbuf, len, TRUE /* generate checksum */))
        return E_NET_IO; /* Route failed */
      retval = len;/* Successfully written len bytes */
      p_pcpyto(pid, (VOID *)cd_statword, (VOID *)&retval, sizeof(WORD));
      p_iosignalbypid(pid);
      return 0; /* Successfully completed */
    case SOCK_STREAM:
      /*ptcb = sockets[sockfd].s_control;*/
      return E_NET_UNIMPLEMENTED;
  }
  return E_NET_BADF;
}


GLDEF_C INT socket_sendto(UWORD pid, UWORD sockfd, UWORD cd_buff, UWORD len,
                          UWORD flags, UWORD cd_to, UWORD tolen, 
                          UWORD cd_statword)
{
struct upq *pupq;
/*struct tcb *ptcb;*/
struct buf *pbuf;
struct ip  *pip;
struct udp *pudp;
byte       *ppayload;
struct sockaddr_in fip;
int retval;
IPaddr tempaddr;
  switch (sockets[sockfd].s_inuse) {
    case SOCK_DGRAM:
      pupq = sockets[sockfd].s_control;
      p_pcpyfr(pid, (VOID *)cd_to, (VOID *)&fip, tolen);
      if (fip.sin_family != AF_INET)
        return E_NET_AFNOSUPPORT;
      /* Allocate a buf to copy the data to */
      pbuf = buf_alloc(len + IPMHLEN + U_HLEN);
      if (pbuf == NULL)
        return E_NET_NOBUFS;
      pip = (struct ip *)pbuf->data;
      pudp = (struct udp *)pip->ip_data;
      ppayload = (byte *)pudp->udp_data;
      p_pcpyfr(pid, (VOID *)ppayload, (VOID *)cd_buff, len);
      /* Get round FUBAR compiler */
      p_bcpy(tempaddr, &fip.sin_addr, sizeof (IPaddr));
      if (!udp_send(tempaddr, fip.sin_port, pupq->upq_lport, 
                    pbuf, len, TRUE /* generate checksum */))
        return E_NET_IO; /* Route failed */
      retval = len;/* Successfully written len bytes */
      p_pcpyto(pid, (VOID *)cd_statword, (VOID *)&retval, sizeof(WORD));
      p_iosignalbypid(pid);
      return 0; /* Successfully completed */
    case SOCK_STREAM:
      /*ptcb = sockets[sockfd].s_control;*/
      return E_NET_UNIMPLEMENTED;
  }
  return E_NET_BADF;
}


GLDEF_C INT socket_setsockopt(UWORD pid, UWORD sockfd, UWORD level, 
                              UWORD optname, UWORD cd_optval, UWORD optlen, 
                              UWORD cd_statword)
{
  if (sockets[sockfd].s_inuse == 0)
    return E_NET_BADF;
  return E_NET_UNIMPLEMENTED;
}


GLDEF_C INT socket_shutdown(UWORD pid, UWORD sockfd, UWORD how, 
                            UWORD cd_statword)
{
  if (sockets[sockfd].s_inuse == 0)
    return E_NET_BADF;
  return E_NET_UNIMPLEMENTED;
}


GLDEF_C INT socket_socket(UWORD pid, UWORD domain, UWORD type, 
                          UWORD protocol, UWORD cd_statword)
{
register struct socket *sock;
struct upq *pupq;
struct tcb *ptcb;
int sockno;
  /* Domain will usually be PF_INET; this doesn't really need to be 
     communicated to the stack in this release (since we're only TCP/IP), 
     but this may change later, when the AppleTalk and IrDA people want
     to add to the stack, so pass it anyway.. */
  switch (domain) {
    case PF_INET:
      break;
    /*
    case PF_IRDA:
    case PF_APPLETALK:
    case PF_AX25: ... etc.
    */
    default:
      return E_NET_PROTONOSUPPORT;
  }
    
  /* Find a free socket */
  for (sockno=0; sockno<sockets_max; sockno++)
    if (sockets[sockno].s_inuse == 0)
      break;
  if (sockno==sockets_max) {
    return E_NET_MFILE;
  }
  
  sock = &sockets[sockno];

  /* Socket type is one of the SOCK_x values.
     Perhaps SOCK_RAWINTERFACE should come under its own protocol family, 
     not PF_INET? 
     Protocol is usually zero, but I ignore it for now...
     There should probably be some code to verify and choose parameters,
     using the well-known identifiers for protocols, in the case that
     you're using PF_INET/SOCK_DGRAM or PF_INET/SOCK_STREAM. This is a
     bit long-winded; those wanting to add other protocol familes etc. 
     should look into this. For now, the stack chooses UDP or TCP based 
     on the type, and ignores the rest.
  */
  switch (type) {
    case SOCK_DGRAM:
      /* Search for free UPQ */
      pupq = upq_alloc();
      if (pupq != NULL) {
        sock->s_inuse = type;
        sock->s_control = pupq;
        p_pcpyto(pid, (VOID *)cd_statword, (VOID *)&sockno, sizeof(WORD));
        p_iosignalbypid(pid);
        return 0; /* Successful completion */
      }
      return E_NET_NFILE;
    case SOCK_STREAM:
      /* Search for free TCB */
      ptcb = tcb_alloc();
      if (ptcb != NULL) {
        sock->s_inuse = type;
        sock->s_control = ptcb;
        p_pcpyto(pid, (VOID *)cd_statword, (VOID *)&sockno, sizeof(WORD));
        p_iosignalbypid(pid);
        return 0; /* Successful completion */
      }
      return E_NET_NFILE;
    case SOCK_RAW:
    case SOCK_RAWINTERFACE:
      /* Not sure how we'll handle these yet */
      return E_NET_PROTONOSUPPORT; 
      break;
    default:
      return E_NET_PROTONOSUPPORT;
  }
}

