/*******************************************************************************
***
*** Filename         : stack.c
*** Purpose          : Top-level functions for the PsiStack stack process.
*** Author           : Matt J. Gumbley
*** Created          : 23/07/97
*** Last updated     : 14/04/98
***
********************************************************************************
***
*** Modification Record
***
*******************************************************************************/

#include "stack.h"

GLREF_D UWORD DatLocked;            /* non-zero when application "busy" */

GLDEF_D char stack_active = TRUE; /* For the main loop to be active */
/*GLDEF_D TEXT *szSTACK = "STACK";  ** eventually */
GLDEF_D TEXT *szSTACK = NULL;

#define BR_300 0
#define BR_600 1
#define BR_1200 2
#define BR_2400 3
#define BR_4800 4
#define BR_9600 5
#define BR_19200 6
#define BR_38400 7
#define BR_57600 8
#define BR_115200 9
LOCAL_D TEXT *szBaudRates[]= {
  "300", "600", "1200", "2400", "4800", "9600",
  "19200", "38400", "57600", "115200"
};

LOCAL_D TEXT *szYes = "Yes";
LOCAL_D TEXT *szNo = "No";
LOCAL_D TEXT *szIgnore = "Ignore";
LOCAL_D TEXT *szUnimplemented = "Not implemented yet";

LOCAL_D BYTE stack_usecount;

/*******************************************************************************
***
*** Function         : stack_stats()
*** Preconditions    : 
*** Postconditions   : 
***
*******************************************************************************/

GLDEF_C VOID stack_stats()
{
  timer_stats();
  udp_stats();
  icmp_stats();
  ip_stats();
  nif_stats();
  slip_stats();
  ppp_stats();
  buf_stats();
}


/*******************************************************************************
***
*** Function         : stack_close()
*** Preconditions    : 
*** Postconditions   : 
***
*******************************************************************************/

GLDEF_C VOID stack_close(INT num)
{
  if (flags.misc_flag & MISC_FLAG_ENDSTATS) {
    console_puts("Generating stats...");
    stack_stats();
  }
  /* FIXME: need to do udp_destroy(), tcp_destroy(), socket_destroy() here */
  config_save();
  nif_destroy();
  buf_destroy();
  timer_cancel();
  msg_cancel();
  db_unload();
  p_sleep(10); /* Just to give you a chance to see any final error messages */
  p_exit(num);     /* Goodnight Vienna */
}



/*******************************************************************************
***
*** Function         : main
*** Preconditions    : Eh-oh!
*** Postconditions   : Teletubbies go to sleep.
***
*******************************************************************************/

GLDEF_C VOID main(VOID)
{
HANDLE hDBMS;
static UWORD largestbuf=0;

  /* We have to be really careful not to cause a context switch here, before we
   * initialise messaging, otherwise the stack_load() API call won't return
   * immediately after the stack has finished initialising (and loading the
   * DBMS). So, initialise messaging as soon as we can */

  /* We're running at priority 0x81 if started by the stack API, but
     let's just make sure, in case the user starts us... */
  p_setpri(p_getpid(), 0x81);

  p_unmarka();

  console_setprefix(szSTACK); /* Doesn't context switch */

  if (!msg_init())
    fatal("Couldn't init messaging");
  /* If msg_init fails, then it's OK to context switch to the console, 
     since without messaging, everything will be completely pear-shaped 
     anyway :-)

  /* Which type of machine are we running on? */
  machine_check();
  if (machine_type == MACH_UNKNOWN)
    fatal("Unknown machine!");

  console_printf("PsiStack TCP/IP v%d.%02d%c %s\r\n", version/100, version%100, 
          releaselevel, szNickname);
  console_puts("Built on " __DATE__);

  if ((hDBMS=dbms_load())<0) {
    console_printf("STACK: couldn't load DBMS, returned %d\r\n", hDBMS); 
    fatal("Couldn't load DBMS");
  }

  if (timer_open()<0)
    fatal("Couldn't open timer device");
  timer_queue(); 


  if (!buf_init()) 
    fatal("Couldn't init memory manager");

  asy_init();
  isp_init();
  script_init();

  if (!nif_init()) 
    fatal("Couldn't init network I/F's");

  config_load(); /* May alter StatWinState etc. */
  console_timestamp(flags.log_flag & LOG_FLAG_TIMESTAMP);
    

  if (!udp_init()) 
    fatal("Couldn't init UDP");

  if (!tcp_init()) 
    fatal("Couldn't init TCP");

  if (!socket_init())
    fatal("Couldn't init sockets");

  /* No users so far */
  stack_usecount = 0;
  console_puts("All OK... off we go...");

  /* Drop to the same priority as the setup program / user app */
  p_setpri(p_getpid(), 0x80);

  /* Main Loop */
  
  while (stack_active) {
    p_iowait();

    /* Serial? */
    if (asy.asy_serH) { /* If a port is open on this interface */
      switch (asy.asy_serStatus) {

        case E_FILE_PENDING:
          /* Request still outstanding - do nothing */
          break;

        case E_FILE_LINE:
          /* DTR or DCD was dropped */
          beep();
          nif_close(netif);
          console_puts("Line was dropped");
          break;

        case E_FILE_PARITY:
          beep();
          console_puts("Parity error");
          asy_queue();
          break;

        case E_FILE_FRAME:
          beep();
          console_puts("Frame error: check baud rate");
          asy_queue();
          break;

        case E_FILE_OVERRUN:
          beep();
          console_puts("Overrun error");
          asy_queue();
          break;

        case 0:
          /* Handle serial data */
          p_tickle();

          /* Collect any further outstanding serial data */
          asy_process();

          /* Send all this data to an appropriate handler */
          (*asy.asy_serHandler)();

          /* And queue another read */
          asy_queue();
          break;

        default: /* Some other asynchronous response code */
          console_printf("main_loop: Unknown P_FREAD completion code: %d\r\n", 
                          asy.asy_serStatus);
          asy_queue();
          break;
      }
    }

    /* Timer queue event expiry? */
    if (timer_status != E_FILE_PENDING) {
      timer_process();
      timer_queue();
    }

    /* Messages from clients? */
    if (msg_status != E_FILE_PENDING) {
      msg_process();
      msg_queue();
    }

    /* Any other attention grabbing events signalled since last time */
    if (attn != 0)
      attn_process();
  }
  stack_close(0);
}

