/*******************************************************************************
***
*** Filename         : stackmsg.h
*** Purpose          : IPC message format and types for PsiStack. This file is
***                    a C header file, but can also be used by the 
***                    experimental AUTOGEN program (a Perl script) which takes
***                    these definitions and converts them into C and OPL code 
***                    to perform the message passing.
*** Author           : Matt J. Gumbley, Patrick Robbe
*** Created          : 28/06/97
*** Last updated     : 24/08/98
***
********************************************************************************
***
*** Modification Record
***
*******************************************************************************/

#ifndef STACKMSG_H
#define STACKMSG_H

/* Needs EPOC.H */

/* The format of messages between PsiStack and any clients */
typedef struct stkmsg {
  UWORD  sm_uwords[7];
} STKMSG;

/* Message types, stored in STKMSG.sm_e_message.type (STKMSG.sm_type) */

/* Notes on the syntax of the embedded AUTOGEN commands
   ====================================================
   AUTOGEN commands are always embedded in C comments so that the C compiler
   ignores them. 

   When parsing this file, AUTOGEN ignores everything until it sees a line that
   starts with the word AUTOGEN. That allows me to write AUTOGEN commands in
   this comment block and have them ignored. The AUTOGEN command would be placed
   in a C comment.... Any more AUTOGEN commands turn processing off.

   OPLINCLUDE filename, used to include a block of OPL from another file, 
   without parsing that OPL code. This is used to include the stack_sendmsg code
   which we have to write by hand, stack_load, etc.

   CINCLUDE filename, does the same for the hand-written C.

   #define SMT_[CODE_]??????????? ID, triggers generation of a function, in 
   either C or OPL, depending on which language was selected. AUTOGEN can
   generate code for the STACK (no CODE), SOCKETS (CODE=SL), DNS (CODE=NDB),
   CONSOLE (CODE=CONSOLE??) libraries. If the SMT_ name does not match the
   code, this #define is skipped. The SMT_???? code is turned into the function
   name.

   The comment following the #define defines the parameters for the function,
   their types and order. AUTOGEN counts the inputs, and builds a function
   with that number of inputs, storing each input in turn in the STKMSG.

   If any of the inputs contain the string 'statusword', then this variable
   will be pre-initialised to E_FILE_PENDING.

   Inputs are declared with an 'Inputs:' line, each input is declared as
   sm_uwords[x] = variable type, name. comment
   The number denoted here by [x] gives the place of this input in the 
   sm_uwords array. The spacing, the equals sign, the comma after the variable
   name, and the full stop after the name are important.

   If there are no inputs, the line must look like:
   Inputs: None
   You will get a VOID parameter list for that.

   Input building stops when the Outputs: line is seen. The function is then
   written in the chosen language.

   OK, enough of this, let's let AUTOGEN loose:
AUTOGEN
   It is now listening for commands...
*/


   
/******************************************************************************/
/*************                                                 ****************/
/*************  S T A C K   C O N T R O L   F U N C T I O N S  ****************/
/*************                                                 ****************/
/******************************************************************************/

#define SMT_free0 0
/* This used to be a message, but doesn't do anything now. Still have to tell
   AUTOGEN about it though...
   Inputs: None
   Output: nothing, only the blank spaces and Output: are used
*/

#define SMT_QUEUE_EVENT_REQUEST 1
/* Processes send this to the stack to queue a request to be notified of 
   certain events. This is done using asynchronous messaging.
   When these events occur, the stack modifies the caller's status word, and
   if the event warrants it, copies data to the caller's buffer.
   The client can specify a NULL buffer, in which case, only the status word 
   will be set.
   Inputs: sm_uwords[0] = UWORD, eventmask. Interest mask, EVT_??? flag above...
           sm_uwords[1] = WORD *, pstatusword. Address of status word in 
                          client memory
           sm_uwords[2] = VOID *, pbuffer. Address of data buffer in client 
                          memory or NULL
           sm_uwords[3] = UWORD, length. Length of client data buffer
           sm_uwords[4] = UWORD *, plenbuffer. Address of returned-length 
                          buffer in client memory or NULL
   Output: TRUE if the interest was registered, or FALSE if 
   there was no more space in the kernel interest table.
*/


#define SMT_CANCEL_EVENT_REQUEST 2
/* Processes send this to the stack to cancel the queued event request.
   Inputs: None
   Output: TRUE if the process had registered any interests, FALSE 
   if they hadn't.
*/


#define SMT_UNLOAD 3
/* Sent by the UI to force an unload of the stack, for instance, when a 
   client crashes and leaves the open count positive, so that the stack 
   would never close.
   Inputs: None
   Output: TRUE.
*/


#define SMT_GET_NETIFSTATE 4
/* Get the current state of the network interface
   Inputs: None
   Output: NIS_UP, NIS_DOWN, NIS_TERM, etc. i.e. netif->ni_state
*/

#define SMT_OPENCLOSE_NETIF 5
/* Open the network interface 
   Inputs: sm_uwords[0] = TRUE for open; FALSE for close.
           sm_uwords[1] = New state for open, NIS_
   Output: TRUE or FALSE, depending on whether the interface opened or not.
*/

#define SMT_SET_MISC 6
/* Set the miscellaneous flags 
   Inputs: sm_uwords[0] = new misc flags
   Output: TRUE.
*/

#define SMT_GET_MISC 7
/* Get the misc flags
   Inputs: None
   Output: the current misc flags
*/

#define SMT_SET_NETIF 8
/* Change the internal settings of the network interface.
   Inputs: sm_uwords[0] = address of a nif structure in client memory
   Output: TRUE or FALSE, depending on whether the change is 
   OK or not; i.e. the change can only take place when the network 
   interface is down.
*/

#define SMT_GET_NETIF 9
/* Get the internal settings of the network interface.
   Inputs: sm_uwords[0] = address of a nif structure in client memory
   Output: TRUE.
*/


#define SMT_SET_ASY 10
/* Set the serial settings of the network interface device.
   Inputs: sm_uwords[0] = address of an asy structure in client memory
   Output: TRUE or FALSE, depending on whether the change is 
   OK or not; i.e. the change can only take place when the network 
   interface is down.
*/

#define SMT_GET_ASY 11
/* Get the internal settings of the serial device.
   Inputs: sm_uwords[0] = address of an asy structure in client memory
   Output: TRUE.
*/

#define SMT_OPENCLOSE_ASY 12
/* Open or close the serial port
   Inputs: sm_uwords[0] = TRUE for open; FALSE for close; 0xBEEF for 
   asy_sethandler()
   Output: TRUE/negative EPOC error code if opened OK/BAD. 
           TRUE for close
           TRUE/FALSE for sethandler. (FALSE if bad link protocol, or I/F state)
*/

#define SMT_SET_SC 13
/* Set the script interpreter settings.
   Inputs: sm_uwords[0] = address of a sc structure in client memory
   Output: TRUE or FALSE, depending on whether the change is 
   OK or not; i.e. the change can only take place when the network 
   interface is down.
*/

#define SMT_GET_SC 14
/* Get the script interpreter settings.
   Inputs: sm_uwords[0] = address of a sc structure in client memory
   Output: TRUE.
*/


#define SMT_SET_ISPC 15
/* Set the ISP settings.
   Inputs: sm_uwords[0] = address of an ispc structure in client memory
   Output: TRUE or FALSE, depending on whether the change is 
   OK or not; i.e. the change can only take place when the network 
   interface is down.
*/

#define SMT_GET_ISPC 16
/* Get the ISP settings.
   Inputs: sm_uwords[0] = address of an ispc structure in client memory
   Output: TRUE.
*/


#define SMT_SET_LOG 17
/* Set the logging settings 
   Inputs: sm_uwords[0] = new logging flags
   Output: TRUE.
*/

#define SMT_GET_LOG 18
/* Get the logging settings
   Inputs: None
   Output: the current logging flags.
*/

#define SMT_GET_PING 19
/* Get the ping settings
   Inputs: sm_uwords[0] = address of a ping_info structure in client memory
   Output: TRUE.
*/

#define SMT_SET_PING 20
/* Set the ping settings.
   Inputs: sm_uwords[0] = address of a ping_info structure in client memory
   Output: TRUE
*/


#define SMT_SEND_PING 21
/* Sends an ICMP echo request to an IP address.
   Inputs: None
   Output: TRUE
*/


#define SMT_SEND_KEY 22
/* Sends a keypress to an interface in TERM mode.
   Inputs: sm_uwords[0] = keypress
   Output: TRUE or FALSE, depending on whether the network interface 
   is in TERM mode or the write succeeded.
*/


#define SMT_STATS 23
/* Trigger generation of statistics output
   Inputs: None
   Output: TRUE
*/

#define SMT_SEND_RAWDATA 24
/* Sends a block of raw data down the line. Interface must be in TERM or UP 
   mode.
   Inputs: sm_uwords[0] = address of a block of data in client memory
           sm_uwords[1] = length of the block
   Output: FALSE (until implemented)
   (not yet implemented in library or kernel)
*/
   

#define SMT_GET_UDP_CONFIG 25
/* Get the UDP settings
   Inputs: sm_uwords[0] = address of a udp_config structure in client memory
   Output: TRUE.
*/

#define SMT_SET_UDP_CONFIG 26
/* Set the UDP settings.
   Inputs: sm_uwords[0] = address of a udp_config structure in client memory
   Output: TRUE or FALSE, depending on whether the change is 
   OK or not; i.e. the change can only take place when the network 
   interface is down.
*/

#define SMT_GET_TCP_CONFIG 27
/* Get the TCP settings
   Inputs: sm_uwords[0] = address of a tcp_config structure in client memory
   Output: TRUE.
*/

#define SMT_SET_TCP_CONFIG 28
/* Set the TCP settings.
   Inputs: sm_uwords[0] = address of a tcp_config structure in client memory
   Output: TRUE or FALSE, depending on whether the change is 
   OK or not; i.e. the change can only take place when the network 
   interface is down.
*/

#define SMT_GET_PPP_CONFIG 29
/* Get the PPP settings.
   Inputs: sm_uwords[0] = address of a ppp_config structure in client memory
   Output: TRUE.
*/

#define SMT_SET_PPP_CONFIG 30
/* Set the PPP settings.
   Inputs: sm_uwords[0] = address of a ppp_config structure in client memory
   Output: TRUE or FALSE, depending on whether the change is 
   OK or not; i.e. the change can only take place when the network 
   interface is down.
*/

/******************************************************************************/
/*************                                                 ****************/
/************* S O C K E T   L I B R A R Y   F U N C T I O N S ****************/
/*************                                                 ****************/
/******************************************************************************/

#define SMT_SL_ACCEPT 128
/* Accept a connection on a socket.
   Inputs: sm_uwords[0] = Socket descriptor
           sm_uwords[1] = Address of a struct sockaddr in client memory
           sm_uwords[2] = Length of the sockaddr
           sm_uwords[3] = Address of status word in client memory
   Output: Reply is zero on success, or negative error code.
*/
   
#define SMT_SL_BIND 129
/* Bind an address to a socket.
   Inputs: sm_uwords[0] = Socket descriptor
           sm_uwords[1] = Address of a struct sockaddr in client memory
           sm_uwords[2] = Length of the sockaddr
           sm_uwords[3] = Address of status word in client memory
   Output: Reply is zero on success, or negative error code.
*/

#define SMT_SL_CONNECT 130
/* Initiate a connection on a socket.
   Inputs: sm_uwords[0] = Socket descriptor
           sm_uwords[1] = Address of a struct sockaddr in client memory
           sm_uwords[2] = Length of the sockaddr
           sm_uwords[3] = Address of status word in client memory
   Output: Reply is zero on success, or negative error code.
*/

#define SMT_SL_LISTEN  131
/* Listen for connections on a socket.
   Inputs: sm_uwords[0] = Socket descriptor
           sm_uwords[1] = Number of queueable connection requests
           sm_uwords[2] = Address of status word in client memory
   Output: Reply is zero on success, or negative error code.
*/

#define SMT_SL_RECVFROM 132
/* Receive a message from a socket
   Inputs: sm_uwords[0] = Socket descriptor
           sm_uwords[1] = Address of buffer in client memory
           sm_uwords[2] = Length of buffer
           sm_uwords[3] = flags
           sm_uwords[4] = Address of sender's sockaddr in client memory or NULL
           sm_uwords[5] = Length of that sockaddr or NULL
           sm_uwords[6] = Address of status word in client memory
   Output: Reply is positive on success, or negative error code.
*/

/* There is no SMT_SL_RECV: this uses RECVFROM with two NULLS */

#define SMT_SL_RECVMSG 133

#define SMT_SL_SEND 134 
/* Send a message from a socket 
   Inputs: sm_uwords[0] = Socket descriptor
           sm_uwords[1] = Address of buffer in client memory
           sm_uwords[2] = Length of buffer
           sm_uwords[3] = flags
           sm_uwords[4] = Address of status word in client memory
   Output: Reply is positive on success, or negative error code.
*/

#define SMT_SL_SENDMSG 135

#define SMT_SL_SENDTO 136
/* Send a message to a socket
   Inputs: sm_uwords[0] = Socket descriptor
           sm_uwords[1] = Address of buffer in client memory
           sm_uwords[2] = Length of buffer
           sm_uwords[3] = flags
           sm_uwords[4] = Address of recipient's sockaddr struct in client mem
           sm_uwords[5] = Length of the sockaddr
           sm_uwords[6] = Address of status word in client memory
   Output: Reply is positive on success, or negative error code.
*/

#define SMT_SL_SHUTDOWN 137
/* Shutdown a connection
   Inputs: sm_uwords[0] = Socket descriptor
           sm_uwords[1] = How code: 0, 1, 2 (no more rx, no more tx, neither)
           sm_uwords[2] = Address of status word in client memory
   Output: Reply is zero on success, or negative error code.
*/

#define SMT_SL_SOCKET 138
/* Create a socket endpoint.
   Inputs: sm_uwords[0] = Protocol family, usually PF_INET, could be PF_IRDA 
                          (later)
           sm_uwords[1] = Socket type, SOCK_x
           sm_uwords[2] = Protocol number, usually 0
           sm_uwords[3] = Address of status word in client memory
   Output: Reply is positive, and is a socket descriptor, or negative, and 
           is an error code, or zero, and the request started successfully.
*/

#define SMT_SL_CLOSE 139
/* Close an open socket
   Inputs: sm_uwords[0] = Socket descriptor
           sm_uwords[1] = Address of status word in client memory
   Output: Reply is zero on success, or negative error code.
*/
   
#define SMT_SL_GETSOCKOPT 140
/* Get socket options
   Inputs: sm_uwords[0] = Socket descriptor
           sm_uwords[1] = Protocol number
           sm_uwords[2] = Option to get
           sm_uwords[3] = Address of option buffer in client memory
           sm_uwords[4] = Length of that buffer in bytes
           sm_uwords[5] = Address of status word in client memory
   Output: Reply is zero on success, or negative error code.
*/

#define SMT_SL_SETSOCKOPT 141
/* Set socket options
   Inputs: sm_uwords[0] = Socket descriptor
           sm_uwords[1] = Protocol number
           sm_uwords[2] = Option to set
           sm_uwords[3] = Address of option buffer in client memory
           sm_uwords[4] = Length of that buffer in bytes
           sm_uwords[5] = Address of status word in client memory
   Output: Reply is zero on success, or negative error code.
*/

#define SMT_SL_GETSOCKNAME 142
/* Get name of socket
   Inputs: sm_uwords[0] = Socket descriptor
           sm_uwords[1] = Address of a struct sockaddr in client memory
           sm_uwords[2] = Length of the sockaddr
           sm_uwords[3] = Address of status word in client memory
   Output: Reply is zero on success, or negative error code.
*/

#define SMT_SL_GETPEERNAME 143
/* Get name of connected peer 
   Inputs: sm_uwords[0] = Socket descriptor
           sm_uwords[1] = Address of a struct sockaddr in client memory
           sm_uwords[2] = Length of the sockaddr
           sm_uwords[3] = Address of status word in client memory
   Output: Reply is zero on success, or negative error code.
*/


/******************************************************************************/
/*************                                                 ****************/
/*************   N E T W O R K   D A T A B A S E   C A L L S   ****************/
/*************                                                 ****************/
/******************************************************************************/

#define SMT_NDB_GETHOSTBYNAME 200
/* gethostbyname request. 
   sm_uwords[0] = address of name
   sm_uwords[1] = strlen(name)
   sm_uwords[2] = hostent buffer for result
   sm_uwords[3] = sizeof(hostent)
   Reply is netdb code or EPOC error. */

#define SMT_NDB_GETHOSTBYADDR 201  /* gethostbyaddr request */
#define SMT_NDB_SETHOSTENT 202  /* sethostent request */
#define SMT_NDB_ENDHOSTENT 203  /* endhostent request */



#endif /* STACKMSG_H */

