/*******************************************************************************
***
*** Filename         : tcp.c
*** Purpose          : Transport Control Protocol for PsiStack (odd functions)
*** Author           : Matt J. Gumbley
*** Created          : 11/04/97
*** Last updated     : 19/04/98
***
********************************************************************************
***
*** Modification Record
***
*******************************************************************************/

#include "stack.h"

/* There are two sets of TCP configuration settings.
   This one may be changed by the TCP dialog, and gets stored in the config
   file. */

struct tcp_config tcp_config = { INIT_TCP_BSDURG, INITNTCBS };

/* This one is copied from tcp_config when the config info is loaded, and is
   the one used by the stack. Reason being:
   If you have a config file which says there are 5 TCBs, and use
   that variable to loop through the queues, then increase the number of
   blocks using the dialog, then you'll step over the end of the allocated
   TCB array. */
struct tcp_config tcp_config_used = { INIT_TCP_BSDURG, INITNTCBS };

struct tcb   *tcbtab = NULL;                  /* TCP transport control blocks */
struct uqe   *uqtab = NULL;                              /* Urgent queue data */
struct queue *tcp_iqueue = NULL; /* NB: an iqueue, not a queue */
struct queue *tcp_oqueue = NULL; /* NB: an iqueue, not a queue */

int tcps_lqsize = 5; /* Default listenq size */
int uqidone = FALSE;
static char nowt[]="---";

/*******************************************************************************
***
*** Function         : tcp_init - Initialise TCP 
*** Preconditions    : 
*** Postconditions   : 
***
*******************************************************************************/

int tcp_init(void)
{
static char szFail[] = "Couldn't create TCP %s queue";
register int i;
register struct tcb *ptcb;

  /* Copy tcp_config into tcp_config_used */
  p_bcpy(&tcp_config_used, &tcp_config, sizeof (struct tcp_config));

  tcbtab = NULL;
  tcbtab = p_alloc(sizeof(struct tcb) * tcp_config_used.tcpc_NTCBS);
  if (tcbtab == NULL) {
    console_puts("Couldn't create TCBs");
    goto initfail;
  }
  for (i=0; i<tcp_config_used.tcpc_NTCBS; i++) {
    ptcb = &tcbtab[i];
    /* FIXME: Do some initialisation stuff here */
    ptcb->tcb_istate = TCPIS_FREE;
  }

  uqtab = NULL;
  uqtab = p_alloc(sizeof(struct uqe) * UQTSIZE);
  if (uqtab == NULL) {
    console_printf(szFail,(UINT)"urgent");
    goto initfail;
  }

  /* Allocate two integer queues */
  if ((tcp_iqueue=iqueue_create(TCPQLEN)) == NULL) {
    console_printf(szFail,(UINT)"in");
    goto initfail;
  }

  if ((tcp_oqueue=iqueue_create(TCPQLEN)) == NULL) {
    console_printf(szFail,(UINT)"out");
    goto initfail;
  }
  return TRUE;

initfail:
  if (tcbtab != NULL)
    p_free(tcbtab);
  if (uqtab != NULL)
    p_free(uqtab);
  if (tcp_iqueue != NULL)
    iqueue_delete(tcp_iqueue);
  if (tcp_oqueue != NULL)
    iqueue_delete(tcp_oqueue);
  return FALSE;
}



/*******************************************************************************
***
*** Function         : tcp_wkp_name(wkp)
*** Preconditions    : wkp is a port number
*** Postconditions   : A static pointer to the name of this well-known-port is
***                    returned, or if it isn't well-known, "?" is returned.
***
*******************************************************************************/

static char *tcp_wkp_name(word16 wkp)
{
  switch (wkp) {
    case 0: return "reserved";
    case 7: return "echo";
    case 9: return "discard";
    case 11: return "systat";
    case 13: return "daytime";
    case 15: return "netstat";
    case 17: return "qotd";
    case 19: return "chargen";
    case 20: return "ftp-data";
    case 21: return "ftp";
    case 23: return "telnet";
    case 25: return "smtp";
    case 37: return "time";
    case 42: return "nameserver";
    case 43: return "whois";
    case 53: return "domain";
    case 67: return "bootp server";
    case 68: return "bootp client";
    case 69: return "tftp";
    case 79: return "finger";
    case 109: return "pop/pop2";
    case 110: return "pop3";
    case 111: return "sunrpc";
    case 119: return "nntp";
    case 123: return "ntp";
    case 161: return "SNMP net monitor";
    case 162: return "SNMP traps";
    case 512: return "biff";
    case 513: return "UNIX rwho daemon";
    case 514: return "system log";
    case 525: return "time daemon";
    default: return "?";
  }
}


/*******************************************************************************
***
*** Function         : tcp_dump outputs an interpreted datagram trace to the 
***                    log device
*** Preconditions    : b is a buffer holding a TCP segment 
*** Postconditions   : A trace is output
***
*******************************************************************************/

void tcp_dump(struct buf *b)
{
struct ip *i;
struct tcp *t;
static char eightlx[12];
  i = (struct ip *)b->data;
  t = (struct tcp *)i->ip_data;

  console_printf("tcp_dump: len 0x%04x (offset 0x%04x in dump) hdr len 0x%04x\r\n", 
             i->ip_len - IP_HLEN(i), IP_HLEN(i), t->tcp_offset >> 2);
  console_printf("src# 0x%04x (%s) ",
             t->tcp_sport, tcp_wkp_name(t->tcp_sport));
  console_printf("dst# 0x%04x (%s)\r\n",
             t->tcp_dport, tcp_wkp_name(t->tcp_dport));

  p_atos(eightlx, "0x%08lx", t->tcp_seq);
  console_printf("Seq %s Ack ", (UINT)eightlx);
  p_atos(eightlx, "0x%08lx", t->tcp_ack);
  console_printf("%s Code 0x%02x\r\n", (UINT)eightlx, t->tcp_code);

  console_printf("Flags %s %s %s %s %s %s window 0x%04x cksum 0x%04x urgptr 0x%04x\r\n",
             (UINT) (t->tcp_code & TCPF_URG ? "URG" : nowt),
             (UINT) (t->tcp_code & TCPF_ACK ? "ACK" : nowt),
             (UINT) (t->tcp_code & TCPF_PSH ? "PSH" : nowt),
             (UINT) (t->tcp_code & TCPF_RST ? "RST" : nowt),
             (UINT) (t->tcp_code & TCPF_SYN ? "SYN" : nowt),
             (UINT) (t->tcp_code & TCPF_FIN ? "FIN" : nowt),
             t->tcp_window, t->tcp_cksum, t->tcp_urgptr);
  buf_hexdump(b);

}



/*******************************************************************************
***
*** Function         : tcb_dump
*** Preconditions    : 
*** Postconditions   : Dumps a TCB
***
*******************************************************************************/

void tcb_dump(struct tcb *t)
{
char *istate, *ostate, *serv;
static char errstate[]="unknown (bug)";
TEXT buf[100];

  switch (t->tcb_istate) {
    case TCPIS_FREE: istate = "FREE"; break;
    case TCPIS_CLOSED: istate = "CLOSED"; break;
    case TCPIS_LISTEN: istate = "LISTEN"; break;
    case TCPIS_SYNSENT: istate = "SYNSENT"; break;
    case TCPIS_SYNRCVD: istate = "SYNRCVD"; break;
    case TCPIS_ESTABLISHED: istate = "ESTABLISHED"; break;
    case TCPIS_FINWAIT1: istate = "FINWAIT1"; break;
    case TCPIS_FINWAIT2: istate = "FINWAIT2"; break;
    case TCPIS_CLOSEWAIT: istate = "CLOSEWAIT"; break;
    case TCPIS_LASTACK: istate = "LASTACK"; break;
    case TCPIS_CLOSING: istate = "CLOSING"; break;
    case TCPIS_TIMEWAIT: istate = "TIMEWAIT"; break;
    default: istate = errstate; break;
  }
  switch (t->tcb_ostate) {
    case TCPOS_IDLE: ostate = "IDLE"; break;
    case TCPOS_PERSIST: ostate = "PERSIST"; break;
    case TCPOS_XMIT: ostate = "XMIT"; break;
    case TCPOS_REXMT: ostate = "REXMT"; break;
    default: ostate = errstate; break;
  }
  switch (t->tcb_type) {
    case READERS: serv = "READERS"; break;
    case WRITERS: serv = "WRITERS"; break;
    default: serv = errstate; break;
  }
  
  p_atos(buf,"TCB Istate: %s Ostate: %s Type: %s", (UINT) istate, ostate, serv);
  console_puts(buf);
  p_atos(buf,"Code: 0x%04x Flags: %s %s %s %s %s %s %s %s %s %s %s",
             (UINT) (t->tcb_code),
             (UINT) (t->tcb_flags & TCBF_NEEDOUT ? "NDO" : nowt),
             (UINT) (t->tcb_flags & TCBF_FIRSTSEND ? "1ST" : nowt),
             (UINT) (t->tcb_flags & TCBF_GOTFIN ? "GFN" : nowt),
             (UINT) (t->tcb_flags & TCBF_RDONE ? "RDN" : nowt),
             (UINT) (t->tcb_flags & TCBF_SDONE ? "SDN" : nowt),
             (UINT) (t->tcb_flags & TCBF_DELACK ? "DLK" : nowt),
             (UINT) (t->tcb_flags & TCBF_BUFFER ? "BFR" : nowt),
             (UINT) (t->tcb_flags & TCBF_PUSH ? "PSH" : nowt),
             (UINT) (t->tcb_flags & TCBF_SNDFIN ? "SFN" : nowt),
             (UINT) (t->tcb_flags & TCBF_RUPOK ? "ROK" : nowt),
             (UINT) (t->tcb_flags & TCBF_SUPOK ? "SOK" : nowt));
  console_puts(buf);

  p_atos(buf,"Error: 0x%04x pNif: 0x%04x", t->tcb_error, t->tcb_pni);
  console_puts(buf);

  p_atos(buf,"Remote %s:0x%04x Local ", szConvertIPaddr(t->tcb_rip), 
                                        t->tcb_rport);
  console_print(buf);
  p_atos(buf,"%s:0x%04x", szConvertIPaddr(t->tcb_lip), t->tcb_lport);
  console_puts(buf);

  p_atos(buf,"suna: 0x%08lx snext: 0x%08lx slast: 0x%08lx swindow: 0x%08lx", 
              t->tcb_suna,  t->tcb_snext,  t->tcb_slast,  t->tcb_swindow);
  console_puts(buf);

  p_atos(buf,"lwseq: 0x%08lx lwack: 0x%08lx cwnd: 0x%04x ssthresh: 0x%04x", 
              t->tcb_lwseq,  t->tcb_lwack,  t->tcb_cwnd, t->tcb_ssthresh);
  console_puts(buf);

  p_atos(buf,"smss: 0x%04x iss: 0x%08lx srt: 0x%04x rtde: 0x%04x", 
              t->tcb_smss, t->tcb_iss,  t->tcb_srt, t->tcb_rtde);
  console_puts(buf);

  p_atos(buf,"persist: 0x%04x keep: 0x%04x rexmt: 0x%04x rexmtcount: 0x%04x", 
              t->tcb_persist, t->tcb_keep, t->tcb_rexmt, t->tcb_rexmtcount);
  console_puts(buf);

  p_atos(buf,"rnext: 0x%08lx", t->tcb_rnext);
  console_puts(buf);

  p_atos(buf,"sndbuf: 0x%04x sbstart: 0x%04x sbcount: 0x%04x sbsize: 0x%04x", 
              t->tcb_sndbuf, t->tcb_sbstart, t->tcb_sbcount, t->tcb_sbsize);
  console_puts(buf);

  p_atos(buf,"rcvbuf: 0x%04x rbstart: 0x%04x rbcount: 0x%04x rbsize: 0x%04x", 
              t->tcb_rcvbuf, t->tcb_rbstart, t->tcb_rbcount, t->tcb_rbsize);
  console_puts(buf);

  p_atos(buf,"rmss: 0x%04x cwin: 0x%08lx finseq: 0x%08lx pushseq: 0x%08lx", 
              t->tcb_rmss, t->tcb_cwin,  t->tcb_finseq,  t->tcb_pushseq);
  console_puts(buf);
}


/*------------------------------------------------------------------------
 * tcb_alloc - allocate a Transmission Control Block
 *------------------------------------------------------------------------
 */
struct tcb *tcb_alloc()
{
register struct tcb *ptcb;
int slot;

  /* look for a free TCB */
  for (ptcb=&tcbtab[0], slot=0; slot<tcp_config_used.tcpc_NTCBS; ++slot, ++ptcb)
    if (ptcb->tcb_istate == TCPIS_FREE)
      break;
  if (slot < tcp_config_used.tcpc_NTCBS) {
    ptcb->tcb_istate = TCPIS_CLOSED;
    /* Shouldn't the queues be created here? 
       "We will see... later..." Comer v2 p202 */
  } 
  else {
    console_puts("tcb_alloc: none free");
    ptcb = (struct tcb *)NULL;
  }
  return ptcb;
}


/*------------------------------------------------------------------------
 *  tcb_free - free a TCB and free its resources
 *  ASSUMES ptcb->tcb_mutex HELD
 *------------------------------------------------------------------------
 */
int tcb_free(struct tcb *ptcb)
{
console_puts("tcb_free");
  /* FIXME heavily */
  if (ptcb->tcb_istate == TCPIS_FREE)
    return TRUE;
  switch (ptcb->tcb_type) {
    case TCPT_CONNECTION:
      timer_killtimers(ptcb);
      if (ptcb->tcb_sndbuf)
        p_free(ptcb->tcb_sndbuf);
      ptcb->tcb_sndbuf = NULL;
      if (ptcb->tcb_rcvbuf)
        p_free(ptcb->tcb_rcvbuf);
      ptcb->tcb_rcvbuf = NULL;
      /* Free segment fragment queue */
      if (ptcb->tcb_rsegq != NULL)
        queue_delete(ptcb->tcb_rsegq);  /* FIXME: is this a 'queue' ? */
      break;
    case TCPT_SERVER:
      if (ptcb->tcb_listenq)
        queue_delete(ptcb->tcb_listenq);
      break;
    default:
      /* signal(ptcb->tcb_mutex); */
      return FALSE;
  }
  ptcb->tcb_istate = TCPIS_FREE;
  /* sdelete(ptcb->tcb_mutex); */
  return TRUE;
}


/*------------------------------------------------------------------------
 *  tcp_net2h -  convert TCP header fields from net to host byte order
 *------------------------------------------------------------------------
 */
struct tcp *tcp_net2h(struct tcp *ptcp)
{
  /* NOTE: does not include TCP options */
  ptcp->tcp_sport = net2hs(ptcp->tcp_sport);
  ptcp->tcp_dport = net2hs(ptcp->tcp_dport);
  ptcp->tcp_seq = net2hl(ptcp->tcp_seq);
  ptcp->tcp_ack = net2hl(ptcp->tcp_ack);
  ptcp->tcp_window = net2hs(ptcp->tcp_window);
  ptcp->tcp_urgptr = net2hs(ptcp->tcp_urgptr);
  return ptcp;
}


/*------------------------------------------------------------------------
 *  tcp_h2net -  convert TCP header fields from host to net byte order
 *------------------------------------------------------------------------
 */
struct tcp *tcp_h2net(struct tcp *ptcp)
{
  /* NOTE: does not include TCP options */
  ptcp->tcp_sport = hs2net(ptcp->tcp_sport);
  ptcp->tcp_dport = hs2net(ptcp->tcp_dport);
  ptcp->tcp_seq = hl2net(ptcp->tcp_seq);
  ptcp->tcp_ack = hl2net(ptcp->tcp_ack);
  ptcp->tcp_window = hs2net(ptcp->tcp_window);
  ptcp->tcp_urgptr = hs2net(ptcp->tcp_urgptr);
  return ptcp;
}


/*------------------------------------------------------------------------
 *  tcp_cksum -  compute a TCP pseudo-header checksum
 *------------------------------------------------------------------------
 */
word16 tcp_cksum(struct ip *pip)
{
struct tcp *ptcp = (struct tcp *)pip->ip_data;
word16 *sptr, len;
word32 tcksum;
int i;

  tcksum = 0;

  sptr = (word16 *) pip->ip_src;
  /* 2*IP_ALEN octets = IP_ALEN shorts... */
  /* they are in net order.    */
  for (i=0; i<IP_ALEN; ++i)
    tcksum += *sptr++;
  sptr = (word16 *)ptcp;
  len = pip->ip_len - IP_HLEN(pip);
  tcksum += hs2net(IPT_TCP + len);
  if (len % 2) {
    ((char *)ptcp)[len] = 0;  /* pad */
    len += 1;  /* for the following division */
  }
  len >>= 1;  /* convert to length in shorts */

  for (i=0; i<len; ++i)
    tcksum += *sptr++;
  tcksum = (tcksum >> 16) + (tcksum & 0xffff);
  tcksum += (tcksum >> 16);

  return (word16)(~tcksum & 0xffff);
}


/*------------------------------------------------------------------------
 *  tcp_demux -  do TCP port demultiplexing
 *------------------------------------------------------------------------
 */
struct tcb *tcp_demux(struct buf *pbuf)
{
struct ip *pip = (struct ip *)pbuf->data;
struct tcp *ptcp = (struct tcp *)pip->ip_data;
int tcbn, lstcbn;

TEXT buf[100];
console_puts("tcp_demux: in");
  for (tcbn=0, lstcbn = -1; tcbn<tcp_config_used.tcpc_NTCBS; ++tcbn) {
console_printf("tcp_demux: checking TCB %d\r\n", tcbn);

    if (tcbtab[tcbn].tcb_istate == TCPIS_FREE) {
console_printf("tcp_demux: TCB %d is free\r\n", tcbn);
      continue;
    }
console_printf("tcp_demux: PACKET dport 0x%04x == TCB lport 0x%04x\r\n", ptcp->tcp_dport, tcbtab[tcbn].tcb_lport);
console_printf("tcp_demux: PACKET sport 0x%04x == TCB rport 0x%04x\r\n", ptcp->tcp_sport, tcbtab[tcbn].tcb_rport);

p_atos(buf,"tcp_demux: PACKET src %s == TCB rip ",szConvertIPaddr(pip->ip_src));
console_print(buf);
p_atos(buf,"%s", szConvertIPaddr(tcbtab[tcbn].tcb_rip));
console_puts(buf);

p_atos(buf,"tcp_demux: PACKET dst %s == TCB lip ",szConvertIPaddr(pip->ip_dst));
console_print(buf);
p_atos(buf,"%s", szConvertIPaddr(tcbtab[tcbn].tcb_lip));
console_puts(buf);

    if (ptcp->tcp_dport == tcbtab[tcbn].tcb_lport &&
        ptcp->tcp_sport == tcbtab[tcbn].tcb_rport &&
        p_bcmp(pip->ip_src, IP_ALEN, tcbtab[tcbn].tcb_rip, IP_ALEN)==0 &&
        p_bcmp(pip->ip_dst, IP_ALEN, tcbtab[tcbn].tcb_lip, IP_ALEN)==0) {
console_printf("tcp_demux: TCB %d matches\r\n", tcbn);
      break;
    }
    if (tcbtab[tcbn].tcb_istate == TCPIS_LISTEN &&
        ptcp->tcp_dport == tcbtab[tcbn].tcb_lport) {
console_printf("tcp_demux: TCB %d is listening on port 0x%04x\r\n", tcbn, ptcp->tcp_dport);
      lstcbn = tcbn;
    }
  }
  if (tcbn >= tcp_config_used.tcpc_NTCBS) {
console_puts("tcp_demux: not found - using last listener");
    if (ptcp->tcp_code & TCPF_SYN)
      tcbn =  lstcbn;
    else
      tcbn = -1;
  }
  if (tcbn < 0) {
console_puts("tcp_demux: no, nothing found that matches - returning NULL");
    return NULL;
  }
  if (tcbtab[tcbn].tcb_istate == TCPIS_FREE) {
    console_puts("tcp_demux: lost the TCB somehow!");
    return NULL;      /* OOPS! Lost it... */
  }
console_printf("tcp_demux: returning TCB %d\r\n", tcbn);
  return &tcbtab[tcbn];
}


/*------------------------------------------------------------------------
 *  tcp_ok -  determine if a received segment is acceptable
 *------------------------------------------------------------------------
 */
int tcp_ok(struct tcb *ptcb, struct buf *pbuf)
{
struct ip *pip = (struct ip *)pbuf->data;
struct tcp *ptcp = (struct tcp *)pip->ip_data;
int seglen, rwindow;
tcpseq wlast, slast;
int rv;

  if (ptcb->tcb_istate < TCPIS_SYNRCVD)
    return TRUE;
  seglen = pip->ip_len - IP_HLEN(pip) - TCP_HLEN(ptcp);

  /* add SYN and FIN */
  if (ptcp->tcp_code & TCPF_SYN)
    ++seglen;
  if (ptcp->tcp_code & TCPF_FIN)
    ++seglen;
  rwindow = ptcb->tcb_rbsize - ptcb->tcb_rbcount; 
  if (rwindow == 0 && seglen == 0)
    return ptcp->tcp_seq == ptcb->tcb_rnext;
  if (ptcp->tcp_code & TCPF_URG)
    (void) tcp_rcvurg(ptcb, pbuf);
  wlast = ptcb->tcb_rnext + rwindow - 1;
  rv = (ptcp->tcp_seq - ptcb->tcb_rnext) >= 0 &&
    (ptcp->tcp_seq - wlast) <= 0;
  if (seglen == 0)
    return rv;
  slast = ptcp->tcp_seq + seglen - 1;
  rv |= (slast - ptcb->tcb_rnext) >= 0 && (slast - wlast) <= 0;

  /* If no window, strip data but keep ACK, RST and URG    */
  if (rwindow == 0)
    pip->ip_len = IP_HLEN(pip) + TCP_HLEN(ptcp);
  return rv;
}


/*------------------------------------------------------------------------
 *  tcprcvurg -  handle TCP urgent data
 *------------------------------------------------------------------------
 */
int tcp_rcvurg(struct tcb *ptcb, struct buf *pbuf)
{
struct  ip  *pip = (struct ip *)pbuf->data;
struct  tcp  *ptcp = (struct tcp *)pip->ip_data;
struct  uqe  *puqe;
int    datalen, offset;

return FALSE; /* FIXME heavily: looks like we can't do urgent data yet */

  datalen = pip->ip_len - IP_HLEN(pip) - TCP_HLEN(ptcp);
  /*
   * Berkeley machines treat the urgent pointer as "last+1", but
   * RFC 1122 says just "last." 
   */
  if (tcp_config.tcpc_bsdurg)
    ptcp->tcp_urgptr -= 1;
  if (ptcp->tcp_urgptr >= datalen || ptcp->tcp_urgptr >= TCPMAXURG)
    ptcp->tcp_urgptr = datalen-1;  /* quietly fix it  */
  puqe = uqalloc();
  if (puqe == (struct uqe *)NULL) {
    /* FIXME: do some diagnostic stuff */
    return FALSE;    /* out of buffer space!!!  */
  }
  puqe->uq_seq = ptcp->tcp_seq;
  puqe->uq_len = ptcp->tcp_urgptr + 1;
  puqe->uq_data = (char *)p_alloc(576 /*Net.netpool*/); /* FIXME */
  if (puqe->uq_data == (char *)NULL) {
    /* puqe->uq_data = 0; * Sheesh, this code.. */
    uqfree(puqe);
    return FALSE;
  }
  p_bcpy(puqe->uq_data, &pip->ip_data[TCP_HLEN(ptcp)], puqe->uq_len);
  if (ptcb->tcb_rudq < 0) {
    ptcb->tcb_rudseq = ptcb->tcb_rnext;
    ptcb->tcb_rudq = queue_create(TCPUQLEN /* FIXME: QF_WAIT */ );
    if (ptcb->tcb_rudq < 0) {
      uqfree(puqe);
      return FALSE;  /* treat it like normal data  */
    }
  }
  if (datalen > puqe->uq_len) {
    /* some non-urgent data left, edit the packet    */

    ptcp->tcp_seq += puqe->uq_len;
    offset = TCP_HLEN(ptcp) + puqe->uq_len;
    p_bcpy(&pip->ip_data[TCP_HLEN(ptcp)],
           &pip->ip_data[offset], datalen-puqe->uq_len);
    ptcp->tcp_urgptr = 0;
    pip->ip_len -= puqe->uq_len;
    /* checksums already checked, so no need to redo them  */
  }
  /* FIXME: this won't work, 'cos we're using uqe's, not buf's */
  if (queue_enqueue_ordered(ptcb->tcb_rudq,puqe,ptcb->tcb_rudseq-puqe->uq_seq) < 0) {
    uqfree(puqe);
    return FALSE;
  }
  return TRUE;
}


/*******************************************************************************
***
*** Function         : tcp_shskip
*** Preconditions    : 
*** Postconditions   : Urgent data holed in send buffer are skipped
***
*******************************************************************************/

int tcp_shskip(struct tcb *ptcb, int datalen, int *poff)
{
struct   uqe  *puqe;
tcpseq    lseq;
int    resid = 0;
console_puts("tcp_shskip");
  if (ptcb->tcb_suhq == NULL)
    return datalen;
  puqe = (struct uqe *)queue_dequeue(ptcb->tcb_suhq);
  while (puqe) {
    lseq = ptcb->tcb_snext + datalen - 1;
    if (SEQCMP(lseq, puqe->uq_seq) > 0)
      resid = (int) (lseq - puqe->uq_seq);
    /* chop off at urgent boundary, but save the hole...  */
    if (resid < datalen) {
      datalen -= resid;
      /* FIXME: this won't work, cos we're using uqe's, not buf's */
      if (queue_enqueue_ordered(ptcb->tcb_suhq, puqe, SUHK(ptcb, puqe->uq_seq)) < 0)
        uqfree(puqe);  /* shouldn't happen */
      return datalen;
    }
    /*
     * else, we're skipping a hole now adjust the beginning
     * of the segment.
     */
    lseq = puqe->uq_seq + puqe->uq_len;
    resid = (int) (lseq - ptcb->tcb_snext);
    if (resid > 0) {
      /* overlap... */
      *poff += resid;
      if (*poff > ptcb->tcb_sbsize)
        *poff -= ptcb->tcb_sbsize;
      datalen -= resid;
      if (datalen < 0)
        datalen = 0;
    }
    uqfree(puqe);
    puqe = (struct uqe *)queue_dequeue(ptcb->tcb_suhq);
  }
  /* puqe == 0 (No more holes left) */
  queue_delete(ptcb->tcb_suhq);
  ptcb->tcb_suhq = NULL;
  return datalen;
}


/*******************************************************************************
***
*** Function         : tcp_sndurg
*** Preconditions    : 
*** Postconditions   : Pending urgent data is sent.
***
*******************************************************************************/

int tcp_sndurg(int tcbnum)
{
struct tcb  *ptcb = &tcbtab[tcbnum];
console_puts("tcp_sndurg");
  ptcb->tcb_code |= TCPF_URG;
  while (ptcb->tcb_sudq > 0)
    tcp_send(tcbnum, TSF_NEWDATA);
  ptcb->tcb_code &= ~TCPF_URG;
  return TRUE;
}



/*------------------------------------------------------------------------
 *  uqalloc -  allocate a TCP urgent queue entry
 *------------------------------------------------------------------------
 */
struct uqe *uqalloc()
{
  static  int  last = 0;
  struct  uqe  *puqe;
  int    count;

  if (!uqidone)
    uqinit();
  /*wait(uqmutex);*/
  for (count=0; count<UQTSIZE; ++count) {
    if (++last >= UQTSIZE)
      last = 0;
    puqe = &uqtab[last];
    if (puqe->uq_state == UQS_FREE) {
      puqe->uq_state = UQS_ALLOC;
      /*signal(uqmutex);*/
      return puqe;
    }
  }
  /*signal(uqmutex);*/
  return (struct uqe *)NULL;
}


/*------------------------------------------------------------------------
 *  uqfree -  free an urgent queue entry
 *------------------------------------------------------------------------
 */
int uqfree(struct  uqe  *puqe)
{
  /*wait(uqmutex);*/
  if (puqe->uq_data)
    p_free(puqe->uq_data);
  puqe->uq_state = UQS_FREE;
  /*signal(uqmutex);*/
  return TRUE;
}

/*------------------------------------------------------------------------
 *  uqinit -  initialize TCP urgent queue data
 *------------------------------------------------------------------------
 */
int uqinit()
{
  int  i;

  uqidone = TRUE;
  /*uqmutex = screate(0);*/
  for (i=0; i<UQTSIZE; ++i)
    uqtab[i].uq_state = UQS_FREE;
  /*signal(uqmutex);*/
  return TRUE;
}

/*------------------------------------------------------------------------
 *  uqprint -  print a queue of urgent pointers (for debugging)
 *------------------------------------------------------------------------
 */
void uqprint(struct queue *q)
{
  struct  uqe  *puqe;

  console_print("\nUQ ");
  while ((puqe = (struct uqe *)seeq(q)))
    console_printf("(l %d s %X) ", (UINT) puqe->uq_len, puqe->uq_seq);
  console_puts("");
}

word16 tcp_nextport(void)
{
static word16 lastport=1; 
word16 i, start;
  for (start = lastport++; start != lastport; ++lastport) {
    if (lastport == TLPORT)
      lastport = 1;
    for (i=0; i<tcp_config_used.tcpc_NTCBS; i++)
      if (tcbtab[i].tcb_istate != TCPIS_FREE &&
          tcbtab[i].tcb_lport == lastport)
        break;
    if (i == tcp_config_used.tcpc_NTCBS)
      break;
  }
  if (lastport == start) 
    console_puts("tcp_nextport: out of TCP ports");
  return lastport;
}

