/*******************************************************************************
***
*** Filename         : tcpin.c
*** Purpose          : TCP input processing for PsiStack
*** Author           : Matt J. Gumbley
*** Created          : 12/08/97
*** Last updated     : 14/06/98
***
********************************************************************************
***
*** Modification Record
***
*******************************************************************************/

#include "stack.h"

/* Some stats about the TCP input layer */
struct logstat tcpi_stat[] = {
  "checksum", 0,                                     /* # TCP Checksum errors */
  "noTCB", 0,                       /* # datagrams for which there was no TCB */
  "estabresets", 0,                                  /* # established resets? */
  "currestab", 0,                                 /* # currently established? */
  "attemptfails", 0,      /* attempt to get correct ACK in tcpi_synsent fails */
  "passiveopens", 0,                                       /* # passive opens */
  "freeswitch", 0,                                /* # switches in FREE state */
  NULL, 0
};
#define tcpi_stat_checksum tcpi_stat[0].ls_value
#define tcpi_stat_noTCBexists tcpi_stat[1].ls_value
#define tcpi_stat_estabresets tcpi_stat[2].ls_value
#define tcpi_stat_currestab tcpi_stat[3].ls_value
#define tcpi_stat_attemptfails tcpi_stat[4].ls_value
#define tcpi_stat_passiveopens tcpi_stat[5].ls_value
#define tcpi_stat_freeswitch tcpi_stat[6].ls_value


/* TCP input state processing */

int tcpi_free(), tcpi_closed(), tcpi_listen(), tcpi_synsent(), tcpi_synrcvd(),
  tcpi_established(), tcpi_fin1(), tcpi_fin2(), tcpi_closewait(),
  tcpi_closing(), tcpi_lastack(), tcpi_timewait();

int (*tcpi_switch[NTCPISTATES])(struct tcb *, struct buf *) = { 
  tcpi_free,                                                    /* TCPIS_FREE */
  tcpi_closed,                                                /* TCPIS_CLOSED */
  tcpi_listen,                                                /* TCPIS_LISTEN */
  tcpi_synsent,                                              /* TCPIS_SYNSENT */
  tcpi_synrcvd,                                              /* TCPIS_SYNRCVD */
  tcpi_established,                                      /* TCPIS_ESTABLISHED */
  tcpi_fin1,                                                /* TCPIS_FINWAIT1 */
  tcpi_fin2,                                                /* TCPIS_FINWAIT2 */
  tcpi_closewait,                                          /* TCPIS_CLOSEWAIT */
  tcpi_lastack,                                              /* TCPIS_LASTACK */
  tcpi_closing,                                              /* TCPIS_CLOSING */
  tcpi_timewait,                                            /* TCPIS_TIMEWAIT */
};

/* Externals */
GLDEF_C INT socket_unblockpassiveopen(struct tcb *ptcb);

/*******************************************************************************
***
*** Function         : tcpi_process
*** Preconditions    : pbuf is a TCP segment received on the local interface, 
***                    passed to this, the TCP input pseudo-process, from IP 
***                    (well, the local loopback interface, actually, as 
***                    that's the 'higher-level protocol' entry point)
*** Postconditions   : The datagram is handled. It is the responsibility of
***                    this function and its minions to free the buffer.
***
*******************************************************************************/

void tcpi_process(struct buf *pbuf)
{
struct ip *pip;
struct tcp *ptcp;
struct tcb *ptcb;
static TEXT *innames[] = {
"FREE", "CLOSED", "LISTEN", "SYNSENT", "SYNRCVD", "ESTABLISHED", "FINWAIT1",
"FINWAIT2", "CLOSEWAIT", "LASTACK", "CLOSING", "TIMEWAIT" };

  pip = (struct ip *)pbuf->data;
  if (tcp_cksum(pip)) {
    if (flags.log_flag & LOG_FLAG_FLOWTRACE) 
      console_puts("tcpi_process: bad cksum");
    tcpi_stat_checksum++;
    buf_free(pbuf);
    return;
  }
  ptcp = (struct tcp *)pip->ip_data;
  tcp_net2h(ptcp); /* convert all fields to host order */
  if (flags.log_flag & LOG_FLAG_UPPERDUMP) {
    console_puts("tcpi_process: dump...");
    tcp_dump(pbuf);
  }
  ptcb = tcp_demux(pbuf);
  if (ptcb == NULL) {
    if (flags.log_flag & LOG_FLAG_FLOWTRACE) 
      console_puts("tcpi_process: no TCB");
    tcpi_stat_noTCBexists++;
    tcp_reset(pbuf);
    buf_free(pbuf);
    return;
  }
  if (flags.log_flag & LOG_FLAG_UPPERDUMP) {
    console_puts("tcpi_process: TCB details:");
    tcb_dump(ptcb);
  }
  if (!tcp_ok(ptcb, pbuf)) {
    if (flags.log_flag & LOG_FLAG_UPPERDUMP) 
      console_puts("tcpi_process: unacceptable window");
    tcp_ackit(ptcb, pbuf); /* Send correctly rx'ed seq & curr. window size */
  }
  else {
    if (flags.log_flag & LOG_FLAG_UPPERDUMP) 
      console_puts("tcpi_process: processing options");
    tcp_opts(ptcb, pbuf);
    if (flags.log_flag & LOG_FLAG_UPPERDUMP) 
      console_printf("tcpi_process: calling %s state handler\r\n", innames[ptcb->tcb_istate]);
    tcpi_switch[ptcb->tcb_istate](ptcb, pbuf);
  }
  buf_free(pbuf);
}


/*******************************************************************************
***
*** Function         : tcpi_free
*** Preconditions    : The TCB is in FREE state
*** Postconditions   : An error is noted; this shouldn't happen?
***
*******************************************************************************/

int tcpi_free(struct tcb *ptcb, struct buf *pbuf)
{
  tcpi_stat_freeswitch++;
  return NULL;
}


/*******************************************************************************
***
*** Function         : tcpi_closed
*** Preconditions    : The TCB is in CLOSED state
*** Postconditions   : CLOSED state processing has been done
***
*******************************************************************************/

int tcpi_closed(struct tcb *ptcb, struct buf *pbuf)
{
  tcp_reset(pbuf);
  return NULL;
}


/*******************************************************************************
***
*** Function         : tcp_wait(tcb)
*** Preconditions    : tcb is a pointer to a TCB.
*** Postconditions   : A DELETE event is (re)scheduled for 2MSL from now
***
*******************************************************************************/

int tcp_wait(struct tcb *ptcb)
{
int tcbnum = ptcb - &tcbtab[0];

  timer_killtimers(ptcb);
  timer_set(MKEVENT(TQ_TCP_DELETE, tcbnum), TCP_TWOMSL);
  return TRUE;
}


/*******************************************************************************
***
*** Function         : tcpi_timewait
*** Preconditions    : The TCB is in TIME_WAIT state
*** Postconditions   : TIME_WAIT state processing has been done.
***
*******************************************************************************/

int tcpi_timewait(struct tcb *ptcb, struct buf *pbuf)
{
struct ip *pip = (struct ip *)pbuf->data;
struct tcp *ptcp = (struct tcp *)pip->ip_data;

  if (ptcp->tcp_code & TCPF_RST)
    return tcb_free(ptcb);
  if (ptcp->tcp_code & TCPF_SYN) {
    tcp_reset(pbuf);
    return tcb_free(ptcb);
  }
  tcp_acked(ptcb, pbuf);
  tcp_data(ptcb, pbuf);    /* just ACK any packets */
  tcp_wait(ptcb);
  return TRUE;
}


/*******************************************************************************
***
*** Function         : tcpi_closing
*** Preconditions    : The TCB is in CLOSED state
*** Postconditions   : CLOSED state input processing has been done.
***
*******************************************************************************/

int tcpi_closing(struct tcb *ptcb, struct buf *pbuf)
{
struct ip *pip = (struct ip *)pbuf->data;
struct tcp *ptcp = (struct tcp *)pip->ip_data;

  if (ptcp->tcp_code & TCPF_RST)
    return tcb_free(ptcb);
  if (ptcp->tcp_code & TCPF_SYN) {
    tcp_reset(pbuf);
    return tcb_free(ptcb);
  }
  tcp_acked(ptcb, pbuf);
  if ((ptcb->tcb_code & TCPF_FIN) == 0) {
    ptcb->tcb_istate = TCPIS_TIMEWAIT;
    /* FIXME: signal(ptcb->tcb_ocsem);    * wake closer  */
    tcp_wait(ptcb);
  }
  return TRUE;
}


/*******************************************************************************
***
*** Function         : tcpi_fin2
*** Preconditions    : The TCB is in FIN_WAIT_2 state.
*** Postconditions   : FIN_WAIT_2 state input processing has been done.
***
*******************************************************************************/

int tcpi_fin2(struct tcb *ptcb, struct buf *pbuf)
{
struct ip *pip = (struct ip *)pbuf->data;
struct tcp *ptcp = (struct tcp *)pip->ip_data;

  if (ptcp->tcp_code & TCPF_RST)
    return tcp_abort(ptcb, E_NET_CONNRESET);
  if (ptcp->tcp_code & TCPF_SYN) {
    tcp_reset(pbuf);
    return tcp_abort(ptcb, E_NET_CONNRESET);
  }
  if (tcp_acked(ptcb, pbuf) == FALSE)
    return TRUE;
  tcp_data(ptcb, pbuf);  /* for data + FIN ACKing */

  if (ptcb->tcb_flags & TCBF_RDONE) {
    ptcb->tcb_istate = TCPIS_TIMEWAIT;
    tcp_wait(ptcb);
  }
  return TRUE;
}


/*******************************************************************************
***
*** Function         : tcpi_fin1
*** Preconditions    : The TCB is in FIN_WAIT_1 state
*** Postconditions   : FIN_WAIT_1 input state processing has been done.
***
*******************************************************************************/

int tcpi_fin1(struct tcb *ptcb, struct buf *pbuf)
{
struct ip *pip = (struct ip *)pbuf->data;
struct tcp *ptcp = (struct tcp *)pip->ip_data;

  if (ptcp->tcp_code & TCPF_RST)
    return tcp_abort(ptcb, E_NET_CONNRESET);
  if (ptcp->tcp_code & TCPF_SYN) {
    tcp_reset(pbuf);
    return tcp_abort(ptcb, E_NET_CONNRESET);
  }
  if (tcp_acked(ptcb, pbuf) == FALSE)
    return TRUE;
  tcp_data(ptcb, pbuf);
  tcp_swindow(ptcb, pbuf);

  if (ptcb->tcb_flags & TCBF_RDONE) {
    if (ptcb->tcb_code & TCPF_FIN)    /* FIN not ACKed */
      ptcb->tcb_istate = TCPIS_CLOSING;
    else {
      ptcb->tcb_istate = TCPIS_TIMEWAIT;
      /* FIXME: signal(ptcb->tcb_ocsem);  * wake closer  */
      tcp_wait(ptcb);
    }
  }
  else if ((ptcb->tcb_code & TCPF_FIN) == 0) {
    /* FIXME: signal(ptcb->tcb_ocsem);    * wake closer  */
    ptcb->tcb_istate = TCPIS_FINWAIT2;
  }
  return TRUE;
}


/*******************************************************************************
***
*** Function         : tcpi_closewait
*** Preconditions    : The TCB is in CLOSE_WAIT state
*** Postconditions   : CLOSE_WAIT input state processing has been done.
***
*******************************************************************************/

int tcpi_closewait(struct tcb *ptcb, struct buf *pbuf)
{
struct ip *pip = (struct ip *)pbuf->data;
struct tcp *ptcp = (struct tcp *)pip->ip_data;

  if (ptcp->tcp_code & TCPF_RST) {
    tcpi_stat_estabresets++;
    tcpi_stat_currestab--;
    return tcp_abort(ptcb, E_NET_CONNRESET);
  }
  if (ptcp->tcp_code & TCPF_SYN) {
    tcpi_stat_estabresets++;
    tcpi_stat_currestab--;
    tcp_reset(pbuf);
    return tcp_abort(ptcb, E_NET_CONNRESET);
  }
  tcp_acked(ptcb, pbuf);
  tcp_swindow(ptcb, pbuf);
  return TRUE;
}


/*******************************************************************************
***
*** Function         : tcpi_lastack
*** Preconditions    : The TCB is in LAST_ACK state
*** Postconditions   : LAST_ACK input state processing has been done.
***
*******************************************************************************/

int tcpi_lastack(struct tcb *ptcb, struct buf *pbuf)
{
struct ip *pip = (struct ip *)pbuf->data;
struct tcp *ptcp = (struct tcp *)pip->ip_data;

  if (ptcp->tcp_code & TCPF_RST)
    return tcp_abort(ptcb, E_NET_CONNRESET);
  if (ptcp->tcp_code & TCPF_SYN) {
    tcp_reset(pbuf);
    return tcp_abort(ptcb, E_NET_CONNRESET);
  }
  tcp_acked(ptcb, pbuf);
  if ((ptcb->tcb_code & TCPF_FIN) == 0) {
    /* FIXME signal(ptcb->tcb_ocsem);  * close() deallocs  */
  }
  return TRUE;
}


/*******************************************************************************
***
*** Function         : tcpi_established
*** Preconditions    : The TCB is in ESTABLISHED state.
*** Postconditions   : ESTABLISHED input state processing has been done.
***
*******************************************************************************/

int tcpi_established(struct tcb *ptcb, struct buf *pbuf)
{
struct ip *pip = (struct ip *)pbuf->data;
struct tcp *ptcp = (struct tcp *)pip->ip_data;

  if (ptcp->tcp_code & TCPF_RST) {
    tcpi_stat_estabresets++;
    tcpi_stat_currestab--;
    return tcp_abort(ptcb, E_NET_CONNRESET);
  }
  if (ptcp->tcp_code & TCPF_SYN) {
    tcpi_stat_estabresets++;
    tcpi_stat_currestab--;
    tcp_reset(pbuf);
    return tcp_abort(ptcb, E_NET_CONNRESET);
  }
  if (tcp_acked(ptcb, pbuf) == FALSE)
    return TRUE;
  tcp_data(ptcb, pbuf);
  tcp_swindow(ptcb, pbuf);
  if (ptcb->tcb_flags & TCBF_RDONE)
    ptcb->tcb_istate = TCPIS_CLOSEWAIT;
  return TRUE;
}


/*******************************************************************************
***
*** Function         : tcp_data
*** Preconditions    : The TCB has data queued on it.
*** Postconditions   : The input segment's data section is processed.
***
*******************************************************************************/

int tcp_data(struct tcb *ptcb, struct buf *pbuf)
{
struct ip *pip = (struct ip *)pbuf->data;
struct tcp *ptcp = (struct tcp *)pip->ip_data;
tcpseq first, last, wlast;
int datalen, rwindow, i, pp, pb;

console_puts("tcp_data in");
  /* Odd. The book deals with urgent data here...  with this code...
  if (ptcp->tcp_code & TCPF_URG) {
    int rup = ptcp->tcp_seq + ptcp->tcp_urgptr;
      if (tcp_config.tcpc_bsdurg)
        rup--;
      if (!ptcb->tcb_flags & TCBF_RUPOK) ||
          SEQCMP(rup, ptcb->tcb_rupseq) > 0) {
        ptcb->tcb_rupseq = rup;
        ptcb->tcb_flags |= TCBF_RUPOK;
      }
  }
  */

  if (ptcp->tcp_code & TCPF_SYN) {
    ptcb->tcb_rnext++;
    ptcb->tcb_flags |= TCBF_NEEDOUT;
    ++ptcp->tcp_seq;  /* so we start with data */
  }
  datalen = pip->ip_len - IP_HLEN(pip) - TCP_HLEN(ptcp);
  rwindow = ptcb->tcb_rbsize - ptcb->tcb_rbcount; 
  wlast = ptcb->tcb_rnext + rwindow-1;
  first = ptcp->tcp_seq;
  last = first + datalen - 1;
  if (SEQCMP(ptcb->tcb_rnext, first) > 0) {
    datalen -= ptcb->tcb_rnext - first;
    first = ptcb->tcb_rnext;
  }
  if (SEQCMP(last, wlast) > 0) {
    datalen -= last - wlast;
    ptcp->tcp_code &= ~TCPF_FIN;  /* cutting it off */
  }
  pb = ptcb->tcb_rbstart + ptcb->tcb_rbcount; /* == rnext, in buf */
  pb += first - ptcb->tcb_rnext;    /* distance in buf  */
  pb %= ptcb->tcb_rbsize;      /* may wrap    */
  pp = (int) (first - ptcp->tcp_seq);    /* distance in packet  */
console_printf("tcp_data: copying %d from data[%d] to rcvbuf[%d]\r\n", datalen, pp, pb);
  for (i=0; i<datalen; ++i) {
    ptcb->tcb_rcvbuf[pb] = ptcp->tcp_data[pp++];
    if (++pb >= ptcb->tcb_rbsize)
      pb = 0;
  }
console_puts("tcp_data: done copying data");
  tcp_dodat(ptcb, ptcp, first, datalen);  /* deal with it    */
  if (ptcb->tcb_flags & TCBF_NEEDOUT)
    tcp_kick(ptcb);
console_puts("tcp_data out");
  return TRUE;
}


/*******************************************************************************
***
*** Function         : tcp_dodat
*** Preconditions    : 
*** Postconditions   : The rest of the input data processing is done.
***
*******************************************************************************/

int tcp_dodat(struct tcb *ptcb, struct tcp *ptcp, tcpseq first, int datalen)
{
int wakeup = 0;
console_puts("tcp_dodat in");
  if (ptcb->tcb_rnext == first) {
console_puts("rnext == first");
    if (datalen > 0) {
console_puts("coalescing");
      tcp_tfcoalesce(ptcb, datalen, ptcp);
      ptcb->tcb_flags |= TCBF_NEEDOUT;
      wakeup++;
    }
    if (ptcp->tcp_code & TCPF_FIN) {
console_puts("got a FIN");
      ptcb->tcb_flags |= TCBF_RDONE|TCBF_NEEDOUT;
      ptcb->tcb_rnext++;
      wakeup++;
    }
    /* The book uses (TCPF_PSH | TCPF_URG) in this 'if' */
    if (ptcp->tcp_code & TCPF_PSH) {
console_puts("got a PUSH");
      ptcb->tcb_flags |= TCBF_PUSH;
      wakeup++;
    }
    if (wakeup) {
console_puts("waking readers");
      tcp_wakeup(READERS, ptcb);
    }
  }
  else {
console_puts("processing delayed controls");
    /* process delayed controls */
    if (ptcp->tcp_code & TCPF_FIN) {
console_puts("got a FIN");
      ptcb->tcb_finseq = ptcp->tcp_seq + datalen;
    }
    /* The book uses (TCPF_PSH | TCPF_URG) in this 'if' */
    if (ptcp->tcp_code & TCPF_PSH) {
console_puts("got a PUSH");
      ptcb->tcb_pushseq = ptcp->tcp_seq + datalen;
    }
    ptcp->tcp_code &= ~(TCPF_FIN|TCPF_PSH);
    /* FIXME: check for return FALSE here, if we get OOM */
console_puts("inserting");
    tcp_tfinsert(ptcb, first, datalen);
  }
console_puts("tcp_dodat out");
  return TRUE;
}


/*******************************************************************************
***
*** Function         : tcp_tfinsert
*** Preconditions    : 
*** Postconditions   : The TCP segment fragment is added to the TCP sequence q.
***
*******************************************************************************/

int tcp_tfinsert(struct tcb *ptcb, tcpseq seq, int datalen)
{
register struct tcpfrag *tf;
register struct buf *tfbuf;

  if (datalen == 0)
    return TRUE;
  tfbuf = buf_alloc(sizeof(struct tcpfrag));
  if (tfbuf == NULL)
    return FALSE; /* Comer doesn't check for NULL... tch tch... */
                  /* not sure what might happen... */
  tf = (struct tcpfrag *)tfbuf->data;
  tf->tf_seq = seq;
  tf->tf_len = datalen;
  if (ptcb->tcb_rsegq == NULL)
    ptcb->tcb_rsegq = queue_create(NTCPFRAG);
  if (queue_enqueue_ordered(ptcb->tcb_rsegq, tfbuf, -tf->tf_seq) == FALSE) {
    buf_free(tf);
    return FALSE; /* Comer doesn't check for queue full, I've added this */
  }
  return TRUE;
}


/*******************************************************************************
***
*** Function         : tcp_tfcoalesce
*** Preconditions    : 
*** Postconditions   : TCP segment fragments are coalesced.
***
*******************************************************************************/

int tcp_tfcoalesce(struct tcb *ptcb, int datalen, struct tcp *ptcp)
{
register struct buf *tfbuf;
register struct tcpfrag *tf;
int new;

  ptcb->tcb_rnext += datalen;
  ptcb->tcb_rbcount += datalen;
  if (ptcb->tcb_rnext == ptcb->tcb_finseq)
    goto alldone;
  if ((ptcb->tcb_rnext - ptcb->tcb_pushseq) >= 0) {
    ptcp->tcp_code |= TCPF_PSH;
    ptcb->tcb_pushseq = 0;
  }
  if (ptcb->tcb_rsegq < 0)  /* see if this closed a hole */
    return TRUE;
  tfbuf = queue_dequeue(ptcb->tcb_rsegq);
  tf = (struct tcbfrag *)tfbuf->data;
  while ((tf->tf_seq - ptcb->tcb_rnext) <= 0) {
    new = tf->tf_len - (ptcb->tcb_rnext - tf->tf_seq);
    if (new > 0) {
      ptcb->tcb_rnext += new;
      ptcb->tcb_rbcount += new;
    }
    if (ptcb->tcb_rnext == ptcb->tcb_finseq)
      goto alldone;
    if ((ptcb->tcb_rnext - ptcb->tcb_pushseq) >= 0) {
      ptcp->tcp_code |= TCPF_PSH;
      ptcb->tcb_pushseq = 0;
    }
    buf_free(tfbuf);
    tfbuf = queue_dequeue(ptcb->tcb_rsegq);
    if (tfbuf == NULL) {
      queue_delete(ptcb->tcb_rsegq);
      ptcb->tcb_rsegq = NULL;
      return TRUE;
    }
    else
      tf = (struct tcbfrag *)tfbuf->data;
  }
  /* got one too many  */
  queue_enqueue_ordered(ptcb->tcb_rsegq, tfbuf, -tf->tf_seq);
  return TRUE;
alldone:
  queue_delete(ptcb->tcb_rsegq);
  ptcb->tcb_rsegq = NULL;
  ptcp->tcp_code |= TCPF_FIN;
  return TRUE;
}


/*******************************************************************************
***
*** Function         : tcp_abort
*** Preconditions    : 
*** Postconditions   : The active TCP connection is terminated.
***
*******************************************************************************/

int tcp_abort(struct tcb *ptcb, int error)
{
  timer_killtimers(ptcb);
  ptcb->tcb_flags |= TCBF_RDONE|TCBF_SDONE;
  ptcb->tcb_error = error;
  tcp_wakeup(READERS|WRITERS, ptcb);
  return TRUE;
}


/*******************************************************************************
***
*** Function         : tcp_sync
*** Preconditions    : The TCB is partially initialised.
*** Postconditions   : The TCB is initialised for a new connection request.
***                    tcp_sync returns TRUE or FALSE depending on whether 
***                    the TCB was correctly initialised.
***
*******************************************************************************/

int tcp_sync(struct tcb *ptcb)
{
console_puts("tcp_sync: in");
  ptcb->tcb_istate = TCPIS_CLOSED;
  ptcb->tcb_type = TCPT_CONNECTION;

  ptcb->tcb_iss = tcp_iss();
  ptcb->tcb_lwack = ptcb->tcb_suna = ptcb->tcb_snext = ptcb->tcb_iss;

  ptcb->tcb_sndbuf = (byte *)p_alloc(TCPSBS);
  if (ptcb->tcb_sndbuf == NULL) {
    console_puts("tcp_sync: sndbuf OOM");
    return FALSE;
  }

  ptcb->tcb_sbsize = TCPSBS;
  ptcb->tcb_sbstart = ptcb->tcb_sbcount = 0;
  ptcb->tcb_wrpid = 0;
  ptcb->tcb_wrstatword = 0;

  ptcb->tcb_rcvbuf = (byte *)p_alloc(TCPRBS);
  if (ptcb->tcb_rcvbuf == NULL) {
    console_puts("tcp_sync: rcvbuf OOM");
    p_free(ptcb->tcb_sndbuf);
    return FALSE;
  }
  ptcb->tcb_rbsize = TCPRBS;
  ptcb->tcb_rbstart = ptcb->tcb_rbcount = 0;
  ptcb->tcb_rsegq = NULL;
  ptcb->tcb_rdpid = 0;
  ptcb->tcb_rdstatword = 0;

  ptcb->tcb_rudq = ptcb->tcb_ruhq = NULL;
  ptcb->tcb_sudq = ptcb->tcb_suhq = NULL;

  /* FIXME ptcb->tcb_ocsem = screate(0);*/
  ptcb->tcb_ocpid = NULL;
  ptcb->tcb_ocstatword = NULL;

  /* timer stuff */
  /* FIXME: the PsiStack timer isn't in 100ths/sec... */
  ptcb->tcb_srt = 0;    /* in sec/100  */
  ptcb->tcb_rtde = 0;    /* in sec/100  */
  ptcb->tcb_rexmt = 50;    /* in sec/100  */
  ptcb->tcb_rexmtcount = 0;
  ptcb->tcb_keep = 12000;    /* in sec/100  */

  ptcb->tcb_code = TCPF_SYN;
  ptcb->tcb_flags = 0;
console_puts("tcp_sync: out ok - here is the filled in TCB");
tcb_dump(ptcb);
  return TRUE;
}


/*******************************************************************************
***
*** Function         : tcp_i_synsent
*** Preconditions    : The TCB is in SYN_SENT state
*** Postconditions   : SYN_SENT input state processing has been done.
***
*******************************************************************************/

int tcpi_synsent(struct tcb *ptcb, struct buf *pbuf)
{
struct ip *pip = (struct ip *)pbuf->data;
struct tcp *ptcp = (struct tcp *)pip->ip_data;
console_puts("tcpi_synsent: in");

  if ((ptcp->tcp_code & TCPF_ACK) &&
      ((ptcp->tcp_ack - ptcb->tcb_iss <= 0) ||
      (ptcp->tcp_ack - ptcb->tcb_snext) > 0)) {
console_puts("ACK received for incorrect sequence: resetting...");
    return tcp_reset(pbuf);
  }
  if (ptcp->tcp_code & TCPF_RST) {
console_puts("RST received: going CLOSED and informing client");
    ptcb->tcb_istate = TCPIS_CLOSED;
    ptcb->tcb_error = E_NET_CONNREFUSED;
    tcpi_stat_attemptfails++;
    timer_killtimers(ptcb);
    p_pcpyto(ptcb->tcb_ocpid, (VOID *)ptcb->tcb_ocstatword, &ptcb->tcb_error, 
             sizeof(UWORD));
    p_iosignalbypid(ptcb->tcb_ocpid);
    return TRUE;
  }
  if ((ptcp->tcp_code & TCPF_SYN) == 0) {
console_puts("no SYN...OK");
    return TRUE;
  }
  ptcb->tcb_swindow = ptcp->tcp_window;
  ptcb->tcb_lwseq = ptcp->tcp_seq;
  ptcb->tcb_rnext = ptcp->tcp_seq;
  ptcb->tcb_cwin = ptcb->tcb_rnext + ptcb->tcb_rbsize;
  tcp_acked(ptcb, pbuf);
  tcp_data(ptcb, pbuf);
  ptcp->tcp_code &= ~TCPF_FIN;
  if (ptcb->tcb_code & TCPF_SYN) {   /* our SYN not ACKed  */
console_puts("SYN not ACKed, going to SYNRCVD");
    ptcb->tcb_istate = TCPIS_SYNRCVD;
  }
  else {
console_puts("YAHOO! going to ESTABLISHED");
    tcpi_stat_currestab++;
    ptcb->tcb_istate = TCPIS_ESTABLISHED;
    ptcb->tcb_error = 0; /* OK */
    p_pcpyto(ptcb->tcb_ocpid, (VOID *)ptcb->tcb_ocstatword, &ptcb->tcb_error, 
             sizeof(UWORD));
    p_iosignalbypid(ptcb->tcb_ocpid);
  }
  return TRUE;
}


/*******************************************************************************
***
*** Function         : tcpi_synrcvd
*** Preconditions    : The TCB is in SYN_RCVD state
*** Postconditions   : SYN_RCVD input state processing has been done.
***
*******************************************************************************/

int tcpi_synrcvd(struct tcb *ptcb, struct buf *pbuf)
{
struct ip *pip = (struct ip *)pbuf->data;
struct tcp *ptcp = (struct tcp *)pip->ip_data;
struct tcb *pptcb;

  if (ptcp->tcp_code & TCPF_RST) {
    tcpi_stat_attemptfails++;
    /* For passively opened connections, the TCB is a separate copy of the 
       Parent TCB, so just free the orphan. For actively opened connections,
       record the error in the TCB. */
    if (ptcb->tcb_pptcb != 0)
      return tcb_free(ptcb);
    else
      return tcp_abort(ptcb, E_NET_CONNREFUSED);
  }
  if (ptcp->tcp_code & TCPF_SYN) {
    /* Sorry. Can't handle a SYN in this state. */
    tcpi_stat_attemptfails++;
    tcp_reset(pbuf);
    return tcp_abort(ptcb, E_NET_CONNRESET);
  }
  /* Because TCP only enters the SYN-RECEIVED state after responding to a SYN,
     any incoming segment other than RESET or SYN means that the other side
     views the connection as established. tcp_acked handles acknowledgement
     and the move to the ESTABLISHED state; it also extracts the data
  */
  if (tcp_acked(ptcb, pbuf) == FALSE)
    return TRUE;

  if (ptcb->tcb_pptcb != 0) {                  /* If a "Parent TCB" exists... */

    /**** PASSIVE OPEN (a.k.a. Listen/Accept) ****/

console_puts("tcpi_synrcvd: passive open");
    pptcb = ptcb->tcb_pptcb;         /* "Parent TCB" is stored by tcpi_listen */

    if (pptcb->tcb_istate != TCPIS_LISTEN) {
      tcpi_stat_attemptfails++;
      tcp_reset(pbuf);
      /* FIXME signal(pptcb->tcb_mutex);*/
      return tcb_free(ptcb);
    }

    /* If a client process has called accept(), their PID and status word 
       will be recorded in the TCB. We then create a new socket for this 
       connection, and return the new socket descriptor to them in their 
       status word. If no client has yet called accept, then enqueue this 
       connection until one does.
    */
    if (ptcb->tcb_ocpid != 0) {
console_puts("tcpi_synrcvd: unblocking server process");
      /* Unblock process with the new connection */
      if (! socket_unblockpassiveopen(ptcb)) {
console_puts("tcpi_synrcvd: unblock failed");
        tcpi_stat_attemptfails++;
        return tcb_free(ptcb);
      }
      /* socket_unblockpassiveopen does an iosignalbypid - 
         will this reschedule? */
    }
    else {
      /* No client process has accept()ed on this socket yet. Can't block 
         an input process, so enqueue connection identifier on listen queue.
         socket_accept will pick up the last entry enqueued on the listen queue.
         ptcb has already been opened for this connection.
      */
      if (queue_full(pptcb->tcb_listenq)) {
        tcpi_stat_attemptfails++;
console_puts("tcpi_synrcvd: no accept outstanding");
        /*  FIXME signal(pptcb->tcb_mutex);*/
        return tcb_free(ptcb);
      }
console_puts("tcpi_synrcvd: tcb enqueued on listenq");
      queue_enqueue(pptcb->tcb_listenq, (int)ptcb);
    }
  }

  else {
console_puts("tcpi_synrcvd: active open");
    /**** ACTIVE OPEN (a.k.a. Connect) ****/
    /*  FIXME signal(ptcb->tcb_ocsem);*/
    /* Wake the waiting process, i.e. write to the process' status word. */
  }
console_puts("tcpi_synrcvd: changing to established");
  tcpi_stat_currestab++;
  ptcb->tcb_istate = TCPIS_ESTABLISHED;
  tcp_data(ptcb, pbuf);
  if (ptcb->tcb_flags & TCBF_RDONE)
    ptcb->tcb_istate = TCPIS_CLOSEWAIT;
  return TRUE;
}


/*******************************************************************************
***
*** Function         : tcpi_listen
*** Preconditions    : The TCB is in LISTEN state
*** Postconditions   : LISTEN state input processing has been done.
***
*******************************************************************************/

int tcpi_listen(struct tcb *ptcb, struct buf *pbuf)
{
struct tcb *newptcb;
struct ip *pip = (struct ip *)pbuf->data;
struct tcp *ptcp = (struct tcp *)pip->ip_data;

  if (flags.log_flag & LOG_FLAG_UPPERDUMP) 
    console_puts("tcpi_listen...");

  if (ptcp->tcp_code & TCPF_RST)
    return TRUE;    /* "parent" TCB still in LISTEN */
  if ((ptcp->tcp_code & TCPF_ACK) ||
      (ptcp->tcp_code & TCPF_SYN) == 0)
    return tcp_reset(pbuf);
  if (flags.log_flag & LOG_FLAG_UPPERDUMP) 
    console_puts("Creating new TCB");
  newptcb = tcb_alloc();
  if (newptcb == NULL || tcp_sync(newptcb) == FALSE)
    return NULL;
  console_puts("tcpi_listen filling in rest of TCB");
  newptcb->tcb_istate = TCPIS_SYNRCVD;
  newptcb->tcb_ostate = TCPOS_IDLE;
  newptcb->tcb_error = 0;
  newptcb->tcb_pptcb = ptcb;      /* for ACCEPT  */

  p_bcpy(newptcb->tcb_rip, pip->ip_src, IP_ALEN);
  newptcb->tcb_rport = ptcp->tcp_sport;
  p_bcpy(newptcb->tcb_lip, pip->ip_dst, IP_ALEN);
  newptcb->tcb_lport = ptcp->tcp_dport;

  tcp_winit(ptcb, newptcb, pbuf);  /* initialize window data  */

  newptcb->tcb_finseq = newptcb->tcb_pushseq = 0;
  newptcb->tcb_flags = TCBF_NEEDOUT;
  tcpi_stat_passiveopens++;
  ptcp->tcp_code &= ~TCPF_FIN;  /* don't process FINs in LISTEN */
  if (flags.log_flag & LOG_FLAG_UPPERDUMP) {
    console_puts("new TCB created OK:");
    tcb_dump(newptcb);
  }
    
  tcp_data(newptcb, pbuf);
  /* FIXME signal(newptcb->tcb_mutex);*/
}


/*******************************************************************************
***
*** Function         : tcp_winit
*** Preconditions    : 
*** Postconditions   : The window data of the TCB is initialised for a new
***                    connection.
***
*******************************************************************************/

int tcp_winit(struct tcb *ptcb, struct tcb *newptcb, struct buf *pbuf)
{
struct ip *pip = (struct ip *)pbuf->data;
struct tcp *ptcp = (struct tcp *)pip->ip_data;
int local, mss, ifnum;

  newptcb->tcb_swindow = ptcp->tcp_window;
  newptcb->tcb_lwseq = ptcp->tcp_seq;
  newptcb->tcb_lwack = newptcb->tcb_iss;  /* set in tcp_sync()  */
console_puts("diag: in tcp_winit");
  ifnum = which_nif(pip->ip_src, FALSE);
  local = (ifnum == NI_LOCAL);
  newptcb->tcb_pni = &nif[ifnum];
  if (local)
    mss = newptcb->tcb_pni->ni_mtu-IPMHLEN-TCPMHLEN;
  else
    mss = 536;  /* RFC 1122 */
  if (ptcb->tcb_smss) {
    newptcb->tcb_smss = min(ptcb->tcb_smss, mss);
    ptcb->tcb_smss = 0;    /* reset server smss  */
  }
  else
    newptcb->tcb_smss = mss;
  newptcb->tcb_rmss = mss;    /* receive mss    */
  newptcb->tcb_cwnd = newptcb->tcb_smss;  /* 1 segment    */
  newptcb->tcb_ssthresh = 0xFFFF;    /* IP max window  */
  newptcb->tcb_rnext = ptcp->tcp_seq;
  newptcb->tcb_cwin = newptcb->tcb_rnext + newptcb->tcb_rbsize;
console_puts("diag: out tcp_winit");
  return TRUE;
}


/*******************************************************************************
***
*** Function         : tcp_ackit
*** Preconditions    : 
*** Postconditions   : An ACK is generated for a received TCP packet. Returns
***                    TRUE or FALSE depending on whether the ACK could be
***                    generated.
***
*******************************************************************************/

int tcp_ackit(struct tcb *ptcb, struct buf *pbufin)
{
struct buf *pbufout;
struct ip  *pipin = (struct ip *)pbufin->data, *pipout;
struct tcp *ptcpin = (struct tcp *)pipin->ip_data, *ptcpout;

  if (ptcpin->tcp_code & TCPF_RST)
    return TRUE;
  if (pipin->ip_len <= IP_HLEN(pipin) + TCP_HLEN(ptcpin) &&
      !(ptcpin->tcp_code & (TCPF_SYN|TCPF_FIN)))
    return TRUE;  /* duplicate ACK */
  pbufout = (struct buf *)buf_alloc(576 /*Net.netpool*/ ); /* FIXME: Needs to be the MTU */
  if (pbufout == NULL)
    return FALSE;
  pbufout->usedsize = 576; /* FIXME */
  pipout = (struct ip *)pbufout->data;
  p_bcpy(pipout->ip_src, pipin->ip_dst, IP_ALEN);
  p_bcpy(pipout->ip_dst, pipin->ip_src, IP_ALEN);
  pipout->ip_proto = IPT_TCP;
  pipout->ip_verlen = (IP_VERSION<<4) | IP_MINHLEN;
  pipout->ip_len = IP_HLEN(pipout) + TCPMHLEN;
  ptcpout = (struct tcp *)pipout->ip_data;
  ptcpout->tcp_sport = ptcpin->tcp_dport;
  ptcpout->tcp_dport = ptcpin->tcp_sport;
  ptcpout->tcp_seq = ptcb->tcb_snext;
  ptcpout->tcp_ack = ptcb->tcb_rnext;
  ptcpout->tcp_code = TCPF_ACK;
  ptcpout->tcp_offset = TCPHOFFSET;
  ptcpout->tcp_window = tcp_rwindow(ptcb);
  ptcpout->tcp_urgptr = 0;
  ptcpout->tcp_cksum = 0;
  tcp_h2net(ptcpout);
  ptcpout->tcp_cksum = tcp_cksum(pipout);
  tcpo_stat_outsegs++;
  ip_send(pipout->ip_dst, pbufout, TCPMHLEN);
  return TRUE;
}


/*******************************************************************************
***
*** Function         : tcp_acked
*** Preconditions    : 
*** Postconditions   : In-bound ACKs and round-trip estimates are done.
***
*******************************************************************************/

int tcp_acked(struct tcb *ptcb, struct buf *pbuf)
{
struct  ip  *pip = (struct ip *)pbuf->data;
struct  tcp *ptcp = (struct tcp *)pip->ip_data;
int     acked, cacked;
/*STATWORD  ps;*/

  if (!(ptcp->tcp_code & TCPF_ACK))
    return FALSE;
  acked = (int)(ptcp->tcp_ack - ptcb->tcb_suna);
  cacked = 0;
  if (acked <= 0)
    return 0;  /* duplicate ACK */ /* FIXME? 0 == NULL? */
  if (SEQCMP(ptcp->tcp_ack, ptcb->tcb_snext) > 0)
    if (ptcb->tcb_istate == TCPIS_SYNRCVD)
      return tcp_reset(pbuf);
    else
      return tcp_ackit(ptcb, pbuf);
  tcp_rtt(ptcb);
  ptcb->tcb_suna = ptcp->tcp_ack;
  if (acked && ptcb->tcb_code & TCPF_SYN) {
    acked--;
    cacked++;
    ptcb->tcb_code &= ~TCPF_SYN;
    ptcb->tcb_flags &= ~TCBF_FIRSTSEND;
  }
  if ((ptcb->tcb_code & TCPF_FIN) &&
      SEQCMP(ptcp->tcp_ack, ptcb->tcb_snext) == 0) {
    acked--;
    cacked++;
    ptcb->tcb_code &= ~TCPF_FIN;
    ptcb->tcb_flags &= ~TCBF_SNDFIN;
  }
  ptcb->tcb_sbstart = (ptcb->tcb_sbstart+acked) % ptcb->tcb_sbsize;
  ptcb->tcb_sbcount -= acked;
  if (acked) {
    /* FIXME
    if (scount(ptcb->tcb_ssema) <= 0)
      signal(ptcb->tcb_ssema);
    */
  }
  tcp_ostate(ptcb, acked+cacked);
  return acked;
}


/*******************************************************************************
***
*** Function         : tcp_opts
*** Preconditions    : 
*** Postconditions   : TCP options for an inbound segment are handled.
***
*******************************************************************************/

int tcp_opts(struct tcb *ptcb, struct buf *pbuf)
{
struct  ip  *pip = (struct ip *)pbuf->data;
struct  tcp *ptcp = (struct tcp *)pip->ip_data;
byte    *popt, *popend;
int    len;

  if (TCP_HLEN(ptcp) == TCPMHLEN)
    return TRUE;
console_puts("tcp_opt: there are options...");
  popt = ptcp->tcp_data;
  popend = &pip->ip_data[TCP_HLEN(ptcp)];
  do {
    switch (*popt) {
    case TPO_NOOP:  popt++;
console_puts("tcp_opt: noop");
      break;
    case TPO_EOOL:  
console_puts("tcp_opt: eool");
      break;
    case TPO_MSS:
console_puts("tcp_opt: mss");
      popt += tcp_smss(ptcb, ptcp, popt);
      break;
    default:
      break;
    }
  } while (*popt != TPO_EOOL && popt<popend);

  /* delete the options */
  len = pip->ip_len-IP_HLEN(pip)-TCP_HLEN(ptcp);
  if (len)
    p_bcpy(ptcp->tcp_data, &pip->ip_data[TCP_HLEN(ptcp)], len);
  pip->ip_len = IP_HLEN(pip) + TCPMHLEN + len;
  ptcp->tcp_offset = TCPHOFFSET;
  return TRUE;
}


/*******************************************************************************
***
*** Function         : tcp_swindow
*** Preconditions    : 
*** Postconditions   : Send window updates from remote are handled
***
*******************************************************************************/

int tcp_swindow(struct tcb *ptcb, struct buf *pbuf)
{
struct ip  *pip = (struct ip *)pbuf->data;
struct tcp *ptcp = (struct tcp *)pip->ip_data;
tcpseq    wlast, owlast;

  if (SEQCMP(ptcp->tcp_seq, ptcb->tcb_lwseq) < 0)
    return TRUE;
  if (SEQCMP(ptcp->tcp_seq, ptcb->tcb_lwseq) == 0 &&
      SEQCMP(ptcp->tcp_ack, ptcb->tcb_lwack) < 0)
    return TRUE;
  /* else, we have a send window update */

  /* compute the last sequences of the new and old windows */

  owlast = ptcb->tcb_lwack + ptcb->tcb_swindow;
  wlast = ptcp->tcp_ack + ptcp->tcp_window;

  ptcb->tcb_swindow = ptcp->tcp_window;
  ptcb->tcb_lwseq = ptcp->tcp_seq;
  ptcb->tcb_lwack = ptcp->tcp_ack;
  if (SEQCMP(wlast, owlast) <= 0)
    return TRUE;
  /* else,  window increased */
  if (ptcb->tcb_ostate == TCPOS_PERSIST) {
    timer_clear(MKEVENT(TQ_TCP_PERSIST, ptcb-&tcbtab[0]));
    ptcb->tcb_ostate = TCPOS_XMIT;
  }
  tcp_kick(ptcb);      /* do something with it */
  return TRUE;
}


/*******************************************************************************
***
*** Function         : tcp_kick
*** Preconditions    : There is a packet to send immediately
*** Postconditions   : It is send, circumventing the timer event queue.
***
*******************************************************************************/

int tcp_kick(struct tcb *ptcb)
{
int  tcbnum = ptcb - &tcbtab[0];  /* for MKEVENT() */
int  tv;
  tv = MKEVENT(TQ_TCP_SEND, tcbnum);
  if (ptcb->tcb_flags & TCBF_DELACK && !timer_left(tv)) {
console_puts("tcp_kick: kicking with an ACK delay");
    timer_set(tv, TCP_ACKDELAY);
  }
  else {
console_printf("tcp_kick: kicking now - enqueueing 0x%04x\r\n",tv);
    if (!iqueue_enqueue(tcp_oqueue, tv)) {  /* send now */
      timer_stat_oqueuefull++;
      console_puts("tcp_kick: output queue full");
    }
    /* Tell the TCP output process that there's something waiting for it */
    attn_signal(ATTN_TCPOUTPUT);
  }
console_puts("tcp_kick: out");
  return TRUE;
}


/*******************************************************************************
***
*** Function         : tcp_rwindow
*** Preconditions    : 
*** Postconditions   : Receive window silly-window-syndrome avoidance is
***                    done for a TCB
***
*******************************************************************************/

int tcp_rwindow(struct tcb *ptcb)
{
int  window;

  window = ptcb->tcb_rbsize - ptcb->tcb_rbcount;
  if (ptcb->tcb_istate < TCPIS_ESTABLISHED)
    return window;
  /*
   *  Receiver-Side Silly Window Syndrome Avoidance:
   *  Never shrink an already-advertised window, but wait for at
   *  least 1/4 receiver buffer and 1 max-sized segment before
   *  opening a zero window.
   */
  if (window*4 < ptcb->tcb_rbsize || window < ptcb->tcb_rmss)
    window = 0;
  window = max(window, (int)(ptcb->tcb_cwin - ptcb->tcb_rnext));
  ptcb->tcb_cwin = ptcb->tcb_rnext + window;
  return window;
}


/*******************************************************************************
***
*** Function         : tcp_ostate
*** Preconditions    : 
*** Postconditions   : TCP output state processing after an ACK is done.
***                    Perhaps this should be in tcpout.c? It isn't used there,
***                    but that seems like a better place for it?
***
*******************************************************************************/

int tcp_ostate(struct tcb *ptcb, int acked)
{
  if (acked <= 0)
    return TRUE;  /* no state change */
  if (ptcb->tcb_ostate == TCPOS_REXMT) {
    ptcb->tcb_rexmtcount = 0;
    ptcb->tcb_ostate = TCPOS_XMIT;
  }
  if (ptcb->tcb_sbcount == 0) {
    ptcb->tcb_ostate = TCPOS_IDLE;
    return TRUE;
  }
  tcp_kick(ptcb);
  return TRUE;
}


/*******************************************************************************
***
*** Function         : tcp_rtt
*** Preconditions    : 
*** Postconditions   : Round trip time esitames and congestion window
***                    processing are preformed.
***
*******************************************************************************/

int tcp_rtt(struct tcb *ptcb)
{
int  rrt,    /* raw round trip    */
     delta;    /* deviation from smoothed  */

  rrt = timer_clear(MKEVENT(TQ_TCP_RETRANSMIT, ptcb-&tcbtab[0]));
  if (rrt != NULL && ptcb->tcb_ostate != TCPOS_REXMT) {
    if (ptcb->tcb_srt == 0)
      ptcb->tcb_srt = rrt<<3;  /* prime the pump */
    /*
     * "srt" is scaled X 8 here, so this is really:
     *  delta = rrt - srt
     */
    delta = rrt - (ptcb->tcb_srt>>3);
    ptcb->tcb_srt += delta;    /* srt = srt + delta/8  */
    if (delta < 0)
      delta = -delta;
    /*
     * "rtde" is scaled X 4 here, so this is really:
     *  rtde = rtde + (|delta| - rtde)/4
     */
    ptcb->tcb_rtde += delta - (ptcb->tcb_rtde>>2);
    /*
     * "srt" is scaled X 8, rtde scaled X 4, so this is:
     *  rto = srt + 2 * rtde
     */
    ptcb->tcb_rexmt = ((ptcb->tcb_srt>>2)+ptcb->tcb_rtde)>>1;
    if (ptcb->tcb_rexmt < TCP_MINRXT)
      ptcb->tcb_rexmt = TCP_MINRXT;
  }
  if (ptcb->tcb_cwnd < ptcb->tcb_ssthresh)
    ptcb->tcb_cwnd += ptcb->tcb_smss;
  else
    ptcb->tcb_cwnd += (ptcb->tcb_smss * ptcb->tcb_smss) /
       ptcb->tcb_cwnd;
  return TRUE;
}


/*******************************************************************************
***
*** Function         : tcp_smss
*** Preconditions    : 
*** Postconditions   : Sender MSS is set from option in incoming segment
***
*******************************************************************************/

int tcp_smss(struct tcb *ptcb, struct tcp *ptcp, char *popt)
{
unsigned int mss, len;

  len = *++popt;
console_printf("tcp_smss: len is %d\r\n", (UINT)len);
  ++popt;    /* skip length field */
  if ((ptcp->tcp_code & TCPF_SYN) == 0)
    return len;
  switch (len-2) {  /* subtract kind & len  */
    case sizeof(char):
      mss = *popt;
console_printf("tcp_smss: char - %d\r\n", mss);
      break;
    case sizeof(short):
      mss = net2hs(*(unsigned short *)popt);
console_printf("tcp_smss: short - %d\r\n", mss);
      break;
    case sizeof(long):
      mss = net2hl(*(unsigned long *)popt);
console_printf("tcp_smss: long - %d\r\n", mss);
      break;
    default:
      mss = ptcb->tcb_smss;
console_printf("tcp_smss: default - %d\r\n", mss);
      break;
  }
  mss -= TCPMHLEN;  /* save just the data buffer size */
  if (ptcb->tcb_smss)
    ptcb->tcb_smss = min(mss, ptcb->tcb_smss);
  else
    ptcb->tcb_smss = mss;
console_printf("tcp_smss: MSS is %d\r\n", ptcb->tcb_smss);
  return len;
}

