/*******************************************************************************
***
*** Filename         : tcpout.c
*** Purpose          : TCP output processing for PsiStack
*** Author           : Matt J. Gumbley
*** Created          : 26/08/97
*** Last updated     : 27/02/98
***
********************************************************************************
***
*** Modification Record
***
*******************************************************************************/

#include "stack.h"

/* Some stats about the TCP output layer */
struct logstat tcpo_stat[] = {
  "checksum", 0,                                     /* # TCP Checksum errors */
  "noTCB", 0,                       /* # datagrams for which there was no TCB */
  "retranssegs", 0,                               /* # retransmitted segments */
  "activeopens", 0,                                         /* # active opens */
  "outsegs", 0,                                     /* # segments transmitted */
  NULL, 0
};
/* #defines are in the tcpout.h file */

/* Output event processing */


int (*tcpo_switch[NTCPOSTATES])(int, int) = { 
  tcpo_idle,                                                    /* TCPOS_IDLE */
  tcpo_persist,                                              /* TCPOS_PERSIST */
  tcpo_xmit,                                                    /* TCPOS_XMIT */
  tcpo_rexmt,                                                  /* TCPOS_REXMT */
};


/*******************************************************************************
***
*** Function         : tcpo_process()
*** Preconditions    : Events are queued on the TCP output queue, and an
***                    attention event has been flagged.
*** Postconditions   : The events are processed.
***
*******************************************************************************/

void tcpo_process()
{
register int i;
struct tcb *ptcb;
static TEXT *tcpo_states[] = {
  "IDLE", "PERSIST", "XMIT", "REXMT" };

console_puts("tcpo_process: in");

  if (iqueue_empty(tcp_oqueue)) {
    console_puts("tcpo_process: nothing to do - bug!");
    return;
  }

  /* Assertion: the queue is not empty, so get the first entry */
  i = iqueue_dequeue(tcp_oqueue);
console_printf("tcpo_process: dequeued 0x%04x\r\n", i);
  if (DATA(i) >= tcp_config_used.tcpc_NTCBS) {
    console_puts("tcpo_process: weird TCB#");
    return;
  }

  if (flags.log_flag & LOG_FLAG_UPPERDUMP) {
    console_printf("tcpo_process: TCB# event: %s data: %d\r\n", 
                    tq_eventnames[(EVENT(i))], DATA(i));
  }

  ptcb = &tcbtab[DATA(i)];
  if (flags.log_flag & LOG_FLAG_UPPERDUMP)
    tcb_dump(ptcb);

  if (ptcb->tcb_istate <= TCPIS_CLOSED) {
    if (flags.log_flag & LOG_FLAG_UPPERDUMP)
      console_puts("TCB deallocated");
    return;    /* TCB deallocated  */
  }
  if (EVENT(i) == TQ_TCP_DELETE) {   /* same for all states  */
    if (flags.log_flag & LOG_FLAG_UPPERDUMP)
      console_puts("Freeing TCB");
    tcb_free(ptcb);
  }
  else {
    if (flags.log_flag & LOG_FLAG_UPPERDUMP)
      console_printf("Passing to state handler %d (%s)\r\n", ptcb->tcb_ostate, tcpo_states[ptcb->tcb_ostate]);
    tcpo_switch[ptcb->tcb_ostate](DATA(i), EVENT(i));
  }
}


/*******************************************************************************
***
*** Function         : tcpo_idle
*** Preconditions    : The connection is IDLE
*** Postconditions   : IDLE output state processing is performed.
***
*******************************************************************************/

int tcpo_idle(int tcbnum, int event)
{
console_puts("tcpo_idle: in");
  if (event == TQ_TCP_SEND) {
console_puts("tcpo_idle: send event, calling tcpo_xmit");
    tcpo_xmit(tcbnum, event);
  }
  return TRUE;
}


/*******************************************************************************
***
*** Function         : tcpo_persist
*** Preconditions    : The send window is closed
*** Postconditions   : Events are handled.
***
*******************************************************************************/

int tcpo_persist(int tcbnum, int event)
{
struct  tcb  *ptcb = &tcbtab[tcbnum];

  if (event != TQ_TCP_PERSIST)
    return TRUE;  /* ignore everything else */
  tcp_send(tcbnum, TSF_REXMT);
  ptcb->tcb_persist = min(ptcb->tcb_persist<<1, TCP_MAXPRS);
  timer_set(MKEVENT(TQ_TCP_PERSIST, tcbnum), ptcb->tcb_persist);
  return TRUE;
}


/*******************************************************************************
***
*** Function         : tcpo_xmit
*** Preconditions    : 
*** Postconditions   : TCP output events while we are transmitting are handled.
***
*******************************************************************************/

int tcpo_xmit(int tcbnum, int event)
{
struct  tcb  *ptcb = &tcbtab[tcbnum];
int    tv, pending;
word32 tosend, window;

TEXT buf[80];

  if (event == TQ_TCP_RETRANSMIT) {
console_puts("tcpo_xmit: RETRANSMIT");
    timer_clear(MKEVENT(TQ_TCP_SEND, tcbnum));
    tcpo_rexmt(tcbnum, event);
    ptcb->tcb_ostate = TCPOS_REXMT;
    return TRUE;
  } /* else SEND */
console_puts("tcpo_xmit: SEND");
console_puts("the tcb is:");
tcb_dump(ptcb);

  tcp_sndurg(tcbnum);    /* send urgent data, if any  */
  tosend = tcp_howmuch(ptcb);
p_atos(buf, "tosend = %ld", tosend);
console_puts(buf);
  if (tosend == 0) {
    if (ptcb->tcb_flags & TCBF_NEEDOUT)
      tcp_send(tcbnum, TSF_NEWDATA);  /* just an ACK */
    if (ptcb->tcb_snext == ptcb->tcb_suna)
      return TRUE;
    /* still unacked data; restart transmit timer  */
    tv = MKEVENT(TQ_TCP_RETRANSMIT, tcbnum);
    if (!timer_left(tv))
      timer_set(tv, ptcb->tcb_rexmt);
    return TRUE;
  }
  else
    if (ptcb->tcb_swindow == 0) {
console_puts("tcpo_xmit: swindow is 0, entering PERSIST out state");
      ptcb->tcb_ostate = TCPOS_PERSIST;
      ptcb->tcb_persist = ptcb->tcb_rexmt;
      tcp_send(tcbnum, TSF_NEWDATA);
      timer_set(MKEVENT(TQ_TCP_PERSIST,tcbnum), ptcb->tcb_persist);
console_puts("tcpo_xmit out");
      return TRUE;
    }  /* else, we have data and window */
console_puts("tcpo_xmit: we have data and window - entering XMIT Out state");
  ptcb->tcb_ostate = TCPOS_XMIT;
  window = min(ptcb->tcb_swindow, ptcb->tcb_cwnd);
  pending = (int)(ptcb->tcb_snext - ptcb->tcb_suna);
  while (tcp_howmuch(ptcb) > 0 && pending <= window) {
p_atos(buf,"tcpo_xmit: sending: %d bytes pending, window is 0x%08lx", pending,window);
console_puts(buf);
    tcp_send(tcbnum, TSF_NEWDATA);
    pending = (int)(ptcb->tcb_snext - ptcb->tcb_suna);
  }
console_puts("tcpo_xmit: scheduling a retransmit timer?");
  tv = MKEVENT(TQ_TCP_RETRANSMIT, tcbnum);
  if (!timer_left(tv)) {
console_puts("tcpo_xmit: yes, scheduling it.");
    timer_set(tv, ptcb->tcb_rexmt);
  }
console_puts("tcpo_xmit out at end");
  return TRUE;
}


/*******************************************************************************
***
*** Function         : tcp_send
*** Preconditions    : 
*** Postconditions   : A TCP segment for the given TCB is computed and sent.
***
*******************************************************************************/

int tcp_send(int tcbnum, int rexmt)
{
struct  tcb  *ptcb = &tcbtab[tcbnum];
struct  buf  *pbuf;
struct  ip  *pip;
struct  tcp  *ptcp;
byte    *pch;
int    i, datalen, tocopy, off, newdata;

TEXT buf[80];

console_puts("tcp_send");
  pbuf = (struct buf *)buf_alloc(576 /*Net.netpool*/); /* FIXME */ /* should be the MTU, I think */
  if (pbuf == NULL) {
    console_puts("tcp_send: OOM");
    return FALSE;
  }
  pip = (struct ip *)pbuf->data;
  p_bcpy(pip->ip_src, ptcb->tcb_lip, IP_ALEN);
  p_bcpy(pip->ip_dst, ptcb->tcb_rip, IP_ALEN);
  pip->ip_proto = IPT_TCP;
  datalen = tcp_sndlen(ptcb, rexmt, &off);  /* get length & offset  */
console_puts("got length");
  ptcp = (struct tcp *)pip->ip_data;
  ptcp->tcp_sport = ptcb->tcb_lport;
  ptcp->tcp_dport = ptcb->tcb_rport;
console_puts("filled in ports");
  if (!rexmt) {
    if (ptcb->tcb_code & TCPF_URG)
      ptcp->tcp_seq = ptcb->tcb_suna + off;
    else
      ptcp->tcp_seq = ptcb->tcb_snext;
  }
  else
    ptcp->tcp_seq = ptcb->tcb_suna;
  ptcp->tcp_ack = ptcb->tcb_rnext;

p_atos(buf,"Seq 0x%08lx Ack 0x%08lx", ptcp->tcp_seq, ptcp->tcp_ack);
console_puts(buf);

  if ((ptcb->tcb_flags & TCBF_SNDFIN) &&
      SEQCMP(ptcp->tcp_seq+datalen, ptcb->tcb_slast) == 0)
    ptcb->tcb_code |= TCPF_FIN;
  ptcp->tcp_code = ptcb->tcb_code;
  ptcp->tcp_offset = TCPHOFFSET;
  if ((ptcb->tcb_flags & TCBF_FIRSTSEND) == 0)
    ptcp->tcp_code |= TCPF_ACK;
  if (ptcp->tcp_code & TCPF_SYN) {
console_puts("calling tcp_rmss");
    tcp_rmss(ptcb, pip);
console_puts("tcp_rmss out");
  }
console_puts("tcb is now");
tcb_dump(ptcb);

  pip->ip_verlen = (IP_VERSION<<4) | IP_MINHLEN;

  /* FIXME: might be wrong but can't see why just now... */
  pbuf->usedsize = pip->ip_len = IP_HLEN(pip) + TCP_HLEN(ptcp) + datalen;
console_printf("usedsize is %d\r\n", pbuf->usedsize);

  if (datalen > 0)
    ptcp->tcp_code |= TCPF_PSH;
  ptcp->tcp_window = tcp_rwindow(ptcb);
  if (ptcb->tcb_code & TCPF_URG)
    if (tcp_config.tcpc_bsdurg)
      ptcp->tcp_urgptr = datalen;  /* 1 past end    */
    else
      ptcp->tcp_urgptr = datalen-1;
  else
    ptcp->tcp_urgptr = 0;
  pch = &pip->ip_data[TCP_HLEN(ptcp)];
console_puts("copying to send buffer");

  i = (ptcb->tcb_sbstart+off) % ptcb->tcb_sbsize;
console_printf("sbsize is %d, sbstart is %d off is %d\r\n", ptcb->tcb_sbsize, ptcb->tcb_sbstart, off);

  for (tocopy=datalen; tocopy > 0; --tocopy) {
console_printf("i is %d, topcopy is %d\r\n", (UINT) i, tocopy);
    *pch++ = ptcb->tcb_sndbuf[i];
    if (++i >= ptcb->tcb_sbsize)
      i = 0;
  }
console_puts("copied");
  ptcb->tcb_flags &= ~TCBF_NEEDOUT;  /* we're doing it  */
  if (rexmt) {
console_puts("rexmt set");
    newdata = (int)(ptcb->tcb_suna + datalen - ptcb->tcb_snext);
    if (newdata < 0)
      newdata = 0;
    tcpo_stat_retranssegs++;
  }
  else {
console_puts("rexmt not set");
    newdata = datalen;
    if (ptcb->tcb_code & TCPF_SYN)
      newdata++; /* SYN is part of the sequence  */
    if (ptcb->tcb_code & TCPF_FIN)
      newdata++; /* FIN is part of the sequence  */
  }
  ptcb->tcb_snext += newdata;
  if (newdata >= 0)
    tcpo_stat_outsegs++;
  if (ptcb->tcb_istate == TCPIS_TIMEWAIT) { /* final ACK    */
console_puts("calling tcp_wait");
    tcp_wait(ptcb);
  }
console_puts("changeing to net word order");
  datalen += TCP_HLEN(ptcp);
  tcp_h2net(ptcp);
  ptcp->tcp_cksum = 0;
  ptcp->tcp_cksum = tcp_cksum(pip);
console_puts("calling ip_send");
  ip_send(pip->ip_dst, pbuf, datalen);
console_puts("tcp_send out");
  return TRUE;
}


/*******************************************************************************
***
*** Function         : tcp_sndlen
*** Preconditions    : 
*** Postconditions   : The packet length and offset in sndbuf is computed.
***
*******************************************************************************/

int tcp_sndlen(struct tcb *ptcb, int rexmt, int *poff)
{
struct   uqe  *puqe, *puqe2;
int    datalen;

console_puts("tcp_sndlen");
  if (rexmt || ptcb->tcb_sudq == NULL) {
    if (rexmt || (ptcb->tcb_code & TCPF_SYN))
      *poff = 0;
    else
      *poff = (int)(ptcb->tcb_snext - ptcb->tcb_suna);
    datalen = ptcb->tcb_sbcount - *poff;
    /* remove urgent data holes */
    if (!rexmt) {
      datalen = tcp_shskip(ptcb, datalen, poff);
      datalen = (int)min(datalen, ptcb->tcb_swindow);
    }
    return min(datalen, ptcb->tcb_smss);
  }
  /* else, URGENT data */

  puqe = (struct uqe *)queue_dequeue(ptcb->tcb_sudq);

/* FIXME: need to look at urgent data processing in more detail, 
especially the uq system.*/

  *poff = (int)(ptcb->tcb_sbstart + puqe->uq_seq - ptcb->tcb_suna);
  if (*poff > ptcb->tcb_sbsize)
    *poff -= ptcb->tcb_sbsize;
  datalen = puqe->uq_len;
  if (datalen > ptcb->tcb_smss) {
    datalen = ptcb->tcb_smss;
    puqe2 = uqalloc();
    if (puqe2 == NULL) {
      uqfree(puqe);    /* bail out and    */
      return ptcb->tcb_smss;  /* try as normal data  */
    }
    puqe2->uq_seq = puqe->uq_seq;
    puqe2->uq_len = datalen;

    /* put back what we can't use */
    puqe->uq_seq += datalen;
    puqe->uq_len -= datalen;
  }
  else {
    puqe2 = puqe;
    puqe = (struct uqe *)queue_dequeue(ptcb->tcb_sudq);
  }
  if (puqe == 0) {
    queue_delete(ptcb->tcb_sudq);
    ptcb->tcb_sudq = -1;
  } 
  else 
    /* FIXME, this is going to take AGES... */
    if (queue_enqueue_ordered(ptcb->tcb_sudq, puqe, SUDK(ptcb, puqe->uq_seq)) < 0)
    uqfree(puqe);  /* shouldn't happen */
  if (ptcb->tcb_suhq == NULL) {
    ptcb->tcb_suhq = queue_create(TCPUQLEN /* FIXME: , QF_WAIT*/ );
    ptcb->tcb_suhseq = puqe2->uq_seq;
  }
  if (queue_enqueue_ordered(ptcb->tcb_suhq, puqe2, SUHK(ptcb, puqe2->uq_seq)) < 0)
    uqfree(puqe2);
  return datalen;
}


/*******************************************************************************
***
*** Function         : tcp_howmuch
*** Preconditions    : 
*** Postconditions   : The amount of data available to send is computed
***
*******************************************************************************/

word32 tcp_howmuch(struct  tcb  *ptcb)
{
word32  tosend;

  tosend = ptcb->tcb_suna + ptcb->tcb_sbcount - ptcb->tcb_snext;
  if (ptcb->tcb_code & TCPF_SYN)
    ++tosend;
  if (ptcb->tcb_flags & TCBF_SNDFIN)
    ++tosend;
  return tosend;
}


/*******************************************************************************
***
*** Function         : tcp_reset
*** Preconditions    : A bad packet has been received
*** Postconditions   : A reset has been generated.
***
*******************************************************************************/

int tcp_reset(struct  buf  *pbufin)
{
struct  buf  *pbufout;
struct  ip  *pipin = (struct ip *)pbufin->data, *pipout;
struct  tcp  *ptcpin = (struct tcp *)pipin->ip_data, *ptcpout;
word16    datalen;

  if (ptcpin->tcp_code & TCPF_RST)
    return TRUE;    /* no RESETs on RESETs */
  pbufout = (struct buf *)buf_alloc(TCPMINSIZE); /* FIXME? */
  if (pbufout == NULL)
    return FALSE;
  
  pipout = (struct ip *)pbufout->data;
  p_bcpy(pipout->ip_src, pipin->ip_dst, IP_ALEN);
  p_bcpy(pipout->ip_dst, pipin->ip_src, IP_ALEN);
  pipout->ip_proto = IPT_TCP;
  pipout->ip_verlen = (IP_VERSION<<4) | IP_MINHLEN;
  pbufout->usedsize = 
    pipout->ip_len = TCPMINSIZE; /* was IPMHLEN + TCPMHLEN, but TCPMINSIZE is
                                     set to that anyway... */

  ptcpout = (struct tcp *)pipout->ip_data;
  ptcpout->tcp_sport = ptcpin->tcp_dport;
  ptcpout->tcp_dport = ptcpin->tcp_sport;
  if (ptcpin->tcp_code & TCPF_ACK) {
    ptcpout->tcp_seq = ptcpin->tcp_ack;
    ptcpout->tcp_code = TCPF_RST;
  }
  else {
    ptcpout->tcp_seq = 0;
    ptcpout->tcp_code = TCPF_RST|TCPF_ACK;
  }
  datalen = 
    pipin->ip_len - IP_HLEN(pipin) - TCP_HLEN(ptcpin);
  
  if (ptcpin->tcp_code & TCPF_FIN)
    datalen++;
  ptcpout->tcp_ack = ptcpin->tcp_seq + datalen;
  ptcpout->tcp_offset = TCPHOFFSET;
  ptcpout->tcp_window = ptcpout->tcp_urgptr = 0;
  tcp_h2net(ptcpout);
  ptcpout->tcp_cksum = 0;
  ptcpout->tcp_cksum = tcp_cksum(pipout);
  tcpo_stat_outsegs++;
  ip_send(pipout->ip_dst, pbufout, TCPMHLEN);
  return TRUE;
}


/*******************************************************************************
***
*** Function         : tcp_wakeup
*** Preconditions    : Some processes may be sleeping, waiting for TCP.
***                    ptcb->tcb_error is set to a meaningful error code.
*** Postconditions   : They're woken. tcb_error is in their status words.
***
*******************************************************************************/

int tcp_wakeup(int type, struct tcb *ptcb)
{
int    freelen;
console_print("tcp_wakeup: ");
if (type & READERS)
  console_print("READERS ");
if (type & WRITERS)
  console_print("WRITERS ");
console_printf(" TCB# %d\r\n", (UINT)(ptcb-tcbtab));


  if (type & READERS) {
    if (((ptcb->tcb_flags & TCBF_RDONE) ||
        ptcb->tcb_rbcount > 0 || ptcb->tcb_rudq >= 0) &&
        ptcb->tcb_rdpid > NULL <= 0) {
      p_pcpyto(ptcb->tcb_rdpid, (VOID *)ptcb->tcb_rdstatword, &ptcb->tcb_error, 
               sizeof(UWORD));
      p_iosignalbypid(ptcb->tcb_rdpid);
    }
  }
  if (type & WRITERS) {
    freelen = ptcb->tcb_sbsize - ptcb->tcb_sbcount;
    if (((ptcb->tcb_flags & TCBF_SDONE) || freelen > 0) &&
        ptcb->tcb_wrpid > NULL) {
      p_pcpyto(ptcb->tcb_wrpid, (VOID *)ptcb->tcb_wrstatword, &ptcb->tcb_error, 
               sizeof(UWORD));
      p_iosignalbypid(ptcb->tcb_wrpid);
    }
    /* special for abort */
    if (ptcb->tcb_error && ptcb->tcb_ocpid > NULL) {
      p_pcpyto(ptcb->tcb_ocpid, (VOID *)ptcb->tcb_ocstatword, &ptcb->tcb_error, 
               sizeof(UWORD));
      p_iosignalbypid(ptcb->tcb_ocpid);
    }
  }
  return TRUE;
}


/*******************************************************************************
***
*** Function         : tcp_iss
*** Preconditions    : 
*** Postconditions   : The initial starting sequence is set for a new 
***                    connection
***
*******************************************************************************/

word32 tcp_iss() /* should really call this tcp_isn ;-)  */
{
static word32 seq = 0;
TEXT buf[50];
console_puts("tcp_iss: in");

  /* FIXME */
  if (seq == 0)
    seq = p_date(); /* time in seconds since 1/1/1970 */
  seq += TCPINCR; /* Now what happens if seq+TCPINCR == 0?
                   * Perhaps initialise the static seq to p_date? */
p_atos(buf, "tcp_iss: p_atos returning 0x%08lx", seq);
console_puts(buf);
  return seq;
}


/*******************************************************************************
***
*** Function         : tcp_rmss
*** Preconditions    : 
*** Postconditions   : Receive MSS option is set
***
*******************************************************************************/

int tcp_rmss(struct tcb *ptcb, struct ip *pip)
{
struct  tcp  *ptcp = (struct tcp *)pip->ip_data;
int    mss, hlen, olen, i;

  hlen = TCP_HLEN(ptcp);
  olen = 2 + sizeof(short);
  pip->ip_data[hlen] = TPO_MSS;    /* option kind    */
  pip->ip_data[hlen+1] = olen;    /* option length  */
  mss = hs2net((short)ptcb->tcb_smss);
  for (i=olen-1; i>1; i--) {
    pip->ip_data[hlen+i] = mss & LOWBYTE;
    mss >>= 8;
  }
  hlen += olen + 3;  /* +3 for proper rounding below */
  /* header length is high 4 bits of tcp_offset, in longs  */
  ptcp->tcp_offset = ((hlen<<2) & 0xf0) | ptcp->tcp_offset & 0xf;
  return TRUE;
}


/*******************************************************************************
***
*** Function         : tcpo_rexmt
*** Preconditions    : 
*** Postconditions   : TCP output events are handled while retransmitting
***
*******************************************************************************/

int tcpo_rexmt(int tcbnum, int event)
{
struct  tcb  *ptcb = &tcbtab[tcbnum];

  if (event != TQ_TCP_RETRANSMIT)
    return TRUE;  /* ignore others while retransmitting  */
  if (++ptcb->tcb_rexmtcount > TCP_MAXRETRIES) {
    tcp_abort(ptcb, E_NET_TIMEDOUT);
    return TRUE;
  }
  tcp_send(tcbnum, TSF_REXMT);
  timer_set(MKEVENT(TQ_TCP_RETRANSMIT, tcbnum),
    min(ptcb->tcb_rexmt<<ptcb->tcb_rexmtcount, TCP_MAXRXT));
  if (ptcb->tcb_ostate != TCPOS_REXMT)
    ptcb->tcb_ssthresh = ptcb->tcb_cwnd;  /* first drop  */
  ptcb->tcb_ssthresh = min(ptcb->tcb_swindow,ptcb->tcb_ssthresh)/2;
  if (ptcb->tcb_ssthresh < ptcb->tcb_smss)
    ptcb->tcb_ssthresh = ptcb->tcb_smss;
  ptcb->tcb_cwnd = ptcb->tcb_smss;
  return TRUE;
}

