/*******************************************************************************
***
*** Filename         : timer.c
*** Purpose          : Timer queue routines for PsiStack
*** Author           : Matt J. Gumbley
*** Created          : 26/07/97
*** Last updated     : 15/08/97
***
********************************************************************************
***
*** Modification Record
***
*******************************************************************************/

#include "stack.h"


#define TIMERGRAN       1               /* timer granularity, secs/10   */


/* Variables for the timer stuff */
GLDEF_D WORD  timer_status;
GLDEF_D ULONG timer_ticks = 0L;
GLDEF_D ULONG timer_interval = 1L;                     /* Every 1/10th second */

LOCAL_D VOID *timer_handle;

/* According to Chapter 10 of the SIBO C SDK's PLIB manual, the highest
   resolution relative timer is available in 1/10th seconds. 
   3c and Siena machines also have access to a system call to return the
   number of systems ticks, which are in 1/32nd seconds. We use that for
   accurate timing for pings, but nowhere else. We can also use p_date to get
   second resolution, which is next to useless.

   Certain threads (PPP, SCRIPT, TCP output) schedule events (stack timer
   events) to complete in the future. For instance, a script can sleep for a 
   few seconds, or a segment retransmission can be scheduled to be 
   automatically re-sent if an ack is not received.

   An ideal system might schedule EPOC relative timer events to occur when
   these stack events are due to occur. Comer's code doesn't neatly fit that
   scheme. Plus, that's more complex. The system used here is almost identical
   to the TCP timer approach in Comer. EPOC relative timer events are queued
   to occur as rapidly as they can be (every 1/10th second). We maintain a
   current 'tick' counter, and use this in determining whether the stack event
   at the head of the queue has occurred.

   OK, this isn't very efficient, and will sap your CPU somewhat. The 'ideal'
   solution needs things like this: When a stack timer event is queued, 
   an EPOC relative timer event may be queued. If this is the only stack timer 
   event outstanding, a new EPOC timer event will be queued. If this stack 
   timer event is scheduled to occur before any event previously scheduled, 
   that event's EPOC timer is cancelled, and a new EPOC timer event generated 
   for this new event.

   Replacing what we have here with the ideal solution is an exercise for the 
   reader :-)


   There is a timer delta list containing descriptions of these scheduled stack
   timer events. The timer thread processes each entry on the delta list when 
   they expire.


   'Processing' means different things for different events.
   e.g. For TCP, the event described in tq_msg is then placed on the TCP output 
   thread's output queue for sending.
   
   Note: our queue structure is usually used to queue struct buf's. The TCP
         output thread uses it to store integers (events).
   Comer's version stored the 'port' (a Xinu queue) and its length here,
   but since it's always going to be the TCP output pseudo-process's
   input queue, I've removed that.... Comer's version only posted TCP segments,
   see Vol II, page 271. We take the other approach, it's not much more complex.
*/

struct  tqent {
  INT            tq_timeleft;                   /* Time to expire (1/10 secs) */
  INT            tq_msg;                        /* Event to send when expired */
  struct tqent  *tq_next;                                 /* Next in the list */
  ULONG          tq_time;                       /* Time this entry was queued */
};

LOCAL_D struct tqent *tqhead = NULL; /* The timer delta queue */
LOCAL_D long lastrun = 0L;

/* Some stats about the timer */
struct logstat timer_stat[] = {
  "oqueuefull", 0,                                 /* # output queue failures */
  "setoom", 0,                 /* # out-of-memory errors when timer_set()ting */
  NULL, 0
};

GLDEF_D TEXT *tq_eventnames[] = {
  "",
  "TCP_SEND",
  "TCP_PERSIST",
  "TCP_RETRANSMIT",
  "TCP_DELETE",
  "SCRIPT_WAKEUP",
  "PPP_RESTART_TIMEOUT",
};

/*******************************************************************************
***
*** Function         : timer_open
*** Preconditions    : 
*** Postconditions   : 
***
*******************************************************************************/

GLDEF_C INT timer_open(VOID)
{
  return p_open(&timer_handle, "TIM:", -1);
}


/*******************************************************************************
***
*** Function         : timer_queue
*** Preconditions    : 
*** Postconditions   : 
***
*******************************************************************************/

GLDEF_C VOID timer_queue(VOID)
{
  p_ioa4(timer_handle, P_FRELATIVE, &timer_status, &timer_interval);
}


/*******************************************************************************
***
*** Function         : timer_cancel
*** Preconditions    : 
*** Postconditions   : 
***
*******************************************************************************/

GLDEF_C VOID timer_cancel(VOID)
{
  if (timer_interval > 0L) {
    p_iow2(timer_handle, P_FCANCEL);
    p_waitstat(&timer_status);
  }
}


/*******************************************************************************
***
*** Function         : timer_process
*** Preconditions    : An EPOC relative timer event has occurred
*** Postconditions   : Any events on the delta queue which have expired are
***                    processed. 
***                    This might mean:
***                    1) Outgoing segments are transferred to the TCP output 
***                       pseudo-process's input queue, and the ATTN_TCPOUTPUT 
***                       flag set, so they'll get sent.
***                    2) The script is woken up.
***                    3) The PPP restart timer expires.
***
*******************************************************************************/

GLDEF_C VOID timer_process(VOID)
{
long now;                                 /* times from system clock */
int  delta;                                      /* time since last iteration */
register struct tqent  *tq;                       /* temporary delta list ptr */

  timer_ticks ++;
console_printf("timer_process: tick %d\r\n",(UINT)timer_ticks&0xFFFF);
timer_qdump();

  if (tqhead == NULL) {
console_puts("<nothing to do>");
    return;
  }

  now = timer_ticks;
  /* delta = (int) (now - lastrun);    * compute elapsed time    */
  delta = 1; 
  /* until we've got a precise way of determining exactly how much
   * time has elapsed since the last tick, we have a rough idea - it's one tick.
   */

  /* Note: check for possible clock reset (time moved 
     backward or delay was over an order of magnitude too
     long)
  */

  if (delta < 0 || delta > TIMERGRAN*100)
    delta = TIMERGRAN*10;  /* estimate the delay  */
  lastrun = now;
console_printf("delta is %d\r\n", delta);
  /* Process all expired events */
  while (tqhead != NULL  &&  tqhead->tq_timeleft <= delta) {

    delta -= tqhead->tq_timeleft;

    /* What kind of event is this? */
    switch (EVENT(tqhead->tq_msg)) {
      case TQ_TCP_SEND:
      case TQ_TCP_RETRANSMIT:
      case TQ_TCP_PERSIST:
        if (flags.log_flag & LOG_FLAG_UPPERDUMP)
          console_puts("timer_process: TCP timer event expired");
        if (!iqueue_enqueue(tcp_oqueue, tqhead->tq_msg)) {
          timer_stat_oqueuefull++;
          console_puts("timer_process: TCP output queue full");
        }
        else {
          /* Tell the TCP output thread that there's something waiting for it */
          attn_signal(ATTN_TCPOUTPUT);
        }
        break;

      case TQ_SCRIPT_WAKEUP:
        if (flags.log_flag & LOG_FLAG_FLOWTRACE)
          console_puts("timer_process: Script wakeup timer expired");
        script_wakeup();
        break;

      case TQ_PPP_RESTART_TIMEOUT:
        if (flags.log_flag & LOG_FLAG_FLOWTRACE)
          console_puts("timer_process: PPP restart timer expired");
        /* Patrick: Call your PPP restart timer timeout routine here! */
        /* Process timed events on PPP interfaces */
        ppp_timer();
        break;
    }

    /* Remove the event from the queue */
    tq = tqhead;
    tqhead = tqhead->tq_next;
    p_free(tq);
  }

  if (tqhead != NULL)
    tqhead->tq_timeleft -= delta;
}


/*******************************************************************************
***
*** Function         : timer_clear(msg)
*** Preconditions    : msg points to a message queued on the timer event queue
*** Postconditions   : It is removed from the queue, and the remaining queued
***                    events are 'shuffled up'
***
*******************************************************************************/

INT timer_clear(INT msg)
{
register struct tqent *prev, *ptq;
register int timespent;

console_printf("timer_clear: msg %d\r\n",(UINT)msg);

  prev = NULL;
  for (ptq = tqhead; ptq != NULL; ptq = ptq->tq_next) {
    if (ptq->tq_msg == msg) {
      timespent = (int) (timer_ticks - ptq->tq_time);
      if (prev != NULL)
        prev->tq_next = ptq->tq_next;
      else
        tqhead = ptq->tq_next;
      if (ptq->tq_next != NULL)
        ptq->tq_next->tq_timeleft += ptq->tq_timeleft;
      p_free(ptq);
      return timespent;
    }
    prev = ptq;
  }
console_puts("timer_clear: not found");
  return NULL;
}


/*******************************************************************************
***
*** Function         : timer_left(msg)
*** Preconditions    : msg is a queued timer event.
*** Postconditions   : The time left for this event is returned (in 1/10ths of
***                    a second)
***
*******************************************************************************/

INT timer_left(INT msg)
{
register struct tqent *tq;
int timeleft = 0;

  if (tqhead == NULL)
    return 0;
  for (tq = tqhead; tq != NULL; tq = tq->tq_next) {
    timeleft += tq->tq_timeleft;
    if (tq->tq_msg == msg) {
      return timeleft;
    }
  }
  return 0;
}


/*******************************************************************************
***
*** Function         : timer_set
*** Preconditions    : msg is an event, time is a delta at which it will expire
*** Postconditions   : TRUE or FALSE, depending on whether the event could be
***                    queued or not
***
*******************************************************************************/

INT timer_set(INT msg, INT time)
{
register struct tqent *ptq, *newtq, *tq;

console_printf("timer_set: msg: 0x%04x time: 0x%04x\r\n", msg, time);

  newtq = (struct tqent *)p_alloc(sizeof(struct tqent));
  if (newtq == NULL) {
    console_puts("timer_set: OOM");
    timer_stat_setoom++;
    return FALSE;
  }
  newtq->tq_timeleft = time;
  newtq->tq_msg = msg;
  newtq->tq_next = NULL;
  newtq->tq_time = timer_ticks;

  /* Clear duplicates */
  (void) timer_clear(msg);

  /* Is the list empty? Stick the event at the front if so. */
  if (tqhead == NULL) {
    tqhead = newtq;

console_puts("new tq is:");
timer_qdump();

    return TRUE;
  }

  /* Not empty, so search the list for our spot */
  for (ptq=0, tq=tqhead; tq != NULL; tq=tq->tq_next) {
    if (newtq->tq_timeleft < tq->tq_timeleft)
      break;
    newtq->tq_timeleft -= tq->tq_timeleft;
    ptq = tq;
  }
  newtq->tq_next = tq;
  if (ptq)
    ptq->tq_next = newtq;
  else
    tqhead = newtq;
  if (tq)
    tq->tq_timeleft -= newtq->tq_timeleft;

console_puts("queued tq is:");
timer_qdump();

  return TRUE;
}


/*******************************************************************************
***
*** Function         : timer_qdump
*** Preconditions    : 
*** Postconditions   : The timer queue is dumped for debugging
***
*******************************************************************************/

VOID timer_qdump(VOID)
{
struct tqent *tq;
int data, event;
TEXT buf[80];
  for (tq=tqhead; tq != NULL; tq=tq->tq_next) {
    data = DATA(tq->tq_msg);
    event = EVENT(tq->tq_msg);
    p_atos(buf,"tq <left=0x%04x,queued=0x%08lx,event=%s,data=0x%04x>", 
      tq->tq_timeleft,
      tq->tq_time, 
      tq_eventnames[event],data);
    console_puts(buf);
  }
}


/*******************************************************************************
***
*** Function         : timer_stats()
*** Preconditions    : none
*** Postconditions   : The timer stats are output.
***
*******************************************************************************/

VOID timer_stats(VOID) 
{
  console_puts("timer_stats:");
  console_writestats(timer_stat);
}


/*******************************************************************************
***
*** Function         : timer_killtimers
*** Preconditions    : 
*** Postconditions   : All outstanding timers for the TCP are killed
***
*******************************************************************************/

INT timer_killtimers(struct tcb *ptcb)
{
int  tcbnum = ptcb - &tcbtab[0];
console_puts("timer_killtimers");

  /* clear all possible pending timers */
  timer_clear(MKEVENT(TQ_TCP_SEND, tcbnum));
  timer_clear(MKEVENT(TQ_TCP_RETRANSMIT, tcbnum));
  timer_clear(MKEVENT(TQ_TCP_PERSIST, tcbnum));
console_puts("...killed");
  return TRUE;
}

