/*******************************************************************************
***
*** Filename         : udp.c
*** Purpose          : User Datagram Protocol for Psion-TCP/IP
*** Author           : Matt J. Gumbley
*** Created          : 11/04/97
*** Last updated     : 14/04/98
***
********************************************************************************
***
*** Modification Record
***
*******************************************************************************/

#include "stack.h"

/* There are two sets of UDP configuration settings.
   This one may be changed by the UDP dialog, and gets stored in the config
   file. */
struct udp_config udp_config = { INITUPPS, INITUPPLEN };

/* This one is copied from udp_config when the config info is loaded, and is
   the one used by the stack. Reason being:
   If you have a config file which says there are 5 demux queues, and use
   that variable to loop through the queues, then increase the number of
   queues using the dialog, then you'll step over the end of the allocated
   demux queue array. */
struct udp_config udp_config_used = { INITUPPS, INITUPPLEN };

struct upq *upqs;

struct logstat udp_stat[] = {
  "checksum", 0,                                     /* # UDP Checksum errors */
  "inerrors", 0,                 /* # datagrams dropped 'cos demux queue full */
  "indatagrams", 0,                      /* # datagrams enqueued successfully */
  "outdatagrams", 0,                     /* # datagrams enqueued successfully */
  "noports", 0,                          /* # datagrams dropped 'cos no ports */
  NULL, 0
};
#define udpstat_checksum udp_stat[0].ls_value
#define udpstat_inerrors udp_stat[1].ls_value
#define udpstat_indatagrams udp_stat[2].ls_value
#define udpstat_outdatagrams udp_stat[3].ls_value
#define udpstat_noports udp_stat[4].ls_value

/*******************************************************************************
***
*** Function         : udp_init
*** Preconditions    : The configuration file has been loaded, and udp_config 
***                    loaded from the configuration file.
*** Postconditions   : udp_config_used is initialised from udp_config, the 
***                    upqs are initialised, and udp_init is TRUE.
***                    OOM, and FALSE.
***
*******************************************************************************/

int udp_init(void)
{
struct upq *pup;
int i;
  /* Copy udp_config into udp_config_used */
  p_bcpy(&udp_config_used, &udp_config, sizeof (struct udp_config));

  upqs = NULL;
  upqs = p_alloc(sizeof(struct upq) * udp_config_used.udpc_UPPS);
  if (upqs == NULL) {
    console_puts("Couldn't allocate UDP ports");
    goto initfail;
  }

  for (i=0; i<udp_config_used.udpc_UPPS; i++) {
    pup = &upqs[i];
    pup->upq_inuse = FALSE;
    pup->upq_lport = -1;
    pup->upq_fport = -1;
    p_bcpy(pup->upq_fip,ip_anyaddr,IP_ALEN);
    pup->upq_q = (struct queue *)NULL;
  }
  return TRUE;

initfail:
  if (upqs != NULL) 
    p_free(upqs);
  return FALSE;
}


/*******************************************************************************
***
*** Function         : upq_alloc
*** Preconditions    : 
*** Postconditions   : A queue is allocated, and a pointer to it is returned.
***                    There is no free queue, and NULL is returned.
***
*******************************************************************************/

struct upq *upq_alloc(void)
{
register struct upq *pup;
int i;

  for (i=0; i<udp_config_used.udpc_UPPS; i++) {
    pup = &upqs[i];
    if (!pup->upq_inuse) {
      pup->upq_lport = -1;
      pup->upq_fport = -1;
      pup->upq_ocpid = 
      pup->upq_ocstatword = 
      pup->upq_wrpid = 
      pup->upq_wrbuf = 
      pup->upq_wrlen = 
      pup->upq_wrstatword = 
      pup->upq_rdpid = 
      pup->upq_rdbuf = 
      pup->upq_rdlen = 
      pup->upq_rdstatword = NULL;

      p_bcpy(pup->upq_fip,ip_anyaddr,IP_ALEN);
      pup->upq_q = (struct queue *)NULL;
      pup->upq_q = queue_create(udp_config_used.udpc_UPPLEN);
      if (pup->upq_q == NULL) {
        console_puts("Couldn't allocate UDP demux q");
        return NULL;
      }
      pup->upq_inuse = TRUE;
      return pup;
    }
  }
  return NULL;
}


/*******************************************************************************
***
*** Function         : udp_net2h(u)
*** Preconditions    : u is a UDP message in network byte order
*** Postconditions   : u is in host byte order
***
*******************************************************************************/

void udp_net2h(struct udp *u)
{
  /* Convert the header fields into host byte order */
  u->udp_srcport = net2hs(u->udp_srcport);
  u->udp_dstport = net2hs(u->udp_dstport);
  u->udp_len = net2hs(u->udp_len);
}


/*******************************************************************************
***
*** Function         : udp_h2net(u)
*** Preconditions    : u is a UDP message in host byte order
*** Postconditions   : u is in network byte order
***
*******************************************************************************/

void udp_h2net(struct udp *u)
{
  /* Convert the header fields into network byte order */
  u->udp_srcport = hs2net(u->udp_srcport);
  u->udp_dstport = hs2net(u->udp_dstport);
  u->udp_len = hs2net(u->udp_len);
}


/*******************************************************************************
***
*** Function         : udp_cksum(b)
*** Preconditions    : 
*** Postconditions   : The UDP pseudo-header checksum is returned
***
*******************************************************************************/

#define UDP_ALEN IP_ALEN

word16 udp_cksum(struct ip *pip)
{
struct udp *pudp = (struct udp *)pip->ip_data;
word16 *psh;
unsigned long sum;
int len = net2hs(pudp->udp_len);
int i;

  sum = 0;

  psh = (word16 *) pip->ip_src;
  for (i=0; i<UDP_ALEN; ++i)
    sum += *psh++;

  psh = (word16 *)pudp;
  sum += hs2net(pip->ip_proto + len);
  if (len & 0x1) {
    ((char *)pudp)[len] = 0;        /* pad */
    len += 1;       /* for the following division */
  }
  len /= 2;       /* convert to length in shorts */

  for (i=0; i<len; ++i)
    sum += *psh++;
  sum = (sum >> 16) + (sum & 0xffff);
  sum += (sum >> 16);

  return (short)(~sum & 0xffff);
}


/*******************************************************************************
***
*** Function         : udp_unblock(pid, cd_buf, buflen, cd_statword, ubuf)
*** Preconditions    : A client sent a recv message to the kernel. No datagrams
***                    had been demultiplexed and queued for this port at that 
***                    time, so the client's details were stored in the demux 
***                    queue waiting for data to arrive. Although the recv 
***                    message was replied to, the client will be blocked on 
***                    its status word.
***                    Now, the UDP input process has received a datagram, and 
***                    is passing it back to the client, unblocking it.
*** Postconditions   : The datagram buffer is freed; the client is signalled.
***
*******************************************************************************/

GLDEF_C VOID udp_unblock(UWORD pid, UWORD cd_buf, UWORD buflen,
                        UWORD cd_statword, struct buf *b)
{
UWORD maxlen;
struct ip *pip;
struct udp *pudp;

  pip = (struct ip *)b->data;
  pudp = (struct udp *)pip->ip_data;

  /* Copy the data from b to the client process's data area */
  maxlen = buflen < pudp->udp_len ? buflen : pudp->udp_len;
  p_pcpyto(pid, (VOID *)cd_buf, (VOID *)pudp->udp_data, maxlen);
  buf_free(b);

  /* Set up the client's status word to be the length of the copied data */
  p_pcpyto(pid, (VOID *)cd_statword, (VOID *)&maxlen, sizeof(UWORD));

  /* Signal the app */
  p_iosignalbypid(pid);
}

/*******************************************************************************
***
*** Function         : udp_process processes an UDP segment
*** Preconditions    : b is a buffer chain holding a UDP packet in network 
***                    byte order.
*** Postconditions   : It has been converted to host byte order and dealt with
***                    and udp_process returns TRUE. Else an error occurred
***                    and udp_process returns FALSE.
***
*******************************************************************************/

int udp_process(struct buf *b)
{
register struct upq *pup;
register int j;
struct ip *pip;
struct udp *pudp;
word16 dst,c;

  pip = (struct ip *)b->data;
  pudp = (struct udp *)pip->ip_data;

  if (pudp->udp_cksum && (c=udp_cksum(pip))) {
    buf_free(b);
    udpstat_checksum++;
    if (flags.log_flag & LOG_FLAG_FLOWTRACE) {
      console_printf("udp_process: bad cksum 0x%04x should be 0x%04x\r\n",
             c, pudp->udp_cksum);
    }
    return FALSE; 
  }

  udp_net2h(pudp);
  dst = pudp->udp_dstport;

  if (flags.log_flag & LOG_FLAG_UPPERDUMP)
    udp_dump(b);

  /* For all available ports */
  for (j=0, pup=upqs; j<udp_config_used.udpc_UPPS; j++, pup++) {

    if (pup->upq_lport == dst) { /* pup->upq_lport in network byte order? */
      if (flags.log_flag & LOG_FLAG_FLOWTRACE)
        console_printf("udp_process: datagram arrived at port %d\r\n", dst);

      /* This doesn't deal with the case where a client requests data and 
         it's already queued (i.e. by the else part of the following if)
         - the message handling code already checks for that before storing 
         pupq->upq_rdpid etc. */

      /* If a client is waiting on this data (i.e. has sent the kernel an IPC 
         message requesting data, which has blocked), copy this datagram to
         the client's data buffer, and reply to the IPC message. */
      if (pup->upq_rdpid != NULL) {
        udp_unblock(pup->upq_rdpid, pup->upq_rdbuf, pup->upq_rdlen, 
                    pup->upq_rdstatword, b);
        /* Reset the waiting process descriptor to NULL. It'll get
           set back again when the next recvfrom socket call comes through. */
        pup->upq_rdpid = NULL;
        return TRUE;
      }
      else {
        /* Nobody's waiting for this data, so queue it if there's room. */
        if (queue_full(pup->upq_q)) {
          buf_free(b);
          udpstat_inerrors++; 
          console_puts("udp_process: Couldn't queue UDP datagram");
          return FALSE;
        }
        queue_enqueue(pup->upq_q, b);
        udpstat_indatagrams++;
        return TRUE;
      }
    }
  }
  /* No such port */
  udpstat_noports++;
  icmp_send(ICT_DESTUR, ICC_PORTUR, pip->ip_src, (struct buf *)b, 0); /* Reuse datagram */
  if (flags.log_flag & LOG_FLAG_FLOWTRACE)
    console_puts("udp_process: port unreachable");
  return TRUE;
}


/*******************************************************************************
***
*** Function         : udp_nextport
*** Preconditions    : 
*** Postconditions   : Returns the next available UDP local "port" number.
***
*******************************************************************************/

word16 udp_nextport(void)
{
static word16 lastport = ULPORT;
int inuse = TRUE;
struct upq *pup;
int i;

  while (inuse) {
    lastport++;
    if (lastport == 0)
      lastport = ULPORT;
    inuse = FALSE;
    for (i=0; !inuse && i<udp_config_used.udpc_UPPS; i++) {
      pup = &upqs[i];
      inuse = pup->upq_inuse && pup->upq_lport == lastport;
    }
  }
  return lastport;
}


/*******************************************************************************
***
*** Function         : udp_send
*** Preconditions    : 
*** Postconditions   : 
***
*******************************************************************************/

int udp_send(IPaddr fip, word16 fport, word16 lport, struct buf *b,
            int datalen, int docksum)
{
struct ip *pip = (struct ip *)b->data;
struct udp *pudp = (struct udp *)pip->ip_data;
int rt_ifnum;

  pip->ip_proto = IPT_UDP;
  pudp->udp_srcport = lport;
  pudp->udp_dstport = fport;
  pudp->udp_len = U_HLEN+datalen;
  pudp->udp_cksum = 0;
  udp_h2net(pudp);

  if (docksum) {

    rt_ifnum = which_nif(fip, TRUE); /* TRUE => Locally generated */
    if (flags.log_flag & LOG_FLAG_FLOWTRACE) {
      console_printf("udp_send: which_nif returned %d\r\n", rt_ifnum);
    }
    if (rt_ifnum == -1) {
      /*ip_mib.ip_unknetwork++;*/
      buf_free(b);
      if (flags.log_flag & LOG_FLAG_FLOWTRACE)
        console_puts("udp_send: datagram route failed");
      return FALSE;
    }
    p_bcpy(pip->ip_dst, fip, IP_ALEN);
    if (rt_ifnum == NI_LOCAL)
      p_bcpy(pip->ip_src, pip->ip_dst, IP_ALEN);
    else
      p_bcpy(pip->ip_src, nif[rt_ifnum].ni_ip, IP_ALEN);

    pudp->udp_cksum = udp_cksum(pip);
    if (pudp->udp_cksum == 0)
      pudp->udp_cksum = ~0;
  }

  udpstat_outdatagrams++;
  ip_send(fip, b, U_HLEN+datalen);
  return TRUE;
}


/*******************************************************************************
***
*** Function         : udp_wkp_name(wkp)
*** Preconditions    : wkp is a port number
*** Postconditions   : A static pointer to the name of this well-known-port is
***                    returned, or if it isn't well-known, "?" is returned.
***
*******************************************************************************/

static char *udp_wkp_name(word16 wkp)
{
  switch (wkp) {
    case 0: return "reserved";
    case 7: return "echo";
    case 9: return "discard";
    case 11: return "active users";
    case 13: return "daytime";
    case 15: return "who is up or netstat";
    case 17: return "quote of the day";
    case 19: return "character generator";
    case 37: return "time";
    case 42: return "host name server";
    case 43: return "who is";
    case 53: return "domain name server";
    case 67: return "bootstrap protocol server";
    case 68: return "bootstrap protocol client";
    case 69: return "trivial file transfer";
    case 111: return "Sun microsystems RPC";
    case 123: return "network time protocol";
    case 161: return "SNMP net monitor";
    case 162: return "SNMP traps";
    case 512: return "UNIX comsat";
    case 513: return "UNIX rwho daemon";
    case 514: return "system log";
    case 525: return "time daemon";
    default: return "?";
  }
}


/*******************************************************************************
***
*** Function         : udp_dump outputs an interpreted datagram trace to the 
***                    log device
*** Preconditions    : b is a buffer holding a UDP datagram 
*** Postconditions   : A trace is output
***
*******************************************************************************/

void udp_dump(struct buf *b)
{
struct ip *i;
struct udp *u;
word16 c;
int len;
TEXT *cksumok;
  i = (struct ip *)b->data;
  u = (struct udp *)i->ip_data;

  /* Is it duff? UDP checksum computation is tricker than this */
  len = i->ip_len - IP_HLEN(i);
  if (u->udp_cksum != 0 && (c=cksum((word16 *)u, u->udp_len)) != 0) 
    cksumok="BAD";
  else
    cksumok="OK";

  console_printf("udp_dump: datagram len 0x%04x (offset 0x%04x in dump)\r\n", 
                 len, IP_HLEN(i));
  console_printf("UDP src#: 0x%04x (%s) ",
                 u->udp_srcport, udp_wkp_name(u->udp_srcport));
  console_printf("dst#: 0x%04x (%s) Len: 0x%04x Checksum: 0x%04x (%s)\r\n",
                 u->udp_dstport, udp_wkp_name(u->udp_dstport),
                 u->udp_len, u->udp_cksum, cksumok);
  hexdump((unsigned char *)u, len);
}


/*******************************************************************************
***
*** Function         : udp_stats()
*** Preconditions    : 
*** Postconditions   : The UDP stats are output
***
*******************************************************************************/

void udp_stats()
{
  console_puts("udp_stats:");
  console_writestats(udp_stat);
}

