/*******************************************************************************
***
*** Filename         : console.c
*** Purpose          : PsiStack Console API Link Library
***                    Gets linked with client apps; responsible for packaging
***                    console messages into buffers, IPC'ing them to the 
***                    console process (if it exists), and returning the result.
***                    Also provides routines for loading and unloading the 
***                    console, and seeing if it's loaded.
*** Author           : Matt J. Gumbley
*** Created          : 14/02/98
*** Last updated     : 21/03/98
***
********************************************************************************
***
*** Modification Record
***
*******************************************************************************/

#include "..\utils\sysincl.h"                  /* Cos this is an EPOC library */
#include "..\..\include\psistack\console.h"          /* Public API definition */
#include "..\..\console\conmsg.h"        /* Private implementation definition */

LOCAL_D TEXT szConsoleName[]="CONSOLE.*";
LOCAL_D TEXT *szPrefix = NULL;
LOCAL_D BYTE timestamped = FALSE;
LOCAL_D HANDLE hCon;

/*******************************************************************************
***
*** Function         : console_setprefix(s)
*** Preconditions    : s is a string, or NULL
*** Postconditions   : The string is printed before any console_... function
***                    or, if NULL, prefixing is disabled.
***
*******************************************************************************/

GLDEF_C VOID console_setprefix(TEXT *s)
{
  szPrefix = s;
}

/*******************************************************************************
***
*** Function         : console_timestamp(b)
*** Preconditions    : b is a boolean
*** Postconditions   : Timestamping is enabled or disabled. Returns previous.
***
*******************************************************************************/

GLDEF_C INT console_timestamp(INT b)
{
INT prev = timestamped;
  timestamped = b;
  return prev;
}

/*******************************************************************************
***
*** Function         : console_mkmsg(buf)
*** Preconditions    : buf is a buffer.
*** Postconditions   : buf is initialised with the prefix (if set) and a 
***                    datestamp (if enabled). A pointer to the last character 
***                    after the timestamp is returned.
***
*******************************************************************************/

LOCAL_C TEXT *console_mkmsg(TEXT *buf)
{
register char *d, *s;

  d = buf;

  /* Add prefix, if used */
  if (szPrefix) {
    for (s=szPrefix; *s; )
      *d++=*s++;
    *d++=':';
  }

  /* Add timestamp, if used */
  if (timestamped) {
    p_now2str(d, "%H%:%T%:%S > ");
    /* Advance d to end of string */
    while (*d++);
    d--;
  }
  return d;
}

/*******************************************************************************
***
*** Function         : console_sendmsg(type, buffer, len)
*** Preconditions    : Type is a CMT_??? flag; there's a buffer and a length
*** Postconditions   : The buffer is sent to the console application, if it is
***                    loaded. The return from p_msendreceivew is returned.
***                    This is 1 on OK, -17 (E_GEN_RECEIVER) on error.
***
*** Note             : This routine won't load the console if it isn't loaded.
***
*******************************************************************************/

LOCAL_C INT console_sendmsg(UINT type, VOID *data, UWORD len)
{
static CONMSG console_msg;

  /* Package parameters into a CONMSG structure */
  console_msg.cm_data = data;
  console_msg.cm_len = len;

  /* This will block if the console has no free message slots... (unlikely) */
  /* ...and will block until the console has finished processing it */
  return p_msendreceivew(hCon, type, (VOID *) &console_msg);

  /* The return from that p_msendreceivew will be:

     E_GEN_RECEIVER (-17) meaning that the console is not loaded, or has
     not yet called p_minit(). 

     Or the console's p_mfree reply value: probably 0
  */
}

GLDEF_C INT console_putch(UINT ch)
{
UINT buf = ch;

  /* Send message if console loaded */
  if ((hCon = console_loaded() ) > 0 ) {
    return console_sendmsg(CMT_TEXT, (VOID *)&buf, 1);
  }
  else return hCon;
}

GLDEF_C INT console_print(TEXT *str)
{
TEXT fstr[CM_MSGMAX];
register char *d, *s, *l;
  /* Send message if console loaded */
  if ((hCon = console_loaded() ) > 0 ) {
    d = console_mkmsg(fstr);
    for (s = str,l = fstr+CM_MSGMAX-2; *s && d<l; s++,d++) 
      *d = *s;
    return console_sendmsg(CMT_TEXT, (VOID *)fstr, d-fstr);
  }
  else return hCon;
}

GLDEF_C INT console_puts(TEXT *str)
{
TEXT fstr[CM_MSGMAX];
register char *d, *s, *l;
  /* Send message if console loaded */
  if ((hCon = console_loaded() ) > 0 ) {
    d = console_mkmsg(fstr);
    for (s = str,l = fstr+CM_MSGMAX-2; *s && d<l; s++,d++) 
      *d = *s;
    *d++ = '\r';
    *d++ = '\n';
    return console_sendmsg(CMT_TEXT, (VOID *)fstr, d-fstr);
  }
  else return hCon;
}

GLDEF_C INT CDECL console_printf(TEXT *fmt, UINT arg,...)
{
TEXT fstr[CM_MSGMAX]; 
register char *d;
register int len;
  /* Send message if console loaded */
  if ((hCon = console_loaded() ) > 0 ) {
    d = console_mkmsg(fstr);
    len = p_atob(d, fmt, &arg);
    *(d+len) = '\0';
    return console_sendmsg(CMT_TEXT, (VOID *)fstr, d- fstr + len);
  }
  else return hCon;
}

GLDEF_C INT console_write_binary(UBYTE *buffer, UWORD len)
{
  /* Send message if console loaded */
  if ((hCon = console_loaded() ) > 0 ) {
    return console_sendmsg(CMT_BINARY, (VOID *)buffer, len);
  }
  else return hCon;
}

GLDEF_C INT console_write_text(UBYTE *buffer, UWORD len)
{
  /* Send message if console loaded */
  if ((hCon = console_loaded() ) > 0 ) {
    return console_sendmsg(CMT_TEXT, (VOID *)buffer, len);
  }
  else return hCon;
}

GLDEF_C INT console_write_ip(UBYTE *buffer, UWORD len)
{
  /* Send message if console loaded */
  if ((hCon = console_loaded() ) > 0 ) {
    return console_sendmsg(CMT_IP, (VOID *)buffer, len);
  }
  else return hCon;
}


/*******************************************************************************
***
*** Function         : console_writestats(s)
*** Preconditions    : s is a name-null-terminated array of status counters
*** Postconditions   : The header and array is output.
***
*******************************************************************************/

void console_writestats(struct logstat *s)
{
struct logstat *l=s;
  console_puts("count description");
  while (l && l->ls_name) {
    console_printf("%5d %s\r\n", l->ls_value, l->ls_name);
    l++;
  }
}


/*******************************************************************************
***
*** Function         : console_writeiostats(s)
*** Preconditions    : s is a name-null-terminated array of I/O status counters
*** Postconditions   : The header and array is output.
***
*******************************************************************************/

void console_writeiostats(struct logiostat *s)
{
struct logiostat *l=s;
  console_puts("input output description");
  while (l && l->ls_name) {
    console_printf("%5d %5d  %s\r\n", l->ls_input, l->ls_output, l->ls_name);
    l++;
  }
}

/*******************************************************************************
***
*** Function         : console_loaded
*** Preconditions    : none
*** Postconditions   : positive result is the process ID of the loaded console
***                    E_FILE_NXIST if the console is not loaded
***
*******************************************************************************/

GLDEF_C HANDLE console_loaded(VOID)
{
  return p_pidfind(szConsoleName);
}

/*******************************************************************************
***
*** Function         : console_load
*** Preconditions    : none
*** Postconditions   : positive result is the process ID of the loaded console
***                    negative result is an EPOC error code
***
*******************************************************************************/

GLDEF_C HANDLE console_load(VOID)
{
  /* May not work, since it's parsed with a related name of .IMG? */
  return p_execc("LOC::M:\\APP\\PSISTACK\\CONSOLE.APP", NULL, 0);
}

/*******************************************************************************
***
*** Function         : console_unload
*** Preconditions    : none
*** Postconditions   : The command is sent to the console application, if it is
***                    loaded. The return from p_msendreceivew is returned.
***                    This is 1 on OK, -17 (E_GEN_RECEIVER) on error.
***
*******************************************************************************/

GLDEF_C INT console_unload(VOID)
{
  if ((hCon = console_loaded() ) > 0 ) {
    return console_sendmsg(CMT_UNLOAD, (VOID *)NULL, 0);
  }
  else return hCon;
}
