/*******************************************************************************
***
*** Filename         : dbms.c
*** Purpose          : PsiStack Database API Link Library
***                    Gets linked with client apps; responsible for packaging
***                    databse access requests into buffers, IPC'ing them to
***                    the database process, and returning the result.
***                    Also provides routines for loading and unloading the 
***                    console, and seeing if it's loaded.
*** Author           : Matt J. Gumbley
*** Created          : 07/02/98
*** Last updated     : 17/02/98
***
********************************************************************************
***
*** Modification Record
***
*******************************************************************************/

#include "..\utils\sysincl.h"                  /* Cos this is an EPOC library */
#include "..\..\include\psistack\dbms.h"             /* Public API definition */
#include "..\..\dbms\dbmsg.h"            /* Private implementation definition */

static char szDBMSName[]="DBMS.*";

/*******************************************************************************
***
*** Function         : dbms_send(table, op, recordnum, record, len)
*** Preconditions    : The four parameters are inserted by the db_xxx macros
***                    defined in ..\..\include\psistack\dbms.h. Constants
***                    are also defined there.
***                    table is a DTN_ table identifier, for table-relative
***                    operations, or -1 for global operations.
***                    op is a DOP_ operation code, and is the only field
***                    guaranteed to be supplied.
***                    recordnum is a record number, only supplied for seek
***                    operations.
***                    record is a pointer to a record structure, as defined in
***                    dbms.h. Supplied in all read/write operations, and
***                    all key-based operations (these two overlap).
***                    len is the length of the record structure.
***                    
*** Postconditions   : The message is packed into a dbmsg structure, and
***                    sent to the DBMS if it is loaded.   Usually the stack 
***                    kernel does the load explicitly.  Success of the call is
***                    indicated by a return of TRUE; failure gives FALSE. In
***                    that case, check db_errno and db_epocerrno for an
***                    explanation.
***
*** Note             : This routine won't load the DBMS if it isn't loaded.
***                    If the stack kernel hasn't loaded it, then you're in 
***                    trouble.
***
*******************************************************************************/

GLDEF_C UINT dbms_send(BYTE table, BYTE op, UINT recno, VOID *record, UWORD len)
{
register HANDLE dbms_pid;
static DBMSG dbms_msg;

  /* Send message if DBMS loaded */
  if ((dbms_pid = dbms_loaded() ) > 0 ) {

    /* Package parameters into a dbmsg structure */
    dbms_msg.dm_op = op;
    dbms_msg.dm_recno = recno;
    dbms_msg.dm_record = record;
    dbms_msg.dm_len = len;

    /* This will block if the DBMS has no free message slots... (unlikely) */
    /* ...and will block until the DBMS has finished processing it */
    return p_msendreceivew(dbms_pid, table, (VOID *) &dbms_msg);

    /* The return from that p_msendreceivew will be:

       E_GEN_RECEIVER (-17) meaning that the DBMS is not loaded, or has
       not yet called p_minit(). 

       Or the DBMS's p_mfree reply value: probably 0
    */
  }
  else return dbms_pid;
}

/*******************************************************************************
***
*** Function         : dbms_loaded
*** Preconditions    : none
*** Postconditions   : positive result is the process ID of the loaded DBMS
***                    E_FILE_NXIST if the DBMS is not loaded
***
*******************************************************************************/

GLDEF_C HANDLE dbms_loaded(VOID)
{
  return p_pidfind(szDBMSName);
}

/*******************************************************************************
***
*** Function         : dbms_load
*** Preconditions    : none
*** Postconditions   : positive result is the process ID of the loaded DBMS
***                    negative result is an EPOC error code
*** Notez Bien       : This function is only to be used by the stack kernel.
***
*******************************************************************************/

GLDEF_C HANDLE dbms_load(VOID)
{
register HANDLE hDBMS = p_execc("LOC::M:\\APP\\PSISTACK\\DBMS.IMG", NULL, 0);
  if (hDBMS>0) {
    /* Set the DBMS to have a higher priority than the stack kernel,
       So it will run until it has finished initialising. Just before it
       enters its message loop, the DBMS lowers its priority to match that
       of the stack kernel */
    p_setpri(hDBMS, 0x82); /* *** RESCHEDULE *** */
    p_presume(hDBMS);
  }
  return hDBMS;
}
