/*******************************************************************************
***
*** Filename         : socket.c
*** Purpose          : PsiStack Sockets Link Library - EPOC style
***                    This gives asynchronous Sockets functionality, a la PLIB.
***                    It talks via IPC to the stack kernel.
***                    This is the "base class" (superclass), of the other
***                    Sockets libraries:
***                    bsdsock.c - BSD style sockets interface
***                    pmisock.c - PsiMail style sockets interface
*** Author           : Matt J. Gumbley
*** Created          : 07/04/98
*** Last updated     : 16/04/98
***
********************************************************************************
***
*** Modification Record
***
*******************************************************************************/

#include "..\utils\sysincl.h"
#include "..\..\kernel\stackmsg.h"

GLREF_C UINT stack_sendmsg(UINT type, STKMSG *msg); /* Private implementation */

#include "..\..\include\psistack\sys\socket.h" /* Public API structs, defines */
#include "..\..\include\psistack\socket.h"           /* Public API prototypes */



/******************************************************************************/
/*************                                                 ****************/
/************* S O C K E T   L I B R A R Y   F U N C T I O N S ****************/
/*************                                                 ****************/
/******************************************************************************/


/*******************************************************************************
***
*** Synchronous fn   : socket_socket(domain, type, protocol)
*** Preconditions    : domain is a protocol family identifier, such as PF_INET
***                    type is a socket type identifier; SOCK_DGRAM, SOCK_RAW,
***                    SOCK_STREAM or SOCK_RAWINTERFACE. protocol is usually 0.
*** Postconditions   : The function blocks until the socket is opened, or until
***                    it fails to open. It returns a socket descriptor (zero 
***                    or greater) or a negative error code (see async version)
***
*******************************************************************************/

GLDEF_C INT socket_socket(INT domain, INT type, INT protocol)
{
STKMSG msg;
INT retval;
WORD statword;
  /* Set the status word to E_FILE_PENDING before the message is sent - you'll 
     be p_iosignalbypid()'ed, and have to check this for completion */

  statword = E_FILE_PENDING;
  msg.sm_uwords[0] = (UWORD)domain;
  msg.sm_uwords[1] = (UWORD)type;
  msg.sm_uwords[2] = (UWORD)protocol;
  msg.sm_uwords[3] = (UWORD)&statword;
  retval = stack_sendmsg(SMT_SL_SOCKET, &msg);
  if (retval == 0) { /* Started OK */
    p_waitstat(&statword);
    return statword;
  }
  else
    return retval;
}

/*******************************************************************************
***
*** Asynchronous fn  : socket_socketa(domain, type, protocol, (UWORD)pstat)
*** Preconditions    : domain is a protocol family identifier, such as PF_INET
***                    type is a socket type identifier; SOCK_DGRAM, SOCK_RAW,
***                    SOCK_STREAM or SOCK_RAWINTERFACE. protocol is usually 0.
***                    pstat is a pointer to a status word
*** Postconditions   : The request is queued, and return passed without waiting
***                    for it to complete. It returns zero if the request was 
***                    started successfully, or a negative error number. When 
***                    the successfully started operation completes, the 
***                    process I/O semaphore is signalled, and *pstat contains 
***                    the completion status. While the operation is pending 
***                    (ie before the I/O semaphore has been signalled), 
***                    *pstat contains E_FILE_PENDING. Asynchronous requests 
***                    in general are described in chapter 8 of the SIBO C SDK.
***                    In summary, you should check the return for zero or error
***                    (negative) and if zero, p_iowait(). Upon waking, the
***                    status word will either be >= 0, representing a socket
***                    descriptor, or a negative error code.
*** Possible errors  : E_NET_PROTONOSUPPORT - bad domain or type.
***                    E_NET_MFILE - no free sockets.
***                    E_NET_NFILE - no free UPQs/TCBs (sockets, effectively)
***
*******************************************************************************/

GLDEF_C INT socket_socketa(INT domain, INT type, INT protocol, WORD *pstat)
{
STKMSG msg;
  /* Set the status word to E_FILE_PENDING before the message is sent - you'll 
     be p_iosignalbypid()'ed, and have to check this for completion */
  *pstat = E_FILE_PENDING;
  msg.sm_uwords[0] = (UWORD)domain;
  msg.sm_uwords[1] = (UWORD)type;
  msg.sm_uwords[2] = (UWORD)protocol;
  msg.sm_uwords[3] = (UWORD)pstat;
  return stack_sendmsg(SMT_SL_SOCKET, &msg);
}

/**** That's the last of the verbose comments for synchronous and asynchronous 
 **** functions!! */


/*******************************************************************************
***
*** Function         : socket_socketpair(domain, type, protocol, sockvec)
*** Preconditions    : As for socket, and sockvec points to two integers.
*** Postconditions   : As for socket. The sockets are connected together.
***
*******************************************************************************/

/* I don't understand how socketpair works yet */

GLDEF_C INT socket_socketpair(int domain, int type, int protocol, int sockvec[2])
{
/*int ret;
  ret = stack_sendmsg(SMT_SL_SOCKET, domain, type, protocol);
  if (ret != -1) {
    sockvec[0] = ret;
    ret = stack_sendmsg(SMT_SL_SOCKET, domain, type, protocol);
    if (ret == -1) {
      * Free first socket *
      return -1;
    }
    sockvec[1] = ret;
    * Connect them together? *
    return 0;
  }
*/
  return -1;
}

GLDEF_C INT socket_socketpaira(int domain, int type, int protocol, int sockvec[2], WORD *pstat)
{
  return -1;
}

/*******************************************************************************
***
*** Function         : socket_bind(sockfd, *saddr, addrlen)
*** Preconditions    : sockfd is a socket descriptor provided by socket_socket 
***                    or socket_socketpair. saddr points to a sockaddr 
***                    structure, and addrlen is the length of the address.
*** Postconditions   : The socket is given the address. It returns 0 on 
***                    success, or a negative error code:
***
*******************************************************************************/

GLDEF_C INT socket_bind(INT sockfd, struct sockaddr *saddr, INT addrlen)
{
STKMSG msg;
INT retval;
WORD statword = E_FILE_PENDING;
  msg.sm_uwords[0] = (UWORD)sockfd;
  msg.sm_uwords[1] = (UWORD)saddr;
  msg.sm_uwords[2] = (UWORD)addrlen;
  msg.sm_uwords[3] = (UWORD)&statword;
  retval = stack_sendmsg(SMT_SL_BIND, &msg);
  if (retval == 0) { /* Started OK */
    p_waitstat(&statword);
    return statword;
  }
  else
    return retval;
}

GLDEF_C INT socket_binda(INT sockfd, struct sockaddr *saddr, INT addrlen,
                         WORD *pstat)
{
STKMSG msg;
  *pstat = E_FILE_PENDING;
  msg.sm_uwords[0] = (UWORD)sockfd;
  msg.sm_uwords[1] = (UWORD)saddr;
  msg.sm_uwords[2] = (UWORD)addrlen;
  msg.sm_uwords[3] = (UWORD)pstat;
  return stack_sendmsg(SMT_SL_BIND, &msg);
}

/*******************************************************************************
***
*** Function         : socket_connect(sockfd, serv_addr, addrlen)
*** Preconditions    : sockfd is a socket descriptor provided by socket_socket 
***                    or socket_socketpair. serv_addr points to a sockaddr 
***                    structure, and addrlen is the length of the address.
*** Postconditions   : A connection on socket sockfd to peer at address
***                    serv_addr (which is addrlen bytes long) is opened. For
***                    connectionless socket types, just set the default address
***                    to send to and the only address from which to accept
***                    transmissions. 0 on success or a negative error code.
***
*******************************************************************************/

GLDEF_C INT socket_connect(INT sockfd, struct sockaddr *serv_addr, INT addrlen)
{
STKMSG msg;
INT retval;
WORD statword = E_FILE_PENDING;
  msg.sm_uwords[0] = (UWORD)sockfd;
  msg.sm_uwords[1] = (UWORD)serv_addr;
  msg.sm_uwords[2] = (UWORD)addrlen;
  msg.sm_uwords[3] = (UWORD)&statword;
  retval = stack_sendmsg(SMT_SL_CONNECT, &msg);
  if (retval == 0) { /* Started OK */
    p_waitstat(&statword);
    return statword;
  }
  else
    return retval;
}

GLDEF_C INT socket_connecta(INT sockfd, struct sockaddr *serv_addr, INT addrlen,
                            WORD *pstat)
{
STKMSG msg;
  *pstat = E_FILE_PENDING;
  msg.sm_uwords[0] = (UWORD)sockfd;
  msg.sm_uwords[1] = (UWORD)serv_addr;
  msg.sm_uwords[2] = (UWORD)addrlen;
  msg.sm_uwords[3] = (UWORD)pstat;
  return stack_sendmsg(SMT_SL_CONNECT, &msg);
}


/*******************************************************************************
***
*** Function         : socket_listen(sockfd, n)
*** Preconditions    : sockfd is a socket descriptor provided by socket_socket 
***                    or socket_socketpair. n is an integer.
*** Postconditions   : The socket is prepared to queue n connection requests
***                    before further requests are refused. 0 on success
***                    or a negative error code.
***
*******************************************************************************/

GLDEF_C INT socket_listen(INT sockfd, INT n)
{
STKMSG msg;
INT retval;
WORD statword = E_FILE_PENDING;
  msg.sm_uwords[0] = (UWORD)sockfd;
  msg.sm_uwords[1] = (UWORD)n;
  msg.sm_uwords[2] = (UWORD)&statword;
  retval = stack_sendmsg(SMT_SL_LISTEN, &msg);
  if (retval == 0) { /* Started OK */
    p_waitstat(&statword);
    return statword;
  }
  else
    return retval;
}

GLDEF_C INT socket_listena(INT sockfd, INT n, WORD *pstat)
{
STKMSG msg;
  *pstat = E_FILE_PENDING;
  msg.sm_uwords[0] = (UWORD)sockfd;
  msg.sm_uwords[1] = (UWORD)n;
  msg.sm_uwords[2] = (UWORD)pstat;
  return stack_sendmsg(SMT_SL_LISTEN, &msg);
}


/*******************************************************************************
***
*** Function         : socket_accept(sockfd, peeraddr, paddrlen)
*** Preconditions    : sockfd is a socket descriptor provided by socket_socket 
***                    or socket_socketpair. peeraddr is the address of a 
***                    sockaddr structure, and paddrlen is the address of an 
***                    integer.
*** Postconditions   : A connection is waited for on the socket. When it
***                    arrives, a new socket is opened to communicate with it,
***                    *peeraddr (which is *paddrlen bytes long) is set to the
***                    address of the connecting peer, and *paddrlen is set
***                    to the address's actual length. The socket's new
***                    descriptor is returned, or a negative error code.
***
*******************************************************************************/

GLDEF_C INT socket_accept(INT sockfd, struct sockaddr *peeraddr, INT *paddrlen)
{
STKMSG msg;
INT retval;
WORD statword = E_FILE_PENDING;
  msg.sm_uwords[0] = (UWORD)sockfd;
  msg.sm_uwords[1] = (UWORD)peeraddr;
  msg.sm_uwords[2] = (UWORD)paddrlen;
  msg.sm_uwords[3] = (UWORD)&statword;
  retval = stack_sendmsg(SMT_SL_ACCEPT, &msg);
  if (retval == 0) { /* Started OK */
    p_waitstat(&statword);
    return statword;
  }
  else
    return retval;
}

GLDEF_C INT socket_accepta(INT sockfd, struct sockaddr *peeraddr, INT *paddrlen,
                           WORD *pstat)
{
STKMSG msg;
  *pstat = E_FILE_PENDING;
  msg.sm_uwords[0] = (UWORD)sockfd;
  msg.sm_uwords[1] = (UWORD)peeraddr;
  msg.sm_uwords[2] = (UWORD)paddrlen;
  msg.sm_uwords[3] = (UWORD)pstat;
  return stack_sendmsg(SMT_SL_ACCEPT, &msg);
}


/*******************************************************************************
***
*** Function         : socket_getsockopt(sockfd, level, optname, *optval, 
***                                      *optlen)
*** Preconditions    : sockfd is a socket descriptor provided by socket_socket 
***                    or socket_socketpair. *optval is a buffer, *optlen bytes
***                    long.
*** Postconditions   : Put the current value for socket sockfd's option 
***                    optname at protocol level into optval and set 
***                    *optlen to the value's actual length. Returns 0 on 
***                    success, or a negative error code.
***
*******************************************************************************/

GLDEF_C INT socket_getsockopt(INT sockfd, INT level, INT optname, VOID *optval,
                              INT *optlen)
{
STKMSG msg;
INT retval;
WORD statword = E_FILE_PENDING;
  msg.sm_uwords[0] = (UWORD)sockfd;
  msg.sm_uwords[1] = (UWORD)level;
  msg.sm_uwords[2] = (UWORD)optname;
  msg.sm_uwords[3] = (UWORD)optval;
  msg.sm_uwords[4] = (UWORD)optlen;
  msg.sm_uwords[5] = (UWORD)&statword;
  retval = stack_sendmsg(SMT_SL_GETSOCKOPT, &msg);
  if (retval == 0) { /* Started OK */
    p_waitstat(&statword);
    return statword;
  }
  else
    return retval;
}

GLDEF_C INT socket_getsockopta(INT sockfd, INT level, INT optname, VOID *optval,
                               INT *optlen, WORD *pstat)
{
STKMSG msg;
  *pstat = E_FILE_PENDING;
  msg.sm_uwords[0] = (UWORD)sockfd;
  msg.sm_uwords[1] = (UWORD)level;
  msg.sm_uwords[2] = (UWORD)optname;
  msg.sm_uwords[3] = (UWORD)optval;
  msg.sm_uwords[4] = (UWORD)optlen;
  msg.sm_uwords[5] = (UWORD)pstat;
  return stack_sendmsg(SMT_SL_GETSOCKOPT, &msg);
}

/*******************************************************************************
***
*** Function         : socket_setsockopt(sockfd, level, optname, *optval, 
***                    optlen)
*** Preconditions    : sockfd is a socket descriptor provided by socket_socket 
***                    or socketpair. 
*** Postconditions   : Set socket sockfd's option optname at protocol level
***                    to *optval (which is optlen bytes long) Returns 0 on
***                    success, or a negative error code.
***
*******************************************************************************/

GLDEF_C INT socket_setsockopt(INT sockfd, INT level, INT optname, VOID *optval,
                              INT optlen)
{
STKMSG msg;
INT retval;
WORD statword = E_FILE_PENDING;
  msg.sm_uwords[0] = (UWORD)sockfd;
  msg.sm_uwords[1] = (UWORD)level;
  msg.sm_uwords[2] = (UWORD)optname;
  msg.sm_uwords[3] = (UWORD)optval;
  msg.sm_uwords[4] = (UWORD)optlen;
  msg.sm_uwords[5] = (UWORD)&statword;
  retval = stack_sendmsg(SMT_SL_SETSOCKOPT, &msg);
  if (retval == 0) { /* Started OK */
    p_waitstat(&statword);
    return statword;
  }
  else
    return retval;
}

GLDEF_C INT socket_setsockopta(INT sockfd, INT level, INT optname, VOID *optval,
                              INT optlen, WORD *pstat)
{
STKMSG msg;
  *pstat = E_FILE_PENDING;
  msg.sm_uwords[0] = (UWORD)sockfd;
  msg.sm_uwords[1] = (UWORD)level;
  msg.sm_uwords[2] = (UWORD)optname;
  msg.sm_uwords[3] = (UWORD)optval;
  msg.sm_uwords[4] = (UWORD)optlen;
  msg.sm_uwords[5] = (UWORD)pstat;
  return stack_sendmsg(SMT_SL_SETSOCKOPT, &msg);
}

/*******************************************************************************
***
*** Function         : socket_getsockname(sockfd, *addr, *paddrlen)
*** Preconditions    : sockfd is a socket descriptor provided by socket_socket 
***                    or socket_socketpair. 
*** Postconditions   : Puts the local address of sockfd into *addr and its
***                    length in *paddrlen.
***
*******************************************************************************/

GLDEF_C INT socket_getsockname(INT sockfd, struct sockaddr *addr, INT *paddrlen)
{
STKMSG msg;
INT retval;
WORD statword = E_FILE_PENDING;
  msg.sm_uwords[0] = (UWORD)sockfd;
  msg.sm_uwords[1] = (UWORD)addr;
  msg.sm_uwords[2] = (UWORD)paddrlen;
  msg.sm_uwords[3] = (UWORD)&statword;
  retval = stack_sendmsg(SMT_SL_GETSOCKNAME, &msg);
  if (retval == 0) { /* Started OK */
    p_waitstat(&statword);
    return statword;
  }
  else
    return retval;
}

GLDEF_C INT socket_getsocknamea(INT sockfd, struct sockaddr *addr, 
                                INT *paddrlen, WORD *pstat)
{
STKMSG msg;
  *pstat = E_FILE_PENDING;
  msg.sm_uwords[0] = (UWORD)sockfd;
  msg.sm_uwords[1] = (UWORD)addr;
  msg.sm_uwords[2] = (UWORD)paddrlen;
  msg.sm_uwords[3] = (UWORD)pstat;
  return stack_sendmsg(SMT_SL_GETSOCKNAME, &msg);
}

/*******************************************************************************
***
*** Function         : socket_getpeername(sockfd, *peer, *paddrlen)
*** Preconditions    : sockfd is a socket descriptor provided by socket_socket 
***                    or socket_socketpair. 
*** Postconditions   : Puts the address of the peer connected to socket sockfd 
***                    into *peer and its length in *paddrlen.
***
*******************************************************************************/

GLDEF_C INT socket_getpeername(INT sockfd, struct sockaddr *peer, INT *paddrlen)
{
STKMSG msg;
INT retval;
WORD statword = E_FILE_PENDING;
  msg.sm_uwords[0] = (UWORD)sockfd;
  msg.sm_uwords[1] = (UWORD)peer;
  msg.sm_uwords[2] = (UWORD)paddrlen;
  msg.sm_uwords[3] = (UWORD)&statword;
  retval = stack_sendmsg(SMT_SL_GETPEERNAME, &msg);
  if (retval == 0) { /* Started OK */
    p_waitstat(&statword);
    return statword;
  }
  else
    return retval;
}

GLDEF_C INT socket_getpeernamea(INT sockfd, struct sockaddr *peer, 
                                INT *paddrlen, WORD *pstat)
{
STKMSG msg;
  *pstat = E_FILE_PENDING;
  msg.sm_uwords[0] = (UWORD)sockfd;
  msg.sm_uwords[1] = (UWORD)peer;
  msg.sm_uwords[2] = (UWORD)paddrlen;
  msg.sm_uwords[3] = (UWORD)pstat;
  return stack_sendmsg(SMT_SL_GETPEERNAME, &msg);
}


/*******************************************************************************
***
*** Function         : socket_send(sockfd, *buff, len, flags)
*** Preconditions    : sockfd is a socket descriptor provided by socket_socket 
***                    or socket_socketpair. *buff is a pointer to a buffer, 
***                    len is the length of the buffer contents, flags are 
***                    flags :-)
*** Postconditions   : len bytes of the buffer are sent to the socket.
***                    Returns the number of bytes sent, or a negative error.
***
*******************************************************************************/

GLDEF_C INT socket_send(INT sockfd, VOID *buff, INT len, UWORD flags)
{
STKMSG msg;
INT retval;
WORD statword = E_FILE_PENDING;
  msg.sm_uwords[0] = (UWORD)sockfd;
  msg.sm_uwords[1] = (UWORD)buff;
  msg.sm_uwords[2] = (UWORD)len;
  msg.sm_uwords[3] = (UWORD)flags;
  msg.sm_uwords[4] = (UWORD)&statword;
  retval = stack_sendmsg(SMT_SL_SEND, &msg);
  if (retval == 0) { /* Started OK */
    p_waitstat(&statword);
    return statword;
  }
  else
    return retval;
}

GLDEF_C INT socket_senda(INT sockfd, VOID *buff, INT len, UWORD flags, 
                         WORD *pstat)
{
STKMSG msg;
  *pstat = E_FILE_PENDING;
  msg.sm_uwords[0] = (UWORD)sockfd;
  msg.sm_uwords[1] = (UWORD)buff;
  msg.sm_uwords[2] = (UWORD)len;
  msg.sm_uwords[3] = (UWORD)flags;
  msg.sm_uwords[4] = (UWORD)pstat;
  return stack_sendmsg(SMT_SL_SEND, &msg);
}


/*******************************************************************************
***
*** Function         : socket_recv(sockfd, *buff, len, flags)
*** Preconditions    : sockfd is a socket descriptor provided by socket_socket 
***                    or socket_socketpair. *buff is a pointer to a buffer, 
***                    len is the length of the buffer, flags are flags :-)
*** Postconditions   : up to len bytes of the buffer are read from the socket.
***                    Returns the number of bytes read, or negative on error.
***
*******************************************************************************/

GLDEF_C INT socket_recv(INT sockfd, VOID *buff, INT len, UWORD flags)
{
STKMSG msg;
INT retval;
WORD statword = E_FILE_PENDING;
  msg.sm_uwords[0] = (UWORD)sockfd;
  msg.sm_uwords[1] = (UWORD)buff;
  msg.sm_uwords[2] = (UWORD)len;
  msg.sm_uwords[3] = (UWORD)flags;
  msg.sm_uwords[4] = (UWORD)NULL;
  msg.sm_uwords[5] = (UWORD)NULL;
  msg.sm_uwords[6] = (UWORD)&statword;
  retval = stack_sendmsg(SMT_SL_RECVFROM, &msg);
  if (retval == 0) { /* Started OK */
    p_waitstat(&statword);
    return statword;
  }
  else
    return retval;
}

GLDEF_C INT socket_recva(INT sockfd, VOID *buff, INT len, UWORD flags, 
                         WORD *pstat)
{
STKMSG msg;
  *pstat = E_FILE_PENDING;
  msg.sm_uwords[0] = (UWORD)sockfd;
  msg.sm_uwords[1] = (UWORD)buff;
  msg.sm_uwords[2] = (UWORD)len;
  msg.sm_uwords[3] = (UWORD)flags;
  msg.sm_uwords[4] = (UWORD)NULL;
  msg.sm_uwords[5] = (UWORD)NULL;
  msg.sm_uwords[6] = (UWORD)pstat;
  return stack_sendmsg(SMT_SL_RECVFROM, &msg);
}

/*******************************************************************************
***
*** Function         : socket_sendto(sockfd, *buff, len, flags, *to, tolen)
*** Preconditions    : sockfd is a socket descriptor provided by socket_socket 
***                    or socket_socketpair. *buff is a pointer to a buffer, 
***                    len is the length of the buffer, flags are flags, to is 
***                    a pointer to a sockaddr structure and tolen is the 
***                    length of that address.
*** Postconditions   : len bytes of *buff are sent to socket sockfd to the peer
***                    at address *to (of length tolen). Returns the number
***                    sent, or negative on failure.
***
*******************************************************************************/

GLDEF_C INT socket_sendto(INT sockfd, VOID *buff, INT len, 
                          UWORD flags, struct sockaddr *to, INT tolen)
{
STKMSG msg;
INT retval;
WORD statword = E_FILE_PENDING;
  msg.sm_uwords[0] = (UWORD)sockfd;
  msg.sm_uwords[1] = (UWORD)buff;
  msg.sm_uwords[2] = (UWORD)len;
  msg.sm_uwords[3] = (UWORD)flags;
  msg.sm_uwords[4] = (UWORD)to;
  msg.sm_uwords[5] = (UWORD)tolen;
  msg.sm_uwords[6] = (UWORD)&statword;
  retval = stack_sendmsg(SMT_SL_SENDTO, &msg);
  if (retval == 0) { /* Started OK */
    p_waitstat(&statword);
    return statword;
  }
  else
    return retval;
}

GLDEF_C INT socket_sendtoa(INT sockfd, VOID *buff, INT len, 
                          UWORD flags, struct sockaddr *to, INT tolen,
                          WORD *pstat)
{
STKMSG msg;
  *pstat = E_FILE_PENDING;
  msg.sm_uwords[0] = (UWORD)sockfd;
  msg.sm_uwords[1] = (UWORD)buff;
  msg.sm_uwords[2] = (UWORD)len;
  msg.sm_uwords[3] = (UWORD)flags;
  msg.sm_uwords[4] = (UWORD)to;
  msg.sm_uwords[5] = (UWORD)tolen;
  msg.sm_uwords[6] = (UWORD)pstat;
  return stack_sendmsg(SMT_SL_SENDTO, &msg);
}

/*******************************************************************************
***
*** Function         : socket_recvfrom(sockfd, *buff, len, flags, *from, 
***                                    *fromlen)
*** Preconditions    : sockfd is a socket descriptor provided by socket_socket 
***                    or socket_socketpair. *buff is a pointer to a buffer, 
***                    len is the length of the buffer, flags are flags, from 
***                    is a pointer to a sockaddr structure and fromlen is the 
***                    length of that address.
*** Postconditions   : up to len bytes of the buffer are read from the socket.
***                    If from is non-NULL, fill in *fromlen bytes of it with
***                    the address of the sender, and store the actual size
***                    of the address in *fromlen.
***                    Returns the number of bytes read, or negative on error.
***
*******************************************************************************/

GLDEF_C INT socket_recvfrom(INT sockfd, VOID *buff, INT len,
                            UWORD flags, struct sockaddr *from, INT *fromlen)
{
STKMSG msg;
INT retval;
WORD statword = E_FILE_PENDING;
  msg.sm_uwords[0] = (UWORD)sockfd;
  msg.sm_uwords[1] = (UWORD)buff;
  msg.sm_uwords[2] = (UWORD)len;
  msg.sm_uwords[3] = (UWORD)flags;
  msg.sm_uwords[4] = (UWORD)from;
  msg.sm_uwords[5] = (UWORD)fromlen;
  msg.sm_uwords[6] = (UWORD)&statword;
  retval = stack_sendmsg(SMT_SL_RECVFROM, &msg);
  if (retval == 0) { /* Started OK */
    p_waitstat(&statword);
    return statword;
  }
  else
    return retval;
}

GLDEF_C INT socket_recvfroma(INT sockfd, VOID *buff, INT len,
                             UWORD flags, struct sockaddr *from, INT *fromlen,
                             WORD *pstat)
{
STKMSG msg;
  *pstat = E_FILE_PENDING;
  msg.sm_uwords[0] = (UWORD)sockfd;
  msg.sm_uwords[1] = (UWORD)buff;
  msg.sm_uwords[2] = (UWORD)len;
  msg.sm_uwords[3] = (UWORD)flags;
  msg.sm_uwords[4] = (UWORD)from;
  msg.sm_uwords[5] = (UWORD)fromlen;
  msg.sm_uwords[6] = (UWORD)pstat;
  return stack_sendmsg(SMT_SL_RECVFROM, &msg);
}

/**************
 Send a message described MESSAGE on socket FD.
   Returns the number of bytes sent, or -1 for errors.  
int sendmsg (int fd, struct msghdr *message,
      unsigned int flags);

   Receive a message as described by MESSAGE from socket FD.
   Returns the number of bytes read or -1 for errors. 
int recvmsg (int fd, struct msghdr *message,
      unsigned int flags);
***************/
 

/*******************************************************************************
***
*** Function         : socket_shutdown(sockfd, how)
*** Preconditions    : sockfd is a socket descriptor provided by socket_socket 
***                    or socket_socketpair. how specifies what to shut down:
***                    0 = no more receptions
***                    1 = no more transmissions
***                    2 = neither
*** Postconditions   : All or part of the connection open on sockfd is shut down
***                    Returns 0 on success, negative on failure
***
*******************************************************************************/

GLDEF_C INT socket_shutdown(INT sockfd, INT how)
{
STKMSG msg;
INT retval;
WORD statword = E_FILE_PENDING;
  msg.sm_uwords[0] = (UWORD)sockfd;
  msg.sm_uwords[1] = (UWORD)how;
  msg.sm_uwords[2] = (UWORD)&statword;
  retval = stack_sendmsg(SMT_SL_SHUTDOWN, &msg);
  if (retval == 0) { /* Started OK */
    p_waitstat(&statword);
    return statword;
  }
  else
    return retval;
}

GLDEF_C INT socket_shutdowna(INT sockfd, INT how, WORD *pstat)
{
STKMSG msg;
  *pstat = E_FILE_PENDING;
  msg.sm_uwords[0] = (UWORD)sockfd;
  msg.sm_uwords[1] = (UWORD)how;
  msg.sm_uwords[2] = (UWORD)pstat;
  return stack_sendmsg(SMT_SL_SHUTDOWN, &msg);
}

/*******************************************************************************
***
*** Function         : socket_close(sockfd)
*** Preconditions    : sockfd is a socket descriptor provided by socket_socket 
***                    or socket_socketpair. 
*** Postconditions   : The socket is closed. Returns 0 on success, or a negative
***                    error.
***
*******************************************************************************/

GLDEF_C INT socket_close(INT sockfd)
{
STKMSG msg;
INT retval;
WORD statword = E_FILE_PENDING;
  msg.sm_uwords[0] = (UWORD)sockfd;
  msg.sm_uwords[1] = (UWORD)&statword;
  retval = stack_sendmsg(SMT_SL_CLOSE, &msg);
  if (retval == 0) { /* Started OK */
    p_waitstat(&statword);
    return statword;
  }
  else
    return retval;
}

GLDEF_C INT socket_closea(INT sockfd, WORD *pstat)
{
STKMSG msg;
  *pstat = E_FILE_PENDING;
  msg.sm_uwords[0] = (UWORD)sockfd;
  msg.sm_uwords[1] = (UWORD)pstat;
  return stack_sendmsg(SMT_SL_CLOSE, &msg);
}

