/*******************************************************************************
***
*** Filename         : stack.c
*** Purpose          : PsiStack Stack Kernel API Link Library
***                    Gets linked with client apps; responsible for packaging
***                    stack messages into buffers, IPC'ing them to the 
***                    stack process (if it exists), and returning the result.
***                    Also provides routines for loading and unloading the 
***                    stack kernel.
*** Author           : Matt J. Gumbley, Patrick Robbe
*** Created          : 23/07/97
*** Last updated     : 24/08/98
***
********************************************************************************
***
*** Modification Record
***
*******************************************************************************/

#include "..\utils\sysincl.h"                 /* 'Cos this is an EPOC library */
#include "..\..\kernel\stackmsg.h"       /* Private implementation definition */
#include "..\..\include\psistack\stack.h"            /* Public API definition */

#include "..\..\include\psistack\console.h"          /* Public API definition */

GLDEF_D HANDLE hStack = NULL;
LOCAL_D TEXT   szStackName[]="STACK.*";

/*******************************************************************************
***
*** Function         : stack_load
*** Preconditions    : none
*** Postconditions   : positive result is the process ID of the loaded stack
***                    negative result is an EPOC error code
***
*******************************************************************************/

GLDEF_C HANDLE stack_load()
{
  hStack = p_execc("LOC::M:\\APP\\PSISTACK\\STACK.IMG", NULL, 0);
  if (hStack > 0) {
    /* Set the stack to have a higher priority than the setup program / user 
       application, so it will run until it has finished initialising. 
       Just before it enters its message loop, the stack lowers its priority 
       to match that of the setup program / user app. */
    p_setpri(hStack, 0x81); /* *** RESCHEDULE *** */
    p_presume(hStack);
  }
  return hStack;
}


/*******************************************************************************
***
*** Function         : stack_loaded()
*** Preconditions    : None
*** Postconditions   : positive result is the process ID of the loaded stack
***                    E_FILE_NXIST if the stack is not loaded
***
*******************************************************************************/

GLDEF_C HANDLE stack_loaded()
{
  return hStack = p_pidfind(szStackName);
}


/*******************************************************************************
***
*** Function         : stack_ensureloaded()
*** Preconditions    : None
*** Postconditions   : Returns positive stack PID, or negative (EPOC) error code
***
*******************************************************************************/

GLDEF_C HANDLE stack_ensureloaded()
{
HANDLE pid;
  if ((pid=stack_loaded()) < 0)
    return stack_load();
  return pid;
}


/*******************************************************************************
***
*** Function         : stack_sendmsg(msgtype, msg)
*** Preconditions    : msgtype is an SMT_??? code; msg is a *STKMSG
***                    You wouldn't normally use this routine in your own code
***                    since this is the guts of the stack API - port this 
***                    routine to OPL, Java, Forth, or whatever...
*** Postconditions   : The return value is an EPOC or E_NET error code. 
***
*******************************************************************************/

GLDEF_C UINT stack_sendmsg(UINT type, STKMSG *msg)
{
UINT retval;

  if (hStack == NULL)
    if ((retval = stack_ensureloaded()) < 0)
      return retval;

  /* This will block if the stack has no free message slots... (unlikely) */
  /* ...and will block until the stack has finished processing it */
  retval = p_msendreceivew(hStack, type, (VOID *) msg);

  /* The return from that p_msendreceivew will be:

     E_GEN_RECEIVER (-17) meaning that the stack is not loaded, or has
     not yet called p_minit(). 

     Or the stack's p_mfree reply value: (message type dependent) 
   */
  return retval;
}


/*******************************************************************************
***
*** Function         : stack_APICALL(params)
*** Preconditions    : The parameters are defined in the SDK.
*** Postconditions   : The call is passed to the stack.
***
*******************************************************************************/

GLDEF_C UINT stack_queue_event_request(UWORD eventmask, WORD *pstatusword, 
                                       VOID *pbuffer, UWORD length, 
                                       UWORD *plenbuffer)
{
STKMSG msg;
  /* Set the status word to E_FILE_PENDING before the message is sent - you'll 
     be p_iosignalbypid()'ed, and have to check this for completion */

  *pstatusword = E_FILE_PENDING;

  msg.sm_uwords[0] = eventmask;
  msg.sm_uwords[1] = (UWORD) pstatusword;
  msg.sm_uwords[2] = (UWORD) pbuffer;
  msg.sm_uwords[3] = length;
  msg.sm_uwords[4] = (UWORD) plenbuffer;
  return stack_sendmsg(SMT_QUEUE_EVENT_REQUEST, &msg);
}

GLDEF_C UINT stack_cancel_event_request(VOID)
{
STKMSG msg; /* Don't really need this */
  return stack_sendmsg(SMT_CANCEL_EVENT_REQUEST, &msg);
}

GLDEF_C UINT stack_unload(VOID)
{
STKMSG msg; /* Don't really need this */
  return stack_sendmsg(SMT_UNLOAD, &msg);
}


GLDEF_C UINT stack_get_netifstate(VOID)
{
STKMSG msg; /* Don't really need this */
UINT retval;
  retval = stack_sendmsg(SMT_GET_NETIFSTATE, &msg);
  return retval;
}

GLDEF_C UINT stack_open_netif(UWORD mode)
{
STKMSG msg; 
  msg.sm_uwords[0] = TRUE;
  msg.sm_uwords[1] = mode;
  return stack_sendmsg(SMT_OPENCLOSE_NETIF, &msg);
}

GLDEF_C UINT stack_close_netif(VOID)
{
STKMSG msg; 
  msg.sm_uwords[0] = FALSE;
  return stack_sendmsg(SMT_OPENCLOSE_NETIF, &msg);
}

GLDEF_C UINT stack_set_misc(UWORD misc)
{
STKMSG msg;
  msg.sm_uwords[0] = misc;
  return stack_sendmsg(SMT_SET_MISC, &msg);
}

GLDEF_C UINT stack_get_misc(VOID)
{
STKMSG msg;
  return stack_sendmsg(SMT_GET_MISC, &msg);
}

GLDEF_C UINT stack_set_netif(struct netif *netif)
{
STKMSG msg;
  msg.sm_uwords[0] = (UWORD) netif;
  return stack_sendmsg(SMT_SET_NETIF, &msg);
}

GLDEF_C UINT stack_get_netif(struct netif *netif)
{
STKMSG msg;
  msg.sm_uwords[0] = (UWORD) netif;
  return stack_sendmsg(SMT_GET_NETIF, &msg);
}

GLDEF_C UINT stack_set_asy(struct asy *asy)
{
STKMSG msg;
  msg.sm_uwords[0] = (UWORD) asy;
  return stack_sendmsg(SMT_SET_ASY, &msg);
}

GLDEF_C UINT stack_get_asy(struct asy *asy)
{
STKMSG msg;
  msg.sm_uwords[0] = (UWORD) asy;
  return stack_sendmsg(SMT_GET_ASY, &msg);
}

GLDEF_C UINT stack_open_asy(VOID)
{
STKMSG msg;
  msg.sm_uwords[0] = TRUE;
  return stack_sendmsg(SMT_OPENCLOSE_ASY, &msg);
}

GLDEF_C UINT stack_close_asy(VOID)
{
STKMSG msg;
  msg.sm_uwords[0] = FALSE;
  return stack_sendmsg(SMT_OPENCLOSE_ASY, &msg);
}

GLDEF_C UINT stack_set_asyhandler(VOID)
{
STKMSG msg;
  msg.sm_uwords[0] = 0xBEEF;
  return stack_sendmsg(SMT_OPENCLOSE_ASY, &msg);
}

GLDEF_C UINT stack_set_sc(struct scriptcontext *sc)
{
STKMSG msg;
  msg.sm_uwords[0] = (UWORD) sc;
  return stack_sendmsg(SMT_SET_SC, &msg);
}

GLDEF_C UINT stack_get_sc(struct scriptcontext *sc)
{
STKMSG msg;
  msg.sm_uwords[0] = (UWORD) sc;
  return stack_sendmsg(SMT_GET_SC, &msg);
}

GLDEF_C UINT stack_set_ispc(struct ISPcontext *ispc)
{
STKMSG msg;
  msg.sm_uwords[0] = (UWORD) ispc;
  return stack_sendmsg(SMT_SET_ISPC, &msg);
}

GLDEF_C UINT stack_get_ispc(struct ISPcontext *ispc)
{
STKMSG msg;
  msg.sm_uwords[0] = (UWORD) ispc;
  return stack_sendmsg(SMT_GET_ISPC, &msg);
}

GLDEF_C UINT stack_set_log(UWORD log)
{
STKMSG msg;
  msg.sm_uwords[0] = log;
  return stack_sendmsg(SMT_SET_LOG, &msg);
}

GLDEF_C UINT stack_get_log(VOID)
{
STKMSG msg;
  return stack_sendmsg(SMT_GET_LOG, &msg);
}



GLDEF_C UINT stack_set_ping(struct ping_info *ping_info)
{
STKMSG msg;
  msg.sm_uwords[0] = (UWORD) ping_info;
  return stack_sendmsg(SMT_SET_PING, &msg);
}

GLDEF_C UINT stack_get_ping(struct ping_info *ping_info)
{
STKMSG msg;
  msg.sm_uwords[0] = (UWORD) ping_info;
  return stack_sendmsg(SMT_GET_PING, &msg);
}


GLDEF_C UINT stack_send_ping(VOID)
{
STKMSG msg;
  return stack_sendmsg(SMT_SEND_PING, &msg);
}

GLDEF_C UINT stack_send_key(UWORD key)
{
STKMSG msg;
  msg.sm_uwords[0] = key;
  return stack_sendmsg(SMT_SEND_KEY, &msg);
}

GLDEF_C UINT stack_stats(VOID)
{
STKMSG msg;
  return stack_sendmsg(SMT_STATS, &msg);
}


GLDEF_C UINT stack_set_udp_config(struct udp_config *udp_config)
{
STKMSG msg;
  msg.sm_uwords[0] = (UWORD) udp_config;
  return stack_sendmsg(SMT_SET_UDP_CONFIG, &msg);
}

GLDEF_C UINT stack_get_udp_config(struct udp_config *udp_config)
{
STKMSG msg;
  msg.sm_uwords[0] = (UWORD) udp_config;
  return stack_sendmsg(SMT_GET_UDP_CONFIG, &msg);
}


GLDEF_C UINT stack_set_tcp_config(struct tcp_config *tcp_config)
{
STKMSG msg;
  msg.sm_uwords[0] = (UWORD) tcp_config;
  return stack_sendmsg(SMT_SET_TCP_CONFIG, &msg);
}

GLDEF_C UINT stack_get_tcp_config(struct tcp_config *tcp_config)
{
STKMSG msg;
  msg.sm_uwords[0] = (UWORD) tcp_config;
  return stack_sendmsg(SMT_GET_TCP_CONFIG, &msg);
}

GLDEF_C UINT stack_set_ppp_config(struct ppp_config *ppp_config)
{
STKMSG msg;
  msg.sm_uwords[0] = (UWORD) ppp_config;
  return stack_sendmsg(SMT_SET_PPP_CONFIG, &msg);
}

GLDEF_C UINT stack_get_ppp_config(struct ppp_config *ppp_config)
{
STKMSG msg;
  msg.sm_uwords[0] = (UWORD) ppp_config;
  return stack_sendmsg(SMT_GET_PPP_CONFIG, &msg);
}

