/*******************************************************************************
***
*** Filename         : utils.c
*** Purpose          : Utility function library for PsiStack
*** Author           : Matt J. Gumbley
*** Created          : 20/01/97
*** Last updated     : 01/03/98
***
********************************************************************************
***
*** Modification Record
***
*******************************************************************************/

#include "sysincl.h"
#include "..\..\include\psistack\sys\global.h"
#include "..\..\include\psistack\sys\ip.h"

GLDEF_D char hexdigs[]="0123456789ABCDEF";
GLDEF_D char *szBaudRates[]= {
  "300", "600", "1200", "2400", "4800", "9600",
  "19200", "38400", "57600", "115200"
};

/*******************************************************************************
***
*** Function         : diagchar(ch)
*** Preconditions    : ch is an ASCII character
*** Postconditions   : The ASCII name of the character is returned.
***
*******************************************************************************/

char *diagchar(int ch)
{
  static char buf[40];
  char buf1[20];
  char *str;

  switch(ch) {
    case 0 : str="NUL"; break;
    case 1 : str="SOH"; break;
    case 2 : str="STX"; break;
    case 3 : str="ETX"; break;
    case 4 : str="EOT"; break;
    case 5 : str="ENQ"; break;
    case 6 : str="ACK"; break;
    case 0x15 : str="NAK"; break;
    default: p_atos(buf1,"'%c' (%d, 0x%02X)",p_isprint(ch) ? ch : '?', ch, ch); str=buf1; break;
  }
  p_scpy(buf,str);
  return buf;
}

/*******************************************************************************
***
*** Function         : hexdig
*** Preconditions    : 0 <= num <= 15
*** Postconditions   : num is converted into its corresponding hexadecimal digit
***
*******************************************************************************/

char hexdig(int num)
{
  return num < 16 ? hexdigs[num] : '?';
}

/*******************************************************************************
***
*** Function         : hexdig2dec
*** Preconditions    : dig is in the character set {0..9, a..f, A..F}
*** Postconditions   : dig is converted into its corresponding decimal digit
***
*******************************************************************************/

int hexdig2dec(char dig)
{
  if (p_isupper(dig))
    dig = p_tolower(dig);
  if (dig>='0' && dig<='9')
    return dig - '0';
  return dig - 'a' + 10;
}



/*******************************************************************************
***
*** Function         : cksum (p. 72, Comer vol 2)
*** Preconditions    : buf is a buffer, nwords is the number of 16-bit words
***                    contained therein.
*** Postconditions   : cksum is the 16-bit one's complement of the 16-bit
***                    one's complement sum of nwords from buf
***
*******************************************************************************/

word16 cksum(word16 *buf, int nwords)
{
word32 sum;
  for (sum = 0; nwords > 0; nwords--)
    sum += *buf++;
  sum = (sum >> 16) + (sum & 0xffff); /* Add in carry */
  sum += (sum >> 16);
  /* was return ~sum; before EPOC port */
  return (word16)(~sum & 0xffff);
}



/*******************************************************************************
***
*** Function         : szConvertIPaddr(ip)
*** Preconditions    : ip is an IPaddr structure
*** Postconditions   : A pointer to static data is returned, holding the IP
***                    address in dotted-quad ASCII form.
***
*******************************************************************************/

GLDEF_C char *szConvertIPaddr(IPaddr ip)
{
static char buf[20];
  p_atos(buf,"%d.%d.%d.%d",ip[0],ip[1],ip[2],ip[3]);
  return buf;
}


/*******************************************************************************
***
*** Function         : isdqchar(ch)
*** Preconditions    : ch is a character
*** Postconditions   : isdqchar is TRUE iff ch is a digit or a full stop
***
*******************************************************************************/

GLDEF_C int isdqchar(char ch)
{
  return (ch == '.' || p_isdigit(ch));
}


/*******************************************************************************
***
*** Function         : isdottedquad(buf, len)
*** Preconditions    : buf points into a len-byte buffer.
*** Postconditions   : If there is a correctly-formed dotted quad at buf,
***                    TRUE is returned, else, FALSE.
*** Disclaimer       : This routine was a bitch to write. It's a bitch to read.
***
*******************************************************************************/

GLDEF_C int isdottedquad(char *buf, int len)
{
int octetno, digno, octet, i;

  octet = digno = 0;
  octetno = 1;

  for (i = 0; i < len; i++) {
    if (p_isdigit(buf[i])) {
      octet *= 10;
      octet += buf[i] - '0';
      digno++;
      if (octet > 255) {
        /*console_puts("octet>255, NO");*/
        return FALSE;
      }
    }
    else {
      if (buf[i] == '.') {
        if (digno > 0) {
          octetno++;
          octet = digno = 0;
          if (octetno > 4) /* Trailing dot */ {
            /*console_puts("trailing dot, YES");*/
            return TRUE;
          }
        }
        else {
          /*console_puts("start or double dot, NO");*/
          return FALSE;
        }
      }
      else { /* buf[i] != digit or dot */
        /*console_printf("char %d is a non DQ char (%c) octetno=%d\r\n", 
                      i, buf[i], octetno);*/
        if (octetno >= 4 && digno > 0) {
          /*console_puts("non DQ digit after DQ: finished parsing, YES");*/
          return TRUE;
        }
        else {
          /*console_puts("starts with non DQ char, NO");*/
          return FALSE;
        }
      }
    }
  }
  if (octetno >= 4 && digno > 0) {
    /*console_puts("end of data and OK, YES");*/
    return TRUE;
  }
  /*console_puts("end of data and not OK, NO");*/
  return FALSE;
}


/*******************************************************************************
***
*** Function         : IPaddrConvertsz(str, dest)
*** Preconditions    : str is a dotted-quad ASCII string, dest is an IPaddr 
***                    structure
*** Postconditions   : If str is a correctly-formed dotted quad, it is stored
***                    as an IPaddr at dest, and TRUE is returned, else, FALSE.
***
*******************************************************************************/

GLDEF_C int IPaddrConvertsz(char *str, IPaddr dest)
{
IPaddr temp;
int octet,octetNum;
int i;
  for (octetNum = 0, i=0; octetNum < 4; octetNum++) {
    for (octet=0; ; i++) {
      if ((str[i]=='.') || (str[i]=='\0' && octetNum==3)) {
        if (octet>255) 
          return FALSE;
        temp[octetNum]=octet & 0xff;
        i++;
        break;
      }
      if (p_isdigit(str[i])) {
        octet *= 10;
        octet += (str[i]-'0');
        continue;
      }
      if (octetNum != 3) 
        return FALSE;
    }
    if (str[i]=='\0' && octetNum != 3) 
      return FALSE;
  }
  /* All OK */
  for (octetNum = 0; octetNum < 4; octetNum++) 
    dest[octetNum]=temp[octetNum];
  return TRUE;
}


/*******************************************************************************
***
*** Function         : beep()
*** Preconditions    : Blessed silence
*** Postconditions   : Blessed silence
***
*******************************************************************************/

GLDEF_C VOID beep(VOID)
{
  p_sound(-5,320); /* Beep! */
}

